import os

from abc import ABC, abstractmethod
from pandas import DataFrame
from . RxSerializable import RxSerializable
from . RxSerializable import rx_native_call
import revoscalepy.datasource as ds


def validate_argument(source, name: str, types=None, default=None):
    if types is None:
        types = object
    if source is None and default is None:
        raise ValueError("Error, parameter {:s} cannot be None".format(name))
    if source is not None and not isinstance(source, types):
        if isinstance(types, tuple):
            raise ValueError("Parameter {:s} should be one of the following types {:s}".format(
                name, str([item.__name__ for item in types])))
        else:
            raise ValueError("Parameter {:s} should be type '{:s}'".format(name, types.__name__))

    if source is None:
        return default
    return source


def validate_list_argument(source, name: str, types=None, default=None):
    if types is None:
        types = ()
    if source is None and default is None:
        raise ValueError("Error, parameter {:s} cannot be None".format(name))
    if source is not None and any(types) and not any([isinstance(source_item, types) for source_item in source]):
        raise ValueError("Parameter {:s} should be a list of one of the following types {:s}".format(
            name, str([item.__name__ for item in types])))
    if source is None:
        return default
    return source


def validate_list_string_argument(source, name: str, value_list=None):
    if value_list is None:
        value_list = ()
    if source is not None and source.lower() not in value_list:
       raise ValueError("Parameter {:s} should be one of the following value {:s}".format(name, str(value_list)))
    return source


def get_optional_parameter(param_name:str, **kwargs):
    if param_name is None:
        return None
    if param_name in kwargs:
        return kwargs[param_name]
    return None


class RxFunction(RxSerializable):
    """Base class for all RxFunctions"""

    def call(self):
        """Call is our entry point into the BxlServer engine. Functions with the @rx_parameter decorator will be
           serialized and results will be put into the self._results dictionary"""
        #print("Calling function {0}".format(type(self).__name__))

        #print("Serializing parameters: ")
        params = self.serialize_bxl()
        self._results = rx_native_call(self._function_name, params)

    def __init__(self, function_name):
        self._results = []
        self._function_name = function_name

    @abstractmethod
    def print(self):
        """Summary/Print views should be overridden here"""
        pass

    @staticmethod
    def call_function(name: str, params: dict):
        return rx_native_call(name, params)

    @staticmethod
    def assign_or_throw(source, none_acceptable=False, is_of_type=None, not_implemented=False):
        if not_implemented:
            if source is not None:
                raise ValueError ("Parameter is not currently supported")
            else:
                return

        if not none_acceptable and source is None:
            raise ValueError(
                "Error, None parameter passed when parameter cannot be none")

        if is_of_type is not None and source is not None and not isinstance(source, is_of_type):
            raise ValueError("Invalid type passed to function")

        return source

    @staticmethod
    def stop_if_none(obj, name):
        if obj is None:
            raise ValueError("Parameter '{0}' cannot be None".format(name))

    @staticmethod
    def stop_if_not_type(obj, name, type_list, allow_none=True):
        if allow_none and obj is None:
            return
        for t in type_list:
            if isinstance(obj, t):
                return

        raise ValueError(
            "Parameter '{0}' must be one of the following types: {1}".format(name, type_list))

    @staticmethod
    def resolve_datasource(datasource: ds.RxDataSource):
        if isinstance(datasource, DataFrame):
            return None
        if isinstance(datasource, str):
            _, ext = os.path.splitext(datasource)
            if ext in ('.xdf', ''):
                return ds.RxXdfData(datasource).to_serialize
            else:
                raise ValueError("If datasource is a string, it must be a filename with extension '.xdf' or a folder.")
        if isinstance(datasource, ds.RxSqlServerData):
            datasource.set_in_sql_server(False)
        return datasource.to_serialize

    def rx_results(self):
        return self._results

    def __repr__(self):
        self.print()

class RxModelResults(ABC):
    def __init__(self, model_type, formula, params, data, compute_context, call):
        self._model_type = model_type
        self._formula = formula
        self._params = params
        if '..rxDataFrame' in self._params:
            self._params['..rxDataFrame'] = None

        from revoscalepy import RxDataSource
        self._data = data if isinstance(data, RxDataSource) else None

        self._compute_context = compute_context
        self._call = call

    @property
    def model_type(self):
        return self._model_type

    @property
    def formula(self) -> str:
        return self._formula

    @property
    def compute_context(self):
        return self._compute_context

    @property
    def params(self) -> dict:
        return self._params
    
    @property
    def transform_function(self) -> callable:
        return self._params['transformFunc']
    
    @transform_function.setter
    def transform_function(self, value : callable):
        self._params['transformFunc'] = value

    @property
    def call(self):
        return self._call