
import os
import inspect
import textwrap

from abc import ABC, abstractmethod
from typing import Union
from collections import OrderedDict

from pandas import DataFrame
from pandas import Categorical
from pandas import Series
import numpy as np

from . utils.RxOptions import RxOptions
from . utils.RxTrace import rx_runtime_trace
from . utils.RxTrace import RxTraceLevel


class DataFrameDict(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

if os.path.isdir(os.path.join(os.path.dirname(os.path.realpath(__file__)), "rxLibs")):
    from . rxLibs.pylink import px_call
else:
    from pylink import px_call

class RxMissingValues(ABC):
    """
    .. rxtitle::

        Provides different missing (NA) value constants for various scalar types
        used by revoscalepy
        
    .. rxdescription::

        revoscalepy package uses Pandas dataframe as the abstraction to hold the
        data and manipulate it. Pandas dataframe in turn uses NumPy ndarray to
        hold homogeneous data efficiently and provides fast access and manipulation
        functions on the ndarray. Python and NumPy ndarray in particular only has
        numpy.NaN value for float types to indicate missing values. For other primitive
        datatypes, like int, there isn't a missing value notation. If one uses Python's
        None object to represent missing value in a series of data, say int, the whole
        ndarray's dtype changes to object. This makes dealing with data with 
        missing values quite inefficient during processing.
        
        This class provides missing values for various NumPy data types which one
        can use to mark missing values in a sequence of data in ndarray.
        
        It provides missing value for following types:
        
            - int8() for numpy.int8 with value of numpy.iinfo(np.int8).min
            - uint8() for numpy.uint8 with value of numpy.iinfo(np.uint8).max
            - int16() for numpy.int16 with value of numpy.iinfo(np.int16).min
            - uint16() for numpy.uint16 with value of numpy.iinfo(np.uint16).max
            - int32() for numpy.int32 with value of numpy.iinfo(np.int32).min
            - uint32() for numpy.uint32 with value of numpy.iinfo(np.uint32).max
            - int64() for numpy.int64 with value of numpy.iinfo(np.int64).min
            - uint64() for numpy.uint64 with value of numpy.iinfo(np.uint64).max
            - float16() for numpy.float16 with value of numpy.NaN
            - float32() for numpy.float32 with value of numpy.NaN
            - float64() for numpy.float64 with value of numpy.NaN

    .. rxkeywords:: missing values, NA, NaN, nan, none, None

    .. rxexample:: 
        ## Not run:
        from revoscalepy import RxXdfData, RxSqlServerData, RxInSqlServer
        from revoscalepy import RxOptions, rx_logit, RxMissingValues
        from revoscalepy.functions.RxLogit import RxLogitResults
        import os

        def transform_late (data, context):
            import numpy

            #
            # create a new 'Late' column based on 'ArrDelay' int32 column
            #
            data['Late'] = data['ArrDelay'] > 15

            #
            # replace all 'Late' with NaN for 'NA' or missing values in ArrDelay 
            #
            data.loc[data['ArrDelay'] == RxMissingValues.int32(), ('Late')] = numpy.NaN

            return data

        transformVars = ["ArrDelay"]
        sample_data_path = RxOptions.get_option("sampleDataDir")
        xdf_file = os.path.join(sample_data_path, "AirlineDemoSmall.xdf")

        data = RxXdfData(xdf_file)
        model = rx_logit("Late~CRSDepTime + DayOfWeek", data=data, transform_function=transform_late, transform_variables=transformVars)

        assert isinstance(model, RxLogitResults)
        ## End(Not run)
    """
    def int8() -> np.int8:
        return np.iinfo(np.int8).min

    def uint8() -> np.uint8:
        return np.iinfo(np.uint8).max
    
    def int16() -> np.int16:
        return np.iinfo(np.int16).min

    def uint16() -> np.uint16:
        return np.iinfo(np.uint16).max

    def int32() -> np.int32:
        return np.iinfo(np.int32).min

    def uint32() -> np.uint32:
        return np.iinfo(np.uint32).max

    def int64() -> np.int64:
        return np.iinfo(np.int64).min

    def uint64() -> np.uint32:
        return np.iinfo(np.uint64).max

    def float16() -> np.float16:
        return np.NaN

    def float32() -> np.float32:
        return np.NaN
    
    def float64() -> np.float64:
        return np.NaN


def resolve_dataframe(dataframe):
    if isinstance(dataframe, DataFrame):
        ret = OrderedDict()
        rx_var_info = getattr(dataframe, "var_info", None)
        ret['..rxVarInfo'] = {}

        for i in dataframe:
            if rx_var_info is not None and i in rx_var_info.keys():
                ret['..rxVarInfo'][i] = rx_var_info[i]
            if str(dataframe[i].dtype) == 'category':
                # BxlServer needs to be one-based and have bSubtractOne set to true

                # Workaround, empty dataframe needs to be sent as an array to convey type information
                if len(dataframe[i]) == 0:
                    # Default numpy array will be float, categorical variable needs to be int
                    ret[i] = np.array([x + 1 for x in dataframe[i].cat.codes.as_matrix().tolist()]) \
                        .reshape((len(dataframe[i]), 1)).astype(np.int)
                else:
                    ret[i] = [x + 1 for x in dataframe[i].cat.codes.as_matrix().tolist()]
                ret['..rxVarInfo'][i] = {}
                ret['..rxVarInfo'][i]['bIsFactor'] = True
                ret['..rxVarInfo'][i]['bSubtractOne'] = True
                ret['..rxVarInfo'][i]['levels'] = [str(cast) for cast in dataframe[i].cat.categories.tolist()]
            else:
                if len(dataframe[i]) == 0:
                    # Workaround, empty dataframe needs to be sent as an array to convey type information
                    ret[i] = dataframe[i].values.reshape((len(dataframe[i]), 1))
                elif dataframe[i].dtypes == np.object:
                    ret[i] = dataframe[i].values.tolist()
                else:
                    ret[i] = dataframe[i].values
            
            if hasattr(dataframe[i], 'rx_attributes') and isinstance(dataframe[i].rx_attributes, dict):
                for k, v in dataframe[i].rx_attributes.items():
                    if i not in ret['..rxVarInfo']:
                        ret['..rxVarInfo'][i] = {}
                    ret['..rxVarInfo'][i][k] = v

        ret['.rxDataFrameNumberColumns'] = len(dataframe.columns)

        return ret
    return None


def parse_result_dataframe(result):

    # BxlServer will add a rxVarInfo key to any dictionary where
    # the column contains additional attributes like rxdescription
    # lowhigh, factor levels, etc. A pandas dataframe does not support
    # column attributes, so instead this is modeled by adding a dictionary
    # where each column is a key into that dictionary -- this dictionary
    # then contains the attributes for that column.
    # This dictionary is added as an attribute called "var_info". Because
    # a dataframe can be returned as a dictionary -- this also handles that
    # case, adding a var_info attribute to the custom dictionary class (DataFrameDict)

    if isinstance(result, dict):
        var_info = result.pop("..rxVarInfo", None)
        col_order = result.pop("..rxColumnOrder", None)

        if "..rxIsDataFrame" in result and result["..rxIsDataFrame"] == True:
            del result["..rxIsDataFrame"]

            tempOrdered = OrderedDict()
            for x in col_order:
                tempOrdered[x] = result[x]

            result = DataFrame(tempOrdered)

            if var_info is not None:
                for col in list(result.columns.values):
                    if col in var_info and "levels" in var_info[col]:
                        levels = var_info[col]["levels"]
                        #
                        # make sure levels is a list
                        #
                        if not isinstance(levels, list):
                            levels = [levels]
                        #
                        # reset missing values to 0 to make them -1 in next step
                        # to eventually map it to numpy.NaN
                        #
                        result.loc[result[col] == RxMissingValues.int32(), (col)] = 0
                        
                        #
                        # switch to 0 based indexing scheme
                        #
                        result[col] -= 1

                        #
                        # create the categorical column now
                        #
                        result[col] = Categorical.from_codes(result[col], categories=levels)
        else:
            if col_order is not None:
                # if col_order is specified, then make sure the columns are put in the correct order in the
                # dictionary
                ordered_result = OrderedDict()
                for x in col_order:
                    ordered_result[x] = parse_result_dataframe(result[x])
                result = ordered_result
            else:
                result = DataFrameDict(result)
                for k,v in result.items():
                    result[k] = parse_result_dataframe(v)

        # Add the var info as an attribute to the result
        if var_info is not None:
            setattr(result, "var_info", var_info)
    elif isinstance(result, list):
        for idx, item in enumerate(result):
            item = parse_result_dataframe(item)
            result[idx] = item

    return result


def build_transform(original_transform_func: callable):
    if original_transform_func.__name__ == "rxTransformFunction":
        return original_transform_func
    else:
        def rxTransformFunction(dataset, context):
            ret = resolve_dataframe(original_transform_func(parse_result_dataframe(dataset), context))
            return ret
        return rxTransformFunction


def _resolve_transform_function(transform_function: callable):
    #
    # redefine transform function in __main__ module to avoid
    # transform function module name resolution error on remote
    #
    if (callable(transform_function) and
        transform_function.__module__ != '__main__' and
        not transform_function.__module__.startswith('revoscalepy.')):
        exec(textwrap.dedent(inspect.getsource(transform_function)))
        transform_function = locals()[transform_function.__name__]

    return transform_function


def rx_native_call(functionname: str, params: dict, rxTransformFunction: callable=None) -> Union[dict, DataFrame]:
    #
    # the stack frame of this function call is where BxlServer calls back to executes python logic.
    # locals defined in this frame may be used when BxlServer calls back.
    # e.g. rxTransformFunction will be used to perform the data transform if required.
    #

    # Note: tlc dll path has to be determined on the server side.
    if functionname == "TlcBridge":
        import importlib # or pkgutil
        mml_loader = importlib.util.find_spec("microsoftml").loader
        tlc_path = os.path.join(os.path.dirname(mml_loader.path), "mxLibs")
        params["tlcPath"] = tlc_path

    # Note: The BXL side implementation of this expects the variable name of this function
    # to be the same as the name of the function in build_transform.
    if "transformFunc" in params and callable(params['transformFunc']):
        rx_runtime_trace("handle transformFunc", RxTraceLevel.DEBUG)
        rxTransformFunction = build_transform(params['transformFunc'])
        params['transformFunc'] = rxTransformFunction

    rx_runtime_trace("call pylink, function name: " + functionname, RxTraceLevel.DEBUG)
    ret = px_call(functionname, params)

    rx_runtime_trace("parse result dataframe", RxTraceLevel.DEBUG)
    ret = parse_result_dataframe(ret)
    return ret

def rx_is_express_edition() -> bool:
    isExpress = rx_native_call(functionname = "IsExpressEdition", params = {})["isExpressEdition"]
    RxOptions.set_option("isExpress", isExpress)

    return isExpress

class RxParameter(object):

    def __init__(self, serialize_as):
        self._serialize_as = serialize_as

    def __call__(self, func):
        func.serialize = True
        func.serialize_as = self._serialize_as
        return func


class RxSerializable(ABC):
    """"""

    def serialize_bxl(self) -> dict:
        ret = {}
        for item in dir(self):
            fn = self.__getattribute__(item)
            if hasattr(fn, 'serialize'):
                val = fn()
                #print("{0}: {1} {2}".format(fn.serialize_as,
                                            #val,
                                            #'' if val is not None else '- Not Serializing due to None value'))
                try:
                    if issubclass(type(val), RxSerializable):
                        ret[fn.serialize_as] = val.serialize_bxl()
                        #print("{0} was serialized to {1}".format(fn.serialize_as,
                                                                 #ret[fn.serialize_as]))
                    else:
                        ret[fn.serialize_as] = val
                except TypeError:
                    ret[fn.serialize_as] = val

        return ret

