import os
import sys
import re
from pathlib import Path
import warnings
import logging

################################################
# Hadoop/Spark init logic
#
# DO NOT DELETE OR MOVE. THIS LOGIC HAS TO BE
# PLACED AT THE VERY BEGINNING OF THE FILE. AS
# IT WILL SET ENVIRONMENT VARS BEFORE STARTING
# BXLSERVER ENGINE.
################################################


def load_hadoop_env():
    script_directory = os.path.dirname(os.path.abspath(__file__))
    hadoop_comp_path = Path(script_directory).parent.parent.joinpath("common", "hadoop")
    hadoop_comp_path_alt = Path(os.path.realpath("/usr/bin/mrs-hadoop")).parent

    if hadoop_comp_path.is_dir():
        pass
    elif Path("/usr/bin/mrs-hadoop").exists() and hadoop_comp_path_alt.joinpath("workflow").exists():
        hadoop_comp_path = hadoop_comp_path_alt
    else:
        logging.debug("Hadoop components not found in " + str(hadoop_comp_path) + " or " + str(hadoop_comp_path_alt) + " Hadoop not configured")
        return False

    if not check_mrs_dependencies(hadoop_comp_path):
        return False
    os.environ["MRS_HADOOP_COMPONENTS"] = str(hadoop_comp_path)

    hadoopEnvFile = os.path.join(str(hadoop_comp_path), 'RevoHadoopEnvVars.site')  # TODO change this value
    if not os.path.isfile(hadoopEnvFile):
        return True

    set_env_from_envFile(hadoopEnvFile)

    return True


def check_mrs_dependencies(hadoop_comp_path: Path):

    files = ['utils/libhdfs.so', 'mrs-hadoop', 'mrs-spark-submit']

    # if we are running in Cloudera Parcel, Revo64/Revoscript will make
    # sure libjvm.so is accessible, otherwise they quit; this is because
    # a symlink to libjvm.so cannot be created inside a parcel's directory
    # TODO we should not check for a specific platform here, this should
    #      be resolved much earlier in the execution stack, possibly at
    #      installation time or by placing this file in $HOME and sourcing
    #      it in Revo64/Revoscript
    files = [os.path.join(str(hadoop_comp_path), x) for x in files]
    notexist_list = []
    for file in files:
        if not os.path.exists(file):
            notexist_list.append(file)

    if len(notexist_list) > 0:
        warnings.warn('Hadoop dependencies not found: ' + ' '.join(notexist_list))
        return False

    return True


def set_env_from_envFile(hadoopEnvFile: str):
    envFile = open(hadoopEnvFile, 'r')
    for line in envFile:
        name, value = parse_env_var(line)
        if len(name) == 0 or len(value) == 0:
            warnings.warn('env var definition malformed in RevoHadoopEnvVars.site: ', line.strip())
        if name not in os.environ or len(os.environ[name]) == 0:
            os.environ[name] = value
    envFile.close()


def parse_env_var(line: str):
    line = re.sub('^export', '', line.strip(), 1)
    pair = line.strip().split("=", 1)
    name = ""
    value = ""
    if len(pair) == 2:
        name = pair[0].strip()
        value = pair[1].strip()
    return name, value

load_hadoop_env()

################################################
# End of Hadoop/Spark init logic
################################################

################################################
# RevoScalePy's public RX classes and functions 
################################################

from .RxSerializable import RxMissingValues

from .computecontext.RxComputeContext import RxComputeContext
from .computecontext.RxComputeContext import rx_get_compute_context
from .computecontext.RxComputeContext import rx_set_compute_context
from .computecontext.RxComputeContext import rx_set_temp_compute_context
from .computecontext.RxInSqlServer import RxInSqlServer
from .computecontext.RxLocalSeq import RxLocalSeq
from .computecontext.RxLocalParallel import RxLocalParallel
from .computecontext.RxRemoteComputeContext import RxRemoteComputeContext
from .computecontext.RxLocalComputeContext import RxLocalComputeContext
from .computecontext.RxRemoteJob import RxRemoteJob
from .computecontext.RxSpark import RxSpark
from .computecontext.RxSparkApi import rx_spark_connect
from .computecontext.RxSparkApi import rx_spark_disconnect
from .computecontext.RxSparkApi import rx_get_pyspark_connection
from .computecontext.RxJob import rx_get_job_status
from .computecontext.RxJob import rx_cancel_job
from .computecontext.RxJob import rx_get_job_output
from .computecontext.RxJob import rx_cleanup_jobs
from .computecontext.RxJob import rx_get_job_info
from .computecontext.RxJob import rx_get_job_results
from .computecontext.RxJob import rx_get_jobs
from .computecontext.RxJob import rx_wait_for_job
from .computecontext.RxJob import RxRemoteJobStatus

from .datasource.RxDataSource import RxDataSource
from .datasource.RxFileSystem import RxFileSystem
from .datasource.RxFileSystem import RxNativeFileSystem
from .datasource.RxFileSystem import RxHdfsFileSystem
from .datasource.RxOdbcData import RxOdbcData
from .datasource.RxSqlData import RxSqlServerData
from .datasource.RxTextData import RxTextData
from .datasource.RxXdfData import RxXdfData
from .datasource.RxSparkData import RxSparkData
from .datasource.RxHiveData import RxHiveData
from .datasource.RxParquetData import RxParquetData
from .datasource.RxOrcData import RxOrcData
from .datasource.RxSparkDataFrame import RxSparkDataFrame
from .datasource.RxSparkData import rx_spark_remove_data
from .datasource.RxSparkData import rx_spark_list_data
from .datasource.RxSparkData import rx_spark_cache_data

from .etl.RxDataStep import rx_data_step
from .etl.RxImport import rx_import

from .functions.RxCreateColInfo import rx_create_col_info
from .functions.RxExec import rx_exec
from .functions.RxExecBy import rx_exec_by
from .functions.RxDeleteObject import rx_delete_object
from .functions.RxDTree import rx_btrees
from .functions.RxDTree import rx_dtree, RxDTreeControl
from .functions.RxDTree import rx_dforest
from .functions.RxGetInfoXdf import rx_get_info
from .functions.RxGetInfoXdf import rx_get_var_info
from .functions.RxGetInfoXdf import rx_get_var_names
from .functions.RxGetInfoXdf import rx_set_var_info
from .functions.RxLinMod import rx_lin_mod
from .functions.RxListKeys import rx_list_keys
from .functions.RxLogit import rx_logit
from .functions.RxPartition import rx_get_partitions
from .functions.RxPartition import rx_partition
from .functions.RxPredict import rx_predict
from .functions.RxPredict import rx_predict_default
from .functions.RxPredict import rx_predict_rx_dtree
from .functions.RxPredict import rx_predict_rx_dforest
from .functions.RxReadObject import rx_read_object
from .functions.RxReadXdf import rx_read_xdf
from .functions.RxSerializeModel import rx_serialize_model
from .functions.RxSummary import rx_summary
from .functions.RxWriteObject import rx_write_object
from .functions.RxHadoopUtils import rx_hadoop_command
from .functions.RxHadoopUtils import rx_hadoop_file_exists
from .functions.RxHadoopUtils import rx_hadoop_list_files
from .functions.RxHadoopUtils import rx_hadoop_copy_from_local
from .functions.RxHadoopUtils import rx_hadoop_copy_to_local
from .functions.RxHadoopUtils import rx_hadoop_make_dir
from .functions.RxHadoopUtils import rx_hadoop_copy
from .functions.RxHadoopUtils import rx_hadoop_move
from .functions.RxHadoopUtils import rx_hadoop_remove
from .functions.RxHadoopUtils import rx_hadoop_remove_dir

from .utils.RxOptions import RxOptions
from .utils.RxTelemetryLogger import rx_privacy_control
######################################################
# End of RevoScalePy's public RX classes and functions
######################################################

import sys

sys.path.append(os.path.join(os.getcwd(), "rxLibs"))

del sys

RxComputeContext.set_current(RxLocalSeq())
discard = RxOptions.instance()
del discard

# initialize telemetry
from . utils import RxTelemetryLogger
RxTelemetryLogger.__init()
del RxTelemetryLogger

# initialize runtime trace
from . utils.RxTrace import init_runtime_trace
init_runtime_trace()
del init_runtime_trace

del rxLibs

__version__ = '9.2.1'

