#
# Copyright (c) Microsoft Corporation.  All rights reserved.
#
import uuid
from abc import ABC
import platform
import datetime
import dateutil
from typing import Callable
from ..utils.RxOptions import RxOptions
from threading import Lock
from pandas import DataFrame
import warnings
class RxComputeContext(ABC):
    """
    .. rxtitle::

        revoscalepy Compute Contexts: Class Generator

    .. rxdescription::

        Main generator for RxComputeContext classes.

    :return: This is a wrapper to specific class generator functions for the
        revoscalepy compute context classes. For example, the RxInSqlServer class
        uses function RxInSqlServer as a generator. Therefore either
        RxInSqlServer(...) or RxComputeContext("RxInSqlServer", ...) will create an
        RxInSqlServer instance.

    .. rxkeywords:: context

    .. rxseealso::

        :py:func:`RxLocalSeq <revoscalepy.RxLocalSeq>`,
        :py:func:`RxInSqlServer <revoscalepy.RxInSqlServer>`,
        :py:func:`rx_get_compute_context <revoscalepy.rx_get_compute_context>`,
        :py:func:`rx_set_compute_context <revoscalepy.rx_set_compute_context>`.

    .. rxexample::
        :execute:
        :print_output:

        from revoscalepy import RxComputeContext, rx_get_compute_context, rx_set_compute_context

        sql_server_cc = RxComputeContext("RxInSqlServer", "1.0")
        rx_set_compute_context(sql_server_cc)
        rx_get_compute_context()
    """

    _lock = Lock()
    _context_lookup = dict()
    _current = None

    @property
    def description(self) -> str:
        return self._description

    @property
    def version(self) -> str:
        return self._version

    @property
    def id(self) -> str:
        return self._id

    @property
    def console_output(self) -> bool:
        return self._console_output

    @console_output.setter
    def console_output(self, value: bool):
        self._console_output = value

    @property
    def auto_cleanup(self) -> bool:
        return self._auto_cleanup

    @auto_cleanup.setter
    def auto_cleanup(self, value: bool):
        self._auto_cleanup = value

    @staticmethod
    def current() -> 'RxComputeContext':
        with RxComputeContext._lock:
            if hasattr(RxComputeContext.current, "current"):
                return RxComputeContext.current.current
            else:
                return None

    @staticmethod
    def set_current(compute_context : 'RxComputeContext')-> 'RxComputeContext':
        if not isinstance(compute_context, RxComputeContext):
            raise ValueError("Invalid compute context is not derived from RxComputeContext")
        if type(compute_context).__name__ == "RxSpark":
            attached_spark_cc = RxOptions.get_option("attachedSparkCC")
            if attached_spark_cc is None and compute_context.persistent_run:
                RxOptions.set_option("attachedSparkCC", compute_context)
        previouscc = None
        with RxComputeContext._lock:
            if hasattr(RxComputeContext.current, "current"):
                 previouscc = RxComputeContext.current.current
            RxComputeContext.current.current = compute_context
        return previouscc

    @staticmethod
    def lookup(to_lookup: str) -> "RxComputeContext":
        """
        Looks up a ``RxComputeContext`` instance by the specified id

        :param to_lookup: The id for which to retrieve the compute context
        :return: The retrieved compute context
        :raises: ValueError
        """
        if not to_lookup:
            raise ValueError("Id must be specified")

        with RxComputeContext._lock:
            if to_lookup not in RxComputeContext._context_lookup:
                return None

            return RxComputeContext._context_lookup[to_lookup]


    # validate input data source
    def validate_in_data(self, in_data, param_name):
        # import these modules inside the function to avoid circular dependency
        from .. RxFunction import RxFunction
        from .. datasource.RxDataSource import RxDataSource
        from .. RxFunction import validate_argument

        in_data = validate_argument(in_data, param_name, (DataFrame, str, RxDataSource))

        from ..utils.RxFileUtils import file_exists
        if isinstance(in_data, str) and not file_exists(in_data):
            raise ValueError("The path of input data not found {:s}".format(in_data))
        return in_data


    # validate output data source
    def validate_out_data(self, out_data, param_name):
        if out_data is None:
            return None

        from .. RxFunction import RxFunction
        from .. datasource.RxDataSource import RxDataSource
        from .. RxFunction import validate_argument
        out_data = validate_argument(out_data, param_name, (DataFrame, str, RxDataSource))
        return out_data


    def start_engine(self):
        print("{0} compute context is not supported for start_engine".format(type(self).__name__))

    def stop_engine(self):
        print("{0} compute context is not supported for stop_engine".format(type(self).__name__))

    def __init__(self, description: str, version: str, id: str = None):
        self._description = description
        self._version = version
        self._remoteId = -1
        self._console_output = False
        self._auto_cleanup = True

        if id:
            self._id = id
        else:
            self._id = uuid.uuid1().hex

        with RxComputeContext._lock:
            RxComputeContext._context_lookup[self.id] = self

    def get_remote_id(self) -> int:
        return self._remoteId

    def set_remote_id(self, remoteId : int):
        self._remoteId = remoteId

    def is_master_node(self) -> bool:
        return self.get_remote_id() < 1

    def get_node_name(self) -> str:
        return platform.node()

    def format_datetime(self, input : datetime.datetime = None) -> str:
        if input is None:
            input = datetime.datetime.now(dateutil.tz.tzlocal())

        return input.strftime("%Y-%m-%d %H:%M:%S %Z")

    # @abstractmethod
    def serialize_bxl(self) -> dict:
        pass

    def _execute(self, py_module_name, py_function_name, py_args):
        pass

    def execute_tasks(self, function: Callable, args_for_tasks: list, times_to_run: int, task_chunk_size: int,
                      callback: Callable, local_state: dict, continue_on_failure: bool, **kwargs) -> list:
        pass


def rx_get_compute_context() -> 'RxComputeContext':
    """
    .. rxtitle::

        Get and Set the compute context

    .. rxdescription::

        Get or set the active compute context for revoscalepy computations

        rx_set_compute_context(compute_context: 'RxComputeContext')
        rx_get_compute_context()

    :param compute_context: Character string specifying class name or description
        of the specific class to instantiate, or an existing RxComputeContext object.
        Choices include: "RxLocalSeq" or "local", "RxInSqlServer".

    :return: rx_set_compute_context returns the previously active compute context
        invisibly. rx_get_compute_context returns the active compute context.

    .. rxkeywords:: context

    .. rxseealso::

        :py:func:`RxComputeContext <revoscalepy.RxComputeContext>`,
        :py:func:`RxLocalSeq <revoscalepy.RxLocalSeq>`,
        :py:func:`RxInSqlServer <revoscalepy.RxInSqlServer>`,
        :py:func:`rx_set_compute_context <revoscalepy.rx_set_compute_context>`.

    .. rxexample::
        from revoscalepy import RxLocalSeq, RxInSqlServer, rx_get_compute_context, rx_set_compute_context

        local_cc = RxLocalSeq()
        sql_server_cc = RxInSqlServer('Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;')
        previous_cc = rx_set_compute_context(sql_server_cc)
        rx_get_compute_context()

    """
    return RxComputeContext.current()

def rx_set_compute_context(compute_context: 'RxComputeContext') -> 'RxComputeContext':
    """
    .. rxtitle::

        Get and Set the compute context

    .. rxdescription::

        Get or set the active compute context for revoscalepy computations

        rx_set_compute_context(compute_context: 'RxComputeContext')
        rx_get_compute_context()

    :param compute_context: Character string specifying class name or description
        of the specific class to instantiate, or an existing RxComputeContext object.
        Choices include: "RxLocalSeq" or "local", "RxInSqlServer".

    :return: rx_set_compute_context returns the previously active compute context
        invisibly. rx_get_compute_context returns the active compute context.

    .. rxkeywords:: context

    .. rxseealso::

        :py:func:`RxComputeContext <revoscalepy.RxComputeContext>`,
        :py:func:`RxLocalSeq <revoscalepy.RxLocalSeq>`,
        :py:func:`RxInSqlServer <revoscalepy.RxInSqlServer>`,
        :py:func:`rx_get_compute_context <revoscalepy.rx_get_compute_context>`.

    .. rxexample::
        from revoscalepy import RxLocalSeq, RxInSqlServer, rx_get_compute_context, rx_set_compute_context

        local_cc = RxLocalSeq()
        sql_server_cc = RxInSqlServer('Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;')
        previous_cc = rx_set_compute_context(sql_server_cc)
        rx_get_compute_context()

    """
    return RxComputeContext.set_current(compute_context = compute_context)


class _RxTemporaryContext:

    def __init__(self, compute_context: 'RxComputeContext'):
        self._previous = rx_set_compute_context(compute_context)

    # support for the context management protocol
    def __enter__(self) -> 'RxComputeContext':
        return self._previous

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        rx_set_compute_context(self._previous)


def rx_set_temp_compute_context(compute_context: 'RxComputeContext') -> 'RxComputeContext':
    """
    .. rxtitle::

       Set the compute context temporarily

    .. rxdescription::

        Set the active compute context for revoscalepy computations temporarily

        rx_set_temp_compute_context(compute_context: 'RxComputeContext')

    :param compute_context: Character string specifying class name or description
        of the specific class to instantiate, or an existing RxComputeContext object.
        Choices include: "RxLocalSeq" or "local", "RxLocalParallel" or "localpar",
        "RxInSqlServer", and "RxSpark".

    :return: rx_set_temp_compute_context returns the previously active compute context
        invisibly.

    .. rxkeywords:: context

    .. rxseealso::

        :py:func:`RxComputeContext <revoscalepy.RxComputeContext>`,
        :py:func:`RxLocalSeq <revoscalepy.RxLocalSeq>`,
        :py:func:`RxInSqlServer <revoscalepy.RxInSqlServer>`,
        :py:func:`rx_get_compute_context <revoscalepy.rx_get_compute_context>`.
        :py:func:`rx_set_compute_context <revoscalepy.rx_get_compute_context>`.

    .. rxexample::

        import pytest
        from revoscalepy import RxLocalSeq, RxInSqlServer
        from revoscalepy import rx_get_compute_context, rx_set_compute_context
        from revoscalepy import rx_set_temp_compute_context

        local_cc = RxLocalSeq()
        connection_string = "Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;"
        sql_server_cc = RxInSqlServer(connection_string)

        ## from RxInSqlServer to RxLocalSeq
        rx_set_compute_context(sql_server_cc)
        with rx_set_temp_compute_context(local_cc) as previous_cc:
            assert previous_cc == sql_server_cc
            assert rx_get_compute_context() == local_cc
        assert rx_get_compute_context() == sql_server_cc

        ## from RxLocalSeq to RxInSqlServer
        rx_set_compute_context(local_cc)
        with rx_set_temp_compute_context(sql_server_cc) as previous_cc:
            assert previous_cc == local_cc
            assert rx_get_compute_context() == sql_server_cc
        assert rx_get_compute_context() == local_cc

        ## from RxInSqlServer to RxLocalSeq with exception
        rx_set_compute_context(sql_server_cc)
        with pytest.raises(TypeError), rx_set_temp_compute_context(local_cc) as previous_cc:
            assert previous_cc == sql_server_cc
            assert rx_get_compute_context() == local_cc
            raise TypeError
        assert rx_get_compute_context() == sql_server_cc

        ## from RxLocalSeq to RxInSqlServer with exception
        rx_set_compute_context(local_cc)
        with pytest.raises(TypeError), rx_set_temp_compute_context(sql_server_cc) as previous_cc:
            assert previous_cc == local_cc
            assert rx_get_compute_context() == sql_server_cc
            raise TypeError
        assert rx_get_compute_context() == local_cc

    """
    return _RxTemporaryContext(compute_context)
