#
# Copyright (c) Microsoft Corporation.  All rights reserved.
#

import argparse
import collections
import datetime
import getpass
import os
import sys
import time
import traceback
import types
import dill

from . RxJob import RxRemoteJobStatus
from .. utils.RxOptions import RxOptions
from revoscalepy.datasource import RxDataSource, RxOdbcData
from .. utils.RxFileUtils import rx_zip
from .. utils.RxFileUtils import rx_unzip
from .. RxSerializable import rx_native_call
from . RxComputeContext import RxComputeContext
from . RxRemoteComputeContext import RxRemoteComputeContext
from . RxRemoteJob import RxRemoteJob
from typing import Callable


#################################################################################
#
# captures current state into a file and returns
#
def _rx_pauseexecution(
        function : Callable,
        args: dict,
        local_state: dict,
        state_file: str
):
    return _rx_savestate(
        filename=state_file,
        state={"function": function, "args": args, "local_state": local_state}
    )


#################################################################################
#
# prunes remote dictionary
#
def _rx_prunedictionary(input: dict):
    def pruneitem(key: str, item: object):
        if key.startswith("_"):
            return True
        if isinstance(item, types.ModuleType):
            return True

        return False

    pruned = {}
    for key in input:
        if not pruneitem(key, input[key]):
            pruned[key] = input[key]

    return pruned


#################################################################################
#
# saves current state
#
def _rx_savestate(filename: str, state):
    if isinstance(state, dict) or isinstance(state, collections.OrderedDict):
        state = _rx_prunedictionary(input=state)

    with open(filename, "wb") as state_file:
        dill.dump(state, state_file, -1)

    return filename


#################################################################################
#
# loads state from given file
#
def _rx_loadstate(filename: str) -> dict:
    with open(filename, "rb") as state_file:
        return dill.load(state_file)


#################################################################################
#
# patches the remote connection strings to use . for loop back connections
# for SQL server machines where public and private IPs for SQL server machine 
# are different
#
def _rx_fix_sql_connection_string(state: dict) -> dict:
    if "args" in state and "params" in state["args"]:
        params = state["args"]["params"]
        for key in params:
            if isinstance(params[key], dict) and 'class' in params[key] and 'connectionString' in params[key]:
                class_value = params[key]['class']
                if class_value == 'RxSqlServerData' or class_value == 'RxOdbcData':
                    params[key]['connectionString'] = RxOdbcData.fix_connection_string(params[key]['connectionString'])
    
    return state


#################################################################################
#
# resumes execution from saved state
#
def rx_resumeexecution(state_file : str):
    _state = _rx_loadstate(state_file)
    locals().update(_state["local_state"])

    #
    # patch connection strings if any
    #    
    _state = _rx_fix_sql_connection_string(_state)

    #
    # make the requested call
    #
    return _state["function"](**_state["args"])


#################################################################################
#
# SQL server remote job
#
class _SqlServerRemoteJob(RxRemoteJob):
    def __init__(self, compute_context: "RxInSqlServer", job_id: str = None):
        super(_SqlServerRemoteJob, self).__init__(compute_context, job_id)

        self._input_data_query = None
        self._use_sql_satellite = False

    def get_type(self) -> str:
        return "SqlServerJob"

    def get_input_data_query(self) -> str:
        return self._input_data_query

    def set_input_data_query(self, query : str):
        self._input_data_query = query

    def get_use_sql_satellite(self) -> bool:
        return self._use_sql_satellite

    def set_use_sql_satellite(self, use : bool):
        self._use_sql_satellite = use

    def serialize_bxl(self) -> dict:
        return {
            "inputZipFile" : self.input_zip_file,
            "jobGuid" : self.id,
            "userName" : getpass.getuser(),
            "inputDataQuery" : self.get_input_data_query(),
            "useSqlSatellite" : self.get_use_sql_satellite(),
            "inSqlServer" : self.get_use_sql_satellite()
            }


#################################################################################
#
# SQL server compute context
#
class RxInSqlServer(RxRemoteComputeContext):
    """
    .. rxtitle::

        Generate SQL Server In-Database Compute Context

    .. rxdescription::

        Creates a compute context for running revoscalepy analyses inside Microsoft SQL Server.
        Currently only supported in Windows.

    :param connection_string: An ODBC connection string used to connect to the
        Microsoft SQL Server database.
    :param num_tasks: Number of tasks (processes) to run for each computation.
        This is the maximum number of tasks that will be used; SQL Server may start
        fewer processes if there is not enough data, if too many resources are
        already being used by other jobs, or if num_tasks exceeds the MAXDOP
        (maximum degree of parallelism) configuration option in SQL Server. Each of
        the tasks is given data in parallel, and does computations in parallel, and
        so computation time may decrease as num_tasks increases. However, that may
        not always be the case, and computation time may even increase if too many
        tasks are competing for machine resources. Note that
        RxOptions.set_option("NumCoresToUse", n) controls how many cores (actually, threads) are
        used in parallel within each process, and there is a trade-off between
        NumCoresToUse and NumTasks that depends upon the specific algorithm, the
        type of data, the hardware, and the other jobs that are running.
    :param wait: Bool value, if True, the job will be blocking and will not
        return until it has completed or has failed. If False, the job will be
        non-blocking and return immediately, allowing you to continue running other
        Python code. The client connection with SQL Server must be maintained while
        the job is running, even in non-blocking mode.
    :param console_output: Bool value, if True, causes the standard output
        of the Python process started by SQL Server to be printed to the user console.
        This value may be overwritten by passing a non-None bool value to the
        consoleOutput argument provided in rx_exec and rx_get_job_results.
    :param auto_cleanup: Bool value, if True, the default behavior is to
        clean up the temporary computational artifacts and delete the result
        objects upon retrieval. If False, then the computational results are not
        deleted, and the results may be acquired using rx_get_job_results, and the
        output via rx_get_job_output until the rx_cleanup_jobs is used to delete the
        results and other artifacts. Leaving this flag set to False can result in
        accumulation of compute artifacts which you may eventually need to delete
        before they fill up your hard drive.
    :param execution_timeout_seconds: Integer value, defaults to 0 which means
        infinite wait.
    :param packages_to_load: Optional list of strings specifying additional
        packages to be loaded on the nodes when jobs are run in this compute context.

    .. rxkeywords:: sql

    .. rxseealso::

        :py:func:`RxComputeContext <revoscalepy.RxComputeContext>`,
        :py:func:`RxLocalSeq <revoscalepy.RxLocalSeq>`,
        :py:func:`rx_get_compute_context <revoscalepy.rx_get_compute_context>`,
        :py:func:`rx_set_compute_context <revoscalepy.rx_set_compute_context>`.

    .. rxexample::
        ## Not run:
        from revoscalepy import RxSqlServerData, RxInSqlServer, rx_lin_mod

        connection_string="Driver=SQL Server;Server=.;Database=RevoTestDB;Trusted_Connection=True"

        cc = RxInSqlServer(
            connection_string = connection_string,
            num_tasks = 1,
            auto_cleanup = False,
            console_output = True,
            execution_timeout_seconds = 0,
            wait = True
            )

        query="select top 100 [ArrDelay],[CRSDepTime],[DayOfWeek] FROM airlinedemosmall"
        data_source = RxSqlServerData(
            sql_query = "select top 100 * from airlinedemosmall",
            connection_string = connection_string,
            column_info = {
                "ArrDelay" : { "type" : "integer" },
                "DayOfWeek" : {
                    "type" : "factor",
                    "levels" : [ "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday" ]
                }
            })

        formula = "ArrDelay ~ CRSDepTime + DayOfWeek"
        lin_mod = rx_lin_mod(formula, data = data_source, compute_context = cc)
        print(lin_mod)
        ## End(Not run)
    """
    def __init__(
        self,
        connection_string : str,
        num_tasks : int = 1,
        auto_cleanup : bool = True,
        console_output : bool = None,
        execution_timeout_seconds : int = None,
        wait : bool = True,
        packages_to_load : list = None
        ) :
        if connection_string is None:
            raise ValueError("connection_string should be valid")

        super().__init__("sqlserver","1.0", wait = wait)

        self._connection_string = connection_string

        if num_tasks is None:
            num_tasks = RxOptions.get_option("NumTasks")

        self.num_tasks = num_tasks

        self.auto_cleanup = True if auto_cleanup else False
        self.console_output = True if console_output else False

        if execution_timeout_seconds is None:
            execution_timeout_seconds = 0

        self.execution_timeout_seconds = execution_timeout_seconds

        if packages_to_load is None:
            packages_to_load = []

        self.packages_to_load = packages_to_load
        self.job = None

    @property
    def connection_string(self):
        return self._connection_string

    def serialize_bxl(self) -> dict:
        result = self.__dict__.copy()
        result.update({ "class" : "RxInSqlServer" })

        return result

    def create_job(self, inData : RxDataSource = None) -> RxRemoteJob:
        self.job = _SqlServerRemoteJob(self)
        return self.get_job()

    def get_job(self) -> RxRemoteJob:
        return self.job

    def start_job(self, job: RxRemoteJob):

        if not isinstance(job, _SqlServerRemoteJob):
            raise ValueError("job should be SQL server job")

        rx_zip(job.input_dir, job.input_zip_file)

        traceLevel = RxOptions.get_option("traceLevel")
        traceEnabled = traceLevel > 0

        #
        # serialize the request
        #
        params = {
            "scriptExt" : ".py",
            "connectionString" : self._connection_string,
            "traceEnabled" : traceEnabled,
            "traceLevel" : traceLevel,
            "autoCleanup" : self.auto_cleanup,
            "numTasks" : self.num_tasks,
            "consoleOutput": self.console_output,
            "executionTimeoutSeconds" : self.execution_timeout_seconds,
            "wait" : self.wait
            }

        params.update(job.serialize_bxl())

        rx_native_call("StartSqlServerClusterJob", params)

    def poll_for_job_completion(self, job: RxRemoteJob):
        if not isinstance(job, _SqlServerRemoteJob):
            raise ValueError("job should be SQL server job")

        #
        # serialize the request
        #
        params = {
            "scriptExt": ".py",
            "connectionString": self._connection_string,
            "jobIdNum": job.id
            }

        rx_native_call("SqlServerPollForCompletion", params)

    def get_job_state(self, job : RxRemoteJob) -> str:
        if not isinstance(job, _SqlServerRemoteJob):
            raise ValueError("job should be SQL server job")

        #
        # serialize the request
        #
        params = {
            "scriptExt" : ".py",
            "connectionString" : self._connection_string,
            "jobIdNum" : job.id
            }

        return rx_native_call("GetSqlServerJobState", params)

    def write_final_job_result(self, job : RxRemoteJob):
        if not isinstance(job, _SqlServerRemoteJob):
            raise ValueError("job should be SQL server job")

        #
        # serialize the request
        #
        params = {
            "scriptExt" : ".py",
            "connectionString" : self._connection_string,
            "jobIdNum" : job.id,
            "writeDirectory" : job.dir
        }

        rx_native_call("SqlServerWriteFinalJobResults", params)

    def delete_job(self, job : RxRemoteJob):
        if not isinstance(job, _SqlServerRemoteJob):
            raise ValueError("job should be SQL server job")

        #
        # serialize the request
        #
        params = {
            "scriptExt" : ".py",
            "connectionString" : self._connection_string,
            "jobIdNum" : job.id
            }

        rx_native_call("SqlServerRemoveClusterData", params)

        # rx_exec calls get_job to check for existing job before creating.
        # if job is deleted and self.job is not set to none, get_job will return
        # invalid job object
        self.job = None

    # Separate function for testing purposes
    def _disable_mpi(self):
        self.num_tasks = 1

    def execute_tasks(self, function: Callable, args_for_tasks: list, times_to_run: int, task_chunk_size: int,
                      callback: Callable = None, **kwargs) -> list:

        local_state = kwargs.get("local_state", {})
        continue_on_failure = kwargs.get("continue_on_failure", True)

        old_num_tasks = self.num_tasks
        need_disable_mpi = old_num_tasks is not None and old_num_tasks > 1 and function != rx_native_call

        try:
            if need_disable_mpi: self._disable_mpi()
            wait = self.wait
            results = [self._execute_job(function, args_for_tasks[x], local_state,
                                    callback, wait, continue_on_failure) for x in range(times_to_run)]

            if wait:
                results = results if(callback is None or not callable(callback)) else callback(results)
            elif not wait and len(results) == 1:
                # This returns single job objects in case of single execution on non-waiting compute context.
                # This is what our documentation states will happen.
                results = results[0]

            return results

        finally:
            if need_disable_mpi: self.num_tasks = old_num_tasks


    def __str__(self) -> str:
        return  "RxInSqlServer Compute Context:\n" \
                "------------------------------\n" \
                "connection_string: \"{0}\"\n" \
                "num_tasks: {1}\n" \
                "execution_timeout_seconds: {2}\n" \
                "wait: {3}\n" \
                "console_output: {4}\n" \
                "auto_cleanup: {5}\n" \
                "packages_to_load: {6}\n" \
                "description: \"{7}\"\n" \
                "version: \"{8}\"".format(
                self._connection_string,
                self.num_tasks,
                self.execution_timeout_seconds,
                self.wait,
                self.console_output,
                self.auto_cleanup,
                self.packages_to_load,
                self.description,
                self.version)

    def supports_job(self, job: RxRemoteJob) -> bool:
        return job is not None and isinstance(job, _SqlServerRemoteJob)

    def get_job_result(self, job: RxRemoteJob) -> list:
        if not self.supports_job(job):
            raise ValueError("The specified job is invalid for this compute context")

        self.write_final_job_result(job)

        result = []

        if os.path.isfile(job.output_file):
            result.append(_rx_loadstate(filename=job.output_file))

        return result

    def cancel_job(self, job: RxRemoteJob) -> bool:
        if not self.supports_job(job):
            raise ValueError("Job should be SQL job")

        state = self.get_job_state(job)
        status = RxRemoteJobStatus.from_string(state)


        if status == RxRemoteJobStatus.RUNNING or RxRemoteJobStatus.QUEUED:
            params = {
                "jobIdNum": job.id,
            }

            rx_native_call("CancelSqlServerClusterJob", params)

    def get_job_output(self, job: RxRemoteJob):
        if not self.supports_job(job):
            raise ValueError("Job should be a SQL job")

        try:
            self.write_final_job_result(job)
        except Exception as e:
            raise RuntimeError("Could not synchronize the results for the job {} with {}".format(job, e))

        output = os.path.join(job.output_dir, "MergedOut.txt")

        if not os.path.isfile(output):
            output = os.path.join(job.output_dir, "$OutputBackup$")

            if not os.path.isfile(output):
                return ""

        with open(output, "r") as output_file:
            return output_file.read()

    def _execute_job(self, function, parameters, local_state, callback, wait, continue_on_failure):
        '''
         Executes a function using jobs in a RemoteComputeContext. If it is a waiting compute context,
         returns the results of the function. If not, returns the job.
        '''

        job = self.get_job()

        if(job is None):
            job = self.create_job()

        _rx_pauseexecution(
            function = function,
            args = parameters,
            local_state = local_state,
            state_file = job.input_file
        )

        self.start_job(job)
        if not wait:
            job.callback = callback
            job.serialize_job()
            return job

        # Wait for the job results (always available in RxLocalSeq)
        self.poll_for_job_completion(job)

        result = self.get_job_result(job)

        # Remote compute contexts return results as a single item list
        if isinstance(result, list) and len(result) == 0:
            print(self)
            output = self.get_job_output(job)
            print(output)
            result.append(output)
        if isinstance(result, list) and len(result) == 1:
            result = result[0]

        self.delete_job(job)

        return result


#################################################################################
#
# sql session start
#
def rx_sql_satellite_call(
    sessionId : str,
    taskId : int,
    port : int,
    logPath : str,
    pipeName : str,
    sqlSatellitePath : str,
    errorToStdError : bool
    ):
    #
    # serialize the request
    #
    params = {}

    params["sessionId"] = sessionId
    params["taskId"] = taskId
    params["port"] = port
    params["logPath"] = logPath
    params["pipeName"] = pipeName
    params["sqlSatellitePath"] = sqlSatellitePath
    params["errorToStdError"] = errorToStdError
    params["scriptExt"] = ".py"

    #
    # make the C++ call
    #
    try:
        rx_native_call("SqlSatelliteCall", params)
    except:
        #
        # catch and write the internal error to stdout
        # so that actual errors are correctly reported
        # through sqlsatellite abort as query failure message
        #
        sys.stdout.write(traceback.format_exc())


#################################################################################
#
# remote session start
#
def rx_sql_session_start(
    sessionDirectory : str,
    sessionId : str,
    waitTime : int = None
    ):
    if (waitTime is None):
        waitTime = -1

    #
    # serialize the request
    #
    params = {}

    params["sessionDirectory"] = sessionDirectory
    params["sessionId"] = sessionId
    params["waitTime"] = waitTime
    params["scriptExt"] = ".py"

    #
    # make the C++ call
    #
    rx_native_call("SqlSessionStart", params)

#################################################################################
#
# process pool session call
#
def rx_sql_satellite_pool_call(
    sessionDirectory : str,
    sessionId : str,
    waitTime : int = None
    ):
    #
    # wait for the session to start
    #
    rx_sql_session_start(sessionDirectory = sessionDirectory, sessionId = sessionId, waitTime = waitTime)

    #
    # read and execute the script file
    #
    scriptfile = os.path.join(sessionDirectory, sessionId + ".py")
    with open(scriptfile) as inputfile:
        exec(inputfile.read())


#################################################################################
#
# parses command line arguments
#
def rx_parse_command_arguments() -> dict:
    parser = argparse.ArgumentParser()

    parser.add_argument("-rxCommType")
    parser.add_argument("-rxComputeWhere")
    parser.add_argument("-rxNetworkShareJobDirectory")
    parser.add_argument("-rxClusterTarget")
    parser.add_argument("-sessionId")
    parser.add_argument("-port")
    parser.add_argument("-sqlSatellitePath")
    parser.add_argument("-pipeName")
    parser.add_argument("-rxTraceLevel")

    RxOptions.set_option("rxCommArgs", vars(parser.parse_args()))

    return RxOptions.get_option("rxCommArgs")

#################################################################################
#
# sets up mpi execution
#
def rx_mpi_config(args : dict) -> dict:
    if args is None:
        args = rx_parse_command_arguments()

    rxIsMpi = False
    rxClusterType = None

    rxCommType = "Unknown"
    if "rxCommType" in args:
        rxCommType = args["rxCommType"]

    rxNetworkShareJobDirectory = None
    if "rxNetworkShareJobDirectory" in args:
        rxNetworkShareJobDirectory = args["rxNetworkShareJobDirectory"]

    if rxNetworkShareJobDirectory is None:
        return

    zipFile = rxNetworkShareJobDirectory + ".zip"

    if not os.path.isfile(zipFile):
        raise ValueError("{:s} not found".format(zipFile))

    compute_context = RxComputeContext.current()

    if rxCommType == "MPI":
        rxIsMpi = True

        #
        # initialize MPI
        #
        rx_native_call("MpiConfig", { "startMPI" : "default" })

        #
        # initialize remote id
        #
        compute_context.set_remote_id(int(rx_native_call("GetMpiRank", {})))

    rxClusterTarget = None
    if "rxClusterTarget" in args:
        rxClusterTarget = args["rxClusterTarget"]

    if rxClusterTarget != "sqlserver":
        raise ValueError("Invalid rxClusterTarget {:s} ".format(rxClusterTarget))

    unzipDoneFile = os.path.join(rxNetworkShareJobDirectory, "revo-unzip-done")
    #print(unzipDoneFile)

    if compute_context.is_master_node():
        #
        # unzip the file on master node
        #
        rx_unzip(inputfile = zipFile, outputdir = rxNetworkShareJobDirectory)
        with open(unzipDoneFile, 'w') as f:
            pass
    else:
        pollWaitTime = 0.1  # 100 msecs
        maxWaitTime = 60    # 60 secs

        startTime = datetime.datetime.now()

        while not os.path.isfile(unzipDoneFile):
            delta = datetime.datetime.now() - startTime
            if delta.total_seconds() > maxWaitTime:
                break

            time.sleep(pollWaitTime)

    inputfile = "input.dill"
    locations = [ os.path.join(rxNetworkShareJobDirectory, inputfile),
        os.path.join(rxNetworkShareJobDirectory, os.path.basename(rxNetworkShareJobDirectory), inputfile) ]

    inputfilelocation = None

    for inputfilename in locations:
        if os.path.isfile(inputfilename):
            inputfilelocation = inputfilename
            break;

    if inputfilelocation is None:
        raise ValueError("{:s} not found".format(locations[0]))

    args.update({"inputfile" : inputfilelocation})

    return args

#################################################################################
#
# executes a function call request recieved from remote client
#
def rx_remote_call(args : dict):
    #
    # initialize MPI configuration
    #
    args = rx_mpi_config(args = args)

    compute_context = RxComputeContext.current()

    print("====== {:s} ( process {:d} ) has started run at {:s} ======".format(
        compute_context.get_node_name(),
        compute_context.get_remote_id(),
        compute_context.format_datetime()
        ))

    rxClusterTarget = args["rxClusterTarget"]
    guidDir = os.path.dirname(args["rxNetworkShareJobDirectory"])

    # set master node
    RxOptions.set_option("nodeType", "master")

    #
    # enable tracing if required
    #
    if "rxTraceLevel" in args and args["rxTraceLevel"] is not None:
        RxOptions.set_option("traceLevel", int(args["rxTraceLevel"]))

    if RxOptions.get_option("traceLevel", 0) > 0:
        rx_native_call(functionname="Util", params={
            "USE": "InitializeLogging",
            "traceLevel": RxOptions.get_option("traceLevel"),
            "processType": "MASTER" if compute_context.is_master_node() else "WORKER",
            "logDir": guidDir
        })

    inputfile = args["inputfile"]

    #
    # wait for server instructions on worker nodes
    #
    if not compute_context.is_master_node():
        remote_params = _rx_loadstate(filename = inputfile)["args"]["params"]

        params = dict()
        params.update(remote_params)

        params.update(
            {
            'rxClusterType' : rxClusterTarget,
            'guidDir' : guidDir
            })

        rx_native_call(functionname = "ServerWait", params = params)

        return

    #
    # set sql server args
    #
    sqlServerArgs = collections.OrderedDict();
    for attribute in ["sessionId", "sqlSatellitePath", "pipeName"]:
        sqlServerArgs[attribute] = args[attribute]
    sqlServerArgs["logPath"] = ""
    sqlServerArgs["port"] = int(args["port"])
    rx_native_call(functionname = "Util", params = { "USE" : "SaveSqlServerArgs", "sqlServerArgs" : sqlServerArgs })

    #
    # initialize master process
    #
    initParams = {
        #'dataDistType' : 'split',
        'workingDir' : guidDir,
        'contextClass' : 'RxInSqlServer'
        }

    rx_native_call(functionname = "RemoteCallInit", params = initParams)

    inputfile = args["inputfile"]
    results = rx_resumeexecution(state_file = inputfile)

    outputfile = os.path.join(os.path.dirname(inputfile), "output.dill")
    _rx_savestate(filename = outputfile, state = results)

    #
    # shutdown the master process cleanly
    #
    rx_native_call(functionname = "JobDone", params = {})

    print("====== {:s} ( process {:d} ) has completed run at {:s} ======".format(
        compute_context.get_node_name(),
        compute_context.get_remote_id(),
        compute_context.format_datetime()
        ))


