import warnings

import datetime

from enum import Enum, unique
from typing import Union, Any

from .RxRemoteJob import RxRemoteJob
from .RxRemoteComputeContext import RxRemoteComputeContext
from revoscalepy.utils import on_exit_handler
from revoscalepy.utils import telemetry_capture_call


#################################################################################
#
# Remote job status
#
@unique
class RxRemoteJobStatus(Enum):
    """
    .. rxtitle::

        Represents the execution status of a remote job

    .. rxdescription::

        This enumeration represents the execution status of a remote Python job.  The enumeration is returned from the
        rx_get_job_status method and will indicate the current status of the job.  The enumeration has the following
        values indicating job status:

        `FINISHED` - The job has run to a successful completion

        `FAILED` - The job has failed

        `CANCELED` - The job was cancelled before it could run to successful completion by using `rx_cancel_job`

        `UNDETERMINED` - The job's status could not be determined, typically meaning it has been executed on the server
        and the job has already been cleaned up by calling `rx_get_job_results` or `rx_cleanup_jobs`

        `RUNNING` - The distributed computing job is currently executing on the server

        `QUEUED` - The distributed computing job is currently awaiting execution on the server

    .. rxseealso::

        :py:func:`rx_get_job_status <revoscalepy.computecontext.RxJob.rx_get_job_status>`

    .. rxkeywords:: remote,job

    .. rxexample::

        from revoscalepy import RxInSqlServer
        from revoscalepy import rx_exec
        from revoscalepy import RxRemoteJobStatus
        from revoscalepy import rx_get_job_status
        from revoscalepy import rx_wait_for_job

        connection_string = 'Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;'

        # Setting wait to False allows the job to be run asynchronously
        # Setting console_output to True allows us to get the console output of the distributed computing job
        compute_context = RxInSqlServer(connection_string=connection_string,
                                        num_tasks=1,
                                        console_output=True,
                                        wait=False)

        def hello_from_sql():
            import time
            print('Hello from SQL server')
            time.sleep(3)
            return 'We just ran Python code asynchronously on a SQL server!'

        job = rx_exec(function=hello_from_sql, compute_context=compute_context)

        # Poll initial status
        status = rx_get_job_status(job)

        print(status)

        if status == RxRemoteJobStatus.RUNNING or status == RxRemoteJobStatus.QUEUED :
        # Wait for the job to finish or fail whatever the case may be
        rx_wait_for_job(job)

        # Poll final status
        status = rx_get_job_status(job)

        print(status)

    """
    #
    # NOTE: make sure to change the example code in acceptance tests when changing the rxexample above
    #
    FINISHED = "finished"
    FAILED = "failed"
    CANCELED = "canceled"
    UNDETERMINED = "undetermined"
    RUNNING = "running"
    QUEUED = "queued"

    @staticmethod
    def from_string(status: str) -> "RxRemoteJobStatus":
        for member in RxRemoteJobStatus.__members__.values():
            if status == member.value:
                return member

        return RxRemoteJobStatus.UNDETERMINED

@on_exit_handler
def rx_get_job_info(job) -> RxRemoteJob:
    """
    .. rxtitle::

        Get Job Information from Distributed Computing Job

    .. rxdescription::

        The object returned from a non-waiting, distributed computation  contains job information together with other
        information.  The job information is used internally by such functions as `rx_get_job_status`,
        `rx_get_job_output`, and `rx_get_job_results`. It is sometimes useful to extract it for its own sake, as it
        contains complete information on the job's compute context as well as other information needed by the
        distributed computing resources.

        For most users, the principal use of this function is to determine whether a given object actually contains job
        information. If the return value is not `None`, then the object contains job information. Note, however, that
        the structure of the job information is subject to change, so code that attempts to manipulate it directly is
        not guaranteed to be forward-compatible.

    :param job: An object containing jobInfo information, such as that returned from a non-waiting, distributed
        computation.

    :return: The job information, if present, or None.

    .. rxseealso::
        :py:func:`rx_get_job_status <revoscalepy.computecontext.RxJob.rx_get_job_status>`

    .. rxkeywords:: context

    .. rxexample::

        from revoscalepy import RxInSqlServer
        from revoscalepy import rx_exec
        from revoscalepy import rx_get_job_info

        connection_string = 'Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;'

        # Setting wait to False allows the job to be run asynchronously
        # Setting console_output to True allows us to get the console output of the distributed computing job
        compute_context = RxInSqlServer(connection_string=connection_string,
                                        num_tasks=1,
                                        console_output=True,
                                        wait=False)

        def hello_from_sql():
            import time
            print('Hello from SQL server')
            time.sleep(3)
            return 'We just ran Python code asynchronously on a SQL server!'

        job = rx_exec(function=hello_from_sql, compute_context=compute_context)

        #Get the job information
        info = rx_get_job_info(job)
        print(info)

    """
    telemetry_capture_call()
    if isinstance(job, RxRemoteJob):
        return job
    elif isinstance(job, str) and job:
        return RxRemoteJob.deserialize_job(job)

    raise ValueError("Unsupported value for job")


@on_exit_handler
def rx_cancel_job(job_info: RxRemoteJob, console_output: bool = False) -> bool:
    """
    .. rxtitle::

        Causes Python to cancel an existing distributed computing job

    .. rxdescription::

        This function does not attempt to retrieve any output objects; if the output is desired the `console_output`
        flag can be used to display it.  This function does, however, remove all job-related artifacts from the
        distributed computing resources including any job results.

    :param job_info: A job to cancel
    :param console_output: If `None` use the console option that was present when the job was created, if `True`
        output any console output that was present when the the job was cancelled will be displayed.  If `False` then
        all console output will not be displayed.

    :return: `True` if the job is successfully cancelled; `False` otherwise.z

    .. rxseealso::

        :py:func:`rx_get_job_status <revoscalepy.computecontext.RxJob.rx_get_job_results>`

    .. rxkeywords:: cancel

    .. rxexample::

        import time
        from revoscalepy import RxInSqlServer
        from revoscalepy import RxSqlServerData
        from revoscalepy import rx_dforest
        from revoscalepy import rx_cancel_job
        from revoscalepy import RxRemoteJobStatus
        from revoscalepy import rx_get_job_status
        from revoscalepy import rx_cleanup_jobs

        connection_string = 'Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;'
        airline_data_source = RxSqlServerData(sql_query='select * from airlinedemosmall',
                                              connection_string=connection_string,
                                              column_info={
                                                  'ArrDelay': {'type': 'integer'},
                                                  'CRSDepTime': {'type': 'float'},
                                                  'DayOfWeek': {
                                                      'type': 'factor',
                                                      'levels': ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
                                                                 'Friday',
                                                                 'Saturday', 'Sunday']
                                                  }
                                              })

        # Setting wait to False allows the job to be run asynchronously
        compute_context = RxInSqlServer(connection_string=connection_string,
                                        num_tasks=1,
                                        auto_cleanup=False,
                                        wait=False)

        job = rx_dforest(formula='ArrDelay ~ DayOfWeek',
                         seed=1,
                         data=airline_data_source,
                         compute_context=compute_context)

        # Poll until status is RUNNING
        status = rx_get_job_status(job)
        while status == RxRemoteJobStatus.QUEUED:
            time.sleep(1)
            status = rx_get_job_status(job)

        # Only running or queued jobs can be canceled
        if status == RxRemoteJobStatus.RUNNING :
            # Cancel the job
            rx_cancel_job(job)

        # Only canceled or failed jobs can be cleaned up
        if status == RxRemoteJobStatus.CANCELED or status == RxRemoteJobStatus.FAILED :
            # Cleanup after the job
            rx_cleanup_jobs(job)
    """
    telemetry_capture_call()
    ret = False

    if job_info is None:
        raise ValueError("job_info must be specified")

    # Resolve the job if it is a string and perform some validation
    job_info = rx_get_job_info(job_info)

    context = _rx_resolve_context(job_info, True)

    status = rx_get_job_status(job_info)

    if status == RxRemoteJobStatus.RUNNING or status == RxRemoteJobStatus.QUEUED:
        context.cancel_job(job_info)

        ret = True
    else:
        raise RuntimeError("Cannot cancel the job {} with status {}".format(job_info, status.value))

    if console_output or (console_output is None and context.console_output):
        print(context.get_job_output(job_info))

    return ret


@on_exit_handler
def rx_get_job_status(job_info: RxRemoteJob) -> RxRemoteJobStatus:
    """
    .. rxtitle::

        Obtain Distributed Computing Job Status

    .. rxdescription::

        Obtain distributed computing processing status for the specified job.

        rx_get_job_status(job_info)

    :param job_info: A job object as returned by rx_exec or a revoscalepy
        analysis function, if available.

    :return:  A RxRemoteJobStatus enumeration value that designates the status of the remote job

    .. rxseealso::
        :py:func:`rx_get_job_results <revoscalepy.computecontext.RxJob.rx_get_job_results>`
        :py:func:`RxRemoteJobStatus <revoscalepy.computecontext.RxJob.RxRemoteJobStatus>`

    .. rxkeywords:: job, status

    .. rxexample::

        from revoscalepy import RxInSqlServer
        from revoscalepy import rx_exec
        from revoscalepy import rx_get_job_status
        from revoscalepy import rx_wait_for_job

        connection_string = 'Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;'

        # Setting wait to False allows the job to be run asynchronously
        # Setting console_output to True allows us to get the console output of the distributed computing job
        compute_context = RxInSqlServer(connection_string=connection_string,
                                        num_tasks=1,
                                        console_output=True,
                                        wait=False)

        def hello_from_sql():
            import time
            print('Hello from SQL server')
            time.sleep(3)
            return 'We just ran Python code asynchronously on a SQL server!'

        job = rx_exec(function=hello_from_sql, compute_context=compute_context)

        # Poll initial status
        status = rx_get_job_status(job)

        print(status)

        # Wait for the job to finish or fail whatever the case may be
        rx_wait_for_job(job)

        # Poll final status
        status = rx_get_job_status(job)

        print(status)
    """
    telemetry_capture_call()
    if not job_info:
        raise ValueError("job_info must be specified")

    # Resolve the job if it is a string and perform some validation
    job_info = rx_get_job_info(job_info)

    context = _rx_resolve_context(job_info, True)

    try:
        ret = context.get_job_state(job_info)
    except:
        ret = "undetermined"

    return RxRemoteJobStatus.from_string(ret)


@on_exit_handler
def rx_get_jobs(compute_context: RxRemoteComputeContext, exact_match: bool = False, start_time: datetime = None,
                end_time: datetime = None, states: list = None, verbose: bool = True) -> list:
    """
    .. rxtitle::

        Get Distributed Computing Jobs

    .. rxdescription::

        Returns a list of job objects associated with the given compute context and matching the specified parameters.

    :param compute_context: A compute context object.

    :param exact_match: Determines if jobs are matched using the full compute
        context, or a simpler subset. If True, only jobs which use the same context
        object are returned. If False, all jobs which have the same headNode (if
        available) and ShareDir are returned.

    :param start_time: A time, specified as a POSIXct object. If specified, only
        jobs created at or after start_time are returned.

    :param end_time: A time, specified as a POSIXct object. If specified, only
        jobs created at or before end_time are returned.

    :param states: If specified (as a list of strings of states that can include
        "none", "finished", "failed", "canceled", "undetermined" "queued"or
        "running"), only jobs in those states are returned. Otherwise, no filtering
        is performed on job state.

    :param verbose: If True (the default), a brief summary of each job is printed
        as it is found. This includes the current job status as returned by
        rx_get_job_status, the modification time of the job, and the current job ID
        (this is used as the component name in the returned list of job information
        objects). If no job status is returned, the job status shows none.

    :return: Returns a list of job information objects based on the compute context.

    .. rxseealso::

    .. rxkeywords:: get, job

    .. rxexample::

        from revoscalepy import RxInSqlServer
        from revoscalepy import rx_exec
        from revoscalepy import rx_get_jobs

        connection_string = 'Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;'

        # Setting wait to False allows the job to be run asynchronously
        # Setting console_output to True allows us to get the console output of the distributed computing job
        compute_context = RxInSqlServer(connection_string=connection_string,
                                        num_tasks=1,
                                        console_output=True,
                                        wait=False)

        def hello_from_sql():
            import time
            print('Hello from SQL server')
            time.sleep(3)
            return 'We just ran Python code asynchronously on a SQL server!'

        job = rx_exec(function=hello_from_sql, compute_context=compute_context)

        job_list = rx_get_jobs(compute_context)
        print(job_list)
    """
    telemetry_capture_call(compute_context)
    if compute_context is None:
        raise ValueError("The compute_context specified was None")

    candidates = [owned for owned in RxRemoteJob.deserialize_jobs() if owned.context_id == compute_context.id]

    for candidate in candidates:
        if (start_time is None or candidate.start_time >= start_time) and (end_time is None or candidate.end_time <= end_time):
            state = rx_get_job_status(candidate)

            if exact_match or state != "undetermined":
                if not states or state in states:
                    if verbose:
                        print("{}   {}   {}".format(state, candidate.start_time, candidate.id))
                    yield candidate


@on_exit_handler
def rx_cleanup_jobs(job_info_list: Union[RxRemoteJob, list], force: bool = False, verbose: bool = True):
    """
    Removes the artifacts for the specified jobs

    .. rxdescription::
        If `job_info_list` is a `RxRemoteJob` object, `rx_cleanup_jobs` attempts to remove the artifacts. However, if
        the job has successfully completed and `force` is `False`, `rx_cleanup_jobs` issues a warning saying to either
        set `force=True` or use `rx_get_job_results` to get the results and delete the artifacts.

        If `job_info_list` is a list of jobs, `rx_cleanup_jobs` attempts to apply the cleanup rules for a single job to
        each element in the list.

    :param job_info_list: The jobs for which to remove the artifacts, this can be a list of jobs or a single job

    :param force: True indicates the cleanup should happen regardless of whether or not the job status can be determined
        and job results have already been retrieved false indicates that the should be cleaned up by calling
        `rx_get_job_results` or the job should have completed unsuccessfully (such as by using rx_cancel_job).

    :param verbose: If `True`, will print the directories/records being deleted.

    :return:  This function does not return a value

    .. rxseealso::

    .. rxkeywords:: cleanup, job

    .. rxexample::

        import time
        from revoscalepy import RxInSqlServer
        from revoscalepy import RxSqlServerData
        from revoscalepy import rx_dforest
        from revoscalepy import rx_cancel_job
        from revoscalepy import RxRemoteJobStatus
        from revoscalepy import rx_get_job_status
        from revoscalepy import rx_cleanup_jobs

        connection_string = 'Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;'
        airline_data_source = RxSqlServerData(sql_query='select * from airlinedemosmall',
                                              connection_string=connection_string,
                                              column_info={
                                                  'ArrDelay': {'type': 'integer'},
                                                  'CRSDepTime': {'type': 'float'},
                                                  'DayOfWeek': {
                                                      'type': 'factor',
                                                      'levels': ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
                                                                 'Friday',
                                                                 'Saturday', 'Sunday']
                                                  }
                                              })

        # Setting wait to False allows the job to be run asynchronously
        compute_context = RxInSqlServer(connection_string=connection_string,
                                        num_tasks=1,
                                        auto_cleanup=False,
                                        wait=False)

        job = rx_dforest(formula='ArrDelay ~ DayOfWeek',
                         seed=1,
                         data=airline_data_source,
                         compute_context=compute_context)

        # Poll until status is RUNNING
        status = rx_get_job_status(job)
        while status == RxRemoteJobStatus.QUEUED:
            time.sleep(1)
            status = rx_get_job_status(job)

        if status == RxRemoteJobStatus.RUNNING :
            # Cancel the job
            rx_cancel_job(job)

        # Only canceled or failed jobs can be cleaned up
        if status == RxRemoteJobStatus.CANCELED or status == RxRemoteJobStatus.FAILED :
            # Cleanup after the job
            rx_cleanup_jobs(job)
    """
    telemetry_capture_call()
    if job_info_list is None:
        raise ValueError("Invalid value for job_info_list")

    errors = []
    index = -1

    if isinstance(job_info_list, RxRemoteJob) or isinstance(job_info_list, str):
        job_info_list = [rx_get_job_info(job_info_list)]

    for candidate in job_info_list:
        try:
            # Resolve the job if it is a string and perform some validation
            job_info = rx_get_job_info(candidate)
        except:
            job_info = None

        index = index + 1
        use_compute_context = False

        if job_info is None or not isinstance(job_info, RxRemoteJob):
            errors.append(ValueError("The job info list contains an invalid job at index {}".format(index)))
            continue

        context = _rx_resolve_context(job_info, False)

        if context is None or not context.supports_job(job_info):
            if not force:
                errors.append(ValueError(("The job info list has a job ({}) to cleanup at index {} but force "
                                          "was not specified and the job is not supported by the current compute "
                                          "context").format(job_info, index)))
                continue
        else:
            try:
                state = rx_get_job_status(job_info)
            except:
                state = RxRemoteJobStatus.UNDETERMINED

            if state == RxRemoteJobStatus.RUNNING:
                errors.append(ValueError("The job {} specified at index {} is currently running"
                                         .format(job_info, index)))
                continue
            if state == RxRemoteJobStatus.QUEUED:
                errors.append(ValueError("The job {} specified at index {} is currently queued"
                                         .format(job_info, index)))
                continue
            if state == RxRemoteJobStatus.FINISHED and not force:
                warnings.warn("Job completed successfully.  Use rx_get_job_results to retrieve the results and " +
                              "cleanup the job or set force to True.")
                continue
            if state == RxRemoteJobStatus.UNDETERMINED and not force:
                warnings.warn("Unable to retrieve job state from the scheduler.  Use rx_get_job_results to retrieve " +
                              "the results and cleanup the job or set force to True.")
                continue
            else:
                use_compute_context = context is not None

        if verbose:
            print("Removing job " + job_info.id)

        if use_compute_context:
            context.delete_job(job_info)

        job_info.close()

    if any(errors):
        job_status = "multiple jobs" if len(errors) > 1 else "a job"
        raise RuntimeError("Failed to clean up {}:  {}".format(job_status, "\n".join(errors)))


@on_exit_handler
def rx_wait_for_job(job_info: RxRemoteJob):
    """
    .. rxtitle::

        Wait for Distributed Job to Complete

    .. rxdescription::

        Causes Python to block on an existing distributed job until completion.  This effectively turns a non-blocking
        job into a blocking job.

    :param job_info: A jobInfo object.

    .. rxseealso::

    .. rxkeywords:: wait, job

    .. rxexample::

        from revoscalepy import RxInSqlServer
        from revoscalepy import RxSqlServerData
        from revoscalepy import rx_logit
        from revoscalepy import rx_wait_for_job
        from revoscalepy import rx_get_job_results

        connection_string = 'Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;'
        airline_data_source = RxSqlServerData(sql_query='select top 1000 * from airlinedemosmall',
                                              connection_string=connection_string,
                                              column_info={
                                                  'ArrDelay': {'type': 'integer'},
                                                  'CRSDepTime': {'type': 'float'},
                                                  'DayOfWeek': {
                                                      'type': 'factor',
                                                      'levels': ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
                                                                 'Friday',
                                                                 'Saturday', 'Sunday']
                                                  }
                                              })

        # Setting wait to False allows the job to be run asynchronously
        compute_context = RxInSqlServer(connection_string=connection_string, num_tasks=1, wait=False)

        def transform_late(dataset, context):
            dataset['Late'] = dataset['ArrDelay'] > 15
            return dataset

        # Since wait is set to False on the compute_context rx_logit will return a job rather than the model
        job = rx_logit(formula='Late ~ DayOfWeek',
                       data=airline_data_source,
                       compute_context=compute_context,
                       transform_function=transform_late,
                       transform_variables=['ArrDelay'])

        # Wait for the job to finish
        rx_wait_for_job(job)
        model = rx_get_job_results(job)
        print(model)

    """
    telemetry_capture_call()
    if job_info is None:
        raise ValueError("job_info must be specified")

    # Resolve the job if it is a string and perform some validation
    job_info = rx_get_job_info(job_info)

    context = _rx_resolve_context(job_info, True)

    context.poll_for_job_completion(job_info)


def _rx_resolve_context(job_info: RxRemoteJob, throw: bool) -> RxRemoteComputeContext:
    try:
        context = job_info.resolve_context()
    except:
        context = None

    if context is None and throw:
        raise RuntimeError("The context that is associated job {} is no longer available".format(job_info))

    return context


@on_exit_handler
def rx_get_job_output(job_info: RxRemoteJob) -> str:
    """
    .. rxtitle::

        Gets the console output from the various nodes in a non-waiting distributed computing job.

    .. rxdescription::

        During a job run, the state of the output is non-deterministic (that is, it may or may not be on disk, and what
        is on disk at any given point in time may not reflect the actual completion state of a job).

        If `auto_cleanup` has been set to `True` on the distributed computing job's compute context, the console output
        will not persist after the job completes.

        Unlike `rx_get_job_results`, this function does not remove any job information upon retrieval.

    :param job_info:  The distributed computing job for which to retrieve the job output

    :return: `str` that contains the console output for the nodes participating in the distributed computing job

    .. rxseealso::

        :py:func:`rx_get_job_status <revoscalepy.computecontext.RxJob.rx_get_job_status>`

    .. rxkeywords:: job, output

    .. rxexample::

        from revoscalepy import RxInSqlServer
        from revoscalepy import rx_exec
        from revoscalepy import rx_get_job_output
        from revoscalepy import rx_wait_for_job

        connection_string = 'Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;'

        # Setting wait to False allows the job to be run asynchronously
        # Setting console_output to True allows us to get the console output of the distributed computing job
        compute_context = RxInSqlServer(connection_string=connection_string,
                                        num_tasks=1,
                                        console_output=True,
                                        wait=False)

        def hello_from_sql():
            import time
            print('Hello from SQL server')
            time.sleep(3)
            return 'We just ran Python code asynchronously on a SQL server!'

        job = rx_exec(function=hello_from_sql, compute_context=compute_context)

        # Wait for the job to finish or fail whatever the case may be
        rx_wait_for_job(job)

        # Print out what our code printed on sql
        output = rx_get_job_output(job)
        print(output)

    """
    telemetry_capture_call()
    if job_info is None:
        raise ValueError("job_info must be specified")

    # Resolve the job if it is a string and perform some validation
    job_info = rx_get_job_info(job_info)

    context = _rx_resolve_context(job_info, True)

    return context.get_job_output(job_info)


@on_exit_handler
def rx_get_job_results(job_info: RxRemoteJob, console_output: bool = None, auto_cleanup: bool = None) -> list:
    """
    .. rxtitle::

        Obtain Distributed Computing Job Status and Results

    .. rxdescription::

        Obtain distributed computing results and processing status.

        rx_get_job_results(job_info, console_output, auto_cleanup)

    :param job_info: A job object as returned by rx_exec or a revoscalepy
        analysis function, if available.

    :param console_output: None or bool value. If True, the console output
        from all of the processes is printed to the user console. If False, no
        console output is displayed. Output can be retrieved with the function
        rxGetJobOutput for a non-waiting job. If not None, this flag overrides the
        value set in the compute context when the job was submitted. If None, the
        setting in the compute context will be used.

    :param auto_cleanup: None or bool value. If True, the default behavior is
        to clean up any artifacts created by the distributed computing job. If False,
        then the artifacts are not deleted, and the results may be acquired using
        rxGetJobResults, and the console output via rxGetJobOutput until the
        rxCleanupJobs is used to delete the artifacts. If not None, this flag
        overwrites the value set in the compute context when the job was submitted.
        If you routinely set auto_cleanup=False, you will eventually fill your hard
        disk with compute artifacts.If you set auto_cleanup=True and experience
        performance degradation on a Windows XP client, consider setting
        auto_cleanup=False.

    :return:
        Either the results of the run (prepended with console output if the console_output argument is set to True) or
        a message saying that the results are not available because the job has not finished, has failed, or was
        deleted.

    .. rxseealso::

        :py:func:`rx_get_job_status <revoscalepy.computecontext.RxJob.rx_get_job_status>`

    .. rxexample::

        from revoscalepy import RxInSqlServer
        from revoscalepy import rx_exec
        from revoscalepy import rx_wait_for_job
        from revoscalepy import rx_get_job_results

        connection_string = 'Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;'

        # Setting wait to False allows the job to be run asynchronously
        # Setting console_output to True allows us to get the console output of the distributed computing job
        compute_context = RxInSqlServer(connection_string=connection_string,
                                        num_tasks=1,
                                        console_output=True,
                                        wait=False)

        def hello_from_sql():
            import time
            print('Hello from SQL server')
            time.sleep(3)
            return 'We just ran Python code asynchronously on a SQL server!'

        job = rx_exec(function=hello_from_sql, compute_context=compute_context)

        # Wait for the job to finish
        rx_wait_for_job(job)

        # Print out what we returned from SQL
        results = rx_get_job_results(job)
        print(results)

    """
    telemetry_capture_call()
    if job_info is None:
        raise ValueError("job_info must be specified")

    # Resolve the job if it is a string and perform some validation
    job_info = rx_get_job_info(job_info)

    context = _rx_resolve_context(job_info, True)

    try:
        state = rx_get_job_status(job_info)
    except:
        state = RxRemoteJobStatus.UNDETERMINED

    if state == RxRemoteJobStatus.UNDETERMINED:
        raise RuntimeError("Failed to get the job state for job {} from its associated compute context".format(job_info))

    if state == RxRemoteJobStatus.FINISHED:
        ret = context.get_job_result(job_info)

        if job_info.callback is not None and callable(job_info.callback):
            ret = job_info.callback(ret)

        if console_output or (console_output is None and context.console_output):
            print(rx_get_job_output(job_info))

        if auto_cleanup or (auto_cleanup is None and context.auto_cleanup):
            context.delete_job(job_info)
            job_info.close()

        return ret

    elif state == RxRemoteJobStatus.RUNNING or state == RxRemoteJobStatus.QUEUED:
        raise RuntimeError("The job {} is {}, cannot get results.".format(job_info, state))
    else:
        if console_output or (console_output is None and context.console_output):
            print(rx_get_job_output(job_info))

        if auto_cleanup or (auto_cleanup is None and context.auto_cleanup):
            context.delete_job(job_info)
            job_info.close()

        if state == "cancelled":
            raise RuntimeError("No results are available for the job {} as it was cancelled".format(job_info))
        else:
            raise RuntimeError("No results available for the job {} - final job state {}".format(job_info, state))
