from . RxLocalComputeContext import RxLocalComputeContext
from typing import Callable, List
from multiprocessing import get_context, Queue, cpu_count
from io import FileIO
from queue import Empty
from enum import Enum
import os
import dill
import tempfile
import sys
import math
from ..utils.RxOptions import RxOptions


class _ProcessState(Enum):
    NotStarted = 0,
    Waiting = 1,
    Executing = 2


# The master process uses _WorkerProcessInfo objects to interface with and keep track of the state
# of worker processes. When start() is called, a worker process is spawned and it begins executing the
# _wait_for_messages function.
# When the MP puts a task on a worker's task queue, the MP updates the worker's state to Executing.
# When the MP receives a result from a worker on the results queue, the MP updates that worker's state to Waiting.
class _WorkerProcessInfo:

    # Worker processes created will execute this function
    # Function blocks until a task appears on task_queue
    @staticmethod
    def _wait_for_messages(results_queue: Queue, task_queue: Queue, worker_id : int):
        while True:
            # Blocks until message is received
            message = task_queue.get()
            # No more tasks.
            if message.finished:
                return
            # Filename is dilled _TaskInfo object
            with open(message.filename, "rb") as file:
                task_info = _TaskInfo.from_file(file)
                task_id = task_info.task_index
                try:
                    result = task_info.user_function(**task_info.args)
                    if task_info.callback is not None: result = task_info.callback(result)
                except Exception as e:
                    result = e
                results_queue.put(_TaskResult(result, worker_id, task_id))

    def __init__(self, results_queue : Queue, worker_id : int, ctx):
        self._tasks_queue = ctx.Queue()
        self._worker_id = worker_id
        self._results_queue = results_queue
        self._state = _ProcessState.NotStarted
        self._current_task_info_filename = None
        self._process = ctx.Process(target=_WorkerProcessInfo._wait_for_messages,
                                    args=(self._results_queue, self._tasks_queue, self._worker_id))

    def start(self):
        # This spawns the actual worker process; it starts executing _wait_for_messages method
        self._process.start()
        self._state = _ProcessState.Waiting

    def assign_task(self, task_info_filename):
        if self._state != _ProcessState.Waiting:
            raise RuntimeError("Process is not available")
        self._tasks_queue.put(_TaskMessage.run_message(task_info_filename))
        self._state = _ProcessState.Executing
        self._current_task_info_filename = task_info_filename

    def finish(self):
        if self._process.is_alive():
            self._tasks_queue.put(_TaskMessage.finish_message())

    def notify_waiting(self):
        self._state = _ProcessState.Waiting

    def errored(self):
        return self._state == _ProcessState.Executing and not self._process.is_alive()

    @property
    def state(self) -> _ProcessState:
        return self._state

    @property
    def current_task_info_filename(self) -> str:
        return self._current_task_info_filename


class _TaskResult:

    def __init__(self, result, worker_id: int, task_id: int):
        self._result = result
        self._worker_id = worker_id
        self._task_id = task_id

    @property
    def result(self):
        return self._result

    @property
    def worker_id(self) -> int:
        return self._worker_id

    @property
    def task_index(self) -> int:
        return self._task_id


class _TaskInfo:

    def __init__(self, user_function: Callable, args : dict, callback: Callable, task_index: int):
        self._user_function = user_function
        self._args = args
        self._callback = callback
        self._task_index = task_index

    @staticmethod
    def from_file(file: FileIO):
        the_dill = dill.load(file)
        return _TaskInfo(the_dill.user_function, the_dill.args, the_dill.callback, the_dill.task_index)

    @property
    def user_function(self) -> Callable:
        return self._user_function

    @property
    def args(self) -> dict:
        return self._args

    @property
    def callback(self) -> Callable:
        return self._callback

    @property
    def task_index(self) -> int:
        return self._task_index


class _TaskMessage:

    def __init__(self, finished: bool, filename: str = None):
        self._finished = finished
        self._filename = filename

    @staticmethod
    def run_message(filename: str):
        return _TaskMessage(False, filename)

    @staticmethod
    def finish_message():
        return _TaskMessage(True)

    @property
    def filename(self) -> str:
        return self._filename

    @property
    def finished(self) -> bool:
        return self._finished


class RxLocalParallel(RxLocalComputeContext):
    """
    .. rxtitle::

        Generate Local Parallel Compute Context

    .. rxdescription::

        Creates a local compute context object. Computations using rx_exec will be processed in parallel.

            def target_function():
                import os
                print(os.getpid())

            rx_exec(target_function, times_to_run = 5, compute_context = RxLocalParallel())
    
    .. rxkeywords:: context, localpar

    .. rxseealso::

        :py:func:`RxComputeContext <revoscalepy.RxComputeContext>`,
        :py:func:`RxInSqlServer <revoscalepy.RxInSqlServer>`,
        :py:func:`rx_get_compute_context <revoscalepy.rx_get_compute_context>`,
        :py:func:`rx_set_compute_context <revoscalepy.rx_set_compute_context>`.

    .. rxexample:: 
        from revoscalepy import RxLocalParallel
        localpar = RxLocalParallel()
    """

    def __init__(self):
        super().__init__("RxLocalParallel", "1.0-1")

    def _get_filenames(self, times_to_run):
        return [tempfile.NamedTemporaryFile(delete=False).name for x in range(times_to_run)]

    def _spawn_workers(self, results_queue, num, ctx) -> dict:

        main_module = sys.modules['__main__']
        old_main_file = getattr(main_module, '__file__', None)
        old_spec_name = None

        try:
            # This prevents multiprocessing from re-importing the main module
            main_module.__file__ = None
            if main_module.__spec__ is not None and main_module.__spec__.name is not None:
                old_spec_name = main_module.__spec__.name
                main_module.__spec__.name = None

            # Spawn the processes
            workers = {x: _WorkerProcessInfo(results_queue, x, ctx) for x in range(num)}
            for worker_id in workers:
                workers[worker_id].start()
        finally:
            # Ensure attributes of the main module are set back
            main_module.__file__ = old_main_file
            if old_spec_name is not None:
                main_module.__spec__.name = old_spec_name

        return workers

    def _dump_task_info_files(self, filenames, user_function, callback, args_for_tasks, times_to_run):
        # We make .dill files because multiprocessing can't automatically serialize some functions
        task_information = [_TaskInfo(user_function, arg_group, callback, idx) for idx, arg_group in enumerate(args_for_tasks)]
        for task, filename in zip(task_information, filenames):
            with open(filename, "wb") as file:
                dill.dump(task, file, -1)

    def _get_ready_worker(self, workers : List[_WorkerProcessInfo]):
        for worker in workers:
            if worker.state == _ProcessState.Waiting:
                return worker
        return None

    def _get_num_workers(self, num_tasks, task_chunk_size) -> int:
        max_cores = RxOptions.get_option("numCoresToUse")
        if max_cores <= 0:
            max_cores = cpu_count()
        num_workers = num_tasks if task_chunk_size is None else num_tasks // task_chunk_size
        return int(min(max_cores, num_workers))

    def execute_tasks(self, function: Callable, args_for_tasks: list, times_to_run: int, task_chunk_size: int,
                      callback: Callable = None, **kwargs) -> list:
        total_num_procs = self._get_num_workers(times_to_run, task_chunk_size)
        if total_num_procs == 1:
            # Only one processor needed, do not run in parallel
            # return super().execute_tasks(function, args_for_tasks, times_to_run, callback)
            return super().execute_tasks(function=function, args_for_tasks=args_for_tasks,
                                         task_chunk_size=task_chunk_size, times_to_run=times_to_run, callback=callback)

        # Multiprocessing context
        ctx = get_context('spawn')
        results = list()
        workers = list()

        # Creates temporary files
        task_info_filenames = self._get_filenames(times_to_run)
        try:
            # Creates dill files with information on each execution.
            # These dill file names are passed to the worker processes.
            self._dump_task_info_files(task_info_filenames, function, callback, args_for_tasks, times_to_run)

            # Keeps track of the remaining task info files that have not been passed to processes
            remaining = task_info_filenames.copy()

            task_results = []
            results_queue = ctx.Queue()

            workers = self._spawn_workers(results_queue, total_num_procs, ctx)

            while len(task_results) < times_to_run:

                # Pass tasks to all available processes and check for broken processes
                alive_workers = {}
                for worker_id in workers:
                    worker = workers[worker_id]
                    # Worker is waiting for a task
                    if worker.state == _ProcessState.Waiting and len(remaining) > 0:
                        worker.assign_task(remaining.pop())
                    # Worked had uncontrolled failure
                    if worker.errored():
                        remaining.append(worker.current_task_info_filename)
                    else:
                        alive_workers[worker_id] = worker

                # Update workers list with healthy workers
                workers = alive_workers
                if len(workers) == 0:
                    raise Exception("All spawned processes died during local parallel execution.")

                try:
                    # Blocks for 5 seconds or until result is available
                    # This allows the master process to periodically check process health
                    res = results_queue.get(block=True, timeout=5)
                    workers[res.worker_id].notify_waiting()
                    task_results.append(res)
                except Empty as empty:
                    pass

            # Sort results from workers based on idx
            results = [worker_result.result for worker_result in sorted(task_results, key = lambda wres: wres.task_index)]

        # This will not be triggered by exceptions in workers, those are captured and returned as results
        except Exception as e:
            results = [e]

        finally:
            # Tell all processes to finish
            for worker_id in workers:
                workers[worker_id].finish()
            # Delete task info files
            for filename in task_info_filenames:
                os.remove(filename)

            return results
