#
# Copyright (c) Microsoft Corporation.  All rights reserved.
#

import abc
import datetime
import os
import tempfile
import uuid
import shutil
from typing import Callable

import dill
import re

from .RxComputeContext import RxComputeContext


#################################################################################
#
# Remote job
#
class RxRemoteJob(object):
    __metaclass_ = abc.ABCMeta
    __uuid_hex = re.compile("[0-9a-f]{32}\Z", re.I)

    def __init__(self, compute_context: RxComputeContext, job_id: str = None):
        if not job_id:
            job_id = uuid.uuid1().hex

        self._job_id = job_id
        self._job_directory = RxRemoteJob.__generate_job_directory_path(self.id)
        self._job_input_directory = os.path.join(self.dir, "pythoninput")
        self._job_output_directory = self.dir

        if not os.path.isdir(self.dir):
            os.makedirs(self.dir)
            os.makedirs(self.input_dir)

        self._input_file = os.path.join(self.input_dir, "input.dill")
        self._sample_data_file = os.path.join(self.input_dir, "sample-ds1.xdf")
        self._output_file = os.path.join(self.output_dir, "output.dill")

        self._input_zip_file = self.input_dir + ".zip"
        self._output_zip_file = self.output_dir + ".zip"

        self._start_time = datetime.datetime.now()

        if compute_context is None or not isinstance(compute_context, RxComputeContext):
            raise ValueError("The compute context was not specified or was not the correct type.")

        self._context_id = compute_context.id

        self._callback = None

        self.serialize_job()

    def __str__(self):
        return self._job_id

    @property
    def id(self) -> str:
        return self._job_id

    @property
    def start_time(self) -> datetime:
        return self._start_time

    @property
    def dir(self) -> str:
        return self._job_directory

    @property
    def input_file(self) -> str:
        return self._input_file

    @property
    def output_file(self) -> str:
        return self._output_file

    @property
    def input_zip_file(self) -> str:
        return self._input_zip_file

    @property
    def output_zip_file(self) -> str:
        return self._output_zip_file

    @property
    def input_dir(self) -> str:
        return self._job_input_directory

    @property
    def output_dir(self) -> str:
        return self._job_output_directory

    @property
    def sample_data_file(self) -> str:
        return self._sample_data_file

    @property
    def context_id(self) -> str:
        return self._context_id

    @property
    def callback(self) -> Callable:
        return self._callback

    @callback.setter
    def callback(self, value: Callable):
        self._callback = value

    def close(self):
        """
        Closes the remote job, purging all the data associated with the job
        """
        if os.path.isdir(self.dir):
            shutil.rmtree(self.dir)

    def resolve_context(self) -> RxComputeContext:
        """
        Resolves the ``RxComputeContext`` that is associated with the job
        
        :return: The ``RxComputeContext`` that is associated with the job or ``None`` if the compute context isn't valid
        """
        return RxComputeContext.lookup(self.context_id)

    @abc.abstractmethod
    def get_type(self) -> str:
        pass

    @abc.abstractmethod
    def serialize_bxl(self) -> dict:
        pass

    def serialize_job(self, file_path: str = None) -> str:
        if file_path is None:
            file_path = RxRemoteJob.__generate_serialize_file_path(self.dir)

        with open(file_path, "wb") as job_file:
            dill.dump(self, job_file)

        return file_path

    @staticmethod
    def deserialize_jobs() -> list:
        """
        Deserializes the existing jobs that have not been cleaned up by calling close().
        
        :return: The deserialized jobs
        """
        job_directory = RxRemoteJob.__generate_jobs_path()

        uuid_hex = re.compile("[0-9a-f]{32}\Z", re.I)
        for entry in os.scandir(job_directory):
            if entry.is_dir() and uuid_hex.match(entry.name) is not None:
                try:
                    yield RxRemoteJob.deserialize_job(entry.name)
                except Exception as e:
                    print("Error processing job {}:  {}".format(entry.name, e))

    @staticmethod
    def deserialize_job(job_id: str) -> "RxRemoteJob":
        """
        Deserializes a RxRemoteJob given the job id
        
        :return: The job that was deserialized
        """
        if job_id is None:
            raise ValueError("Must specify a jobId to deserialize a job.")

        job = uuid.UUID(job_id)

        path = RxRemoteJob.__generate_job_directory_path(job.hex)

        serialized_file_path = RxRemoteJob.__generate_serialize_file_path(path)

        if not os.path.isfile(serialized_file_path):
            raise ValueError("The job {} does not exist on the current system".format(job_id))

        with open(serialized_file_path, "rb") as jobFile:
            return dill.load(jobFile)

    @staticmethod
    def __generate_job_directory_path(job_id: str) -> str:
        """
        Generates the path to the job directory given the specified job id
        
        :param job_id: The id of the job
        :return: The path on disk to the job
        """
        return os.path.join(RxRemoteJob.__generate_jobs_path(), job_id)

    @staticmethod
    def __generate_jobs_path() -> str:
        return os.path.join(tempfile.gettempdir(), "RevoPythonJobs")

    @staticmethod
    def __generate_serialize_file_path(job_path: str) -> str:
        """
        Generates the path to which to serialize job data
        
        :return: The path to the serialized job data on disk
        """
        return os.path.join(job_path, "jobData.dill")

    @staticmethod
    def matches_job_format(candidate: str) -> bool:
        return candidate is not None and RxRemoteJob.__uuid_hex.match(candidate) is not None
