#
# Copyright (c) Microsoft Corporation.  All rights reserved.
#

import argparse
import copy
import dill
import getpass
import os
import shutil
import subprocess
import sys
import time
import uuid
import warnings

from typing import Callable
from pandas import DataFrame

from . RxRemoteComputeContext import RxRemoteComputeContext
from . RxRemoteJob import RxRemoteJob
from . RxSparkRemoteJob import RxSparkRemoteJob, RxSparkRemoteHpcJob
from . RxPySparkBase import rx_pyspark_check_pyspark_is_loaded
from . RxPySparkBase import rx_pyspark_check_data_is_spark_data_frame

from .. datasource import RxXdfData
from .. datasource.RxHiveData import RxHiveData
from .. datasource import RxSparkData
from .. datasource.RxSparkDataFrame import RxSparkDataFrame
from .. datasource.RxDataSource import RxDataSource
from .. utils.RxOptions import RxOptions
from .. utils.RxFileUtils import file_exists
from .. utils.RxTrace import RxTraceLevel
from .. utils.RxTrace import rx_runtime_trace
from .. RxSerializable import rx_native_call
from .. RxSerializable import _resolve_transform_function
from .. RxSerializable import build_transform
from .. RxFunction import validate_argument


#################################################################################
#
# Execute command
#
def rx_spark_execute_command(command):
    p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output, error = p.communicate()

    out_message, err_message = _extract_message(output.decode('utf-8'))
    if p.returncode > 128:
        raise ValueError(err_message)
    elif p.returncode > 0:
        warnings.warn(str.join("\n", err_message))
    return out_message


def _extract_message(output):
    out_message = []
    error_message = []
    is_stdout_message = True
    for line in output.split('\n'):
        if "%PROTOCOL-STDOUT%" in line:
            continue
        if "%PROTOCOL-STDERR%" in line:
            is_stdout_message = False
            continue
        if "%PROTOCOL-END%" in line:
            break

        if is_stdout_message:
            out_message.append(line)
        else:
            error_message.append(line)
    return out_message, error_message


def _print_error_msg(file_path, lines_to_show: int = 10,
                     archive_dir: str = "/tmp/MLSArchive-{0}".format(getpass.getuser())):
    rx_runtime_trace("start", RxTraceLevel.TRACE)

    print("============================================================")

    with open(file_path) as file:
        output = file.readlines()
    archive = True

    if len(output) < lines_to_show:
        lines_to_show = len(output)
        archive = False
    else:
        print("Job failed, the first 10 lines of the log are shown below:")

    for line in range(0, lines_to_show):
        print(output[line], end='')

    if archive:
        if not os.path.exists(archive_dir):
            os.mkdir(archive_dir, mode=0o700)

        archive_file_path = os.path.join(archive_dir, "MLSLog-{0}.log".format(uuid.uuid1().hex))
        shutil.copy(file_path, archive_file_path)
        print()
        print("For the complete log, please refer to the log file {0}".format(archive_file_path))
    print("============================================================")


def _monitor_hadoop_job(job: RxSparkRemoteJob,
                        execute_command: callable = rx_spark_execute_command,
                        msg_level: str = "INFO",):
    rx_runtime_trace("start msg_level={0}".format(msg_level), RxTraceLevel.TRACE)

    block_file_name = "lastMonitorBlock.txt"
    # remote call to generate the output file block
    command = [
        "/usr/bin/mrs-hadoop-job", "monitor",
        "--dir", job.get_spark_job_work_dir(),
        "--msg-level", msg_level,
        "--max-line", "all",
        "--version", "1.0",
        "--trace-level", str(RxOptions.get_option("traceLevel", 0))
    ]

    execute_command(command)
    file_path = os.path.join(job.get_spark_job_work_dir(), block_file_name)

    # print output
    if msg_level == "INFO":
        with open(file_path) as file:
            output = file.read()
            print(output, end='')
    else:
        _print_error_msg(file_path)


#################################################################################
#
# Spark compute context
#
class RxSpark(RxRemoteComputeContext):
    """
    .. rxtitle::

        Generate Spark Compute Context

    .. rxdescription::

        Creates the compute context for running revoscalepy analysis on Spark.
        Note that the use of rx_spark_connect() is recommended over RxSpark()
        as rx_spark_connect() supports persistent mode with in-memory caching.
        Run help(revoscalepy.rx_spark_connect) for more information.

    """
    @property
    def hdfs_share_dir(self) -> str:
        return self._hdfs_share_dir

    @property
    def share_dir(self) -> str:
        return self._share_dir

    @property
    def user(self) -> str:
        return self._user

    @property
    def name_node(self) -> str:
        return self._name_node

    @property
    def port(self) -> int:
        return self._port

    @property
    def auto_cleanup(self) -> bool:
        return self._auto_cleanup

    @property
    def console_output(self) -> bool:
        return self._console_output

    @property
    def package_to_load(self) -> list:
        return self._packages_to_load

    @property
    def session_pid(self) -> str:
        return self._session_pid

    @property
    def app_id(self) -> str:
        return self._app_id

    @property
    def job_id(self) -> str:
        return self._job_id

    @property
    def hadoop_python_path(self) -> str:
        return self._hadoop_python_path

    @property
    def idle_timeout(self) -> int:
        return self._idle_timeout

    @property
    def num_executors(self) -> int:
        return self._num_executors

    @property
    def executor_cores(self) -> int:
        return self._executor_cores

    @property
    def executor_mem(self) -> str:
        return self._executor_mem

    @property
    def driver_mem(self) -> str:
        return self._driver_mem

    @property
    def executor_overhead_mem(self) -> str:
        return self._executor_overhead_mem

    @property
    def extra_spark_config(self) -> str:
        return self._extra_spark_config

    @property
    def spark_reduce_method(self) -> str:
        return self._spark_reduce_method

    @property
    def tmp_fs_work_dir(self) -> str:
        return self._tmp_fs_work_dir

    @property
    def persistent_run(self) -> bool:
        return self._persistent_run

    @property
    def pyspark_connection(self):
        return self._pyspark_connection

    # For any dynamic default values from RxOptions. Cannot directly assign in signature.
    # Otherwise, the default value will be decided at the time of loading revoscalepy.
    # Then the function will not respect any updated RxOptions values.
    def __init__(
            self,
            hdfs_share_dir: str = os.path.join("/user/RevoShare", getpass.getuser()),
            share_dir: str = os.path.join("/var/RevoShare", getpass.getuser()),
            user: str = getpass.getuser(),
            name_node: str = None,
            port: int = None,
            auto_cleanup: bool = True,
            console_output: bool = False,
            packages_to_load: list = None,
            idle_timeout: int = 3600,
            num_executors: int = None,
            executor_cores: int = None,
            executor_mem: str = None,
            driver_mem: str = None,
            executor_overhead_mem: str = None,
            extra_spark_config: str = "",
            spark_reduce_method: str = "auto",
            tmp_fs_work_dir: str = "/dev/shm",
            persistent_run: bool = False,
            wait: bool = True,
            **kwargs
    ):
        if "from_spark_connect" not in kwargs or kwargs["from_spark_connect"] is not True:
            print("Note that the use of rx_spark_connect() is recommended over RxSpark() "
                  "as rx_spark_connect() supports persistent mode with in-memory caching. "
                  "Run help(revoscalepy.rx_spark_connect) for more information.\n")
        if name_node is None:
            name_node = RxOptions.get_option("hdfsHost")
        if port is None:
            port = RxOptions.get_option("hdfsPort")
        if num_executors is None:
            num_executors = RxOptions.get_option("spark.numExecutors")
        if executor_cores is None:
            executor_cores = RxOptions.get_option("spark.executorCores")
        if executor_mem is None:
            executor_mem = RxOptions.get_option("spark.executorMem")
        if driver_mem is None:
            driver_mem = RxOptions.get_option("spark.driverMem")
        if executor_overhead_mem is None:
            executor_overhead_mem = RxOptions.get_option("spark.executorOverheadMem")
        if packages_to_load is None:
            packages_to_load = []
        session_pid = kwargs.get("session_pid", str(os.getpid()))
        app_id = kwargs.get("app_id", str(rx_native_call(functionname="CreateGUID", params={})))
        job_id = kwargs.get("job_id", "initid")
        hadoop_python_path = kwargs.get("hadoop_python_path", RxOptions.get_option("unixPythonPath"))
        pyspark_connection = kwargs.get("pyspark_connection", None)
        super(RxSpark, self).__init__("spark", "1.0-1", wait=wait)
        self._hdfs_share_dir = hdfs_share_dir
        self._share_dir = share_dir
        self._user = user
        self._name_node = name_node
        self._port = port
        self._auto_cleanup = auto_cleanup
        self._console_output = console_output
        self._packages_to_load = packages_to_load
        self._session_pid = session_pid
        self._app_id = app_id
        self._job_id = job_id
        self._hadoop_python_path = hadoop_python_path
        self._idle_timeout = idle_timeout
        self._num_executors = num_executors
        self._executor_cores = executor_cores
        self._executor_mem = executor_mem
        self._driver_mem = driver_mem
        self._executor_overhead_mem = executor_overhead_mem
        self._extra_spark_config = extra_spark_config
        self._spark_reduce_method = spark_reduce_method
        self._tmp_fs_work_dir = tmp_fs_work_dir
        self._persistent_run = persistent_run
        self._pyspark_connection = pyspark_connection

        self.job = None

    def serialize_bxl(self) -> dict:
        result = self.__dict__.copy()
        result.update({"class": "RxSpark"})

        return result


    # validate input data source
    def validate_in_data(self, in_data, param_name):

        if rx_pyspark_check_pyspark_is_loaded():
            if rx_pyspark_check_data_is_spark_data_frame(in_data):
                in_data = RxSparkDataFrame(in_data)
                in_data._spark_data_frame = None
            # master node will not load pyspark
            # for pyInData, cannot contains any objs from pyspark
            elif isinstance(in_data, RxSparkDataFrame):
                new_in_data = copy.copy(in_data)
                new_in_data._spark_data_frame = None
                in_data = new_in_data

        in_data = validate_argument(in_data, param_name, (DataFrame, str, RxDataSource))
        if isinstance(in_data, str) and not file_exists(in_data):
            raise ValueError("The path of input data not found {:s}".format(in_data))
        return in_data


    # validate output data source
    def validate_out_data(self, out_data, param_name):
        if out_data is None:
            return None

        if rx_pyspark_check_pyspark_is_loaded():
            if isinstance(out_data, RxSparkDataFrame):
                new_out_data = copy.copy(out_data)
                new_out_data._spark_data_frame = None
                out_data = new_out_data

        out_data = validate_argument(out_data, param_name, (DataFrame, str, RxDataSource))
        return out_data


    def handle_rx_spark_data_objs(self, py_args):
        in_data = None
        if 'data' in py_args:
            in_data = py_args['data']
        elif 'input_data' in py_args:
            in_data = py_args['input_data']

        # keep track of the cached RxSparkData obj in _rx_spark_data_objs option in output data
        if in_data is not None and (isinstance(in_data, RxSparkData) or isinstance(in_data, RxXdfData)):
            if self.persistent_run:
                if in_data.cache:
                    spark_data_objs = RxOptions.get_option("_rx_spark_data_objs")
                    if len([x for x in spark_data_objs if x.df_name == in_data.df_name]) == 0:
                        spark_data_objs.append(in_data)
                        RxOptions.set_option("_rx_spark_data_objs", spark_data_objs)
                else:
                    spark_data_objs = RxOptions.get_option("_rx_spark_data_objs")
                    new_ojbs = [x for x in spark_data_objs if x.df_name != in_data.df_name]
                    RxOptions.set_option("_rx_spark_data_objs", new_ojbs)
            else:
                warnings.warn("The input " + in_data.df_type + " data source for this computation, " + in_data.df_source
                              + ", has cache = True. Caching data in Spark memory is only applicable in RxSpark "
                              + "persistent mode. The cache setting is ignored. Please set parameter persistentRun "
                              + "to True in RxSpark() or use rx_spark_connect() if caching is desired.")

    def handle_output_rx_spark_data_objs(self, py_args):
        out_data = None
        if 'output_file' in py_args:
            out_data = py_args['output_file']
        elif 'output_data' in py_args:
            out_data = py_args['output_data']

        # Always remove outData from cache manager. It can avoid the case:
        # RxHiveData loads data into revoscalepy and we cache the converted XDF.
        # Reuse the RxHiveData as output, We write something new back to the hive table.
        # Use the RxHiveData as input again. If with cache, then it will load the old converted XDF file.
        # Our spark engine will clean the cache if the RxHiveData is used as output. Just need to clean python side's
        #  cache record here.
        if out_data is not None and (isinstance(out_data, RxSparkData) or isinstance(out_data, RxXdfData)):
            if not (isinstance(out_data, RxHiveData) and out_data.df_type == "hive-temp-table"):
                if self.persistent_run:
                    spark_data_objs = RxOptions.get_option("_rx_spark_data_objs")
                    new_ojbs = [x for x in spark_data_objs if x.df_name != out_data.df_name]
                    RxOptions.set_option("_rx_spark_data_objs", new_ojbs)
            else:
                # Always cache when output is hive temp table
                if self.persistent_run:
                    spark_data_objs = RxOptions.get_option("_rx_spark_data_objs")
                    if len([x for x in spark_data_objs if x.df_name == out_data.df_name]) == 0:
                        spark_data_objs.append(out_data)
                        RxOptions.set_option("_rx_spark_data_objs", spark_data_objs)
                else:
                    raise TypeError("Hive temp table is an in-memory object. You can only use it in RxSpark with "
                                    + "persistentRun mode")

    def _execute(self, py_module_name, py_function_name, py_args):
        rx_runtime_trace("start", RxTraceLevel.INFO)

        # copy the current compute context to clean pyspark_connection
        # pyspark_connection should not be included for dumping to master node
        compute_context_to_dump = RxSparkBuilder().build_from(self)\
            .set_pyspark_connection(None).set_additional_conf("from_spark_connect", True).create()
        py_args['compute_context'] = compute_context_to_dump

        self.handle_rx_spark_data_objs(py_args)
        self.handle_output_rx_spark_data_objs(py_args)

        job = self.create_job()
        job.create_spark_work_dir(self._share_dir)

        job.prepare_job_metadata(py_function_name, py_args)
        job.dump_job_metadata_json("job.json")
        job.dump_json(rx_spark_get_spark_env_metadata(self), "env.json")

        py_in_data = dict()
        py_in_data['MODULE'] = py_module_name
        py_in_data['FUN'] = py_function_name
        py_in_data['ARGS'] = py_args
        job.dump_job_in_data("job.pyInData", py_in_data)

        self.start_job(job)

        job_state = self.poll_for_job_completion(job)
        rx_runtime_trace("RxSpark job final state : " + str(job_state), RxTraceLevel.INFO)

        result = []
        if job_state == "finished":
            output_dir = os.path.join(job.get_spark_job_work_dir(), job.get_result_dir_name())
            result = self.get_job_result(job, output_dir)
        elif job_state == "failed":
            _monitor_hadoop_job(job, rx_spark_execute_command, "ERROR")
            if self.auto_cleanup:
                self.delete_job(job)
            raise ValueError("Job fails, no results available.")
        else:
            raise ValueError("Invalid job state:" + job_state)

        # todo add try/catch/final to well handle clean up
        if self.auto_cleanup:
            self.delete_job(job)

        rx_runtime_trace("finish", RxTraceLevel.INFO)

        return result


    def _execute_spark_hpc(self, py_function_name, py_args, py_data=dict()):
        job = self.create_spark_hpc_job()
        job.create_spark_work_dir(self._share_dir)

        job.prepare_job_metadata(py_function_name, py_args)
        job.dump_job_metadata_json("job.json")
        job.dump_json(rx_spark_get_spark_env_metadata(self), "env.json")

        job.dump_job_in_data("job.pyInData", py_data)

        self.start_job(job)

        job_state = self.poll_for_job_completion(job)
        rx_runtime_trace("RxSpark job final state : " + str(job_state), RxTraceLevel.INFO)

        result = []
        if job_state == "finished":
            output_dir = os.path.join(job.get_spark_job_work_dir(), job.get_result_dir_name())
            result = self.get_job_result(job, output_dir)
        elif job_state == "failed":
            _monitor_hadoop_job(job, rx_spark_execute_command, "ERROR")
            if self.auto_cleanup:
                self.delete_job(job)
            raise ValueError("Job fails, no results available.")
        else:
            raise ValueError("Invalid job state:" + job_state)

        # todo add try/catch/final to well handle clean up
        if self.auto_cleanup:
            self.delete_job(job)

        return result

    def create_job(self, inData: RxDataSource = None) -> RxSparkRemoteJob:
        self.job = RxSparkRemoteJob(self)
        return self.get_job()

    def create_spark_hpc_job(self) -> RxSparkRemoteJob:
        self.job = RxSparkRemoteHpcJob(self)
        return self.get_job()

    def get_job(self) -> RxSparkRemoteJob:
        return self.job

    def start_job(self, job: RxRemoteJob, execute_command: callable = rx_spark_execute_command):
        rx_runtime_trace("start", RxTraceLevel.INFO)

        if not isinstance(job, RxSparkRemoteJob):
            raise ValueError("job should be Spark job")

        command = [
            "/usr/bin/mrs-hadoop-job", "run",
            "--dir", job.get_spark_job_work_dir(),
            "--version", "1.0",
            "--trace-level", str(RxOptions.get_option("traceLevel", 0)),
            "--server-type", "Python"
        ]
        execute_command(command=command)

    def poll_for_job_completion(self, job: RxRemoteJob, execute_command: callable = rx_spark_execute_command):
        rx_runtime_trace("start", RxTraceLevel.INFO)

        if not isinstance(job, RxSparkRemoteJob):
            raise ValueError("job should be Spark job")

        while True:
            if self.console_output:
                _monitor_hadoop_job(job, execute_command)

            job_state = self.get_job_state(job, execute_command)
            if job_state in ["finished", "canceled", "failed", "undetermined"]:
                return job_state

            time.sleep(1)

    def get_job_state(self, job: RxRemoteJob, execute_command: callable = rx_spark_execute_command) -> str:
        rx_runtime_trace("start", RxTraceLevel.TRACE)

        if not isinstance(job, RxSparkRemoteJob):
            raise ValueError("job should be Spark job")

        command = [
            "/usr/bin/mrs-hadoop-job", "query", "--status",
            "--dir", job.get_spark_job_work_dir(),
            "--version", "1.0",
            "--trace-level", str(RxOptions.get_option("traceLevel", 0))
        ]
        return execute_command(command)[0]

    def cancel_job(self, job: RxRemoteJob) -> bool:
        rx_runtime_trace("start", RxTraceLevel.DEBUG)

        if not self.supports_job(job):
            raise ValueError("Job should be Spark job")

        raise NotImplemented("Not implemented yet")

    def get_job_result(self, job: RxRemoteJob, output_dir):
        rx_runtime_trace("start", RxTraceLevel.DEBUG)

        if not isinstance(job, RxSparkRemoteJob):
            raise ValueError("job should be Spark job")

        output_file_name = os.path.join(output_dir, "job.pyOutData")
        if not os.path.exists(output_file_name):
            raise FileNotFoundError("Cannot find job output data file: " + output_file_name)

        rx_runtime_trace("RxSpark load job_output_file", RxTraceLevel.DEBUG)
        with open(output_file_name, "rb") as job_output_file:
            job_output_object = dill.load(job_output_file)

        rx_runtime_trace("RxSpark exec func to get job result", RxTraceLevel.DEBUG)
        job_output_object['funcParams']['output_dir'] = output_dir
        result = job_output_object['func'](**job_output_object['funcParams'])
        return result

    def delete_job(self, job: RxRemoteJob):
        rx_runtime_trace("start", RxTraceLevel.DEBUG)

        if not isinstance(job, RxSparkRemoteJob):
            raise ValueError("job should be Spark job")

        shutil.rmtree(job.get_spark_job_work_dir(), ignore_errors=True)

    def supports_job(self, job: RxRemoteJob) -> bool:
        return job is not None and isinstance(job, RxSparkRemoteJob)

    def get_job_output(self, job: RxRemoteJob) -> bool:
        if not self.supports_job(job):
            raise ValueError("Job should be Spark job")

        raise NotImplemented("Not implemented yet")

    def execute_tasks(self, function: Callable, args_for_tasks: list, times_to_run: int, task_chunk_size: int,
                      callback: Callable, local_state: dict, continue_on_failure: bool, **kwargs) -> list:
        if task_chunk_size is None:
            task_chunk_size = 1

        py_function_name = "exec_v1"
        py_args = {
            'numTasks': times_to_run,
            'taskChunkSize': task_chunk_size
        }
        py_data = {
            'func': function,
            'funcParams': args_for_tasks,
            "localState": local_state,
            "continueOnFailure": continue_on_failure
        }

        return self._execute_spark_hpc(py_function_name, py_args, py_data)


# currently for internal use only
class RxSparkBuilder(object):
    def __init__(self):
        self._build_conf = {}

    def build_from(self, compute_context):
        self._build_conf['hdfs_share_dir'] = compute_context.hdfs_share_dir
        self._build_conf['share_dir'] = compute_context.share_dir
        self._build_conf['user'] = compute_context.user
        self._build_conf['name_node'] = compute_context.name_node
        self._build_conf['port'] = compute_context.port
        self._build_conf['auto_cleanup'] = compute_context.auto_cleanup
        self._build_conf['console_output'] = compute_context.console_output
        self._build_conf['package_to_load'] = compute_context.package_to_load
        self._build_conf['session_pid'] = compute_context.session_pid
        self._build_conf['app_id'] = compute_context.app_id
        self._build_conf['job_id'] = compute_context.job_id
        self._build_conf['hadoop_python_path'] = compute_context.hadoop_python_path
        self._build_conf['idle_timeout'] = compute_context.idle_timeout
        self._build_conf['num_executors'] = compute_context.num_executors
        self._build_conf['executor_cores'] = compute_context.executor_cores
        self._build_conf['executor_mem'] = compute_context.executor_mem
        self._build_conf['driver_mem'] = compute_context.driver_mem
        self._build_conf['executor_overhead_mem'] = compute_context.executor_overhead_mem
        self._build_conf['extra_spark_config'] = compute_context.extra_spark_config
        self._build_conf['spark_reduce_method'] = compute_context.spark_reduce_method
        self._build_conf['tmp_fs_work_dir'] = compute_context.tmp_fs_work_dir
        self._build_conf['persistent_run'] = compute_context.persistent_run
        self._build_conf['pyspark_connection'] = compute_context.pyspark_connection
        return self

    def set_wait(self, wait):
        self._build_conf['wait'] = wait
        return self

    def set_idle_timeout(self, idle_timeout):
        self._build_conf['idle_timeout'] = idle_timeout
        return self

    def set_persistent_run(self, persistent_run):
        self._build_conf['persistent_run'] = persistent_run
        return self

    def set_pyspark_connection(self, pyspark_connection):
        self._build_conf['pyspark_connection'] = pyspark_connection
        return self

    def set_additional_conf(self, key, value):
        self._build_conf[key] = value
        return self

    def create(self) -> RxSpark:
        return RxSpark(**self._build_conf)


#################################################################################
#
# executes a function call request received from remote client
# This is the entry point of HPC job on worker nodes.
#
def rx_remote_spark_exec_call(args: dict = None):
    raise Exception("rx_remote_spark_exec_call is not implemented yet")


#################################################################################
#
# executes a function call request received from remote client
# This is the entry point of HPA job on master node.
#
def rx_remote_spark_hpa_call(args: dict = None):
    try:
        # Set the node type to master
        RxOptions.set_option("nodeType", "master")

        if args is None:
            args = _parse_rx_remote_spark_hpa_call_arguments()

        # set traceLevel without calling into native
        if "rxTraceLevel" in args:
            RxOptions.instance()["traceLevel"] = int(args["rxTraceLevel"])

        rx_runtime_trace("########## Master Python Process Begin ##########", RxTraceLevel.INFO)

        _native_call_function = rx_native_call

        guid_dir_on_head_node, in_file_name, fro_data_path, hpa_job_dir \
            = _init_and_check_rx_remote_spark_hpa_call_input(args, _native_call_function)

        input_object, compute_context = \
            _init_and_check_rx_remote_spark_hpa_call_input_data(guid_dir_on_head_node, in_file_name)

        hadoop_workflow_dir_path = rx_spark_get_mrs_hadoop_workflow_path()

        input_object_path, final_iro_path, hdfs_share_dir_path, \
            hdfs_share_dir_path_with_prefix, cmd_to_launch_job \
            = _prepare_rx_remote_spark_hpa_call_job(guid_dir_on_head_node,
                                                    hpa_job_dir,
                                                    input_object,
                                                    compute_context,
                                                    hadoop_workflow_dir_path,
                                                    _native_call_function)

        return_object = _execute_rx_remote_spark_hpa_call_job(input_object, input_object_path,
                                                              final_iro_path, hdfs_share_dir_path,
                                                              cmd_to_launch_job, _native_call_function)
        # TODO: rxMDSLog for logging HDI error metric

        _done_rx_remote_spark_hpa_call(fro_data_path, return_object)

    except Exception as e:
        sys.stderr.write("\nMaster Python Process Error:\n{0}\n".format(str(e)))
        return_object = e
    finally:
        # cleanup logic, no need to fail if fail to cleanup.
        # also swallow any stdout and stderr
        try:
            with open(os.devnull, 'w') as FNULL:
                hadoop_cmd = " ".join([RxOptions.get_option("mrsHadoopPath"),
                                       "fs -rm -r -skipTrash",
                                       hdfs_share_dir_path_with_prefix])
                subprocess.call(hadoop_cmd.split(), stdout=FNULL, stderr=FNULL)
        except Exception:
            pass
        # TODO: implement rxDeleteHelperGuidDirFromHdfs to clean hdfs share dir path. It is for tree algorithms
        rx_runtime_trace("########## Master Python Process End ##########", RxTraceLevel.INFO)

    if isinstance(return_object, Exception):
        exit(1)


#################################################################################
#
# parses rx_remote_spark_call command line arguments
#
def _parse_rx_remote_spark_hpa_call_arguments() -> dict:
    parser = argparse.ArgumentParser()

    parser.add_argument("-rxNetworkShareJobDirectory")
    parser.add_argument("-rxJobPyData")
    parser.add_argument("-rxFroDataPath")
    parser.add_argument("-rxHPAJobDir")
    parser.add_argument("-rxTraceLevel")

    return vars(parser.parse_args())


#################################################################################
#
# dump and load transform function
#
def _dump_transform_function(transform_function: callable, work_dir: str):
    rx_runtime_trace("work dir: " + work_dir, RxTraceLevel.DEBUG)
    _transform_func = _resolve_transform_function(transform_function)
    _transform_func_path = os.path.join(work_dir, "transform_func.dill")
    with open(_transform_func_path, "wb") as file:
        dill.dump(_transform_func, file)


def load_transform_function(work_dir: str) -> callable:
    rx_runtime_trace("work dir: " + work_dir, RxTraceLevel.DEBUG)
    transform_func_path = os.path.join(work_dir, "transform_func.dill")
    if os.path.isfile(transform_func_path):
        with open(transform_func_path, "rb") as file:
            return build_transform(dill.load(file))
    return None


def _init_and_check_rx_remote_spark_hpa_call_input(args: dict, _native_call_function):
    if len(args) == 0:
        raise Exception("No input data for rx_remote_spark_call."
                        " Should provide -rxCommType, -rxNetworkShareJobDirectory, -rxJobPyData,"
                        " -rxFroDataPath and -rxTraceLevel as input value.")
    if "rxNetworkShareJobDirectory" in args:
        guid_dir_on_head_node = args["rxNetworkShareJobDirectory"]
    else:
        raise Exception("missing rxNetworkShareJobDirectory")

    if "rxJobPyData" in args:
        in_file_name = args["rxJobPyData"]
    else:
        raise Exception("Hpa job need to specify rxJobPyData")

    if "rxFroDataPath" in args:
        fro_data_path = args["rxFroDataPath"]
    else:
        raise Exception("Hpa job need to specify rxFroDataPath")

    if "rxHPAJobDir" in args:
        hpa_job_dir = args["rxHPAJobDir"]
    else:
        raise Exception("Hpa job need to specify rxHPAJobDir")

    if "rxTraceLevel" in args:
        RxOptions.set_option("traceLevel", int(args["rxTraceLevel"]))

    if RxOptions.get_option("traceLevel", 0) > 0:
        _native_call_function(functionname="Util", params={
            "USE": "InitializeLogging",
            "traceLevel": RxOptions.get_option("traceLevel"),
            "processType": "MASTER",
            "logDir": guid_dir_on_head_node
        })
    return guid_dir_on_head_node, in_file_name, fro_data_path, hpa_job_dir


def _init_and_check_rx_remote_spark_hpa_call_input_data(guid_dir_on_head_node, in_file_name):
    with open(os.path.join(guid_dir_on_head_node, in_file_name), "rb") as in_file:
        input_object = dill.load(in_file)

    packages_to_load = input_object.get("packagesToLoad")
    if packages_to_load and packages_to_load is list:
        for package in packages_to_load:
            __import__(package)

    args_input_object = input_object.get("ARGS")
    if not args_input_object:
        raise ValueError("Invalid format of input_object. Missing ARGS field in file.")

    compute_context = args_input_object.get("compute_context")
    if not compute_context or not isinstance(compute_context, RxSpark):
        raise Exception("Field computeContext in ARGS from input_object file should be RxSpark")

    return input_object, compute_context


def _prepare_rx_remote_spark_hpa_call_job(guid_dir_on_head_node, hpa_job_dir, input_object, compute_context,
                                          hadoop_workflow_dir_path, _native_call_function):
    name_node = compute_context.name_node
    port = compute_context.port
    # TODO: make create guid a util func
    guid = str(_native_call_function(functionname="CreateGUID", params={}))
    hdfs_share_dir_path = os.path.join(compute_context.hdfs_share_dir, guid)
    in_data = _extract_input_data(input_object)
    in_data_file = _extract_data_file(in_data)
    out_data = _extract_output_data(input_object)

    use_rx_spark_data_output = isinstance(out_data, RxSparkData)
    if use_rx_spark_data_output \
            and not input_object.get("FUN") \
            in ["rx_import", "rx_data_step"]:
        raise Exception("This function does not support " +
                        "RxHiveData/RxParquetData/RxOrcData/RxSparkDataFrame as output data source.")

    if "://" in name_node:
        hdfs_share_dir_path_with_prefix = name_node.strip() + hdfs_share_dir_path
    else:
        hdfs_share_dir_path_with_prefix = hdfs_share_dir_path

    input_object_path = os.path.join(hpa_job_dir, ".input")
    final_iro_path = os.path.join(hpa_job_dir, "IRO.iro")

    is_input_xdf = isinstance(in_data, RxXdfData) or isinstance(in_data, DataFrame)
    in_xdf_uuid = "00000000000000000000000000000000"
    if isinstance(in_data, RxXdfData) or isinstance(in_data, RxSparkData):
        in_xdf_uuid = in_data.xdf_uuid

    out_xdf_uuid = "00000000000000000000000000000000"
    if use_rx_spark_data_output:
        out_xdf_uuid = out_data.xdf_uuid

    fifo_to_spark = _build_fifo_path(compute_context, "RevoSparkFifoToSpark")
    fifo_from_spark = _build_fifo_path(compute_context, "RevoSparkFifoFromSpark")

    # prepare transform function
    if "transform_function" in input_object["ARGS"] and callable(input_object["ARGS"]['transform_function']):
        _dump_transform_function(input_object["ARGS"]['transform_function'], hpa_job_dir)

    cmd_to_launch_job = " ".join([
        os.path.join(hadoop_workflow_dir_path, "common", "RevoSparkHPALauncher"),
        fifo_to_spark,
        fifo_from_spark,
        str(is_input_xdf).lower(),
        input_object_path,
        in_data_file,
        final_iro_path,
        compute_context.hadoop_python_path,
        name_node,
        str(port),
        input_object.get("FUN"),
        "spark-no-header-xUwb1",
        "spark-create-xdf-rdd",
        hdfs_share_dir_path,
        compute_context.user,
        compute_context.session_pid,
        compute_context.app_id,
        compute_context.job_id,
        "true",  # python does not support non-wait mode, so always true here
        str(compute_context.idle_timeout),
        str(compute_context.executor_cores),
        compute_context.executor_mem,
        compute_context.driver_mem,
        str(rx_spark_count_mb(compute_context.executor_overhead_mem)),
        str(compute_context.num_executors),
        guid_dir_on_head_node,
        compute_context.spark_reduce_method,
        in_xdf_uuid,
        str(use_rx_spark_data_output).lower(),
        out_xdf_uuid,
        "outDataPathPlaceHolder",
        str(RxOptions.get_option("traceLevel", default=0)),
        "true",
        compute_context.tmp_fs_work_dir,
        "".join(['"', compute_context.extra_spark_config.replace("\"", "\\\\\""), '"'])
    ])
    return input_object_path, final_iro_path, hdfs_share_dir_path,\
        hdfs_share_dir_path_with_prefix, cmd_to_launch_job


def rx_spark_get_mrs_hadoop_workflow_path():
    command = ["/usr/bin/mrs-hadoop-job", "path"]
    out_message = rx_spark_execute_command(command)
    if len(out_message) != 1:
        raise ValueError("'mrs-hadoop-job path' doesn't return correctly")

    return os.path.join(os.path.dirname(out_message[0].strip()), "..")


def _execute_rx_remote_spark_hpa_call_job(input_object, input_object_path,
                                          final_iro_path, hdfs_share_dir_path,
                                          cmd_to_launch_job, _native_call_function):
    rx_runtime_trace("start", RxTraceLevel.INFO)

    rx_runtime_trace("Init Hadoop", RxTraceLevel.INFO)
    _native_call_function(functionname="Util", params={
        "USE": "InitializeHadoop",
        "inputObjectPath": input_object_path,
        "inputInHdfs": False,
        "outputObjectPath": final_iro_path,
        "outputInHdfs": False,
        "hdfsShareDirPath": hdfs_share_dir_path,
        "systemCommand": cmd_to_launch_job,
        "runningInSpark": True,
        "callFromHpaFunMasterExecutor": True
    })

    __import__(input_object["MODULE"])
    mod = sys.modules[input_object["MODULE"]]
    fun = getattr(mod, input_object["FUN"])

    # TODO: handle mds and rxTelemetryPause()
    # TODO: turn HDI telemetry off for hadoop to avoid duplicate value
    # options(mds.inHadoop = TRUE)

    # remove compute_context from ARGS for call back
    # this is to make sure fun runs in local compute context
    rx_runtime_trace("eval fun(**args)", RxTraceLevel.INFO)
    input_object["ARGS"].pop("compute_context", None)
    return_object = fun(**input_object["ARGS"])

    # TODO: turn HDI telemetry on for hadoop

    rx_runtime_trace("Uninit Hadoop", RxTraceLevel.INFO)
    _native_call_function(functionname="Util", params={"USE": "UninitializeHadoop"})

    rx_runtime_trace("finish", RxTraceLevel.INFO)

    return return_object


def _done_rx_remote_spark_hpa_call(fro_data_path, return_object):
    out_file = open(fro_data_path, 'wb')
    dill.dump(return_object, out_file)
    out_file.close()


def _extract_input_data(input_object):
    input_data = input_object["ARGS"].get("data")
    if not input_data:
        input_data = input_object["ARGS"].get("input_data")
        if not input_data:
            input_data = input_object["ARGS"].get("file")
    return input_data


def _extract_output_data(input_object):
    output_data = input_object["ARGS"].get("output_file")
    if not output_data:
        output_data = input_object["ARGS"].get("output_data")
    return output_data


def _extract_data_file(data) -> str:
    if isinstance(data, DataFrame):
        raise TypeError("DataFrame is not supported data source in RxSpark compute context")
    elif issubclass(type(data), RxSparkData):
        return data.df_name
    else:
        return data.file


def _build_fifo_path(compute_context, base_name) -> str:
    return os.path.join(
        compute_context.share_dir,
        "-".join([
            base_name,
            compute_context.user,
            compute_context.session_pid,
            compute_context.app_id]))


def rx_spark_count_mb(mem_size):
    mem_size = str.lower(mem_size)
    if "m" in mem_size:
        return int(mem_size.replace("m", ""))
    if "g" in mem_size:
        return 1024 * int(mem_size.replace("g", ""))


def rx_spark_get_spark_env_metadata(compute_context: RxSpark):
    return {
        "envVersion": "0.1",
        "envType": "spark",
        "revoUtilsDir": "utils",
        "mrsHadoopPath": "/usr/bin/mrs-hadoop",
        "hadoopPyPath": compute_context.hadoop_python_path,
        "shareDir": compute_context.share_dir,
        "user": compute_context.user,
        "pySessionPid": compute_context.session_pid,
        "appId": compute_context.app_id,
        "wait": "true",
        "nameNode": compute_context.name_node,
        "port": compute_context.port,
        "idleTimeout": compute_context.idle_timeout,
        "executorCores": compute_context.executor_cores,
        "executorMem": compute_context.executor_mem,
        "driverMem": compute_context.driver_mem,
        "overheadMem": compute_context.executor_overhead_mem,
        "numExecutors": compute_context.num_executors,
        # TODO: review skipLaunchSpark cases
        "skipLaunchSpark": str(compute_context.pyspark_connection is not None).lower(),
        "extraConfig": compute_context.extra_spark_config,
        "loggingLevel": RxOptions.get_option("traceLevel", 0),
        "persistentRun": str(compute_context.persistent_run).lower(),
        "tmpFSWorkDir": compute_context.tmp_fs_work_dir
    }
