#
# Copyright (c) Microsoft Corporation.  All rights reserved.
#


import getpass
import os
import warnings
from typing import Union

from ..utils.RxOptions import RxOptions
from .RxComputeContext import RxComputeContext
from .RxSpark import RxSpark
from .RxPySparkBase import rx_pyspark_check_pyspark_is_loaded
from .RxPySparkBase import rx_pyspark_get_existing_pyspark_app
from .RxPySparkBase import rx_pyspark_check_pyspark_sc_master_type
from .RxSparkApiBase import rx_spark_get_running_spark_app
from .RxSparkApiBase import rx_spark_stop_engine
from .RxSparkApiBase import rx_spark_register_pyspark_interop_termination_hook
from .RxSparkApiBase import rx_spark_create_pyspark_interop
from .RxSparkApiBase import rx_spark_link_pyspark_interop
from .RxSparkApiBase import rx_spark_start_engine
from .RxSparkApiBase import rx_spark_register_spark_interpreter_termination_hook
from .RxSparkApiBase import rx_spark_connect_pyspark_interop_base
from .RxSparkApiBase import rx_spark_connect_base
from .RxSparkApiBase import rx_spark_disconnect_base
from .RxSparkApiBase import rx_spark_stop_pyspark_interop_spark_app


# For any dynamic default values from RxOptions. Cannot directly assign in signature.
# Otherwise, the default value will be decided at the time of loading revoscalepy.
# Then the function will not respect any updated RxOptions values.
def rx_spark_connect(hdfs_share_dir: str = os.path.join("/user/RevoShare", getpass.getuser()),
                     share_dir: str = os.path.join("/var/RevoShare", getpass.getuser()),
                     user: str = getpass.getuser(),
                     name_node: str = None,
                     port: int = None,
                     auto_cleanup: bool = True,
                     console_output: bool = False,
                     packages_to_load: list = None,
                     idle_timeout: int = None,
                     num_executors: int = None,
                     executor_cores: int = None,
                     executor_mem: str = None,
                     driver_mem: str = None,
                     executor_overhead_mem: str = None,
                     extra_spark_config: str = "",
                     spark_reduce_method: str = "auto",
                     tmp_fs_work_dir: str = "/dev/shm",
                     interop: Union[list, str] = None,
                     reset: bool = False) -> RxSpark:
    """
    .. rxtitle::

        Connect revoscalepy to remote Spark application.

    .. rxdescription::

        Creates the compute context object with RxSpark and then immediately starts the remote Spark application.

    :param hdfs_share_dir: Character string specifying the file sharing location within HDFS.
        You must have permissions to read and write to this location.

    :param share_dir: Character string specifying the directory on the master (perhaps edge) node that is
        shared among all the nodes of the cluster and any client host. You must have permissions to read and write
        in this directory.

    :param user: Character string specifying the username for submitting job to the
        Spark cluster. Defaults to the username of the user running the python (that is, the value of getpass.getuser()).

    :param name_node: Character string specifying the Spark name node hostname or IP address.
        Typically you can leave this at its default value.
        If set to a value other than "default" or the empty string (see below),
        this must be an address that can be resolved by the data nodes and used by them to contact the
        name node. Depending on your cluster, it may need to be set to a private network address
        such as "master.local". If set to the empty string, "", then the master process will set
        this to the name of the node on which it is running, as returned by platform.node().
        Defaults to RxOptions.get_option("hdfsHost").

    :param port: Integer value specifying the port used by the name node for HDFS. Needs
        to be able to be cast to an integer. Defaults to RxOptions.get_option("hdfsPort").

    :param auto_cleanup: Bool value, if True, the default behavior is to clean up the
        temporary computational artifacts and delete the result objects upon retrieval.  If False,
        then the computational results are not deleted. Leaving this
        flag set to False can result in accumulation of compute artifacts which you may
        eventually need to delete before they fill up your hard drive.

    :param console_output: Bool value, if True, causes the standard output
        of the python processes to be printed to the user console.

    :param packages_to_load: Optional character vector specifying additional modules to be
        loaded on the nodes when jobs are run in this compute context.

    :param idle_timeout: Integer value, specifying the number of seconds of being idle
        (i.e., not running any Spark job) before system kills the Spark process.
        Setting a value greater than 600 is recommended. Defaults to 3600.
        Setting a negative value will not timeout. pyspark interoperation will have no timeout.

    :param num_executors: Integer value, specifying the number of executors in Spark, which is
        equivalent to parameter --num-executors in spark-submit app. If not specified,
        the default behavior is to launch as many executors as possible,
        which may use up all resources and prevent other users from sharing the cluster.
        Defaults to RxOptions.get_option("spark.numExecutors").

    :param executor_cores: Integer value, specifying the number of cores per executor, which is
        equivalent to parameter --executor-cores in spark-submit app.
        Defaults to RxOptions.get_option("spark.executorCores").

    :param executor_mem: Character string specifying memory per executor (e.g., 1000M, 2G), which is
        equivalent to parameter --executor-memory in spark-submit app.
        Defaults to RxOptions.get_option("spark.executorMem").

    :param driver_mem: Character string specifying memory for driver (e.g., 1000M, 2G), which is
        equivalent to parameter --driver-memory in spark-submit app.
        Defaults to RxOptions.get_option("spark.driverMem").

    :param executor_overhead_mem: Character string specifying memory overhead per executor (e.g. 1000M, 2G), which is
        equivalent to setting spark.yarn.executor.memoryOverhead in YARN.
        Increasing this value will allocate more memory for the python process and the ScaleR engine process
        in the YARN executors, so it may help resolve job failure or executor lost issues.
        Defaults to RxOptions.get_option("spark.executorOverheadMem").

    :param extra_spark_config: Character string specifying extra arbitrary Spark properties
        (e.g., "--conf spark.speculation=true --conf spark.yarn.queue=aQueue"),
        which is equivalent to additional parameters passed into spark-submit app.

    :param spark_reduce_method: Spark job collects all parallel tasks' results to reduce as final result.
        This parameter decides reduce strategy: oneStage/twoStage/auto.
        oneStage: reduce parallel tasks to one result with one reduce function;
        twoStage: reduce parallel tasks to square root size with the first reduce function,
        then reduce to final result with the second reduce function;
        auto: spark will smartly select oneStage or twoStage.

    :param tmp_fs_work_dir: Character string specifying the temporary working directory in worker nodes.
        It defaults to /dev/shm to utilize in-memory temporary file system for performance gain.
        You can specify a different location if the size of /dev/shm is insufficient.
        Please make sure that YARN run-as user has permission to read and write to this location

    :param interop: None or a character string or list of interop names for revoscalepy interoperation.
        Currently, the "pyspark" package is supported.

    :param reset: Bool value, if True, all cached Spark Data Frames will be freed and all existing Spark applications
        that belong to the current user will be shut down.

    .. rxexample::

        #############################################################################
        from revoscalepy import rx_spark_connect, rx_spark_disconnect
        cc = rx_spark_connect()
        rx_spark_disconnect(cc)

        #############################################################################
        ####### pyspark interoperation example ####################
        from pyspark.sql import SparkSession
        from pyspark import SparkContext
        from revoscalepy import rx_spark_connect, rx_get_pyspark_connection, RxSparkDataFrame, rx_summary, rx_data_step
        spark = SparkSession.builder.master("yarn").getOrCreate()
        cc = rx_spark_connect(interop = "pyspark")

        # sc is equivalent to result of pyspark API call SparkContext.getOrCreate()
        sc = rx_get_pyspark_connection(cc)

        ####### algorithms with input from spark data frame #######
        df = spark.createDataFrame([('Monday',3012),('Friday',1354),('Sunday',5452)], ["DayOfWeek", "Sale"])
        input_df = RxSparkDataFrame(df)
        rx_summary("~.", input_df)

        ####### ETL with output to spark data frame ###############
        output_df = RxSparkDataFrame()
        rx_data_step(input_df, output_df, vars_to_drop = ['DayOfWeek'])
        odf = output_df.spark_data_frame
        odf.take(2)
        # [Row(Sale=3012), Row(Sale=1354)]
    """
    pyspark_interop = (isinstance(interop, list) and "pyspark" in interop) or ("pyspark" is interop)
    if name_node is None:
        name_node = RxOptions.get_option("hdfsHost")
    if port is None:
        port = RxOptions.get_option("hdfsPort")
    if num_executors is None:
        num_executors = RxOptions.get_option("spark.numExecutors")
    if executor_cores is None:
        executor_cores = RxOptions.get_option("spark.executorCores")
    if executor_mem is None:
        executor_mem = RxOptions.get_option("spark.executorMem")
    if driver_mem is None:
        driver_mem = RxOptions.get_option("spark.driverMem")
    if executor_overhead_mem is None:
        executor_overhead_mem = RxOptions.get_option("spark.executorOverheadMem")
    if idle_timeout is None:
        if pyspark_interop:
            idle_timeout = -1
        else:
            idle_timeout = 3600
    else:
        if pyspark_interop and idle_timeout > -1:
            warnings.warn("'idle_timeout' needs to be negative int when pyspark interop is on. Set to -1.")
            idle_timeout = -1
    args = locals()
    args.pop("reset")
    args["persistent_run"] = True
    args["from_spark_connect"] = True

    compute_context = RxSpark(**args)
    existing_pyspark_sc = None
    if rx_pyspark_check_pyspark_is_loaded():
        existing_pyspark_sc = rx_pyspark_get_existing_pyspark_app()
    if pyspark_interop:
        if not rx_pyspark_check_pyspark_is_loaded():
            raise Exception("interop = 'pyspark', but the 'pyspark' module has not been imported. " +
                            "Please import 'pyspark' module.")
        existing_sc = rx_pyspark_get_existing_pyspark_app()
        compute_context = rx_spark_connect_pyspark_interop_base(compute_context,
                                                                existing_sc,
                                                                reset,
                                                                rx_spark_get_running_spark_app,
                                                                rx_spark_stop_engine,
                                                                rx_spark_register_pyspark_interop_termination_hook,
                                                                rx_spark_create_pyspark_interop,
                                                                rx_spark_link_pyspark_interop,
                                                                rx_pyspark_check_pyspark_sc_master_type)
    else:
        compute_context = rx_spark_connect_base(compute_context, existing_pyspark_sc, reset,
                                                rx_spark_get_running_spark_app,
                                                rx_spark_stop_engine,
                                                rx_spark_start_engine,
                                                rx_spark_register_spark_interpreter_termination_hook)
    RxComputeContext.set_current(compute_context)
    return compute_context


def rx_spark_disconnect(compute_context=None):
    """
    .. rxtitle::

        Disconnect revoscalepy from remote Spark application.

    .. rxdescription::

        Shuts down the remote Spark application and switches to a local compute context.
        All rx* function calls after this will run in a local compute context.
        In pyspark-interop mode, if Spark application is started by pyspark APIs,
        rx_spark_disconnect will not shut down the remote Spark application but disassociate from it.
        Run help(revoscalepy.rx_spark_connect) for more information about interop.

    :param compute_context: Spark compute context to be terminated by rx_spark_disconnect.
        If input is None, then current compute context will be used.

    .. rxexample::

        from revoscalepy import rx_spark_connect, rx_spark_disconnect
        cc = rx_spark_connect()
        rx_spark_disconnect(cc)

    """
    if compute_context is None:
        compute_context = RxComputeContext.current()
    rx_spark_disconnect_base(compute_context, rx_spark_stop_engine, rx_spark_stop_pyspark_interop_spark_app)


def rx_get_pyspark_connection(compute_context: RxSpark):
    """
    .. rxtitle::

        Get pyspark spark_context connection from Spark compute context

    .. rxdescription::

        Get a Spark compute context with pyspark interop.
        rx_get_pyspark_connection gets pyspark spark connection from created Spark compute context.

    :param compute_context: Compute context get created by rx_spark_connect.

    :return: Object of python.context.SparkContext.

    .. rxexample::
        from revoscalepy import rx_spark_connect, rx_get_pyspark_connection
        from pyspark.sql import SparkSession
        cc = rx_spark_connect(interop = "pyspark")
        sc = rx_get_pyspark_connection(cc)
        spark = SparkSession(sc)
        df = spark.createDataFrame([('Monday',3012),('Friday',1354),('Sunday',5452)], ["DayOfWeek", "Sale"])

    """
    if not isinstance(compute_context, RxSpark):
        raise Exception("rx_get_pyspark_connection is called with %s " % type(compute_context).__name__ +
                        "object. It is only applicable for RxSpark compute context.")
    if compute_context.pyspark_connection is None:
        raise Exception("No existing pyspark connection in this compute context. " +
                        "To use pyspark interoperation, please set interop='pyspark' in rx_spark_connect.")
    return compute_context.pyspark_connection


