#
# Copyright (c) Microsoft Corporation.  All rights reserved.
#


import os
import warnings
import shutil
import glob
from .RxComputeContext import RxComputeContext
from .RxPySparkBase import rx_pyspark_check_pyspark_is_loaded
from .RxPySparkBase import rx_pyspark_get_existing_pyspark_app
from .RxPySparkBase import get_revoscalepy_spark_jar_path
from .RxSpark import RxSpark
from .RxSpark import RxSparkBuilder
from .RxLocalSeq import RxLocalSeq
from .RxSpark import rx_spark_count_mb
from .RxSpark import rx_spark_get_mrs_hadoop_workflow_path
from .RxSpark import rx_spark_execute_command
from .RxSpark import rx_spark_get_spark_env_metadata
from ..utils.RxOptions import RxOptions


def _parse_spark_extra_config(extra_spark_config):
    extra_spark_config_list = extra_spark_config.strip().split("--conf")
    result = list()
    if len(extra_spark_config_list) > 1:
        # skip the first element, since it will be empty string or invalid string
        for pair_raw in extra_spark_config_list[1:]:
            pair_raw = pair_raw.strip()
            if pair_raw.startswith("\"") and pair_raw.endswith("\""):
                pair_raw = pair_raw.strip("\"")
            if pair_raw.startswith("\'") and pair_raw.endswith("\'"):
                pair_raw = pair_raw.strip("\'")
            pair = pair_raw.split("=")
            if len(pair) == 2:
                key = pair[0].strip()
                value = pair[1].strip()
                result.append((key, value))
            else:
                warnings.warn("Ignore invalid spark extra config: %s" % pair_raw)
    return result


def _fetch_spark_basic_config(compute_context: RxSpark):
    result = list()
    result.append(("spark.executor.cores", str(compute_context.executor_cores)))
    result.append(("spark.executor.instances", str(compute_context.num_executors)))
    result.append(("spark.driver.memory", compute_context.driver_mem))
    result.append(("spark.executor.memory", compute_context.executor_mem))
    result.append(("spark.yarn.executor.memoryOverhead", str(rx_spark_count_mb(compute_context.executor_overhead_mem))))
    return result


def _create_pyspark_conf(compute_context: RxSpark):
    # Not to move import to header. pyspark is optional module
    from pyspark import SparkConf
    user_session_app = compute_context.user + '-' + compute_context.session_pid + '-' + compute_context.app_id
    spark_shell_id = 'pyspark-scaleR-spark-won2r0FHOA-' + user_session_app
    conf = SparkConf()
    revoscalepy_spark_jar_path = get_revoscalepy_spark_jar_path()
    if conf.get('spark.jars') is not None:
        if "scaler-spark_2.11-0.1.0.jar" not in conf.get("spark.jars"):
            conf.set("spark.jars", conf.get("spark.jars") + "," + revoscalepy_spark_jar_path)
    else:
        conf.set("spark.jars", revoscalepy_spark_jar_path)
    conf.setMaster("yarn")
    conf.setAppName(spark_shell_id)
    basic_spark_config_pair_list = _fetch_spark_basic_config(compute_context)
    for pair in basic_spark_config_pair_list:
        conf.set(pair[0], pair[1])
    extra_spark_config_pair_list = _parse_spark_extra_config(compute_context.extra_spark_config)
    for pair in extra_spark_config_pair_list:
        conf.set(pair[0], pair[1])
    return conf


def rx_spark_create_pyspark_interop(compute_context: RxSpark):
    # Not to move import to header. pyspark is optional module
    from pyspark import SparkContext
    from pyspark.sql import SparkSession
    from pyspark.sql import SQLContext
    conf = _create_pyspark_conf(compute_context)
    # build new sc, session, sqlContext and update the default values in pyspark
    new_sc = SparkContext(conf=conf)
    SparkContext._active_spark_context = new_sc
    new_ss = SparkSession(new_sc)
    if getattr(SparkSession, "_instantiatedSession", None) is not None:
        SparkSession._instantiatedSession = new_ss
    elif getattr(SparkSession, "_instantiatedContext", None) is not None:
        # spark 2.0.0, still use _instantiatedContext, but later change to _instantiatedSession
        SparkSession._instantiatedContext = new_ss
    # pyspark bug, have to set the default value to None first.
    # otherwise SQLContext(new_sc) will use the cached _instantiatedContext instead
    SQLContext._instantiatedContext = None
    SQLContext._instantiatedContext = SQLContext(new_sc, new_ss)
    
    return rx_spark_link_pyspark_interop(compute_context=compute_context, pyspark_sc=new_sc,
                                         hadoop_workflow_dir_fetch_fun=rx_spark_get_mrs_hadoop_workflow_path)


def rx_spark_link_pyspark_interop(compute_context: RxSpark, pyspark_sc, hadoop_workflow_dir_fetch_fun):
    # update compute_context with settings from pyspark_sc
    # NOTICE: pyspark interop need to set idle timeout to -1 (infinite)
    #         We should keep the spark application running
    compute_context = RxSparkBuilder().build_from(compute_context).set_idle_timeout(-1)\
        .set_persistent_run(True).set_wait(True).set_pyspark_connection(pyspark_sc)\
        .set_additional_conf("from_spark_connect", True).create()
    user_session_app = compute_context.user + '-' + compute_context.session_pid + '-' + compute_context.app_id
    spark_shell_id = 'pyspark-scaleR-spark-won2r0FHOA-' + user_session_app
    fifo_to_spark = compute_context.share_dir + '/RevoSparkFifoToSpark-' + user_session_app
    fifo_from_spark = compute_context.share_dir + '/RevoSparkFifoFromSpark-' + user_session_app
    _check_head_node_share_dir_exist(compute_context.share_dir)
    _create_fifo(fifo_to_spark)
    _create_fifo(fifo_from_spark)
    arglist = ['--sparkShellId', spark_shell_id,
               '--fifoFromSpark', fifo_from_spark,
               '--fifoToSpark', fifo_to_spark,
               '--interpreterPath', compute_context.hadoop_python_path,
               '--hadoopDirPath', os.path.join(hadoop_workflow_dir_fetch_fun(), ".."),
               '--nameNode', compute_context.name_node,
               '--nameNodePort', str(compute_context.port),
               '--idleTimeout', str(compute_context.idle_timeout),
               '--debug', 'false',
               '--user', compute_context.user,
               '--loggingLevel', str(RxOptions.get_option("traceLevel", default=0)),
               '--tmpFSWorkDir', compute_context.tmp_fs_work_dir,
               '--serverType', 'Python'
               ]
    try:
        pyspark_sc._jvm.com.microsoft.scaler.spark.RREServer.startService(arglist, pyspark_sc._jsc, None)
    except TypeError as e:
        if "'JavaPackage' object is not callable" in str(e):
            jar_path = get_revoscalepy_spark_jar_path()
            raise Exception("%s is not loaded in Spark by setting config spark.jars. " % jar_path + 
                            "Please call rx_spark_connect with reset=True to launch a " +
                            "new spark application or explicitly set spark.jar=%s in SparkConf." % jar_path)
        raise e
    return compute_context


def rx_spark_connect_pyspark_interop_base(compute_context: RxSpark, existing_pyspark_sc, reset: bool,
                                          list_app_fun, stop_app_fun, register_cleanup_hook_fun,
                                          create_pyspark_interop_fun, link_pyspark_interop_fun,
                                          check_pyspark_sc_master_type_fun):
    if reset:
        running_apps = list_app_fun(compute_context, "user")
        if len(running_apps) > 0:
            print("Parameter 'reset' is set to True. Shutting down existing Spark applications ( %s )" %
                  ",".join(running_apps))
            stop_app_fun(compute_context, "user")
    else:
        running_apps = list_app_fun(compute_context, "session")
        if len(running_apps) > 0:
            raise Exception("Parameter 'reset' is set to False. " +
                            "The existing revoscalepy Spark backend application ( %s ) " % (",".join(running_apps)) +
                            "is running. Please set 'reset' to True for switching revoscalepy Spark backend "
                            "to pyspark backend, or call rx_spark_disconnect to shut down the spark application.")
    if existing_pyspark_sc is not None:
        # if has pyspark application is running
        if reset:
            print("Parameter 'reset' is set to True. Shutting down existing pyspark backend application ( %s )." %
                  existing_pyspark_sc.applicationId)
            existing_pyspark_sc.stop()
            print("It may take 1 to 2 minutes to launch a new Spark application.")
            final_compute_context = create_pyspark_interop_fun(compute_context)
        else:
            check_pyspark_sc_master_type_fun(existing_pyspark_sc)
            attached_spark_cc = RxOptions.get_option("attachedSparkCC")
            # if pyspark_sc_match_revoscalepy_cc is false, then that means the current pyspark does not match
            # with the one in attached_spark_cc.
            # if pyspark_sc_match_revoscalepy_cc is true, then the current pyspark should be the same as the one
            # in attached_spark_cc. (only one pyspark sc can run at the same time)
            pyspark_sc_match_revoscalepy_cc = attached_spark_cc is not None and \
                attached_spark_cc.pyspark_connection is not None \
                and attached_spark_cc.pyspark_connection._jsc is not None
            if pyspark_sc_match_revoscalepy_cc:
                print("Parameter 'reset' is set to False. " +
                      "Try to reuse the existing pyspark backend application ( %s )" %
                      existing_pyspark_sc.applicationId)
                final_compute_context = attached_spark_cc
            else:
                print("Trying to link revoscalepy to the existing pyspark backend application ( %s )."
                      % existing_pyspark_sc.applicationId)
                final_compute_context = link_pyspark_interop_fun(compute_context,
                                                                 existing_pyspark_sc,
                                                                 rx_spark_get_mrs_hadoop_workflow_path)
    else:
        print("It may take 1 to 2 minutes to launch a new Spark application.")
        final_compute_context = create_pyspark_interop_fun(compute_context)
    register_cleanup_hook_fun(final_compute_context)
    RxComputeContext.set_current(final_compute_context)
    RxOptions.set_option("attachedSparkCC", final_compute_context)
    return final_compute_context


def rx_spark_connect_base(compute_context: RxSpark, existing_pyspark_sc, reset: bool,
                          list_app_fun, stop_app_fun, start_app_fun, register_termination_hook_fun) -> RxSpark:
    attached_spark_cc = RxOptions.get_option("attachedSparkCC")
    # clean old attachment first
    RxOptions.set_option("attachedSparkCC", None)
    if reset:
        if existing_pyspark_sc is not None:
            print("Parameter 'reset' is set to True. Shutting down existing pyspark backend application ( %s )."
                  % existing_pyspark_sc.applicationId)
            existing_pyspark_sc.stop()
        running_apps = list_app_fun(compute_context, "user")
        if len(running_apps) > 0:
            print("Parameter 'reset' is set to True. Shutting down existing Spark applications ( %s )" %
                  ",".join(running_apps))
            stop_app_fun(compute_context, "user")
            print("It may take 1 to 2 minutes to launch a new Spark application.")
        else:
            print("No Spark applications can be reused. It may take 1 to 2 minutes to launch a new Spark application.")
    else:
        if existing_pyspark_sc is not None:
            print("Found existing pyspark backend application ( %s ). " % existing_pyspark_sc.applicationId +
                  "Parameter 'interop' does not contain 'pyspark', cannot reuse this Spark application.")
        running_apps = list_app_fun(compute_context, "session")
        if len(running_apps) > 1:
            raise Exception("This Python session is running more than one Spark application, which is unexpected. " +
                            "Please call rx_spark_disconnect function or set parameter 'reset' to True to clean" +
                            " up your environment first.")
        elif len(running_apps) > 0:
            if attached_spark_cc is None:
                raise Exception("Failed to find the Spark Compute Context which is attached to Spark application ( %s )" %
                                running_apps[0] + ". Please set parameter 'reset' to True to clean up" +
                                " your environment first.")
            print("The `reset` argument is set to False. Try to reuse the existing Spark application ( %s )" %
                  running_apps[0])
            return attached_spark_cc
        else:
            print("No Spark applications can be reused. It may take 1 to 2 minutes to launch a new Spark application.")
    start_app_fun(compute_context, register_termination_hook_fun)
    return compute_context


def rx_spark_disconnect_base(compute_context: RxSpark, stop_app_fun, pyspark_interop_app_stop_fun):
    if not isinstance(compute_context, RxSpark):
        raise Exception("rx_spark_disconnect is called with %s object," % type(compute_context).__name__ +
                        " It is only applicable for RxSpark compute context.")
    stop_app_fun(compute_context, "session")
    spark_is_killed = pyspark_interop_app_stop_fun(compute_context)
    RxOptions.set_option("attachedSparkCC", None)
    if spark_is_killed:
        print("We have shut down the current Spark application and switched to local compute context.")
    else:
        print("We have disconnect revoscalepy from the current Spark application " +
              "and switched to local compute context.")
    RxComputeContext.set_current(RxLocalSeq())


def rx_spark_stop_pyspark_interop_spark_app(compute_context: RxSpark) -> bool:
    spark_is_killed = True
    if rx_pyspark_check_pyspark_is_loaded():
        existing_pyspark_sc = rx_pyspark_get_existing_pyspark_app()
        if existing_pyspark_sc is not None:
            if "pyspark-scaleR-spark-won2r0FHOA" in existing_pyspark_sc.appName:
                # if the application is owned by revoscalepy, then we should kill it
                existing_pyspark_sc.stop()
                from pyspark import SparkContext
                from pyspark.sql import SparkSession
                from pyspark.sql import SQLContext
                # pyspark does not have good cleanup logic for SparkContext, SparkSession, SQLContext
                # in 2.0.0, SparkContext close while SparkSession.builder.getOrCreate() gets the old sc
                # in 2.1.0, SparkContext close while SparkSession.builder.getOrCreate() creates new sc
                # SQLContext.getOrCreate() always uses the cached sc even though it is closed
                # for simplicity, we should cleanup all default values to let 
                # SparkContext.getOrCreate(), SparkSession.builder.getOrCreate(), 
                # SQLContext.getOrCreate(sc) use new sc rather than closed sc
                # clean the default values in pyspark
                SparkContext._active_spark_context = None
                if getattr(SparkSession, "_instantiatedSession", None) is not None:
                    SparkSession._instantiatedSession = None
                elif getattr(SparkSession, "_instantiatedContext", None) is not None:
                    # spark 2.0.0, still use _instantiatedContext, but later change to _instantiatedSession
                    SparkSession._instantiatedContext = None
                SQLContext._instantiatedContext = None
            else:
                # otherwise, we should call into RREServer to stop service but keep the spark application
                try:
                    existing_pyspark_sc._jvm.com.microsoft.scaler.spark.RREServer.stopService()
                    spark_is_killed = False
                except TypeError as e:
                    if "'JavaPackage' object is not callable" in str(e):
                        jar_path = get_revoscalepy_spark_jar_path()
                        raise Exception("%s is not loaded in Spark by setting config spark.jars. " % jar_path + 
                                        "Please call rx_spark_connect with reset=True to launch a " +
                                        "new spark application or explicitly set spark.jar=%s in SparkConf." % jar_path)
                    raise e

        user_session = compute_context.user + '-' + compute_context.session_pid + "*"
        fifo_pattern = os.path.join(compute_context.share_dir, "RevoSparkFifo*-" + user_session)
        _cleanup_session_all_fifo(fifo_pattern)
        # Clean cleanup hook
        global cleanup_pyspark_interop_compute_context
        cleanup_pyspark_interop_compute_context = None
    return spark_is_killed


def rx_spark_get_running_spark_app(compute_context: RxSpark, scope: str = "session",
                                   execute_command: callable = rx_spark_execute_command):
    if scope not in ["session", "user"]:
        raise Exception("scope must be 'session' or 'user'.")
    job = compute_context.create_job()
    job.create_spark_work_dir(compute_context.share_dir)
    job.dump_json(rx_spark_get_spark_env_metadata(compute_context), "env.json")
    job.dump_json({
        "function": "spark_list_v1",
        "jobType": "spark-hpa",
        "serverType": "python",
        "jobGUID": job.id
    }, "job.json")

    try:
        command = job.build_mrs_hadoop_job_command("applist", {"scope": scope})
        raw = execute_command(command)
        result = []
        for line in raw:
            if "SparkApp:" in line:
                result.append(line.strip().lstrip("SparkApp:").rstrip("EndSparkApp"))
        return result
    finally:
        if compute_context.auto_cleanup:
            job.clean_spark_work_dir()


def rx_spark_stop_engine(compute_context: RxSpark, scope: str = "session",
                         execute_command: callable = rx_spark_execute_command):
    if scope not in ["session", "user"]:
        raise Exception("scope must be 'session' or 'user'.")
    job = compute_context.create_job()
    job.create_spark_work_dir(compute_context.share_dir)
    job.dump_json(rx_spark_get_spark_env_metadata(compute_context), "env.json")
    job.dump_json({
        "function": "spark_disconnect_v1",
        "jobType": "spark-hpa",
        "serverType": "python",
        "jobGUID": job.id
    }, "job.json")

    try:
        command = job.build_mrs_hadoop_job_command("disconnect", {"scope": scope})
        execute_command(command)
        # Clean shutdown hook
        global termination_spark_compute_context
        termination_spark_compute_context = None
    finally:
        if compute_context.auto_cleanup:
            job.clean_spark_work_dir()


def rx_spark_start_engine(compute_context: RxSpark, register_termination_hook_fun,
                          execute_command: callable = rx_spark_execute_command):
    job = compute_context.create_job()
    job.create_spark_work_dir(compute_context.share_dir)
    job.dump_json(rx_spark_get_spark_env_metadata(compute_context), "env.json")
    job.dump_json({
        "function": "spark_connect_v1",
        "jobType": "spark-hpa",
        "serverType": "python",
        "jobGUID": job.id
    }, "job.json")

    try:
        command = job.build_mrs_hadoop_job_command("connect", {})
        execute_command(command)
        register_termination_hook_fun(compute_context)
    finally:
        if compute_context.auto_cleanup:
            job.clean_spark_work_dir()


def rx_spark_register_spark_interpreter_termination_hook(compute_context: RxSpark):
    # register shutdown hook
    def termination_hook():
        global termination_spark_compute_context
        if isinstance(termination_spark_compute_context, RxSpark):
            rx_spark_stop_engine(termination_spark_compute_context, "session")
            termination_spark_compute_context = None

    global termination_spark_compute_context
    termination_spark_compute_context = compute_context
    import atexit
    atexit.register(termination_hook)


def rx_spark_register_pyspark_interop_termination_hook(compute_context: RxSpark):
    # register cleanup hook
    def revoscalepy_spark_fifo_cleanup_hook():
        global cleanup_pyspark_interop_compute_context
        if isinstance(cleanup_pyspark_interop_compute_context, RxSpark):
            user_session = compute_context.user + '-' + compute_context.session_pid + "*"
            fifo_pattern = os.path.join(compute_context.share_dir, "RevoSparkFifo*-" + user_session)
            _cleanup_session_all_fifo(fifo_pattern)
            cleanup_pyspark_interop_compute_context = None

    global cleanup_pyspark_interop_compute_context
    cleanup_pyspark_interop_compute_context = compute_context
    import atexit
    atexit.register(revoscalepy_spark_fifo_cleanup_hook)


def _check_head_node_share_dir_exist(share_dir: str):
    if not os.path.exists(share_dir):
        try:
            os.makedirs(share_dir)
        except PermissionError:
            raise Exception("Failed to create share dir (%s) or the current user (%s) " % (share_dir, os.getlogin()) +
                            "has no permission to access the share dir. " +
                            "You should either contact your administrator " +
                            "to have it created, or specify another directory as directed by your administrator. "
                            "Or use a directory such as $HOME/Revo with access " +
                            "permission for the 'share_dir' location.")


def _create_fifo(path):
    if os.path.isdir(path):
        shutil.rmtree(path)
    if os.path.exists(path):
        os.remove(path)
    os.mkfifo(path)


def _cleanup_session_all_fifo(fifo_pattern):
    try:
        for file_path in glob.glob(fifo_pattern):
            os.remove(file_path)
    except Exception:
        # fifo cleanup should not fail the job
        pass

