import dill
import json
import os
import shutil
import uuid

from . RxRemoteJob import RxRemoteJob, RxComputeContext
from .. datasource.RxDataSource import RxDataSource
from .. utils.RxOptions import RxOptions


class RxSparkRemoteJob(RxRemoteJob):
    def __init__(self, compute_context: RxComputeContext, job_id: str = None):
        # super(_SparkRemoteJob, self).__init__(compute_context, job_id)
        if not job_id:
            job_id = uuid.uuid1().hex
        self._job_id = job_id
        self._spark_job_work_dir = None
        self._job_metadata = {}
        self._result_dir_name = "resultDir"

    def get_type(self) -> str:
        return "SparkJob"

    def create_spark_work_dir(self, root_path):
        work_dir = os.path.join(root_path, "cluster-" + self.id)
        if os.path.exists(work_dir):
            shutil.rmtree(work_dir)
        os.makedirs(work_dir, mode=0o700)
        self._spark_job_work_dir = work_dir

    def clean_spark_work_dir(self):
        if os.path.exists(self.get_spark_job_work_dir()):
            shutil.rmtree(self.get_spark_job_work_dir())
        self._spark_job_work_dir = None

    def get_spark_job_work_dir(self):
        if self._spark_job_work_dir is None:
            raise Exception("Spark work dir has not been created.")
        return self._spark_job_work_dir

    def prepare_job_metadata(self, py_function_name, py_args: dict):
        self._job_metadata['jobGUID'] = self.id
        self._job_metadata['jobWorkingDir'] = self.get_spark_job_work_dir()
        self._job_metadata['jobResultDirName'] = self._result_dir_name
        self._job_metadata['function'] = 'hpa_v1'
        self._job_metadata['jobType'] = 'spark-hpa'
        self._job_metadata['traceLevel'] = RxOptions.get_option("traceLevel", 0)
        self._job_metadata['jobPyData'] = 'job.pyInData'
        self._job_metadata['serverType'] = 'python'

        # enable progress reporting by default unless user chooses to disable it by setting report_progress = 0
        self._job_metadata['reportProgress'] = True
        if 'report_progress' in py_args and type(py_args['report_progress']) is int:
            self._job_metadata['reportProgress'] = py_args['report_progress'] > 0

        args = {}

        in_data = None
        if 'data' in py_args:
            in_data = py_args['data']
        elif 'input_data' in py_args:
            in_data = py_args['input_data']
        if isinstance(in_data, RxDataSource):
            args['inData'] = in_data.extract_info()

        out_data = None
        if 'output_file' in py_args:
                out_data = py_args['output_file']
        if isinstance(out_data, RxDataSource):
            args['outData'] = out_data.extract_info()
            if py_function_name in ("rx_import", "rx_data_step") and 'overwrite' in py_args and py_args['overwrite']:
                args['outData']['overwrite'] = True

        if len(args) > 0:
            self._job_metadata['args'] = args

    def dump_job_metadata_json(self, file_name: str):
        self.dump_json(self._job_metadata, file_name)

    def dump_json(self, data, file_name: str):
        with open(os.path.join(self.get_spark_job_work_dir(), file_name), 'w') as out_file:
            json.dump(data, out_file)

    def dump_job_in_data(self, file_name: str, data):
        with open(os.path.join(self.get_spark_job_work_dir(), file_name), 'wb') as out_file:
            dill.dump(data, out_file)

    def build_mrs_hadoop_job_command(self, command: str, extra_settings: dict):
        base_command = ["/usr/bin/mrs-hadoop-job", command,
                        "--version", "1.0", "--dir", self.get_spark_job_work_dir(),
                        "--trace-level", str(RxOptions.get_option("traceLevel", 0)),
                        "--server-type", "Python"]
        for key in extra_settings:
            base_command.append("--" + key)
            base_command.append(extra_settings[key])

        return base_command

    def serialize_bxl(self) -> dict:
        # TODO: add serialize data
        return {
        }

    def get_job_metadata(self) -> dict:
        return self._job_metadata

    def get_result_dir_name(self) ->str:
        return self._result_dir_name


class RxSparkRemoteHpcJob(RxSparkRemoteJob):
    def __init__(self, compute_context: RxComputeContext, job_id: str = None):
        # super(_SparkRemoteJob, self).__init__(compute_context, job_id)
        super().__init__(
            compute_context=compute_context,
            job_id=job_id
        )

    def prepare_job_metadata(self, py_function_name, py_args: dict):
        self._job_metadata['jobGUID'] = self.id
        self._job_metadata['jobWorkingDir'] = self.get_spark_job_work_dir()
        self._job_metadata['jobResultDirName'] = self._result_dir_name
        self._job_metadata['function'] = py_function_name
        self._job_metadata['jobType'] = 'spark-hpc'
        self._job_metadata['traceLevel'] = RxOptions.get_option("traceLevel", 0)
        self._job_metadata['jobPyData'] = 'job.pyInData'
        self._job_metadata['serverType'] = 'python'
        self._job_metadata['reportProgress'] = True

        if len(py_args) > 0:
            self._job_metadata['args'] = py_args
