from enum import Enum
from pandas import DataFrame
from revoscalepy.RxSerializable import RxSerializable
from revoscalepy.RxSerializable import rx_native_call
from .constants import PropertyConstants as pc
from .constants import FunctionConstants as fc


class OpenType(Enum):
    Read = 0
    ModifyOrAppendCols = 4
    Write = 1
    ReadWrite = 2
    AppendRows = 3
    AllowRead = 5
    AllowWrite = 6
    ReadText = 8


class RxDataSource(RxSerializable):
    """
    .. rxtitle::

        RxDataSource

    .. rxdescription::

        Base class for all revoscalepy data sources.

    """
    def __init__(self,
                 column_info: dict = None):
        self._to_serialize = {}

        # Initialize the variables with default values
        self.add_serialization_property(pc.resource_id, 0)
        self.add_serialization_property(pc.column_classes_id, None)
        self.add_serialization_property(pc.return_data_frame_id, None)
        self.add_serialization_property(pc.strings_as_factors_id, None)
        self.add_serialization_property(pc.rows_or_blocks_per_read_id, 1)
        self.add_serialization_property(pc.column_info_id, column_info)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __hash__(self):
        if self.column_info is not None:
            return hash(tuple(frozenset(sorted(self.column_info.keys()))))
        else:
            return hash(self.id)

    @property
    def id(self) -> str:
        return self.to_serialize[pc.resource_id]

    @property
    def column_classes(self) -> dict:
        return self.to_serialize[pc.column_classes_id]

    @column_classes.setter
    def column_classes(self, value: dict = None):
        self.to_serialize[pc.column_classes_id] = value

    @property
    def column_info(self) -> list:
        return self.to_serialize[pc.column_info_id]

    @column_info.setter
    def column_info(self, value: dict = None):
        self.to_serialize[pc.column_info_id] = value

    @property
    def return_data_frame(self) -> bool:
        return self.to_serialize[pc.return_data_frame_id]

    @property
    def strings_as_factors(self) -> bool:
        return self.to_serialize[pc.strings_as_factors_id]

    @strings_as_factors.setter
    def strings_as_factors(self, value: bool = False):
        self.to_serialize[pc.strings_as_factors_id] = value

    @property
    def rows_or_blocks_per_read(self) -> int:
        return self.to_serialize[pc.rows_or_blocks_per_read_id]

    @rows_or_blocks_per_read.setter
    def rows_or_blocks_per_read(self, value: int):
        self.to_serialize[pc.rows_or_blocks_per_read_id] = value

    @property
    def class_id(self) -> str:
        return self.to_serialize[pc.class_id]

    @property
    def to_serialize(self) -> dict:
        return self._to_serialize

    def add_serialization_property(self, property_name, value):
        self.to_serialize[property_name] = value

    def read_next(self):
        if self.id is not None:
            call_params = {pc.resource_id: self.id,
                           pc.return_data_frame_id: self.return_data_frame}
            ret = rx_native_call(fc.DataSourceReadNext, call_params)
            return DataFrame(ret)
        else:
            print("Please call open on the data source first.")

    def write_next(self, data_frame: DataFrame, verbose: int=0):
        if not self.is_open(OpenType.Write):
            raise ValueError("Output data source has not been opened for write: please call open first.")

        call_params = {pc.data_id: data_frame,
                       pc.resource_id: self.id,
                       pc.verbose: verbose}
        rx_native_call(fc.DataSourceWriteNext, call_params)
        return True

    def open(self, mode: OpenType=OpenType.Read):
        open_type = mode.value if isinstance(mode, OpenType) else mode
        call_params = self.to_serialize
        call_params["openType"] = open_type

        ret = rx_native_call(fc.OpenDataSource, call_params)
        self._to_serialize[pc.resource_id] = ret

    def is_open(self, mode: OpenType=OpenType.Read) -> bool:
        open_type = mode.value if isinstance(mode, OpenType) else mode
        call_params = {pc.resource_id: self.id,
                       pc.open_type_id: open_type}
        status = rx_native_call(fc.IsOpenDataSource, call_params)
        return bool(status)

    def close(self):
        if self.id is None:
            return

        call_params = {pc.resource_id: self._to_serialize[pc.resource_id]}
        rx_native_call(fc.CloseDataSource, call_params)

    def serialize_bxl(self) -> dict:
        return self.to_serialize

    def extract_info(self) -> dict:
        column_infos = []
        if self.column_info is not None:
            for column_name in self.column_info:
                column_info = self.column_info[column_name].copy()
                column_info['name'] = column_name
                column_infos.append(column_info)

        return {'colInfos': column_infos}

    def head(self, num_rows:int=6, report_progress:int=None):
        """
        .. rxdescription::

           Displays the first rows of the datasource.

        :param num_rows: Integer value specifying the number of rows to display starting from the beginning of the dataset.
            If not specified, the default of 6 will be used.

        :param report_progress: Integer value with options:
        * 0: no progress is reported.
        * 1: the number of processed rows is printed and updated.
        * 2: rows processed and timings are reported.
        * 3: rows processed and all timings are reported.

        .. rxkeywords:: head

        .. rxexample::

            import os
            from revoscalepy import RxOptions, RxXdfData
            sample_data_path = RxOptions.get_option("sampleDataDir")
            ds = RxXdfData(os.path.join(sample_data_path, "AirlineDemoSmall.xdf"))
            ds.head(num_rows=4)
        """
        from revoscalepy.etl.RxDataStep import rx_data_step
        results = rx_data_step(self, number_rows_read=num_rows, report_progress=report_progress)
        return results

    def tail(self, num_rows:int=6, report_progress:int=None):
        """
        .. rxdescription::

           Displays the last rows of the datasource.

        :param num_rows: Integer value specifying the number of rows to display starting from the end of the dataset.
            If not specified, the default of 6 will be used.

        :param report_progress: Integer value with options:
        * 0: no progress is reported.
        * 1: the number of processed rows is printed and updated.
        * 2: rows processed and timings are reported.
        * 3: rows processed and all timings are reported.

        .. rxkeywords:: tail

        .. rxexample::

            import os
            from revoscalepy import RxOptions, RxXdfData
            sample_data_path = RxOptions.get_option("sampleDataDir")
            ds = RxXdfData(os.path.join(sample_data_path, "AirlineDemoSmall.xdf"))
            ds.tail(num_rows=4)
        """
        raise NotImplementedError("Tail is not implemented on this DataSource object")
