import os
import re
from typing import Union

from .RxDataSource import RxDataSource
from .RxFileSystem import RxFileSystem
from .RxFileSystem import RxHdfsFileSystem
from .RxFileSystem import RxNativeFileSystem
from .constants import PropertyConstants as pc


class RxFileData(RxDataSource):
    """Base class for all file type data sources"""
    def __init__(self):
        super().__init__()
        self.add_serialization_property(pc.file_id, None)
        self._file_system = RxNativeFileSystem()
        self.add_serialization_property(pc.column_names_id, "")

    @property
    def file_system(self):
        return self.__file_system

    @file_system.setter
    def _file_system(self, file_system: Union[str, RxFileSystem]):
        self.__file_system = RxFileData._get_file_system(file_system)
        self.to_serialize[pc.file_system_id] = self.__file_system.to_serialize

        if isinstance(self.__file_system, RxHdfsFileSystem):
            self._handle_hdfs_file_system(self.__file_system)

    def _handle_hdfs_file_system(self, file_system: RxHdfsFileSystem):
        storage_schema_file_name = ("wasb://" in self.file or "adl://" in self.file or "hdfs://" in self.file)
        storage_schema_in_fs_hostname = ("wasb://" in file_system.host_name() or \
                                         "hdfs://" in file_system.host_name() or \
                                         "adl://" in file_system.host_name())
        if not storage_schema_file_name and storage_schema_in_fs_hostname:
            authority = re.sub("/$", "", file_system.host_name())
            file_path = re.sub("^/", "", self.file)
            self.add_serialization_property(pc.df_source_id, os.path.join(authority, file_path))

    @property
    def file(self) -> str:
        return self.to_serialize[pc.file_id]

    @property
    def column_names(self) -> str:
        return self.to_serialize[pc.column_names_id]

    @column_names.setter
    def column_names(self, value: str = ""):
        self.to_serialize[pc.column_names_id] = value

    @staticmethod
    def _get_file_system(file_system) -> RxFileSystem:
        if not isinstance(file_system, RxFileSystem) and not isinstance(file_system, str):
            raise ValueError("'file_system' should be string ('native'/'hdfs') "
                             "or RxNativeFileSystem/RxHdfsFileSystem")
        elif isinstance(file_system, str):
            if file_system != "hdfs" and file_system != "native":
                raise ValueError("'file_system' should be string ('native'/'hdfs') "
                                 "or RxNativeFileSystem/RxHdfsFileSystem")
            elif file_system == "native":
                return RxNativeFileSystem()
            else:
                return RxHdfsFileSystem()
        else:
            return file_system

    @staticmethod
    def _validate_file_argument(file):
        if file is None:
            raise ValueError("Argument file is missing, with no default")
        if isinstance(file, str) is False:
            raise ValueError("Argument file is an invalid type, must be a string")

