#
# Copyright (c) Microsoft Corporation.  All rights reserved.
#

from .RxSparkData import RxSparkData
from .constants import PropertyConstants as pc


class RxHiveData(RxSparkData):
    """
    .. rxtitle::

        Generate Hive Data Source Object

    .. rxdescription::
        Main generator for class RxHiveData, which extends RxSparkData.

    :param query: Character string specifying a Hive query, e.g. "select * from sample_
        table". Cannot be used with 'table'.

    :param table: Character string specifying the name of a Hive table, e.g. "sample_table".
        Cannot be used with 'query'.

    :param column_info: List of named variable information lists. Each variable
        information list contains one or more of the named elements given below.
        Currently available properties for a column information list are:
        type: Character string specifying the data type for the column.
            Supported types are:
                "bool" (stored as uchar),
                "integer" (stored as int32),
                "int16" (alternative to integer for smaller storage space),
                "float32" (stored as FloatType),
                "numeric" (stored as float64),
                "character" (stored as string),
                "factor" (stored as uint32),
                "Date" (stored as Date, i.e. float64.)
                "POSIXct" (stored as POSIXct, i.e. float64.)
        levels: List of strings containing the levels when type = "factor". If
            the levels property is not provided, factor levels will be determined
            by the values in the source column. If levels are provided, any value
            that does not match a provided level will be converted to a missing
            value.

    :param save_as_temp_table: Bool value, only applicable when using as output with "table" parameter.
        If "TRUE" register a temporary Hive table in Spark memory system otherwise generate a
        persistent Hive table. The temporary Hive table is always cached in Spark memory system.

    :param write_factors_as_indexes: Bool value, If True, when writing to an
        RxHiveData data source, underlying factor indexes will be written instead
        of the string representations.

    :return: object of class :py:class:`RxHiveData <microsoftml.RxHiveData>`.

    .. rxkeywords:: datasource, hive

    .. rxexample::
        import os
        from revoscalepy import rx_data_step, RxOptions, RxHiveData
        sample_data_path = RxOptions.get_option("sampleDataDir")
        colInfo = {"DayOfWeek": {"type":"factor"}}
        ds1 = RxHiveData(table = "AirlineDemo")
        result = rx_data_step(ds1)
        ds2 = RxHiveData(query = "select * from hivesampletable")
        result = rx_data_step(ds2)

    """


    @property
    def query(self) -> str:
        return self.to_serialize[pc.query_id]

    @property
    def table(self) -> str:
        return self.to_serialize[pc.table_id]

    def __init__(
            self,
            query=None,
            table=None,
            save_as_temp_table=False,
            write_factors_as_indexes: bool = False,
            column_info=None):

        if (query is None or len(query) == 0) and (table is None or len(table) == 0):
            raise ValueError("both query and table is None, need to give either query or table value")

        super().__init__(
            column_info=column_info,
            write_factors_as_indexes=write_factors_as_indexes,
            from_spark_data=True
        )

        self.add_serialization_property(pc.class_id, "RxSparkData")

        if query is None:
            if save_as_temp_table is False:
                self.add_serialization_property(pc.df_type_id, "hive-table")
            else:
                self.add_serialization_property(pc.df_type_id, "hive-temp-table")
            self.add_serialization_property(pc.df_source_id, table)
        else:
            self.add_serialization_property(pc.df_type_id, "hive")
            self.add_serialization_property(pc.df_source_id, query)

        self.add_serialization_property(pc.table_id, table)
        self.add_serialization_property(pc.query_id, query)

        if save_as_temp_table is True:
            self.add_serialization_property(pc.cache_id, True)
