import codecs
import pandas
import dill

from pandas import DataFrame
from .RxDataSource import RxDataSource
from .constants import PropertyConstants as pc

class RxOdbcData(RxDataSource):
    """
    .. rxtitle::

        Generate ODBC Data Source Object

    .. rxdescription::

        Main generator for class RxOdbcData, which extends RxDataSource.

    :param connection_string: None or character string specifying the
        connection string.
    :param table: None or character string specifying the table name. Cannot be
        used with sqlQuery.
    :param sql_query: None or character string specifying a valid SQL select
        query. Cannot be used with table.
    :param dbms_name: None or character string specifying the Database
        Management System (DBMS) name.
    :param database_name: None or character string specifying the name of the
        database.
    :param use_fast_read: Bool specifying whether or not to use a direct
        ODBC connection.
    :param trim_space: Bool specifying whether or not to trim the white
        character of string data for reading.
    :param row_buffering: Bool specifying whether or not to buffer rows on
        read from the database. If you are having problems with your ODBC driver,
        try setting this to False.
    :param return_data_frame: Bool indicating whether or not to convert the
        result from a list to a data frame (for use in rxReadNext only). If False,
        a list is returned.
    :param string_as_factors: Bool indicating whether or not to
        automatically convert strings to factors on import. This can be overridden
        by specifying "character" in column_classes and column_info. If True, the
        factor levels will be coded in the order encountered. Since this factor
        level ordering is row dependent, the preferred method for handling factor
        columns is to use column_info with specified "levels".
    :param column_classes: Dictionary of column name to strings specifying the 
        column types to use when converting the data. The element names for the 
        vector are used to identify which column should be converted to which type.

        Allowable column types are:

            "bool" (stored as uchar),
            "integer" (stored as int32),
            "float32" (the default for floating point data for '.xdf' files),
            "numeric" (stored as float64 as in R),
            "character" (stored as string),
            "factor" (stored as uint32),
            "int16" (alternative to integer for smaller storage space),
            "uint16" (alternative to unsigned integer for smaller storage space),
            "Date" (stored as Date, i.e. float64)

        Note for "factor" type, the levels will be coded in the order
        encountered. Since this factor level ordering is row dependent, the
        preferred method for handling factor columns is to use column_info with
        specified "levels".
        Note that equivalent types share the same bullet in the list above; for
        some types we allow both 'R-friendly' type names, as well as our own,
        more specific type names for '.xdf' data.
        Note also that specifying the column as a "factor" type is currently
        equivalent to "string" - for the moment, if you wish to import a column
        as factor data you must use the column_info argument, documented below.

    :param column_info: List of named variable information lists. Each variable
        information list contains one or more of the named elements given below.
        The information supplied for column_info overrides that supplied for
        column_classes.

        Currently available properties for a column information list are:

            type: character string specifying the data type for the column. See
                column_classes argument description for the available types. Specify
                "factorIndex" as the type for 0-based factor indexes. levels must also
                be specified.
            newName: character string specifying a new name for the variable.
            description: character string specifying a description for the variable.
            levels: list of strings containing the levels when type = "factor". If
                the levels property is not provided, factor levels will be determined
                by the values in the source column. If levels are provided, any value
                that does not match a provided level will be converted to a missing
                value.
            newLevels: new or replacement levels specified for a column of type
                “factor”. It must be used in conjunction with the levels argument.
                After reading in the original data, the labels for each level will be
                replaced with the newLevels.
            low: the minimum data value in the variable (used in computations using
                the F() function.
            high: the maximum data value in the variable (used in computations
                using the F() function.

    :param rows_per_read: Number of rows to read at a time.
    :param verbose: integer value. If 0, no additional output is printed. If 1,
        information on the odbc data source type (odbc or odbcFast) is printed.
    :param write_factors_as_indexes: Bool value, If True, when writing to an
        RxOdbcData data source, underlying factor indexes will be written instead
        of the string representations.
    :param kwargs: Additional arguments to be passed directly to the underlying
        functions.

    :return: Object of class RxOdbcData.

    .. rxkeywords:: odbc, datasource

    .. rxexample::
        from revoscalepy import RxOdbcData
        connection_string = 'Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;'
        dest = RxOdbcData(connection_string, table = "dataframe")
    """
    def __init__(
            self,
            connection_string: str = None,
            table: str = None,
            sql_query: str = None,
            dbms_name: str = None,
            database_name:str = None,
            use_fast_read: bool = True,
            trim_space: bool = True,
            row_buffering: bool = True,
            return_data_frame:bool = True,
            string_as_factors:bool = False,
            column_classes:dict = None,
            column_info: dict = None,
            rows_per_read: int = 500000,
            verbose: int = 0,
            write_factors_as_indexes: bool = False,
            **kwargs):
        super().__init__()

        import revoscalepy.RxFunction as arg

        if connection_string is None or len(connection_string) == 0:
            raise ValueError("Parameter 'connection_string' must be a valid string.")

        if (table is None) and (sql_query is None):
            raise ValueError("Either table or sql_query argument has to be specified.")

        if (table is not None) and (sql_query is not None):
            raise ValueError("Both table and sql_query arguments cannot be specified together.")

        if use_fast_read is None:
            use_fast_read = True

        if trim_space is None:
            trim_space = True

        if rows_per_read is not None and rows_per_read <= 0:
            raise ValueError("Parameter 'rows_per_read' must be a positive integer.")

        if column_info is not None and len(column_info) == 0:
            raise ValueError("Parameter 'column_info' must be a valid dictionary.")

        # process optional arguments
        server = arg.get_optional_parameter('server', **kwargs)
        dsn = arg.get_optional_parameter('dsn', **kwargs)
        user = arg.get_optional_parameter('user', **kwargs)
        password = arg.get_optional_parameter('password', **kwargs)
        report_progress = arg.get_optional_parameter('report_progress', **kwargs)
        isolation_level = arg.get_optional_parameter('isolation_level', **kwargs)

        if isolation_level is None:
            isolation_level = 0

        if server is not None and len(server) == 0:
            raise ValueError("Parameter 'server' must be None or a valid string.")

        if dsn is not None and len(dsn) == 0:
            raise ValueError("Parameter 'dsn' must be None or a valid string.")

        if user is not None and len(user) == 0:
            raise ValueError("Parameter 'user' must be None or a valid string.")

        if password is not None and len(password) == 0:
            raise ValueError("Parameter 'password' must be None or a valid string.")

        if report_progress is not None and (report_progress < 1 or report_progress > 3):
            raise ValueError("Parameter 'report_progress' must be between 1 and 3.")

        if verbose > 0:
            print("Data source type: RxOdbcData, use_fast_read={0}.".format(use_fast_read))

        self.add_serialization_property(pc.class_id, "RxOdbcData")
        self.add_serialization_property(pc.connection_string_id, connection_string)
        self.add_serialization_property(pc.table_id, table)
        self.add_serialization_property(pc.sql_query_id, sql_query)
        self.add_serialization_property(pc.dbms_name_id, dbms_name)
        self.add_serialization_property(pc.database_name_id, database_name)
        self.add_serialization_property(pc.use_fast_read, use_fast_read)
        self.add_serialization_property(pc.trim_space_id, trim_space)
        self.add_serialization_property(pc.row_buffering_id, row_buffering)
        self.add_serialization_property(pc.return_data_frame_id, return_data_frame)
        self.add_serialization_property(pc.strings_as_factors_id, string_as_factors)
        self.add_serialization_property(pc.column_classes_id, column_classes)
        self.add_serialization_property(pc.column_info_id, column_info)
        self.add_serialization_property(pc.rows_or_blocks_per_read_id, rows_per_read)
        self.add_serialization_property(pc.write_factors_as_indexes_id, write_factors_as_indexes)
        self.add_serialization_property(pc.isolation_level_id, isolation_level)
        self.add_serialization_property(pc.server, server)
        self.add_serialization_property(pc.dsn, dsn)
        self.add_serialization_property(pc.user, user)
        self.add_serialization_property(pc.password, password)


    @property
    def table(self) -> dict:
        return self.to_serialize[pc.table_id]

    @property
    def connection_string(self) -> str:
        return self.to_serialize[pc.connection_string_id]

    @property
    def sql_query(self) -> str:
        if pc.sql_query_id in self.to_serialize:
            return self.to_serialize[pc.sql_query_id]
        return None

    @property
    def use_fast_read(self):
        return self.to_serialize[pc.use_fast_read]

    @property
    def input_data_query(self) -> str:
        return self.sql_query if not self.sql_query is None else "SELECT * from {0}".format(self.table)

    @property
    def column_info(self) -> dict:
        return self.to_serialize[pc.column_info_id]

    @property
    def rows_per_read(self) -> int:
        return self.to_serialize[pc.rows_or_blocks_per_read_id]

    @staticmethod
    def fix_connection_string(connection_string : str) -> str:
        new_connection_string = ''
        
        for token in connection_string.split(';'):
            token = token.strip()
            if len(token) == 0:
                continue;
            key, value = token.split('=')
            if key.lower() == 'server':
                server_values = value.split('\\')
                if len(server_values) > 1:
                    port_values = server_values[0].split(',')
                    if len(port_values) > 1:
                        server_values[0] = '.' + ',' + port_values[1]
                    else:
                        server_values[0] = '.'
                    value = server_values[0] + '\\' + server_values[1]
                else:
                    port_values = server_values[0].split(',')
                    if len(port_values) > 1:
                        server_values[0] = '.' + ',' + port_values[1]
                    else:
                        server_values[0] = '.'
                    value = server_values[0]
            new_connection_string += (key + '=' + value + ';')
        
        return new_connection_string;

    def rx_read_object(
                    self,
                    key: str,
                    version: str = None,
                    key_name: str = "id",
                    version_name: str = "version",
                    value_name: str = "value",
                    deserialize: bool = True,
                    decompress: str = "zip"
                    ):
        # Import locally to prevent cyclic issues
        from revoscalepy import rx_data_step

        # prepare RxOdbcData to send to DataStep
        if self.sql_query is None:
            if version is None:
                query = "select {0} from {1} where {2} = '{3}'".format(value_name, self.table, key_name, key)
            else:
                query = "select {0} from {1} where {2} = '{3}' and {4} = '{5}'".format(value_name, self.table, key_name, key, version_name, version)
        else:
            query = self.sql_query

        # call RxDataStep to retrieve value
        read_val = rx_data_step(RxOdbcData(self.connection_string, sql_query= query), number_rows_read = 1)

        # Return None if read value=None OR not a dict OR doesn't contain expected key or values,
        if (read_val is None
            or (not isinstance(read_val, dict) and not isinstance(read_val, DataFrame))
            or not value_name in read_val
            or not len(read_val[value_name]) > 0):
            return None

        read_val = read_val[value_name][0]
        # decompress and deserialize if needed
        read_val = codecs.decode(read_val, decompress) if decompress else read_val
        return dill.loads(read_val) if deserialize else read_val

    def rx_write_object(
                    self,
                    key: str,
                    value: str,
                    version: str = None,
                    key_name: str = "id",
                    value_name: str = "value",
                    version_name: str = "version",
                    serialize: bool = True,
                    overwrite: bool = False,
                    compress: str = "zip"
                    ) -> bool:

        # Import locally to prevent cyclic issues
        from revoscalepy import rx_data_step

        if key is None or value is None:
            raise ValueError("key or value cannot be None.")

        if not (serialize or issubclass(type(value), bytes)):
            raise ValueError("Only bytes values are supported when serialization is turned off.")

        value = dill.dumps(value) if serialize else value
        value = codecs.encode(value, compress) if compress else value

        if version is None:
            df = pandas.DataFrame({key_name : [key], value_name: [value]})
        else:
            df = pandas.DataFrame({key_name : [key], value_name: [value], version_name: [version]})

        # If overwrite = True, delete the existing object
        if overwrite:
            self.rx_delete_object(key, version = version, key_name = key_name, version_name = version_name)

        rx_data_step(df, self, append = ['rows'])
        return True

    def rx_list_keys(
                self,
                key: str = None,
                version: str = None,
                key_name: str = "id",
                version_name: str = "version"
                ):
        # Import locally to prevent cyclic issues
        from revoscalepy import rx_data_step

        if self.sql_query is None:
            if key is None:
                query = "select {0} from {1} order by {0}".format(key_name, self.table)
            else:
                if version is None:
                    query = "select {0} from {1} where {2} = '{3}'".format(key_name, self.table, key_name, key)
                else:
                    query = "select {0}, {4} from {1} where {2} = '{3}' and {4} = '{5}'".format(key_name, self.table, key_name, key, version_name, version)
        else:
            query = self.sql_query

        return rx_data_step(RxOdbcData(self.connection_string, sql_query = query))


    def rx_delete_object(
                self,
                key: str = None,
                version: str = None,
                key_name: str = "id",
                version_name: str = "version",
                all: bool = False
                ):

        if not all and key is None:
            raise ValueError("key is required to delete object, or use all=True to remove all objects")

        if all:
            delete_cmd = "delete from {0}".format(self.table)
        else:
            if version is None:
                delete_cmd = "delete from {0} where {1} = '{2}'".format(self.table, key_name, key)
            else:
                delete_cmd = "delete from {0} where {1} = '{2}' and {3} = '{4}'".format(self.table, key_name, key, version_name, version)

        ret = RxOdbcData._rx_execute_sql_dml(self.connection_string, self.table, delete_cmd, report_exceptions = False)
        return ret


    def _print_details(self) -> str:
        return  "connection_string: \"{0}\"\n" \
                "{1}: {2}\n" \
                "use_fast_read: {3}".format(
                self.connection_string,
                "sql_query" if self.table is None else "table",
                self.sql_query if self.table is None else self.table,
                self.use_fast_read)

    def __str__(self) -> str:
        return  "RxOdbcData Source:\n" \
                "------------------\n" \
                "{0}".format(self._print_details())

    @staticmethod
    def _rx_execute_sql_dml(connection_string : str, table: str, dmlquery : str, report_exceptions : bool = True) -> bool:
        # Import locally to prevent cyclic issues
        from revoscalepy.RxSerializable import rx_native_call
        from revoscalepy.datasource.RxDataSource import OpenType

        is_successful = False

        try:
            dmlexecutor = RxOdbcData(table = table, connection_string = connection_string)
            dmlexecutor.open(mode = OpenType.Write)
            # the param names here should match with the names in BxlServer
            params = {
                    "id" : dmlexecutor.id,
                    "sSQLString" : dmlquery
            }
            rx_native_call("ExecuteSQLDML", params)
            is_successful = True
        except Exception as exception :
            if report_exceptions :
                print("While executing query: {0}, following exception occurred : {1}".format(dmlquery, exception))
        finally :
            if dmlexecutor.is_open():
                dmlexecutor.close()

        return is_successful