#
# Copyright (c) Microsoft Corporation.  All rights reserved.
#
import re
import os
import warnings

from .RxDataSource import RxDataSource
from .RxSparkData import RxSparkData
from .RxFileSystem import RxFileSystem
from .. utils.RxOptions import RxOptions
from .RxFileSystem import RxHdfsFileSystem
from .constants import PropertyConstants as pc


class RxParquetData(RxSparkData):
    """
    .. rxtitle::

        Generate Parquet Data Source Object

    .. rxdescription::
        Main generator for class RxParquetData, which extends RxSparkData.

    :param file: Character string specifying the location of the data. e.g.
                 "/tmp/AirlineDemoSmall.parquet".

    :param column_info: List of named variable information lists. Each variable
        information list contains one or more of the named elements given below.
        Currently available properties for a column information list are:
        type: Character string specifying the data type for the column.
            Supported types are:
                "bool" (stored as uchar),
                "integer" (stored as int32),
                "int16" (alternative to integer for smaller storage space),
                "float32" (stored as FloatType),
                "numeric" (stored as float64),
                "character" (stored as string),
                "factor" (stored as uint32),
                "Date" (stored as Date, i.e. float64.)
                "POSIXct" (stored as POSIXct, i.e. float64.)
        levels: List of strings containing the levels when type = "factor". If
            the levels property is not provided, factor levels will be determined
            by the values in the source column. If levels are provided, any value
            that does not match a provided level will be converted to a missing
            value.


    :param file_system: Character string or RxFileSystem object indicating type
        of file system; It supports native HDFS and other HDFS compatible systems,
        e.g., Azure Blob and Azure Data Lake. Local file system is not supported.

    :param write_factors_as_indexes: Bool, if True, when writing to an
        RxParquetData data source, underlying factor indexes will be written instead
        of the string representations.

    :return: Object of class :py:class:`RxParquetData <microsoftml.RxParquetData>`.

    .. rxkeywords:: datasource, parquet

    .. rxexample::
        import os
        from revoscalepy import rx_data_step, RxOptions, RxParquetData
        sample_data_path = RxOptions.get_option("sampleDataDir")
        colInfo = {"DayOfWeek": {"type":"factor"}}
        ds = RxParquetData(os.path.join(sample_data_path, "AirlineDemoSmall.parquet"))
        result = rx_data_step(ds)

    """


    @property
    def file(self) -> str:
        return self.to_serialize[pc.file_id]

    @property
    def file_system(self) -> str:
        return self._file_system

    def _rx_set_file_system(self, fs):
        if fs is None:
            current_file_system = RxOptions.get_option("FileSystem")
            if isinstance(current_file_system, RxHdfsFileSystem):
                self._file_system = current_file_system
            else:
                self._file_system = RxHdfsFileSystem()
                warnings.warn("Current file system is not Hdfs file system, automatically \
                 change to default Hdfs file system for Parquet data")
        elif not isinstance(fs, RxHdfsFileSystem):
            raise TypeError("file system should be HDFS File system for Parquet data source, please use RxHdfsFileSystem")
        else:
            self._file_system = fs

        self.add_serialization_property(pc.file_system_id, self._file_system.to_serialize)

    def __init__(
        self,
        file: str = None,
        column_info=None,
        file_system: RxFileSystem = None,
        write_factors_as_indexes: bool = False):

        if file is None or len(file) is 0:
            raise ValueError("'file' must be a non-empty character string")

        super().__init__(
            column_info=column_info,
            write_factors_as_indexes = write_factors_as_indexes,
            from_spark_data=True
        )

        self._rx_set_file_system(file_system)

        self.add_serialization_property(pc.file_id, file)
        self.add_serialization_property(pc.class_id, "RxSparkData")
        self.add_serialization_property(pc.df_source_id, file)
        self.add_serialization_property(pc.df_type_id, "parquet")

        if isinstance(self.file_system, RxHdfsFileSystem):
            self._handle_hdfs_file_system(self.file_system)

    def _handle_hdfs_file_system(self, file_system: RxHdfsFileSystem):
        storage_schema_file_name = ("wasb://" in self.file or "adl://" in self.file or "hdfs://" in self.file)
        storage_schema_in_fs_hostname = ("wasb://" in file_system.host_name() or \
                                         "hdfs://" in file_system.host_name() or \
                                         "adl://" in file_system.host_name())
        if not storage_schema_file_name and storage_schema_in_fs_hostname:
            authority = re.sub("/$", "", file_system.host_name())
            file_path = re.sub("^/", "", self.file)
            self.add_serialization_property(pc.df_source_id, os.path.join(authority, file_path))


