#
# Copyright (c) Microsoft Corporation.  All rights reserved.
#
import uuid
import pandas
import warnings

from . RxDataSource import RxDataSource
from . RxXdfData import RxXdfData
from . constants import PropertyConstants as pc
from .. utils.RxOptions import RxOptions
from .. computecontext.RxComputeContext import RxComputeContext


class RxSparkData(RxDataSource):
    """
    .. rxtitle::

        RxSparkData

    .. rxdescription::

        Base class for all revoscalepy Spark data sources,
        including RxHiveData, RxOrcData, RxParquetData and RxSparkDataFrame.
        It is intended to be called from those subclasses
        instead of directly.

    """
    @property
    def df_name(self) -> str:
        return self.to_serialize[pc.df_name_id]

    @property
    def xdf_uuid(self) -> str:
        return self.to_serialize[pc.xdf_uuid_id]

    @property
    def df_source(self) -> str:
        return self.to_serialize[pc.df_source_id]

    @property
    def df_type(self) -> str:
        return self.to_serialize[pc.df_type_id]

    @property
    def cache(self) -> bool:
        return self.to_serialize[pc.cache_id]

    @cache.setter
    def cache(self,cache):
        self.to_serialize[pc.cache_id] = cache

    @property
    def write_factors_as_indexes(self) -> bool:
        return self.to_serialize[pc.write_factors_as_indexes_id]

    @property
    def column_names(self) -> str:
        return self.to_serialize[pc.column_names_id]

    @column_names.setter
    def column_names(self, value: str = ""):
        self.to_serialize[pc.column_names_id] = value

    def __init__(self,
                 column_info: dict = None,
                 df_source: str = None,
                 df_type: str = None,
                 write_factors_as_indexes: bool = False,
                 **kwargs):
        """
        Constructs an 'RxSparkData' object.
        :param file: character string specifying the location of the data.

        """

        if "from_spark_data" not in kwargs or kwargs["from_spark_data"] is not True:
            raise ValueError(
                "RxSparkData is the base class for all revoscalepy Spark data sources, "
                "including RxHiveData, RxOrcData, and RxParquetData. "
                "It is intended to be called from those subclasses instead of directly.\n")

        super().__init__(
            column_info=column_info
        )

        self.add_serialization_property(pc.xdf_uuid_id, str(uuid.uuid4()))
        self.add_serialization_property(pc.df_name_id, "df-" + str(uuid.uuid4()))
        self.add_serialization_property(pc.cache_id, False)
        self.add_serialization_property(pc.write_factors_as_indexes_id, write_factors_as_indexes)
        self.add_serialization_property(pc.df_type_id, df_type)
        self.add_serialization_property(pc.df_source_id, df_source)
        self.add_serialization_property(pc.column_names_id, "")

        self.rx_validate_spark_data_col_info()

    def read_next(self):
        raise Exception("Not supported")

    def write_next(self, to: RxDataSource):
        raise Exception("Not supported")

    def open(self, mode: str):
        raise Exception("Not supported")

    def close(self):
        raise Exception("Not supported")

    def extract_info(self) -> dict:
        infos = super().extract_info()
        infos['sourceType'] = self.df_type
        infos['dfName'] = self.df_name
        infos['xdfId'] = self.xdf_uuid
        infos['source'] = self.df_source
        infos['cache'] = self.cache
        infos['writeFactorsAsIndexes'] = self.write_factors_as_indexes
        return infos

    def rx_validate_spark_data_col_info(self):
        if self.column_info is None:
            return True
        for column_name in self.column_info:
            expected = {"type", "levels"}
            if not all(key in expected for key in self.column_info[column_name]):
                warnings.warn("colInfo variable information lists can only support 'type' and 'levels', " +
                              "all other properties will be ignored.")
                break
        return True


def _generate_listdata_output(global_data_pairs: dict, objs: list):
    output = pandas.DataFrame(columns=["Name", "Type", "Source"])
    index = 0
    for obj in objs:
        variable_names = {k for k in global_data_pairs if global_data_pairs[k].df_name == obj.df_name}
        df_type = obj.df_type
        df_source = obj.df_source
        if len(variable_names)>0:
            output.loc[index] = [variable_names, df_type, df_source]
        else:
            output.loc[index] = ["N/A", df_type, df_source]
        index += 1
    return output


def rx_spark_list_data(show_description: bool, compute_context: RxComputeContext = None):
    """
    .. rxtitle::

        Manage Cached Data in Spark

    .. rxdescription::
        Use these functions to manage the objects cached in the Spark memory system. These functions are only applicable
         when using RxSpark compute context.

    :param show_description: Bool value, indicating whether or not to print out the detail to console.

    :param compute_context: RxSpark compute context object.

    :return: List of all objects cached in Spark memory system for rxSparkListData.

    .. rxkeywords:: spark, data

    .. rxexample::        
        from revoscalepy import RxOrcData, rx_spark_connect, rx_spark_list_data, rx_lin_mod, rx_spark_cache_data
        rx_spark_connect()
        col_info = {"DayOfWeek": {"type": "factor"}}
        df = RxOrcData(file = "/share/sample_data/AirlineDemoSmallOrc", column_info = col_info)
        df = rx_spark_cache_data(df, True)

        # After the first run, a Spark data object is added into the list
        rx_lin_mod("ArrDelay ~ DayOfWeek", data = df)
        rx_spark_list_data(True)

    """

    from .. computecontext.RxSpark import RxSpark
    if compute_context is None:
        compute_context=RxComputeContext.current()
    if not isinstance(compute_context,RxSpark):
        raise TypeError("rx_spark_list_data can only be used in RxSpark compute context.")
    if show_description:
        all_variables = globals()
        all_data_pairs = {k:all_variables[k] for k in all_variables if isinstance(all_variables[k], RxXdfData) or isinstance(all_variables[k], RxSparkData)}
        spark_data_objs = RxOptions.get_option("_rx_spark_data_objs" )
        if len(spark_data_objs) > 0:
            output = _generate_listdata_output(all_data_pairs,spark_data_objs)
            print(output)
        else:
            print("No Cached Df.\n")
    else:
        return RxOptions.get_option("_rx_spark_data_objs" )


def rx_spark_cache_data(in_data: RxDataSource, cache: bool):
    """
    .. rxtitle::

        Set the Cache Flag in Spark Compute Context

    .. rxdescription::
        Use this function to set the cache flag to control whether data objects should be cached in Spark
         memory system, applicable for RxXdfData, RxHiveData, RxParquetData, RxOrcData and RxSparkDataFrame.

    :param in_data: A data source that can be RxXdfData, RxHiveData, RxParquetData, RxOrcData or RxSparkDataFrame.
        RxTextData is currently not supported.

    :param cache: Bool value controlling whether the data should be cached or not. If TRUE the data will
        be cached in the Spark memory system after the first use for performance enhancement.

    :return: Data object with new cache flag.

    .. rxkeywords:: spark, data, cache

    .. rxexample::
        from revoscalepy import RxHiveData, rx_spark_connect, rx_spark_cache_data
        hive_data = RxHiveData(table = "mytable")

        # The cache flag of hiveData is now set to TRUE.
        hive_data = rx_spark_cache_data(hive_data, True)

        # set cache value to False
        hive_data = rx_spark_cache_data(hive_data, False)


    """


    if in_data is None or (not isinstance(in_data, RxXdfData) and not isinstance(in_data, RxSparkData)):
        raise TypeError("Please set inData as RxHiveData, RxParquetData, RxOrcData, RxSparkDataFrame or RxXdfData")
    elif in_data.df_type == "xdf" and ".xdf" in in_data.df_source:
        raise TypeError("rx_spark_cache_data doesn't support single xdf file. " +
                        "This RxXdfData will not be cached in spark.")
    elif in_data.df_type == "hive-temp-table":
        if cache is False:
            warnings.warn("Hive temp table is an in-memory object. You cannot uncache it. " +
                          "If you would like to remove it from memory, please use rx_spark_remove_data instead.")
            return in_data

    in_data.cache = cache
    return in_data


def reset_spark_data_objs(remove_list):
    spark_data_objs = RxOptions.get_option("_rx_spark_data_objs")
    if isinstance(remove_list, list):
        new_objs = spark_data_objs
        for obj in remove_list:
            new_objs = [x for x in new_objs if x.df_name != obj.df_name]
            if obj.df_type == "hive-temp-table":
                warnings.warn("Hive temp table " + obj.df_source + " is destroyed. It cannot be reused in the future" +
                 "methods")
    elif isinstance(remove_list, RxXdfData) or isinstance(remove_list, RxSparkData):
        new_objs = [x for x in spark_data_objs if x.df_name != remove_list.df_name]
        if remove_list.df_type == "hive-temp-table":
            warnings.warn("Hive temp table " + remove_list.df_source + " is destroyed. It cannot be reused in the" +
                          "future methods.")
    RxOptions.set_option("_rx_spark_data_objs", new_objs)


def check_and_prepare_remove_data_spark_job(remove_list):
    job_data = list()
    if isinstance(remove_list, list):
        for data in remove_list:
            if not isinstance(data, RxXdfData) and not isinstance(data, RxSparkData):
                raise TypeError("rx_spark_remove_data can only accept" +
                                " RxHiveData/RxParquetData/RxOrcData/RxSparkDataFrame/RxXdfData"
                                + "object or list of these objects.")
            if data.xdf_uuid is None or data.df_name is None:
                raise ValueError(
                    "Found broken input data object. " +
                    "Input RxHiveData/RxParquetData/RxOrcData/RxSparkDataFrame/RxXdfData object must have"
                    + "not null dfName and not null xdfUuid.")
            job_data.append(data.extract_info())
    elif not isinstance(remove_list, RxXdfData) and not isinstance(remove_list, RxSparkData):
        raise TypeError("rx_spark_remove_data can only accept " +
                        "RxHiveData/RxParquetData/RxOrcData/RxSparkDataFrame/RxXdfData object"
                        + "or list of these objects.")
    else:
        job_data.append(remove_list.extract_info())
    return job_data


def rx_spark_remove_data(remove_list, compute_context: RxComputeContext = None):
    """
    .. rxtitle::

        Manage Cached Data in Spark

    .. rxdescription::
        Use these functions to manage the objects cached in the Spark memory system. These functions are only applicable
         when using RxSpark compute context.

    :param remove_list: Cached object or a list of cached objects need to be deleted.

    :param compute_context: RxSpark compute context object.

    :return: No return objs for rxSparkRemoveData.

    .. rxkeywords:: spark, data

    .. rxexample::
        from revoscalepy import RxOrcData, rx_spark_connect, rx_spark_disconnect, rx_spark_list_data, rx_spark_cache_data, rx_spark_remove_data, rx_lin_mod
        cc=rx_spark_connect()
        col_info = {"DayOfWeek": {"type": "factor"}}

        df = RxOrcData(file = "/share/sample_data/AirlineDemoSmallOrc", column_info = col_info)
        df = rx_spark_cache_data(df, True)

        # After the first run, a Spark data object is added into the list
        rx_lin_mod("ArrDelay ~ DayOfWeek", data = df)

        rx_spark_remove_data(df)
        rx_spark_list_data(True)
        rx_spark_disconnect(cc)

    """


    if compute_context is None:
        compute_context = RxComputeContext.current()

    if compute_context.__class__.__name__ != 'RxSpark':
        raise TypeError("rx_spark_remove_data can only be used in RxSpark compute context.")

    if remove_list is None or (isinstance(remove_list, list) and len(remove_list) == 0):
        raise ValueError("Input remove_list is empty")

    py_args = {"removeDataList": check_and_prepare_remove_data_spark_job(remove_list)}
    py_function_name = "remove_data_v1"

    result = compute_context._execute_spark_hpc(py_function_name, py_args)

    reset_spark_data_objs(remove_list)

    return result
