#
# Copyright (c) Microsoft Corporation.  All rights reserved.
#
from .RxSparkData import RxSparkData
from ..computecontext.RxPySparkBase import rx_pyspark_check_pyspark_is_loaded
from ..computecontext.RxPySparkBase import rx_pyspark_get_existing_pyspark_app
from ..computecontext.RxPySparkBase import rx_pyspark_check_data_spark_context_still_alive
from ..computecontext.RxPySparkBase import rx_pyspark_check_data_is_spark_data_frame
from ..computecontext.RxPySparkBase import get_revoscalepy_spark_jar_path
from .constants import PropertyConstants as pc


class RxSparkDataFrame(RxSparkData):
    """
    .. rxtitle::

        Generate Spark Data Frame Data Source Object

    .. rxdescription::
        Main generator for class RxSparkDataFrame, which extends RxSparkData.

    :param data: Spark data frame obj from pyspark.sql.DataFrame.

    :param column_info: List of named variable information lists. Each variable
        information list contains one or more of the named elements given below.
        Currently available properties for a column information list are:
        type: Character string specifying the data type for the column.
            Supported types are:
                "bool" (stored as uchar),
                "integer" (stored as int32),
                "int16" (alternative to integer for smaller storage space),
                "float32" (stored as FloatType),
                "numeric" (stored as float64),
                "character" (stored as string),
                "factor" (stored as uint32),
                "Date" (stored as Date, i.e. float64.)
                "POSIXct" (stored as POSIXct, i.e. float64.)
        levels: List of strings containing the levels when type = "factor". If
            the levels property is not provided, factor levels will be determined
            by the values in the source column. If levels are provided, any value
            that does not match a provided level will be converted to a missing
            value.

    :param write_factors_as_indexes: Bool value, if True, when writing to an
        RxSparkDataFrame data source, underlying factor indexes will be written instead
        of the string representations.

    :return: Object of class :py:class:`RxSparkDataFrame <microsoftml.RxSparkDataFrame>`.

    .. rxkeywords:: datasource

    .. rxexample::
        import os
        from revoscalepy import rx_data_step, RxSparkDataFrame
        from pyspark.sql import SparkSession
        spark = SparkSession.builder.getOrCreate()
        df = spark.createDataFrame([('Monday',3012),('Friday',1354),('Sunday',5452)], ["DayOfWeek", "Sale"])
        col_info = {"DayOfWeek": {"type":"factor"}}
        ds = RxSparkDataFrame(df, column_info=col_info)
        out_ds = RxSparkDataFrame(write_factors_as_indexes=True)
        rx_data_step(ds, out_ds)
        out_df = out_ds.spark_data_frame
        out_df.take(2)
        # [Row(DayOfWeek=0, Sale=3012), Row(DayOfWeek=2, Sale=1354)]

    """
    def _put_spark_df_to_rre_server(self, spark_data_frame):
        if not rx_pyspark_check_data_spark_context_still_alive(spark_data_frame):
            raise Exception("The input spark data frame is from a stopped Spark application.")
        sc = rx_pyspark_get_existing_pyspark_app()
        if sc is None:
            raise Exception("No Spark application is running. "
                            "Spark Data Frame only can work in active Spark Application.")
        try:
            sc._jvm.com.microsoft.scaler.spark.RREServer.putSparkDataFrame(self.df_name, spark_data_frame._jdf)
        except TypeError as e:
            if "'JavaPackage' object is not callable" in str(e):
                jar_path = get_revoscalepy_spark_jar_path()
                raise Exception("%s is not loaded in Spark by setting config spark.jars. " % jar_path +
                                "Please call rx_spark_connect with reset=True to launch a " +
                                "new spark application or explicitly set spark.jar=%s in SparkConf." % jar_path)
            raise e

    def _get_spark_df_from_rre_server(self):
        sc = rx_pyspark_get_existing_pyspark_app()
        if sc is None:
            raise Exception("No Spark application is running. "
                            "Spark Data Frame only can work in active Spark Application.")
        try:
            spark_data_frame_jdf = sc._jvm.com.microsoft.scaler.spark.RREServer.getSparkDataFrame(self.df_name)
        except TypeError as e:
            if "'JavaPackage' object is not callable" in str(e):
                jar_path = get_revoscalepy_spark_jar_path()
                raise Exception("%s is not loaded in Spark by setting config spark.jars. " % jar_path +
                                "Please call rx_spark_connect with reset=True to launch a " +
                                "new spark application or explicitly set spark.jar=%s in SparkConf." % jar_path)
            raise e
        return spark_data_frame_jdf


    @property
    def spark_data_frame(self):
        spark_data_frame_jdf = self._get_spark_df_from_rre_server()
        if spark_data_frame_jdf is None:
            return None

        if self._spark_data_frame is not None and spark_data_frame_jdf == self._spark_data_frame._jdf:
            # if _jdf in scala is the same as the one in RxSparkDataFrame.
            # That means they are in sync
            return self._spark_data_frame
        # _jdf in scala is out of sync with _spark_data_frame in RxSparkDataFrame.
        # this RxSparkDataFrame obj has been used as output data source.
        # Need to update.
        from pyspark import SQLContext
        from pyspark.sql import DataFrame
        sc = rx_pyspark_get_existing_pyspark_app()
        sql_ctx = SQLContext(sc)
        # reconstruct the pyspark dataframe with
        # the returned java dataframe and the current sql context
        new_df = DataFrame(spark_data_frame_jdf, sql_ctx)
        self._spark_data_frame = new_df
        return new_df




    def __init__(
            self,
            data=None,
            column_info=None,
            write_factors_as_indexes: bool = False):
        if not rx_pyspark_check_pyspark_is_loaded():
            raise Exception("RxSparkDataFrame can only be used in pyspark interoperation, "
                            "while 'pyspark' module has not been imported. " +
                            "Please import 'pyspark' module.")
        if data is not None and not rx_pyspark_check_data_is_spark_data_frame(data):
            raise Exception("Parameter 'data' needs to be python.sql.DataFrame or None.")

        super().__init__(
            column_info=column_info,
            write_factors_as_indexes=write_factors_as_indexes,
            from_spark_data=True
        )
        self.add_serialization_property(pc.class_id, "RxSparkData")
        self.add_serialization_property(pc.df_type_id, "spark-data-frame")
        self._spark_data_frame = data
        if data is None:
            self.add_serialization_property(pc.df_source_id, "N/A")
        else:
            self.add_serialization_property(pc.df_source_id, "spark-rdd %s" % str(data.rdd.id()))
            self._put_spark_df_to_rre_server(data)
        self.add_serialization_property(pc.cache_id, False)
        self._object_id = id(self)

    def __del__(self):
        try:
            # only allow the initial RxSparkDataFrame obj to cleanup df
            # To avoid dill load obj or copy obj to do cleanup
            if self._object_id != id(self):
                return
            sc = rx_pyspark_get_existing_pyspark_app()
            if sc is not None:
                sc._jvm.com.microsoft.scaler.spark.RREServer.removeSparkDataFrame(self.df_name)
        except:
            # ignore any exceptions if fail to cleanup df in RREServer when __del__
            pass



