from .RxOdbcData import RxOdbcData
from .constants import PropertyConstants as pc


class RxSqlServerData(RxOdbcData):
    """
    .. rxtitle::

        Generate SqlServer Data Source Object

    .. rxdescription::

        Main generator for class RxSqlServerData, which extends RxDataSource.

    :param connection_string: None or character string specifying the 
        connection string.
    :param table: None or character string specifying the table name. Cannot be 
        used with sql_query.
    :param sql_query: None or character string specifying a valid SQL select 
        query. Cannot contain hidden characters such as tabs or newlines. Cannot be 
        used with table. If you want to use TABLESAMPLE clause in sqlQuery, set 
        row_buffering argument to False.
    :param row_buffering: Bool value specifying whether or not to buffer rows on 
        read from the database. If you are having problems with your ODBC driver, 
        try setting this to False.
    :param return_data_frame: Bool value indicating whether or not to convert the 
        result from a list to a data frame (for use in rxReadNext only). If False, 
        a list is returned.
    :param string_as_factors: Bool value indicating whether or not to 
        automatically convert strings to factors on import. This can be overridden 
        by specifying "character" in column_classes and column_info. If True, the 
        factor levels will be coded in the order encountered. Since this factor 
        level ordering is row dependent, the preferred method for handling factor 
        columns is to use column_info with specified "levels".
    :param column_classes: Dictionary of column name to strings specifying the 
        column types to use when converting the data. The element names for the 
        vector are used to  identify which column should be converted to which type.
        Allowable column types are:
        "bool" (stored as uchar),
        "integer" (stored as int32),
        "float32" (the default for floating point data for ‘.xdf’ files),
        "numeric" (stored as float64 as in R),
        "character" (stored as string),
        "factor" (stored as uint32),
        "int16" (alternative to integer for smaller storage space),
        "uint16" (alternative to unsigned integer for smaller storage space),
        "Date" (stored as Date, i.e. float64)

        Note for "factor" type, the levels will be coded in the order 
        encountered. Since this factor level ordering is row dependent, the 
        preferred method for handling factor columns is to use column_info with 
        specified "levels".
        Note that equivalent types share the same bullet in the list above; for 
        some types we allow both ‘R-friendly’ type names, as well as our own, 
        more specific type names for ‘.xdf’ data.
        Note also that specifying the column as a “factor” type is currently 
        equivalent to “string” - for the moment, if you wish to import a column 
        as factor data you must use the column_info argument, documented below.

    :param column_info: List of named variable information lists. Each variable 
        information list contains one or more of the named elements given below. 
        The information supplied for column_info overrides that supplied for 
        column_classes.
        Currently available properties for a column information list are:
        type: Character string specifying the data type for the column. See 
            column_classes argument description for the available types. Specify 
            "factorIndex" as the type for 0-based factor indexes. levels must also 
            be specified.
        newName: Character string specifying a new name for the variable.
        description: character string specifying a description for the variable.
        levels: List of strings containing the levels when type = "factor". If 
            the levels property is not provided, factor levels will be determined 
            by the values in the source column. If levels are provided, any value 
            that does not match a provided level will be converted to a missing 
            value.
        newLevels: New or replacement levels specified for a column of type 
            “factor”. It must be used in conjunction with the levels argument. 
            After reading in the original data, the labels for each level will be 
            replaced with the newLevels.
        low: The minimum data value in the variable (used in computations using 
            the F() function.
        high: The maximum data value in the variable (used in computations 
            using the F() function.
    :param rows_per_read: Number of rows to read at a time.
    :param verbose: Integer value. If 0, no additional output is printed. If 1, 
        information on the odbc data source type (odbc or odbcFast) is printed.
    :param use_fast_read: Bool value, specifying whether or not to use a direct 
        ODBC connection. On Linux systems, this is the only ODBC connection 
        available.
    :param server: Target SQL Server instance. Can also be specified in the 
        connection string with the Server keyword.
    :param database_name: Database on the target SQL Server instance. Can also 
        be specified in the connection string with the Database keyword.
    :param user: SQL login to connect to the SQL Server instance. SQL login can 
        also be specified in the connection string with the uid keyword.
    :param password: Password associated with the SQL login. Can also be 
        specified in the connection string with the pwd keyword.
    :param write_factors_as_indexes: Bool value, if True, when writing to an 
        RxOdbcData data source, underlying factor indexes will be written instead 
        of the string representations.
    :param kwargs: Additional arguments to be passed directly to the underlying 
        functions.
    
    :return: object of class :py:class:`RxSqlServerData <microsoftml.RxSqlServerData>`.
    
    .. rxkeywords:: datasource, sql

    .. rxexample:: 
        from revoscalepy import RxSqlServerData, rx_data_step

        connection_string="Driver=SQL Server;Server=.;Database=RevoTestDB;Trusted_Connection=True"
        query="select top 100 [ArrDelay],[CRSDepTime],[DayOfWeek] FROM airlinedemosmall"

        ds = RxSqlServerData(sql_query = query, connection_string = connection_string)
        df = rx_data_step(ds)
        df.head()

    """

    def __init__(
            self,
            connection_string: str = None,
            table: str = None,
            sql_query: str = None,
            row_buffering: bool = True,
            return_data_frame: bool = True,
            string_as_factors: bool = False,
            column_classes: dict = None,
            column_info: dict = None,
            rows_per_read: int = 50000,
            verbose: bool = False,
            use_fast_read: bool = True,
            server: str = None,
            database_name: str = None,
            user: str = None,
            password: str = None,
            write_factors_as_indexes: bool = False,
            **kwargs):

        import revoscalepy.RxFunction as arg

        if arg.get_optional_parameter("dsn", **kwargs) is not None:
            raise ValueError("Parameter 'dsn' is not supported ")

        super().__init__(
            connection_string=connection_string,
            table=table,
            sql_query=sql_query,
            row_buffering=row_buffering,
            return_data_frame=return_data_frame,
            string_as_factors=string_as_factors,
            column_classes=column_classes,
            column_info=column_info,
            rows_per_read=rows_per_read,
            verbose=verbose,
            use_fast_read=use_fast_read,
            server=server,
            database_name=database_name,
            user=user,
            password=password,
            write_factors_as_indexes=write_factors_as_indexes,
            **kwargs
        )

        self.add_serialization_property(pc.class_id, "RxSqlServerData")
        self.add_serialization_property(pc.use_sql_satellite_id, self.to_serialize[pc.use_fast_read])
        self.add_serialization_property(pc.in_sql_server_id, False)

    def set_in_sql_server(self, in_sql_server: bool):
        self.to_serialize[pc.in_sql_server_id] = in_sql_server

    def __str__(self) -> str:
        return  "RxSqlServerData Source:\n" \
                "-----------------------\n" \
                "{0}".format(self._print_details())