import os
import uuid
import re
from typing import Union

from .. RxSerializable import rx_native_call
from .. computecontext.RxLocalSeq import RxLocalSeq
from .. computecontext.RxComputeContext import RxComputeContext
from .RxFileData import RxFileData
from .RxFileSystem import RxFileSystem
from .RxFileSystem import RxHdfsFileSystem
from .constants import PropertyConstants as pc

class RxXdfData(RxFileData):
    """
    .. rxtitle::

        Generate Xdf Data Source Object

    .. rxdescription::

        Main generator for class RxXdfData, which extends RxDataSource.

    :param file: Character string specifying the location of the data. For single 
        Xdf, it is a ‘.xdf’ file. For composite Xdf, it is a directory like 
        ‘/tmp/airline’. 
    :param vars_to_keep: List of strings of variable names to keep around during 
        operations. If None, argument is ignored. Cannot be used with vars_to_drop.
    :param vars_to_drop: list of strings of variable names to drop from 
        operations. If None, argument is ignored. Cannot be used with vars_to_keep.
    :param return_data_frame: Bool value indicating whether or not to convert the 
        result to a data frame.
    :param strings_as_factors: Bool value indicating whether or not to convert 
        strings into factors (for reader mode only).
    :param blocks_per_read: Number of blocks to read for each chunk of data read 
        from the data source.
    :param file_system: Character string or RxFileSystem object indicating type 
        of file system; "native" or RxNativeFileSystem object can be used for the 
        local operating system. If None, the file system will be set to that in 
        the current compute context, if available, otherwise the fileSystem option.
    :param create_composite_set: Bool value or None. Used only when writing. 
        If True, a composite set of files will be created instead of a single ‘.xdf’ 
        file. Subdirectories ‘data’ and ‘metadata’ will be created. In the ‘data’ 
        subdirectory, the data will be split across a set of ‘.xdfd’ files (see 
        blocks_per_composite_file below for determining how many blocks of data will be 
        in each file). In the ‘metadata’ subdirectory there is a single ‘.xdfm’ file, 
        which contains the meta data for all of the ‘.xdfd’ files in the ‘data’ 
        subdirectory.
    :param create_partition_set: Bool value or None. Used only when writing. 
        If True, a set of files for partitioned Xdf will be created when assigning 
        this RxXdfData object for outData of rxPartition. Subdirectories ‘data’ and 
        ‘metadata’ will be created. In the ‘data’ subdirectory, the data will be 
        split across a set of ‘.xdf’ files (each file stores data of a single data 
        partition, see rxPartition for details). In the ‘metadata’ subdirectory there 
        is a single ‘.xdfp’ file, which contains the meta data for all of the ‘.xdf’ 
        files in the ‘data’ subdirectory. The partitioned Xdf object is currently 
        supported only in rxPartition and rxGetPartitions
    :param blocks_per_composite_file: Integer value. If 
        create_composite_set=True, this will be the number of blocks put into each 
        '.xdfd' file in the composite set.

    :return: Object of class :py:class:`RxXdfData <microsoftml.RxXdfData>`.

    .. rxkeywords:: datasource, xdf

    .. rxexample:: 
        import os
        from revoscalepy import rx_data_step, RxOptions, RxXdfData
        sample_data_path = RxOptions.get_option("sampleDataDir")
        ds = RxXdfData(os.path.join(sample_data_path, "kyphosis.xdf"))
        kyphosis = rx_data_step(ds)

    """


    def __init__(self,
                 file: str,
                 vars_to_keep=None,
                 vars_to_drop=None,
                 return_data_frame=True,
                 strings_as_factors=False,
                 blocks_per_read=1,
                 file_system: Union[str, RxFileSystem]="native",
                 create_composite_set=None,
                 create_partition_set=None,
                 blocks_per_composite_file=3):
        """
        Constructs an 'RxXdfData' object.

        :param file: Character string specifying the location of the data.
        :param vars_to_keep: List of strings of variable names to keep around during operations. If None, argument is ignored. Cannot be used with varsToDrop.
        :param vars_to_drop: List of strings of variable names to drop from operations. If None, argument is ignored. Cannot be used with varsToKeep.
        :param return_data_frame: Bool value indicating whether or not to convert the result to a data frame when reading with read_next. If False, a list is returned when reading read_next.
        :param strings_as_factors: Bool value indicating whether or not to convert strings into factors.
        :param blocks_per_read: Number of blocks to read for each chunk of data from the data source.
        :param file_system: Character string indicating the type of file system; 'native'. If None, the file system will be set to that in the current compute context, if available, otherwise the fileSystem option.
        :param create_composite_set: Bool or None. Used only when writing. If True, a composite set of files will be created instead of a single '.xdf' file.
        :param create_partition_set: Bool or None. User only when writing. If True, As set of files for partitioned Xdf will be created when assigning this RxXdfData object for outData of rxPartition.
        :param blocks_per_composite_file: Integer value. If createCompositeSet=True, and if the compute context is not RxHadoopMR, this will be the number of blocks put into each '.xdfd' file in the composite set.
                When importing is being done on Hadoop using MapReduce, the number of rows assigned to each MapReduceTask, ans the number of blocks per '.xdfd' file is therefore determined by rowsPerRead.
        """

        if type(blocks_per_read) is not int or blocks_per_read < 0:
            raise ValueError("'blocks_per_read' must be a positive integer")

        super().__init__()

        # Add fixed properties as serialization properties
        self.add_serialization_property(pc.file_id, file)
        self.add_serialization_property(pc.return_data_frame_id, return_data_frame)
        self.add_serialization_property(pc.strings_as_factors_id, strings_as_factors)
        self.add_serialization_property(pc.rows_or_blocks_per_read_id, blocks_per_read)
        self.add_serialization_property(pc.class_id, "RxXdfData")
        self.add_serialization_property(pc.create_composite_set_id, create_composite_set)
        self.add_serialization_property(pc.create_partition_set_id, create_partition_set)
        self.add_serialization_property(pc.blocks_per_composite_file_id, blocks_per_composite_file)
        self.add_serialization_property(pc.cache_id, False)
        self.add_serialization_property(pc.xdf_uuid_id, str(uuid.uuid4()))
        self.add_serialization_property(pc.df_name_id, "df-" + str(uuid.uuid4()))
        self.add_serialization_property(pc.df_type_id, "xdf")
        self.add_serialization_property(pc.df_source_id, file)
        self.add_serialization_property(pc.read_by_block_id, True)

        self._file_system = file_system

        if vars_to_keep is not None or vars_to_drop is not None:
            from revoscalepy.utils.RxUtils import _rx_resolve_var_list
            column_names = _rx_resolve_var_list(self, vars_to_keep, vars_to_drop)
            self.column_names = ','.join(column_names)

    def _update_info(self):
        self._info = {}
        self._info = rx_native_call("GetInfo", {"File": self.file, "ReturnObject": True,
                                                "fileSystem": self.file_system.to_serialize})
    @property
    def number_rows(self):
        self._update_info()
        return None if "numRows" not in self._info else self._info["numRows"]

    @property
    def df_name(self) -> str:
        return self.to_serialize[pc.df_name_id]

    @property
    def xdf_uuid(self) -> str:
        return self.to_serialize[pc.xdf_uuid_id]

    @property
    def df_type(self) -> str:
        return self.to_serialize[pc.df_type_id]

    @property
    def df_source(self) -> str:
        return self.to_serialize[pc.df_source_id]

    @df_source.setter
    def _df_source(self, source) -> str:
        self.to_serialize[pc.df_source_id] = source

    @property
    def cache(self) -> bool:
        return self.to_serialize[pc.cache_id]

    @cache.setter
    def cache(self,cache):
        self.to_serialize[pc.cache_id] = cache

    @property
    def read_by_block(self) -> bool:
        return self.to_serialize[pc.read_by_block_id]

    @read_by_block.setter
    def read_by_block(self, value):
        self.to_serialize[pc.read_by_block_id] = value

    @property
    def create_partition_set(self) -> bool:
        return self.to_serialize[pc.create_partition_set_id]

    def head(self, num_rows:int=6, report_progress:int=None):
        from revoscalepy.etl.RxDataStep import rx_data_step
         
        # Use Local Compute context to execute head.
        old_compute_context = RxComputeContext.set_current(RxLocalSeq())
        try:
            results = rx_data_step(self, number_rows_read=num_rows, report_progress=report_progress)
        except Exception:
            raise
        finally:
            RxComputeContext.set_current(old_compute_context)
        return results

    def tail(self, num_rows: int=6, report_progress: int=None):
        from revoscalepy.functions.RxGetInfoXdf import rx_get_info
        from revoscalepy.etl.RxDataStep import rx_data_step
        from revoscalepy.RxSerializable import _resolve_transform_function

        # Use Local Compute context to execute tail.
        old_compute_context = RxComputeContext.set_current(RxLocalSeq())
        try:
            info = rx_get_info(self)
            source_num_rows = info['numRows']
            if source_num_rows < num_rows:
                num_rows = source_num_rows

            start_row = int(source_num_rows - num_rows)
            end_row = int(start_row + num_rows)

            # Add 1 to start_row since R is 1-based
            results = rx_data_step(self, start_row=start_row + 1,
                                  number_rows_read=num_rows,
                                  report_progress=report_progress)

            # Range is exclusive (<) for end_row
            results.index = range(int(start_row), int(end_row))
        except Exception: 
            raise
        finally:
            RxComputeContext.set_current(old_compute_context)
        return results

    def extract_info(self) -> dict:
        infos = super().extract_info()
        infos["sourceType"] = self.df_type
        infos["dfName"] = self.df_name
        infos["xdfId"] = self.xdf_uuid
        infos["source"] = self.df_source
        infos["cache"] = self.cache

        return infos







