from typing import Union
from collections import OrderedDict
from pandas import DataFrame

from revoscalepy import RxComputeContext, RxInSqlServer, RxRemoteComputeContext
from revoscalepy import RxXdfData, RxOdbcData, RxTextData, RxDataSource, RxSqlServerData
from revoscalepy.utils.RxUtils import _rx_set_verbose_flags

from revoscalepy.computecontext.RxSpark import RxSpark
from revoscalepy.RxFunction import RxFunction, validate_argument
from revoscalepy.RxSerializable import rx_native_call, resolve_dataframe, _resolve_transform_function

from revoscalepy.utils.RxUtils import rx_get_out_data_source
from revoscalepy.utils import on_exit_handler, telemetry_capture_call
from revoscalepy.etl.RxImport import _rx_create_sample_data_file

@on_exit_handler
def rx_data_step(input_data: Union[RxDataSource, DataFrame, str] = None,
                    output_file: Union[str, RxXdfData, RxTextData] = None,
                    vars_to_keep: list = None,
                    vars_to_drop: list = None,
                    row_selection: str = None,
                    transforms: dict = None,
                    transform_objects: list = None,
                    transform_function=None,
                    transform_variables: list = None,
                    transform_packages: list = None,
                    transform_environment=None,
                    append: list = None,
                    overwrite: bool = False,
                    row_variable_name: str = None,
                    remove_missings_on_read: bool = False,
                    remove_missings: bool = False,
                    compute_low_high: bool = True,
                    max_rows_by_cols: int = 3000000,
                    rows_per_read: int = -1,
                    start_row: int = 1,
                    number_rows_read: int = -1,
                    return_transform_objects: bool = False,
                    blocks_per_read: int = None,
                    report_progress: int = None,
                    xdf_compression_level: int = 0,
                    strings_as_factors: bool = None,
                    **kwargs) -> Union[RxXdfData, DataFrame]:
    """
    .. rxtitle::

        Data Step for RevoScaleR data sources

    .. rxdescription::

        Transform data from an input data set to an output data set

    :param input_data: A character string with the path for the data to import
        (delimited, fixed format, ODBC, or XDF). Alternatively, a data source
        object representing the input data source can be specified.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param output_file: A character string representing the output ‘.xdf’ file,
        or a RxXdfData object. If None, a data frame will be returned in memory.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object.
    :param vars_to_keep: list of strings of variable names to include when
        reading from the input data file. If None, argument is ignored. Cannot be
        used with vars_to_drop. Not supported for ODBC or fixed format text files.
    :param vars_to_drop: List of strings of variable names to exclude when
        reading from the input data file. If None, argument is ignored. Cannot be
        used with vars_to_keep. Not supported for ODBC or fixed format text files.
    :param row_selection: None. Not currently supported, reserved for future use.
    :param transforms: None. Not currently supported, reserved for future use.
    :param transform_objects: None. Not currently supported, reserved for
        future use.
    :param transform_function: Variable transformation function. See
        rxTransform for details.
    :param transform_variables: List of strings of input data set variables
        needed for the transformation function. See rxTransform for details.
    :param transform_packages: None. Not currently supported, reserved for future use.
    :param transform_environment: None. Not currently supported, reserved for future use.
    :param append: Either "none" to create a new ‘.xdf’ file or "rows" to
        append rows to an existing ‘.xdf’ file. If output_file exists and append is
        "none", the overwrite argument must be set to True. Ignored if a data frame
        is returned.
    :param overwrite: Bool value. If True, the existing outData will be
        overwritten. Ignored if a dataframe is returned.
    :param row_variable_name: Character string or None. If inData is a data.frame:
        If None, the data frame's row names will be dropped. If a character string, an
        additional variable of that name will be added to the data set containing the
        data frame's row names. If a data.frame is being returned, a variable with the
        name row_variable_name will be removed as a column from the data frame and will be
        used as the row names.
    :param remove_missings_on_read: Bool value. If True, rows with missing
        values will be removed on read.
    :param remove_missings: Bool value. If True, rows with missing values will
        not be included in the output data.
    :param compute_low_high: Bool value. If False, low and high values will not
        automatically be computed. This should only be set to False in special
        circumstances, such as when append is being used repeatedly. Ignored for data
        frames.
    :param max_rows_by_cols: The maximum size of a data frame that will be returned
        if output_file is set to None and inData is an ‘.xdf’ file , measured by the number
        of rows times the number of columns. If the number of rows times the number of
        columns being created from the ‘.xdf’ file exceeds this, a warning will be
        reported and the number of rows in the returned data frame will be truncated.
        If max_rows_by_cols is set to be too large, you may experience problems from
        loading a huge data frame into memory.
    :param rows_per_read: Number of rows to read for each chunk of data read from
        the input data source. Use this argument for finer control of the number of
        rows per block in the output data source. If greater than 0, blocks_per_read is
        ignored. Cannot be used if input_data is the same as output_file. The default value of
        -1 specifies that data should be read by blocks according to the
        blocks_per_read argument.
    :param start_row: The starting row to read from the input data source. Cannot
        be used if input_data is the same as output_file.
    :param number_rows_read: Number of rows to read from the input data source. If
        remove_missings are used, the output data set may have fewer
        rows than specified by number_rows_read. Cannot be used if input_data is the same
        as output_file.
    :param return_transform_objects: Bool value. If True, the list of
        transformObjects will be returned instead of a data frame or data source
        object. If the input transformObjects have been modified, by using .rxSet or
        .rxModify in the transformFunc, the updated values will be returned. Any data
        returned from the transformFunc is ignored. If no transformObjects are used,
        None is returned. This argument allows for user-defined computations within a
        transform_function without creating new data.
    :param blocks_per_read: Number of blocks to read for each chunk of data read
        from the data source. Ignored for data frames or if rows_per_read is positive.
    :param report_progress: Integer value with options:
        0: no progress is reported.
        1: the number of processed rows is printed and updated.
        2: rows processed and timings are reported.
        3: rows processed and all timings are reported.
    :param xdf_compression_level: Integer in the range of -1 to 9. The higher the
        value, the greater the amount of compression - resulting in smaller files but a
        longer time to create them. If xdf_compression_level is set to 0, there will be
        no compression and files will be compatible with the 6.0 release of Revolution
        R Enterprise. If set to -1, a default level of compression will be used.
    :param strings_as_factors: Bool indicating whether or not to
        automatically convert strings to factors on import. This can be overridden
        by specifying "character" in column_classes and column_info. If True, the factor
        levels will be coded in the order encountered. Since this factor level
        ordering is row dependent, the preferred method for handling factor columns
        is to use column_info with specified "levels".
    :param kwargs: Additional arguments to be passed to the input data source.

    :return: If an output_file is not specified, an output data frame is
        returned. If an output_file is specified, an RxXdfData data source is
        returned that can be used in subsequent revoscalepy analysis.

    .. rxkeywords:: datasource

    .. rxexample::
        import os
        from revoscalepy import RxOptions, RxXdfData, rx_data_step
        sample_data_path = RxOptions.get_option("sampleDataDir")
        kyphosis = RxXdfData(os.path.join(sample_data_path, "kyphosis.xdf"))
        kyphosis_df = rx_data_step(kyphosis)
        kyphosis_df.head()
    """
    telemetry_capture_call()

    input_data = RxComputeContext.current().validate_in_data(input_data, "input_data")
    output_file = RxComputeContext.current().validate_out_data(output_file, "output_file")

    if transform_function is not None:
        transform_function = _resolve_transform_function(transform_function=transform_function)

    py_args = locals()
    for key in kwargs:
        py_args[key] = kwargs[key]
    py_args.pop("kwargs", None)
    py_function_name = "rx_data_step"
    py_module_name = __name__

    compute_context = RxComputeContext.current()

    # todo need to refactor the spark workflow with Keith
    if isinstance(compute_context, RxSpark):
        return compute_context._execute(py_module_name, py_function_name, py_args)

    # handle duplicate columns in varsToKeep, varsToDrop 
    if vars_to_keep is not None:
        vars_to_keep = list(set(vars_to_keep))
    if vars_to_drop is not None:
        vars_to_drop = list(set(vars_to_drop))

    function_name = "DataStep"
    call_parameters = {}

    compute_context = RxComputeContext.current()

    append_eval = ["none"] if append is None else RxFunction.assign_or_throw(append, False, list)

    call_parameters['DataFrameEnvironment'] = None #TODO look at this
    call_parameters['OutFile'] = None #looks unused
    call_parameters['InFile'] = None #looks unused
    call_parameters['ColSelection'] = None if vars_to_keep is None or len(vars_to_keep) == 0 else ','.join(vars_to_keep)
    call_parameters['ColsToDrop'] = None if vars_to_drop is None or len(vars_to_drop) == 0 else ','.join(vars_to_drop)
    call_parameters['ReportMissingVars'] = True  # todo check this
    call_parameters['CreateNewOutFile'] = False
    call_parameters['StartBlock'] = 1
    call_parameters['blocksPerRead'] = blocks_per_read
    call_parameters['startRow'] = start_row
    call_parameters['numRows'] = number_rows_read
    call_parameters['rowSelectionVar'] = row_selection
    call_parameters['hasRowSelection'] = row_selection is None
    call_parameters['RemoveMissingsOnRead'] = remove_missings_on_read
    call_parameters['RemoveMissingsOnWrite'] = remove_missings
    call_parameters['xdfCompressionLevel'] = xdf_compression_level
    call_parameters['ComputeLowHigh'] = compute_low_high
    call_parameters['overwrite'] = overwrite

    _rx_set_verbose_flags(call_parameters, verbose = 0, report_progress = report_progress)

    call_parameters['appendRows'] = append_eval is not None and "rows" in append_eval
    call_parameters['appendCols'] = append_eval is not None and "cols" in append_eval
    call_parameters['transforms'] = RxFunction.assign_or_throw(transforms, not_implemented=True)
    call_parameters['transformObjects'] = RxFunction.assign_or_throw(transform_objects, True)
    call_parameters['transformFunc'] = RxFunction.assign_or_throw(transform_function, True)
    call_parameters['transformVars'] = RxFunction.assign_or_throw(transform_variables, True)
    call_parameters['transformEnvir'] = RxFunction.assign_or_throw(transform_environment, True)
    call_parameters['transformPackages'] = RxFunction.assign_or_throw(transform_packages, True)

    if output_file is not None:
        output_file = rx_get_out_data_source(output_file)
        call_parameters['outSource'] = RxFunction.resolve_datasource(output_file)

    call_parameters['ReturnDataFrame'] = output_file is None and return_transform_objects is False
    call_parameters['MaxRowsByCols'] = RxFunction.assign_or_throw(max_rows_by_cols, False, int)

    if isinstance(input_data, DataFrame):
        call_parameters['..rxDataFrame'] = resolve_dataframe(input_data)

        faddcp_param = OrderedDict()
        faddcp_param['bComputeLowHigh'] = True
        call_parameters['FaddCP'] = faddcp_param

        call_parameters['readVars'] = input_data.columns.tolist()
    elif issubclass(type(input_data), RxOdbcData):
        dict = OrderedDict()
        dict['table'] = input_data.table
        dict['sqlQuery'] = input_data.sql_query
        dict['connectionString'] = input_data.connection_string
        dict['class'] = 'RxOdbcData'
        dict['useFastRead'] = True
        call_parameters['inSource'] = dict
    else:
        call_parameters['inSource'] = RxFunction.resolve_datasource(input_data)

    if 'inSource' in call_parameters:
        data = call_parameters['inSource']
    elif '..rxDataFrame' in call_parameters:
        data = call_parameters['..rxDataFrame']
    else:
        raise ValueError()

    # strings_as_factors
    if strings_as_factors is not None \
       and isinstance(input_data, RxDataSource):
        strings_as_factors = validate_argument(strings_as_factors, "strings_as_factors", (bool), False)
        input_data.strings_as_factors = strings_as_factors

    # rows_per_read
    rows_per_read = validate_argument(rows_per_read, "rows_per_read", int, -1)
    if rows_per_read > 0:
        if isinstance(input_data, RxDataSource):
            input_data.rows_or_blocks_per_read = rows_per_read

        if isinstance(input_data, RxXdfData):
            input_data.read_by_block = False
    call_parameters['rowsPerRead'] = rows_per_read

    #
    # capture sample data if its a remote compute context call
    #
    # print("Using compute context {0}".format(type(compute_context).__name__))

    if isinstance(compute_context, RxRemoteComputeContext):
        job = compute_context.create_job()
        _rx_create_sample_data_file(
            input_data=input_data,
            sample_data_file=job.sample_data_file
            )
        if isinstance(data, RxSqlServerData):
            job.set_input_data_query(data.input_data_query)
            job.set_use_sql_satellite(data.useFastRead)
            data.set_in_sql_server(True)

    #print("Calling function {0}".format(function_name))
    #print("Serializing parameters: ")

    #
    # make the C++ call
    #
    #print(compute_context)
    if isinstance(compute_context, RxRemoteComputeContext):
        def process_result(ret):
            ret = ret if len(ret) == 0 else ret[0]

            return ret

        from revoscalepy import rx_exec
        results = rx_exec(function=rx_native_call, args={'functionname': function_name, 'params': call_parameters},
                          compute_context=compute_context, callback=process_result)
    else:
        results = rx_native_call(
            functionname=function_name, params=call_parameters)

    return results




