import warnings

from typing import Union
from pandas import DataFrame
from pathlib import Path

from .. RxSerializable import rx_native_call
from .. RxSerializable import resolve_dataframe, _resolve_transform_function
from .. RxFunction import RxFunction
from .. RxFunction import validate_argument
from .. utils.RxOptions import RxOptions
from .. utils.RxTrace import RxTraceLevel, rx_runtime_trace

from revoscalepy.computecontext.RxRemoteComputeContext import RxComputeContext
from revoscalepy.computecontext.RxSpark import RxSpark
from revoscalepy.datasource import RxDataSource, RxFileData, RxXdfData, RxSparkData
from revoscalepy.utils.RxUtils import rx_get_in_data_source, rx_get_out_data_source
from revoscalepy.utils.RxUtils import _rx_resolve_var_list, _rx_set_verbose_flags

from revoscalepy.utils import on_exit_handler
from revoscalepy.utils import telemetry_capture_call

#
# import data source
#
@on_exit_handler
def rx_import(input_data: Union[RxDataSource, DataFrame, str],
              output_file=None,
              vars_to_keep: list = None,
              vars_to_drop: list = None,
              row_selection: str = None,
              transforms: dict = None,
              transform_objects: dict = None,
              transform_function: callable = None,
              transform_variables: dict = None,
              transform_packages: dict = None,
              transform_environment: dict = None,
              append: str = None,
              overwrite: bool = False,
              number_rows: int = None,
              strings_as_factors: bool = None,
              column_classes: dict = None,
              column_info: dict = None,
              rows_per_read: int = None,
              type: str = None,
              max_rows_by_columns: int = None,
              report_progress: int = None,
              verbose: int = None,
              xdf_compression_level: int = None,
              create_composite_set: bool = None,
              blocks_per_composite_file: int = None):
    """
    .. rxtitle::

        Import Data to .xdf or data frame

    .. rxdescription::

        Import data into an ‘.xdf’ file or data.frame.

    :param input_data: A character string with the path for the data to import
        (delimited, fixed format, ODBC, or XDF). Alternatively, a data source
        object representing the input data source can be specified.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param output_file: A character string representing the output ‘.xdf’ file
        or an RxXdfData object.
        If None, a data frame will be returned in memory.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object.
    :param vars_to_keep: List of strings of variable names to include when
        reading from the input data file. If None, argument is ignored. Cannot be
        used with vars_to_drop. Not supported for ODBC or fixed format text files.
    :param vars_to_drop: List of strings of variable names to exclude when
        reading from the input data file. If None, argument is ignored. Cannot be
        used with vars_to_keep. Not supported for ODBC or fixed format text files.
    :param row_selection: None. Not currently supported, reserved for future use.
    :param transforms: None. Not currently supported, reserved for future use.
    :param transform_objects: None. Not currently supported, reserved for
        future use.
    :param transform_function: Variable transformation function.
    :param transform_variables: List of strings of input data set variables
        needed for the transformation function.
    :param transform_packages: None. Not currently supported, reserved for future use.
    :param transform_environment: None. Not currently supported, reserved for future use.
    :param append: Either "none" to create a new ‘.xdf’ file or "rows" to
        append rows to an existing ‘.xdf’ file. If output_file exists and append is
        "none", the overwrite argument must be set to True. Ignored if a data frame
        is returned.
    :param overwrite: Bool value. If True, the existing output_file will be
        overwritten. Ignored if a dataframe is returned.
    :param number_rows: Integer value specifying the maximum number of rows to
        import. If set to -1, all rows will be imported.
    :param strings_as_factors: Bool value indicating whether or not to
        automatically convert strings to factors on import. This can be overridden
        by specifying "character" in column_classes and column_info. If True, the factor
        levels will be coded in the order encountered. Since this factor level
        ordering is row dependent, the preferred method for handling factor columns
        is to use column_info with specified "levels".
    :param column_classes: Dictionary of column name to strings specifying the column types
        to use when converting the data. The element names for the vector are used to
        identify which column should be converted to which type.
                Allowable column types are:
                    "bool" (stored as uchar),
                    "integer" (stored as int32),
                    "float32" (the default for floating point data for ‘.xdf’ files),
                    "numeric" (stored as float64 as in R),
                    "character" (stored as string),
                    "factor" (stored as uint32),
                    "ordered" (ordered factor stored as uint32. Ordered factors are
        treated the same as factors in RevoScaleR analysis functions.),
                    "int16" (alternative to integer for smaller storage space),
                    "uint16" (alternative to unsigned integer for smaller storage space),
                    "Date" (stored as Date, i.e. float64. Not supported for import
        types "textFast", "fixedFast", or "odbcFast".)
                    "POSIXct" (stored as POSIXct, i.e. float64. Not supported for
        import types "textFast", "fixedFast", or "odbcFast".)
                Note for "factor" and "ordered" types, the levels will be coded in the
        order encountered. Since this factor level ordering is row dependent,
        the preferred method for handling factor columns is to use column_info with
        specified "levels".
                Note that equivalent types share the same bullet in the list above; for
        some types we allow both ‘R-friendly’ type names, as well as our own,
        more specific type names for ‘.xdf’ data.
                Note also that specifying the column as a “factor” type is currently
        equivalent to “string” - for the moment, if you wish to import a column
        as factor data you must use the column_info argument, documented below.
    :param column_info: List of named variable information lists. Each variable
        information list contains one or more of the named elements given below.
        When importing fixed format data, either column_info or an an ‘.sts’ schema
        file should be supplied. For fixed format text files, only the variables
        specified will be imported. For all text types, the information supplied
        for column_info overrides that supplied for column_classes.
        Currently available properties for a column information list are:
            type: Character string specifying the data type for the column. See
                column_classes argument description for the available types. If the
                type is not specified for fixed format data, it will be read as
                character data.
            newName: Character string specifying a new name for the variable.
                description: character string specifying a description for the
                variable.
            levels: List of strings containing the levels when type =
                "factor". If the levels property is not provided, factor levels
                will be determined by the values in the source column. If levels
                are provided, any value that does not match a provided level will
                be converted to a missing value.
            newLevels: New or replacement levels specified for a column of type
                “factor”. It must be used in conjunction with the levels argument.
                After reading in the original data, the labels for each level will
                be replaced with the newLevels.
            low: The minimum data value in the variable (used in computations
                using the F() function.)
            high: The maximum data value in the variable (used in computations
                using the F() function.)
            start: The left-most position, in bytes, for the column of a fixed
                format file respectively. When all elements of column_info have start,
                the text file is designated as a fixed format file. When none of
                the elements have it, the text file is designated as a delimited
                file. Specification of start must always be accompanied by
                specification of width.
            width: The number of characters in a fixed-width character column
                or the column of a fixed format file. If width is specified for a
                character column, it will be imported as a fixed-width character
                variable. Any characters beyond the fixed width will be ignored.
                Specification of width is required for all columns of a fixed
                format file.
            decimalPlaces: The number of decimal places.
    :param rows_per_read: Number of rows to read at a time.
    :param type: Character string set specifying file type of input_data. This is
        ignored if input_data is a data source. Possible values are:
        "auto": File type is automatically detected by looking at file
            extensions and argument values.
        "textFast": Delimited text import using faster, more limited import
            mode. By default variables containing the values True and False or T
            and F will be created as bool variables.
        "text": Delimited text import using enhanced, slower import mode. This
            allows for importing Date and POSIXct data types, handling the
            delimiter character inside a quoted string, and specifying decimal
            character and thousands separator. (See RxTextData.)
        "fixedFast": Fixed format text import using faster, more limited import
            mode. You must specify a ‘.sts’ format file or column_info specifications
            with start and width for each variable.
        "fixed": Fixed format text import using enhanced, slower import mode.
            This allows for importing Date and POSIXct data types and specifying
            decimal character and thousands separator.
            You must specify a ‘.sts’ format file or column_info specifications with
            start and width for each variable.
        "odbcFast": ODBC import using faster, more limited import mode.
        "odbc": ODBC import using slower, enhanced import on Windows. (See
            RxOdbcData.)
    :param max_rows_by_columns: The maximum size of a data frame that will be
        read in if output_file is set to None, measured by the number of rows times the
        number of columns. If the number of rows times the number of columns being
        imported exceeds this, a warning will be reported and a smaller number of
        rows will be read in than requested. If max_rows_by_columns is set to be too
        large, you may experience problems from loading a huge data frame into
        memory.
    :param report_progress: Integer value with options:
        0: no progress is reported.
        1: the number of processed rows is printed and updated.
        2: rows processed and timings are reported.
        3: rows processed and all timings are reported.
    :param verbose: Integer value. If 0, no additional output is printed. If 1,
        information on the import type is printed if type is set to auto.
    :param xdf_compression_level: Integer in the range of -1 to 9. The higher
        the value, the greater the amount of compression - resulting in smaller
        files but a longer time to create them. If xdfCompressionLevel is set to 0,
        there will be no compression and files will be compatible with the 6.0
        release of Revolution R Enterprise. If set to -1, a default level of
        compression will be used.
    :param create_composite_set: Bool value or None. If True, a composite
        set of files will be created instead of a single ‘.xdf’ file. A directory
        will be created whose name is the same as the ‘.xdf’ file that would
        otherwise be created, but with no extension. Subdirectories ‘data’ and
        ‘metadata’ will be created. In the ‘data’ subdirectory, the data will be
        split across a set of ‘.xdfd’ files (see blocks_per_composite_file below for
        determining how many blocks of data will be in each file). In the
        ‘metadata’ subdirectory there is a single ‘.xdfm’ file, which contains the
        meta data for all of the ‘.xdfd’ files in the ‘data’ subdirectory.
    :param blocks_per_composite_file: Integer value. If
        create_composite_set=True, this will be the number of blocks put into
        each ‘.xdfd’ file in the composite set. If the output_file is an RxXdfData
        object, set the value for blocks_per_composite_file there instead.
    :param kwargs: Additional arguments to be passed directly to the underlying
        data source objects to be imported.

    :return: If an output_file is not specified, an output data frame is
        returned. If an output_file is specified, an RxXdfData data source is
        returned that can be used in subsequent revoscalepy analysis.

    .. rxkeywords:: import, datasource

    .. rxexample::
        import os
        from revoscalepy import rx_import, RxOptions, RxXdfData
        sample_data_path = RxOptions.get_option("sampleDataDir")
        ds = RxXdfData(os.path.join(sample_data_path, "kyphosis.xdf"))
        kyphosis = rx_import(input_data = ds)
        kyphosis.head()

    """
    telemetry_capture_call()
    input_data = RxComputeContext.current().validate_in_data(input_data, "input_data")
    output_file = RxComputeContext.current().validate_out_data(output_file, "output_file")

    if transform_function is not None:
        transform_function = _resolve_transform_function(transform_function=transform_function)

    py_args = locals()
    py_function_name = "rx_import"
    py_module_name = __name__

    compute_context = RxComputeContext.current()

    # todo need to refactor the spark workflow with Keith
    if isinstance(compute_context, RxSpark):
        return compute_context._execute(py_module_name, py_function_name, py_args)

    input_data = validate_argument(input_data, "input_data", (RxDataSource, DataFrame, str))
    input_data = rx_get_in_data_source(input_data)
    append = validate_argument(append, "append", str, "none")

    if output_file is not None:
        output_file = rx_get_out_data_source(output_file)

        if isinstance(output_file, RxFileData):
            test = Path(output_file.file)
            if test.is_file() and overwrite is not True and append != 'rows':
                raise RuntimeError("File {} exists and overwrite=True or append='row' not specified.".format(output_file.file))

    if vars_to_drop is not None or vars_to_keep is not None:
        column_names = _rx_resolve_var_list(input_data, vars_to_keep, vars_to_drop)
        rx_runtime_trace("column names: " + ','.join(column_names), RxTraceLevel.DEBUG)
        if isinstance(input_data, DataFrame):
            input_data = input_data[column_names]
        elif isinstance(input_data, (RxFileData, RxSparkData)):
            input_data.column_names = ','.join(column_names)
        else:
            warnings.warn("'vars_to_keep' and 'vars_to_drop' are not supported.")

    row_selection = row_selection
    transforms = validate_argument(transforms, "transforms", dict, {})
    transform_objects = validate_argument(transform_objects, "transform_objects", dict, {})
    transform_variables = validate_argument(transform_variables, "transform_variables", dict, {})
    transform_packages = validate_argument(transform_packages, "transform_packages", dict, {})
    transform_environment = validate_argument(transform_environment, "transform_environment", dict, {})
    overwrite = validate_argument(overwrite, "overwrite", bool, False)
    number_rows = validate_argument(number_rows, "number_rows", None, -1)

    # column_classes
    if column_classes is not None and isinstance(input_data, RxDataSource):
        #column_classes = validate_argument(column_classes, "column_classes", (dict), {})
        input_data.column_classes = column_classes

    # column_info
    if column_info is not None and isinstance(input_data, RxDataSource):
        column_info = validate_argument(column_info, "column_info", (dict), {})
        input_data.column_info = column_info

    # strings_as_factors
    if strings_as_factors is not None and isinstance(input_data, RxDataSource):
        strings_as_factors = validate_argument(strings_as_factors, "strings_as_factors", bool, False)
        input_data.strings_as_factors = strings_as_factors
    else:
        strings_as_factors = validate_argument(strings_as_factors, "strings_as_factors", bool, False)

    # rows_per_read
    if rows_per_read is not None:
        if rows_per_read <= 0:
            raise ValueError("Parameter 'rows_per_read' must be a positive integer.")

        rows_per_read = validate_argument(rows_per_read, "rows_per_read", int, -1)
        if isinstance(input_data, RxDataSource):
            input_data.rows_or_blocks_per_read = rows_per_read

        if isinstance(input_data, RxXdfData):
            input_data.read_by_block = False
    else:
        rows_per_read = validate_argument(rows_per_read, "rows_per_read", int, -1)

    column_classes = validate_argument(column_classes, "column_classes", dict, {})
    column_info = validate_argument(column_info, "column_info", dict, {})
    type = validate_argument(type, "type", str, "auto")
    max_rows_by_columns = validate_argument(max_rows_by_columns, "max_rows_by_columns", (int), {})

    xdf_compression_level = validate_argument(xdf_compression_level, "xdf_compression_level", (int),
                                              RxOptions.get_option("XdfCompressionLevel"))
    create_composite_set = validate_argument(create_composite_set, "create_composite_set", bool, False)
    blocks_per_composite_file = validate_argument(blocks_per_composite_file, "blocks_per_composite_file", int, 3)

    #
    # serialize the request
    #
    params = dict()

    params["transformFunc"] = transform_function
    params["transformVars"] = transform_variables

    if isinstance(input_data, RxDataSource):
        params["inSource"] = RxFunction.resolve_datasource(input_data)
    else:
        params["inSource"] = {'..rxDataFrame': resolve_dataframe(input_data)}

    params["outSource"] = {}
    if output_file is not None:
        if not isinstance(output_file, (RxXdfData, RxSparkData)):
            raise ValueError("rx_import currently only supports " +
                             "RxXdfData, RxHiveData, RxParquetData, RxOrcData, RxSparkDataFrame"
                             + " as output. Use rx_data_step for other output files.")
        params["outSource"] = output_file.serialize_bxl()

    if append == "none":
        params["append"] = 0
    else:
        params["append"] = 1

    params["env"] = {}

    params["NumRows"] = number_rows
    params["MaxRowsByCols"] = max_rows_by_columns

    _rx_set_verbose_flags(params, verbose = verbose, report_progress = report_progress)

    params["xdfCompressionLevel"] = float(xdf_compression_level)
    params["createCompositeSet"] = create_composite_set
    params["blocksPerCompositeFile"] = float(blocks_per_composite_file)
    params["overwrite"] = overwrite

    #
    # make the C++ call
    #
    result = rx_native_call(functionname="ImportDataSource", params=params)

    return result

def _rx_create_sample_data_file(
    input_data,
    sample_data_file : str
    ):
    try:
        old_strings_as_factors = input_data.strings_as_factors
        input_data.strings_as_factors = False
        rx_import(
            input_data=input_data,
            output_file=RxXdfData(file=sample_data_file),
            number_rows=100,
            verbose=0,
            report_progress=0
            )
    finally:
        input_data.strings_as_factors = old_strings_as_factors

