#############################################################
# Copyright (c) Microsoft Corporation.  All rights reserved.
#############################################################
from collections import OrderedDict
from pandas import DataFrame
from typing import Union
from numpy import float64
import warnings

from .. datasource.RxDataSource import RxDataSource
from .. RxFunction import validate_argument
from .. utils.RxTelemetryLogger import telemetry_capture_call
from .  RxGetInfoXdf import GetVarInfoResults, rx_get_var_info

# Retrieve subset of information from the var_info for the given column
def _get_column_info(column_name : str,
                     var_info : dict,
                     include_low_high : bool = False,
                     factors_only : bool = False,
                     sort_levels : bool = False,
                     use_factor_index : bool = False) -> dict:

    if var_info is None or 'varType' not in var_info.keys() :
        raise ValueError("Could not get var info for : {0}".format(column_name))

    if factors_only and var_info['varType'] != 'factor' :
        return None

    column_info = dict()

    column_info['type'] = var_info['varType']
    if var_info['storage'] == 'float32' :
        column_info['type'] = 'float32'
    elif use_factor_index and var_info['varType'] == 'factor' :
        column_info['type'] = 'factorIndex'
        if sort_levels :
            warnings.warn("'sort_levels' = True cannot be used with 'use_factor_index' = True, resetting 'sort_levels' = False")
            sort_levels = False

    if include_low_high :
        if 'low' in var_info.keys() :
            column_info['low'] = var_info['low']
            if isinstance(column_info['low'], float64) :
                column_info['low'] = float(column_info['low'])
        if 'high' in var_info.keys() :
            column_info['high'] = var_info['high']
            if isinstance(column_info['high'], float64) :
                column_info['high'] = float(column_info['high'])

    if 'levels' in var_info.keys() and len(var_info['levels']) > 0 :
        column_info['levels'] = var_info['levels']

    if sort_levels and var_info['varType'] == 'factor' and 'levels' in var_info.keys() :
        try :
            # try converting to integer values and then sorting
            # level is the factor label
            int_levels = [ int(level) for level in var_info['levels'] ]
            int_levels.sort()
            column_info['levels'] = [ str(level) for level in int_levels ]
        except Exception as e:
            column_info['levels'] = sorted(var_info['levels'],
                                           key = lambda val : (val is None, val))

    return column_info

#############################################################
def rx_create_col_info(data : Union[RxDataSource, str, DataFrame, GetVarInfoResults],
                       include_low_high : bool = False,
                       factors_only : bool = False,
                       vars_to_keep : list = None,
                       sort_levels : bool = False,
                       compute_info : bool = True,
                       use_factor_index : bool = False) -> OrderedDict :
    """
    .. rxtitle::

        Function to generate a 'column_info' ordered dictionary from a data source

    .. rxdescription::

        From the data source, generates an ordered dictionary of dictionaries with variable names as keys 
        and column information dictionaries as values. It can be used in rx_import or an RxDataSource constructor.
        This function can be used to ensure consistent categorical (factor) levels when importing 
        a series of text files to xdf. It is also useful for repeated analysis on non-xdf data sources.

    :param data : An RxDataSource object, a character string containing an '.xdf' file name, or a data frame.
         An object returned from rx_get_var_info is also supported.
    :param include_low_high : Bool value. If True, the low/high values will be included in the column_info object.
         Note that this will override any actual low/high values in the data set if the column_info object
         is applied to a different data source.
    :param factors_only : Bool value. If True, only column information for categorical (factor) variables
         will be included in the output.
    :param vars_to_keep : None to include all variables, or list of strings of variable names to include
    :param sort_levels : Bool value. If True, categorical (factor) levels will be sorted. 
         If categorical levels represent integers, they will be put in numeric order.
    :param compute_info : Bool value. If True, a pass through the data will be taken for non-xdf data sources to compute
         categorical (factor) levels and low/high values.
    :param use_factor_index : Bool value. If True, the factor_index variable type will be used instead of factor

    :return : Ordered dict column_info of named variable information dicts. It can be used as input for rx_import
        and in data sources such as RxTextData and RxSqlServerData
        Each variable information dict contains one or more of the named elements given below..
            type: Character string specifying the data type for the column.
            levels: List of strings containing the levels when type = "factor".
            low: The minimum data value in the variable (used in computations using the F() function).
                 For factors, it is the index of the lowest level.
            high: The maximum data value in the variable (used in computations using the F() function).
                  For factors, it is the index of the highest level.

    .. rxseealso::

        :py:class:`RxDataSource <revoscalepy.datasource.RxDataSource.RxDataSource>`,
        :py:class:`RxTextData <revoscalepy.datasource.RxTextData.RxTextData>`,
        :py:class:`RxSqlServerData <revoscalepy.datasource.RxSqlData.RxSqlServerData>`,
        :py:class:`RxOdbcData <revoscalepy.datasource.RxOdbcData.RxOdbcData>`,
        :py:class:`RxXdfData <revoscalepy.datasource.RxXdfData.RxXdfData>`,
        :py:func:`rx_import <revoscalepy.functions.RxImport.rx_import>`.

    .. rxkeywords:: column, information, col, info

    .. rxexample::
        :execute:
        :print_output:

        import os
        from revoscalepy import RxOptions, RxTextData, rx_create_col_info, rx_import, rx_logit, rx_predict
        # Get the low/high values and factor levels before using a data source
        # for rx_import or analysis

        # Create a text data source, specifying that 'yearsEmploy' should be a factor
        mort = os.path.join(RxOptions.get_option("sampleDataDir"), "mortDefaultSmall2000.csv")
        mort_ds = RxTextData(file = mort, column_classes = {'yearsEmploy' : "factor", 'default' : "logical"})

        # By default, rx_create_col_info will make a pass through the data to compute factor levels
        # and low/high values.  We'll also request that the levels be sorted
        mort_column_info = rx_create_col_info(data = mort_ds, include_low_high = True, sort_levels = True)

        # Re-create the data source, now using the computed colInfo
        mort_ds = RxTextData(file = mort, column_info = mort_column_info)
        # Import the data
        mort_df = rx_import(mort_ds)
        print(mort_df['yearsEmploy'])

        # Or use the text data source directly in an analysis 
        # (not needing a pass through the data to compute the factor levels)
        logit_obj = rx_logit('default~yearsEmploy', data = mort_ds)

        ##############################################################################################

        # Train a model on one imported data set, then score using another
        # Train a model on the first year of the data, importing it from text to a data frame

        mort1 = os.path.join(RxOptions.get_option("sampleDataDir"), "mortDefaultSmall2000.csv")
        mort1_ds = RxTextData(file = mort1, column_classes = {'yearsEmploy' : "factor", 'default' : "logical"})
        # Since we haven't specified factor levels, they will be created 'first come, first serve'
        mort1_df = rx_import(mort1_ds)
        mort1_df['yearsEmploy'].cat.categories
        # Estimate a logit model
        logit_obj = rx_logit('default~yearsEmploy', data = mort1_df)

        # Now import the second year of data
        mort2 = os.path.join(RxOptions.get_option("sampleDataDir"), "mortDefaultSmall2001.csv")
        mort2_ds = RxTextData(file = mort2, column_classes = {'yearsEmploy' : "factor", 'default' : "logical"})
        mort2_df = rx_import(mort2_ds)
        # The levels are in a different order
        mort2_df['yearsEmploy'].cat.categories

        # If we try to use the model estimated from the first data set to predict on the seoond,
        # pred_out = rx_predict(logit_obj, data = mort2_df)
        # We will get an error
        #ERROR: order of factor levels in the data are inconsistent with
        #the order of the model coefficients

        # Instead, we can extract the col_info from the first data set
        mort_col_info = rx_create_col_info(data = mort1_df)

        # And use it when importing the second
        mort2_ds = RxTextData(file = mort2, column_info = mort_col_info)
        mort2_df = rx_import(mort2_ds)
        pred_out = rx_predict(logit_obj, data = mort2_df)
        pred_out.head()

       ###############################################################
       # Generating Column Information for a SQL Server Data Source 
       ###############################################################
       from revoscalepy import RxSqlServerData, rx_create_col_info, rx_set_temp_compute_context, RxInSqlServer
       connection_string="Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=TRUE;"
       sql_source = RxSqlServerData(connection_string = connection_string,
                                    table="AirlineDemoSmall",
                                    rows_per_read=500000,
                                    string_as_factors = True)

       with rx_set_temp_compute_context(RxInSqlServer(connection_string = connection_string)):
           results = rx_create_col_info(sql_source)
    """

    telemetry_capture_call()

    data = validate_argument(source = data, name = 'data', types = (DataFrame, str, RxDataSource, GetVarInfoResults))

    valid_columns_info = OrderedDict([])

    try :
        if isinstance(data, GetVarInfoResults) :
            vars_info = data
        else :
            vars_info = rx_get_var_info(data = data, vars_to_keep = vars_to_keep, compute_info = compute_info)

        assert isinstance(vars_info, GetVarInfoResults)

        all_columns_info = OrderedDict([(key,
                                        _get_column_info(column_name = key,
                                                         var_info = vars_info[key],
                                                         include_low_high = include_low_high,
                                                         factors_only = factors_only,
                                                         sort_levels = sort_levels,
                                                         use_factor_index = use_factor_index))
                                        for index, key in enumerate(vars_info)])

        # Remove those columns which have none as their dict values
        for column_name in all_columns_info.keys() :
            if all_columns_info[column_name] is not None :
                valid_columns_info[column_name] = all_columns_info[column_name]
    except ValueError as v:
        print('Exception occurred while creating column information : {0}'.format(v))

    return valid_columns_info