from array import array
from math import sqrt
from math import floor
from collections import OrderedDict
from random import SystemRandom
import struct
import warnings
from copy import deepcopy

from numpy import diag
from numpy import dot
from numpy import matrix
from numpy import maximum
from numpy import sum
from numpy import ndarray
from numpy import transpose
from numpy import ceil
from pandas import DataFrame
from pandas import concat

from revoscalepy.datasource import RxXdfData
from revoscalepy.datasource import RxSqlServerData
from revoscalepy.datasource.RxDataSource import RxDataSource

from .. RxFunction import RxFunction
from .. RxFunction import RxModelResults
from .. RxSerializable import RxSerializable
from .. RxSerializable import RxParameter
from .. computecontext.RxComputeContext import RxComputeContext
from .. computecontext.RxSpark import RxSpark
from .. computecontext.RxLocalSeq import RxLocalSeq
from .. computecontext.RxInSqlServer import RxRemoteComputeContext
from .. RxSerializable import rx_native_call
from .. RxSerializable import resolve_dataframe, _resolve_transform_function
from ..functions.RxExec import rx_exec
from .. etl.RxImport import rx_import
from .. utils.RxOptions import RxOptions
from .. utils.RxUtils import _rx_construct_read_vars, _rx_get_formula_components, \
    _rx_handle_formula, rx_get_out_data_source
from revoscalepy.utils import on_exit_handler
from revoscalepy.utils import telemetry_capture_call
from revoscalepy.etl.RxImport import _rx_create_sample_data_file

# Python's sys.maxsize is not suitable for our uses because
# we need the MaxInt available to bxlserver (C/C++ IntMax is not the same as python)
MAX_INT = 2 ** (struct.Struct('i').size * 8 - 1) - 1
RANDOM = SystemRandom()


def __format_dtree_results__(results):
    """Formatting code where we don't want to incur cost for each time the property is accessed"""
    isplit = results['isplit']
    dsplit = results['dsplit']
    inode = results['inode']
    dnode = results['dnode']

    if isinstance(results['xvars'], list):
        xvars = ["<leaf>"] + results['xvars']
    else:
        xvars = ["<leaf>"] + [results['xvars']]


    newcols = ['CP', 'nsplit', 'rel error', 'xerror', 'xstd']

    if len(results['cptable'].shape) > 1 and results['cptable'].shape[1] == len(newcols):
        results['cptable'] = DataFrame(transpose(results['cptable']))

        replace_columns = results['cptable'].columns.tolist()

        for i in range(0, len(replace_columns)):
            replace_columns[i] = newcols[i]

        results['cptable'].columns = replace_columns
        results['cptable'].low_high = None
        results['cptable'].type_info = None

    splits_data = OrderedDict([('count', isplit[:, 1]),
                               ('ncat', isplit[:, 2]),
                               ('improve', dsplit[:, 0]),
                               ('index', dsplit[:, 1]),
                               ('adj', dsplit[:, 2])])

    results['splits'] = DataFrame(data=splits_data, index=dsplit[:, 1])

    results['splits'].index = [xvars[int(name)] for name in isplit[:, 0]]

    is_ordered = DataFrame([results['isord']])
    is_ordered.columns = [results['xvars']]
    number_categories = results['cats']
    number_ordered_splits = sum([is_ordered.as_matrix()[0][int(x - 1)] for x in (isplit[:, 0])])
    number_categorical_splits = results['ncat']

    # TODO: if number_ordered_splits > 0 {}
    # else

    if results['nsplit'] == 0:
        node_id = results['inode'][:, 0]

        frame_data = OrderedDict([('var', '<leaf>'),
                                  ('n', inode[:, 4]),
                                  ('wt', dnode[:, 2]),
                                  ('dev', dnode[:, 0]),
                                  ('yval', dnode[:, 3]),
                                  ('complexity', dnode[:, 1]),
                                  ('ncompete', [max(0, inode[:, 2][x] - 1)
                                                          for x in range(0, len(inode[:, 2]))]),
                                  ('nsurrogate', inode[:, 3])])

        results['frame'] = DataFrame(data=frame_data, index=node_id)
    else:
        if(isinstance(results['csplit'], ndarray)):
            results['csplit'] += 2  # why do we do this?
            results['csplit'] = DataFrame(results['csplit'])
            results['csplit'].low_high = None
            results['csplit'].type_info = None


        pri_split_index = results['inode'][:, 1]
        var_splits = []

        for x in pri_split_index:
            if int(x) == 0:
                var_splits.append(xvars[0])
            else:
                var_splits.append(xvars[int(isplit[int(x) - 1, 0])])

        node_id = results['inode'][:, 0]

        frame_data = OrderedDict([('var', var_splits),
                                  ('n', inode[:, 4]),
                                  ('wt', dnode[:, 2]),
                                  ('dev', dnode[:, 0]),
                                  ('yval', dnode[:, 3]),
                                  ('complexity', dnode[:, 1]),
                                  ('ncompete', [max(0, inode[:, 2][x] - 1)
                                                          for x in range(0, len(inode[:, 2]))]),
                                  ('nsurrogate', inode[:, 3])])
        results['frame'] = DataFrame(data=frame_data, index=node_id)

    if results['method'] == 'class':
        dnodedf = DataFrame(results['dnode'])
        nylevels = len(results['ylevels'])
        numresp = results['numresp']
        nodeprob = dnodedf.iloc[:, numresp + 2] / dnodedf.iloc[0, numresp + 2]
        temp = dot(dnodedf.iloc[:, 4:4+nylevels], diag(results['parms']['prior'] * sum(results['counts']) / maximum(1, results['counts'])))

        divsor = DataFrame([sum(temp, axis=1)]).transpose()
        for loop in range(1, temp.shape[1]):
            divsor[loop] = divsor[0]

        yprob = temp / divsor
        yval2 = dnodedf.iloc[:, 3:3+nylevels+1]
        yval2 = concat([yval2, yprob, nodeprob], axis=1)

        newcols = []
        for x in range(0, len(yval2.columns)):
            if x is len(yval2.columns) - 1:
                newcols.append("yval2.nodeprob")
            else:
                newcols.append("yval2.V" + str(x))

        yval2.columns = newcols
        yval2.index = node_id

        results['frame'] = concat([results['frame'], yval2], axis=1)
    else:
        if 'parms' in results:
            results['parms'] = None

    # TODO sort
    sort_importance = False
    if 'variable.importance' in results:
        if sort_importance:
            results['variable.importance'][0].sort().reverse()
        # identify if the importance item is a list/dataframe or scalar
        if hasattr(results['variable.importance'], '__iter__'):
            results['variable.importance'] = DataFrame(data = results['variable.importance'], index = results['xvars'])
        else:
            results['variable.importance'] = DataFrame(data = [results['variable.importance']], index = [results['xvars']])

class RxDTreeControl(RxSerializable):
    @RxParameter("minsplit")
    def min_split(self):
        if self._min_split is not None:
            return self._min_split
        elif self._min_bucket is not None:
            return self._min_bucket * 3
        elif self._number_observations is not None:
            return max(20, floor(sqrt(self._number_observations)))
        else:
            return 20

    @RxParameter("minbucket")
    def min_bucket(self):
        if self._min_bucket is not None:
            return self._min_bucket
        else:
            return int(round(self.min_split() / 3))

    @RxParameter("maxdepth")
    def max_depth(self):
        return self._max_depth

    @RxParameter("cp")
    def cp(self):
        return self._cp

    @RxParameter("maxcompete")
    def max_compete(self):
        return self._max_compete

    @RxParameter("maxsurrogate")
    def max_surrogate(self):
        return self._max_surrogate

    @RxParameter("usesurrogate")
    def use_surrogate(self):
        return self._use_surrogate

    @RxParameter("surrogatestyle")
    def surrogate_style(self):
        return self._surrogate_style

    @RxParameter("xval")
    def x_val(self):
        return self._x_val

    def __init__(self,
                 number_observations=None,
                 min_split=None,
                 min_bucket=None,
                 max_depth=10,
                 cp=0,
                 max_compete=0,
                 max_surrogate=0,
                 use_surrogate=2,
                 surrogate_style=0,
                 x_val=2):
        self._number_observations = number_observations
        self._min_split = min_split
        self._min_bucket = min_bucket
        self._max_depth = max_depth
        self._cp = cp
        self._max_compete = max_compete
        self._max_surrogate = max_surrogate
        self._use_surrogate = use_surrogate
        self._surrogate_style = surrogate_style
        self._x_val = x_val

        if max_depth < 1:
            raise ValueError("max_depth must be >= 1")


class RxDTreeResults(RxModelResults):
    def __init__(self, formula, data, results, params, format_override: bool = False):
        super().__init__(
            model_type = 'rxDTree' if 'model_type' not in params else params['model_type'],
            formula = formula,
            params = params,
            data = data,
            compute_context = None,
            call = None
            )

        if format and not format_override:
            __format_dtree_results__(results)

        self._results = results

    @property
    def valid_observations(self):
        if 'valid.obs' in self._results:
            return self._results['valid.obs']
        else:
            return None

    @property
    def missing_observations(self):
        if 'missing.obs' in self._results:
            return self._results['missing.obs']
        else:
            return None

    @property
    def cp_table(self):
        if 'cptable' in self._results:
            return self._results['cptable']
        else:
            return None

    @property
    def splits(self):
        if 'splits' in self._results:
            return self._results['splits']
        else:
            return None

    @property
    def csplit(self):
        if 'csplit' in self._results:
            return self._results['csplit']
        else:
            return None

    @property
    def frame(self):
        if 'frame' in self._results:
            return self._results['frame']
        else:
            return None

    @property
    def where(self):
        if 'which' in self._results:
            return self._results['which']
        else:
            return None

    @property
    def cptable(self) -> DataFrame:
        if 'cptable' in self._results:
            ret = DataFrame(self._results['cptable'].transpose())
            ret.columns = ['CP', 'nsplit', 'relative error']
            return ret
        else:
            return None

    @property
    def importance(self) -> DataFrame:
        if 'variable.importance' in self._results:
            return self._results['variable.importance']
        else:
            return None

    @property
    def method(self) -> str:
        if 'method' in self._results:
            return self._results['method']

    @property
    def control(self):
        return self._params['control']

    @property
    def ordered(self):
        if 'isord' in self._results:
            ret = DataFrame([self._results['isord']]) # todo should this be unwrapped this much?
            ret.columns = [self._results['xvars']]
            return ret
        else:
            return None

    @property
    def x_levels(self):
        if 'xlevels' in self._results:
            return self._results['xlevels']
        else:
            return None

    @property
    def number_categorical_splits(self):
        if 'ncat' in self._results:
            return self._results['ncat']
        else:
            return None

    @property
    def parms(self):
        if 'parms' not in self._results or self._results['parms'] is None:
            return None

        parms = deepcopy(self._results['parms'])

        if 'split' in parms:
            if parms['split'] == "gini":
                parms['split'] = 1
            elif parms['split'] == "information":
                parms['split'] = 2
            else:
                parms['split'] = None

        if 'loss' in parms and 'ylevels' in self._results:
            cols = len(self._results['ylevels'])
            rows = int((parms['loss'].size) / cols)
            if (parms['loss'].shape[0] != cols):
                parms['loss'] = parms['loss'].reshape(rows, cols).transpose()

        return parms

    @property
    def ylevels(self):
        if 'ylevels' in self._results:
            return self._results['ylevels']
        else:
            return None

    def print(self):
        print(self)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        ret = "Number of valid observations: {}".format(self.valid_observations)
        ret += "\nNumber of missing observations: {}".format(self.missing_observations)
        ret += "\nImportance:\n{}".format(self.importance)
        ret += "\n\nTree representation:"
        ret += "\nn={}".format(self.valid_observations)

        nodes = self.frame.index
        split_nodes = []
        var_names = []
        categorical_splits = []
        index = self.splits['index']
        ncat = self.splits['ncat']

        for i in nodes:
            if self.frame['var'][i] != "<leaf>":
                split_nodes.append(i)
                var_names.append(self.frame['var'][i])

        left_split = [''] * len(split_nodes)
        right_split = [''] * len(split_nodes)

        handle_categorical = False
        for x in range(0, len(self.splits['ncat'])):
            if self.splits['ncat'].as_matrix()[x] > 1:
                handle_categorical = True
                categorical_splits.append(x)

        categorical_indicies = []
        x_level_keys = list(self.x_levels.keys())
        for x in var_names:
            for y in range(0, len(x_level_keys)):
                if x == x_level_keys[y]:
                    categorical_indicies.append(y)
                    break

        if handle_categorical:
            csplit = self._results['csplit']
            categorical_row = index[categorical_splits]
            for i in range(0, len(categorical_splits)):
                j = categorical_splits[i]
                blah = categorical_row.as_matrix()[i]
                split = csplit.iloc[[int(blah) - 1]]
                left_cat = []
                right_cat = []
                for x in range(0, len(split.columns)):
                    if split.as_matrix()[:, x][0] == 1:
                        left_cat.append(x)
                    if split.as_matrix()[:, x][0] == 3:
                        right_cat.append(x)

                to_join = []
                for x in left_cat:
                    to_join.append(self.x_levels[x_level_keys[categorical_indicies[i]]][x])
                left_split[j] = '=' + ','.join(to_join)

                to_join = []
                for x in right_cat:
                    to_join.append(self.x_levels[x_level_keys[categorical_indicies[i]]][x])
                right_split[j] = '=' + ','.join(to_join)

        parents = []
        for x in nodes:
            try:
                item = split_nodes.index(int(int(x) / 2))
                if x % 2 == 0:
                    parents.append(var_names[item] + left_split[item])
                else:
                    parents.append(var_names[item] + right_split[item])
            except ValueError:
                parents.append(-1)

        parents[0] = 'root'

        if not handle_categorical:
            parent_index = 0
            split_index = 0
            last_seen = -1
            stack = []
            operator_stack = []
            for x in nodes:
                if int(x) == 1:
                    parent_index += 1
                    continue

                if int(x) > last_seen + 1:
                    value = self._results['splits']['index'].iloc[split_index]
                    operator, push_operator = ('< ', '>= ') if int(self._results['splits']['ncat'].iloc[split_index]) is -1 else ('>= ', '< ')
                    parents[parent_index] = "{}{}{}".format(parents[parent_index], operator, value)
                    stack.append(value)
                    operator_stack.append(push_operator)
                    last_seen = int(x)
                    split_index += 1
                elif int(x) == last_seen + 1:
                    parents[parent_index] = "{}{}{}".format(parents[parent_index], operator_stack.pop(), stack.pop())
                    last_seen = int(x)
                else:
                    parents[parent_index] = "{}{}{}".format(parents[parent_index], operator_stack.pop(), stack.pop())
                    last_seen = int(x)
                parent_index += 1

        df_data = OrderedDict([('node', nodes),
                               ('split', parents),
                               ('n', self.frame['n']),
                               ('deviance', self.frame['dev']),
                               ('yval', self.frame['yval'])])
        df = DataFrame(data=df_data)

        return "{}\n\n{}".format(ret, df.__str__())


class RxDForestResults(RxDTreeResults):
    def __init__(self, formula, data, results, params):
        if format:
            self._forest = []
            for tree in results['forest']:
                self._forest.append(RxDTreeResults(formula, data, tree, params))

        super().__init__(formula, data, results, params, True)

        self._confusion = None

        self.process()

    @property
    def forest(self) -> list:
        return self._forest

    @property
    def ntree(self) -> int:
        return self._results['ntree']

    @property
    def type(self) -> str:
        return self._results['type']

    @property
    def mtry(self) -> int:
        return self._results['mtry']

    @property
    def cutoff(self) -> int:
        if 'cutoff' in self._results:
            return self._results['cutoff']
        else:
            return None

    @property
    def oob_err(self) -> DataFrame:
        if 'oob.err' in self._results:
            return self._results['oob.err']
        else:
            return None

    @property
    def confusion(self) -> DataFrame:
        return self._confusion

    @property
    def importance(self) -> DataFrame:
        if 'importance' in self._results:
            return self._results['importance']
        else:
            return None

    def __str__(self):
        ret = ''

        ret += 'Type of decision forest:  {}\n'.format(self.type)
        ret += 'Number of trees:  {}\n'.format(self.ntree)
        ret += 'Number of variables tried at each split:  {}\n'.format(self.mtry)

        if self.type == 'class' and 'confusion' in self._results:
            ret += 'OOB estimate of error rate:  {}\n'.format(round(self._results['oob.err']['oob.err'][self.ntree - 1] * 100, 2))
            ret += 'Confusion matrix:\n{}\n'.format(self.confusion)
        elif self.type == 'anova' and self.ntree > 0 and 'oob.err' in self._results:
            ret += 'Mean of squared residuals:  {}\n'.format(self._results['oob.err']['oob.err'][self.ntree - 1])
            ret += '% var explained:  {}\n'.format(round(self._results['oob.err']['oob.rsq'][self.ntree - 1], 2) * 100)

        ret += 'Importance:\n{}\n'.format(self.importance)

        return ret


    @staticmethod
    def create_results(formula, data, results, params) -> RxDTreeResults:
        if 'forest' in results:
            return RxDForestResults(formula, data, results, params)
        else:
            return RxDTreeResults(formula, data, results, params)

    def process(self):
        if 'confusion' in self._results and self._results['confusion'] is not None and len(self.forest) >= 1 and self.forest[0].ylevels:
            # @TODO: The labels of the confusion matrix are the y values of the trees so use those until dim information is available
            labels = self.forest[0].ylevels
            dim = len(labels)
            raw = self._results['confusion'].reshape(dim, dim).transpose()

            self._confusion = DataFrame(raw)
            self._confusion.columns = labels
            self._confusion.index = labels
            self._confusion['class.error'] = 1 - (diag(raw) / sum(raw, axis=1))
        elif self.oob_err is not None:
            self.oob_err['oob.rsq'] = 1 - (self.oob_err['oob.err'] / self.oob_err['oob.var'])
        if 'importance' in self._results:
            self._results['importance'].index = [self._forest[0]._results['xvars']]


@on_exit_handler
def rx_dforest(formula,
                  data,
                  output_file=None,
                  write_model_vars=False,
                  overwrite=False,
                  pweights=None,
                  fweights=None,
                  cost=None,
                  min_split=None,
                  min_bucket=None,
                  max_depth=10,
                  cp=0,
                  max_compete=0,
                  max_surrogate=0,
                  use_surrogate=2,
                  surrogate_style=0,
                  n_tree=10,
                  m_try=None,
                  replace=True,
                  cutoff=None,
                  strata=None,
                  sample_rate=None,
                  importance=False,
                  seed=None,
                  compute_cp_table=False,
                  compute_oob_error=1,
                  max_num_bins=None,
                  max_unordered_levels=32,
                  remove_missings=False,
                  use_sparse_cube=False,
                  find_splits_in_parallel=True,
                  row_selection=None,
                  transforms=None,
                  transform_objects=None,
                  transform_function=None,
                  transform_variables=None,
                  transform_packages=None,
                  transform_environment=None,
                  blocks_per_read=1,
                  report_progress=2,
                  verbose=0,
                  compute_context=None,
                  xdf_compression_level=1,
                  **kwargs):
    """
    .. rxtitle::

        Parallel External Memory Algorithm for Classification and Regression Decision Forests

    .. rxdescription::

        Fit classification and regression decision forests on an ‘.xdf’ file or
        data frame for small or large data using parallel external memory algorithm.

    :param formula: Statistical model using symbolic formulas.
    :param data: Either a data source object, a character string specifying a
        ‘.xdf’ file, or a data frame object.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param output_file: Either an RxXdfData data source object or a character
        string specifying the ‘.xdf’ file for storing the resulting node indices.
        If None, then no node indices are stored to disk. If the input data is a
        data frame, the node indices are returned automatically.
    :param write_model_vars: Bool value. If True, and the output file is
        different from the input file, variables in the model will be written to
        the output file in addition to the node numbers. If variables from the
        input data set are transformed in the model, the transformed variables will
        also be written out.
    :param overwrite: Bool value. If True, an existing output_file with an
        existing column named outColName will be overwritten.
    :param pweights: Character string specifying the variable of numeric values
        to use as probability weights for the observations.
    :param fweights: Character string specifying the variable of integer values
        to use as frequency weights for the observations.
    :param cost: A vector of non-negative costs, containing one element for
        each variable in the model. Defaults to one for all variables. When
        deciding which split: to choose, the improvement on splitting on a variable
        is divided by its cost.
    :param min_split: The minimum number of observations that must exist in a
        node before a split is attempted. By default, this is sqrt(num of obs). For
        non-XDF data sources, as (num of obs) is unknown in advance, it is wisest
        to specify this argument directly.
    :param min_bucket: The minimum number of observations in a terminal node
        (or leaf). By default, this is min_split/3.
    :param cp: Numeric scalar specifying the complexity parameter. Any split
        that does not decrease overall lack-of-fit by at least cp is not attempted.
    :param max_compete: The maximum number of competitor splits retained in the
        output. These are useful model diagnostics, as they allow you to compare
        splits in the output with the alternatives.
    :param max_surrogate: The maximum number of surrogate splits retained in
        the output. See the Details for a description of how surrogate splits are
        used in the model fitting. Setting this to 0 can greatly improve the
        performance of the algorithm; in some cases almost half the computation
        time is spent in computing surrogate splits.
    :param use_surrogate: An integer specifying how surrogates are to be used
        in the splitting process:
        0: Display-only; observations with a missing value for the primary
            split variable are not sent further down the tree.
        1: Use surrogates,in order, to split observations missing the primary
            split variable. If all surrogates are missing, the observation is not
            split.
        2: Use surrogates, in order, to split observations missing the primary
            split variable. If all surrogates are missing or max_surrogate=0, send
            the observation in the majority direction.
        The 0 value corresponds to the behavior of the tree function, and 2
            (the default) corresponds to the recommendations of Breiman et al.
    :param surrogate_style: An integer controlling selection of a best
        surrogate. The default, 0, instructs the program to use the total number of
        correct classifications for a potential surrogate, while 1 instructs the
        program to use the percentage of correct classification over the
        non-missing values of the surrogate. Thus, 0 penalizes potential surrogates
        with a large number of missing values.
    :param n_tree: A positive integer specifying the number of trees to grow.
    :param m_try: A positive integer specifying the number of variables to
        sample as split candidates at each tree node. The default values is
        sqrt(num of vars) for classification and (num of vars)/3 for regression.
    :param replace: A bool value specifying if the sampling of observations
        should be done with or without replacement.
    :param cutoff: (Classification only) A vector of length equal to the number
        of classes specifying the dividing factors for the class votes. The default
        is 1/(num of classes).
    :param strata: A character string specifying the (factor) variable to use
        for stratified sampling.
    :param sample_rate: A scalar or a vector of positive values specifying the
        percentage(s) of observations to sample for each tree:
        for unstratified sampling: A scalar of positive value specifying the
            percentage of observations to sample for each tree. The default is 1.0
            for sampling with replacement (i.e., replace=True) and 0.632 for
            sampling without replacement (i.e., replace=False).
        for stratified sampling: A vector of positive values of length equal to
            the number of strata specifying the percentages of observations to
            sample from the strata for each tree.
    :param importance: A bool value specifying if the importance of
        predictors should be assessed.
    :param seed: An integer that will be used to initialize the random number
        generator. The default is random. For reproducibility, you can specify the
        random seed either using set.seed or by setting this seed argument as part
        of your call.
    :param compute_oob_error: An integer specifying whether and how to compute
        the prediction error for out-of-bag samples:
        <0: never. This option may reduce the computation time.
        =0: only once for the entire forest.
        >0: once for each addition of a tree. This is the default.
    :param max_num_bins: The maximum number of bins to use to cut numeric data.
        The default is min(1001, max(101, sqrt(num of obs))). For non-XDF data
        sources, as (num of obs) is unknown in advance, it is wisest to specify
        this argument directly. If set to 0, unit binning will be used instead of
        cutting. See the ‘Details’ section for more information.
    :param max_unordered_levels: The maximum number of levels allowed for an
        unordered factor predictor for multiclass (>2) classification.
    :param remove_missings: Bool value. If True, rows with missing values
        are removed and will not be included in the output data.
    :param use_sparse_cube: Bool value. If True, sparse cube is used.
    :param find_splits_in_parallel: Bool value. If True, optimal splits for
        each node are determined using parallelization methods; this will typically
        speed up computation as the number of nodes on the same level is increased.
    :param row_selection: None. Not currently supported, reserved for future use.
    :param transforms: None. Not currently supported, reserved for future use.
    :param transform_objects: None. Not currently supported, reserved for future use.
    :param transform_function: Variable transformation function. The variables used
        in the transformation function must be specified in transform_variables if they
        are not variables used in the model.
    :param transform_variables: List of strings of input data set variables needed
        for the transformation function.
    :param transform_packages: None. Not currently supported, reserved for future use.
    :param transform_environment: None. Not currently supported, reserved for future use.
    :param blocks_per_read: number of blocks to read for each chunk of data read from
        the data source.
    :param report_progress: Integer value with options:
        0: No progress is reported.
        1: The number of processed rows is printed and updated.
        2: Rows processed and timings are reported.
        3: Rows processed and all timings are reported.
    :param verbose: Integer value. If 0, no additional output is printed. If 1,
        additional summary information is printed.
    :param compute_context: A RxComputeContext object for prediction.
    :param kwargs: Additional parameters

    :return: A RxDForestResults object of dtree model.

    .. rxkeywords:: learner, tree

    .. rxseealso::

        :py:func:`rx_predict <revoscalepy.rx_predict>`,
        :py:func:`rx_predict_rx_dforest <revoscalepy.rx_predict_rx_dforest>`.

    .. rxexample::
        import os
        from revoscalepy import rx_dforest, rx_import, RxOptions, RxXdfData
        sample_data_path = RxOptions.get_option("sampleDataDir")
        ds = RxXdfData(os.path.join(sample_data_path, "kyphosis.xdf"))
        kyphosis = rx_import(input_data = ds)

        # classification
        formula = "Kyphosis ~ Number + Start"
        method = "class"
        parms = {'prior': [0.8, 0.2], 'loss': [0, 2, 3, 0], 'split': "gini"}

        dforest = rx_dforest(formula, data = kyphosis, pweights = "Age", method = method,
                             parms = parms, cost = [2, 3], max_num_bins = 100, importance = True,
                             n_tree = 3, m_try = 2, sample_rate = 0.5,
                             replace = False, seed = 0, compute_oob_error = 1)

        # regression
        formula = "Age ~ Number + Start"
        method = "anova"
        parms = {'prior': [0.8, 0.2], 'loss': [0, 2, 3, 0], 'split': "gini"}

        dforest = rx_dforest(formula, data = kyphosis, pweights = "Kyphosis", method = method,
                             parms = parms, cost = [2, 3], max_num_bins = 100, importance = True,
                             n_tree = 3, m_try = 2, sample_rate = 0.5,
                             replace = False, seed = 0, compute_oob_error = 1)
    """
    telemetry_capture_call(compute_context)

    # TODO: R Sample starts at 1. Check to see what BxlServer supports
    input_seed = RANDOM.randint(1, MAX_INT) if seed is None else seed
    method = None if 'method' not in kwargs else kwargs['method']
    kwargs.pop('method', None)
    # NOTE: temp file won't get cleaned up if (compute_oob_error < 0 && allow_disk_write == False),
    #   as there won't be any good stage to do the cleanup, which we are trying to avoid here.
    if compute_oob_error < 0 and "allow_disk_write" not in kwargs:
        kwargs['allow_disk_write'] = False
    if compute_oob_error < 0 and output_file is not None:
        warnings.warn("The 'output_file' argument is ignored as compute_oob_error < 0 thus nothing to write.")
        output_file = None
    response = rx_dtree(formula=formula, data=data, output_file=output_file, write_model_vars=write_model_vars,
                           overwrite=overwrite, pweights=pweights, fweights=fweights, method = method, cost=cost, min_split=min_split,
                           min_bucket=min_bucket, max_depth=max_depth, cp=cp, max_compete=max_compete,
                           max_surrogate=max_surrogate, use_surrogate=use_surrogate, surrogate_style=surrogate_style,
                           x_val=1, n_tree=n_tree, m_try=m_try, replace=replace, cutoff=cutoff, strata=strata,
                           sample_rate=sample_rate, importance=importance, seed=input_seed,
                           compute_cp_table=compute_cp_table, compute_oob_error=compute_oob_error,
                           max_num_bins=max_num_bins, max_unordered_levels=max_unordered_levels,
                           remove_missings=remove_missings, use_sparse_cube=use_sparse_cube,
                           find_splits_in_parallel=find_splits_in_parallel, row_selection=row_selection,
                           transforms=transforms, transform_objects=transform_objects, transform_function = transform_function,
                           transform_variables = transform_variables, transform_packages = transform_packages,
                           transform_environment = transform_environment, blocks_per_read=blocks_per_read,
                           report_progress=report_progress, verbose=verbose, compute_context=compute_context,
                           xdf_compression_level=xdf_compression_level, model_type='rxDForest', **kwargs)

    return response


@on_exit_handler
def rx_btrees(formula,
                 data,
                 output_file=None,
                 write_model_vars=False,
                 overwrite=False,
                 pweights=None,
                 fweights=None,
                 cost=None,
                 min_split=None,
                 min_bucket=None,
                 max_depth=1,
                 cp=0,
                 max_compete=0,
                 max_surrogate=0,
                 use_surrogate=2,
                 surrogate_style=0,
                 n_tree=10,
                 m_try=None,
                 replace=False,
                 strata=None,
                 sample_rate=None,
                 importance=False,
                 seed=None,
                 loss_function="bernoulli",
                 learning_rate=0.1,
                 max_num_bins=None,
                 max_unordered_levels=32,
                 remove_missings=False,
                 compute_obs_node_id=None,
                 use_sparse_cube=False,
                 find_splits_in_parallel=True,
                 row_selection=None,
                 transforms=None,
                 transform_objects=None,
                 transform_function=None,
                 transform_variables=None,
                 transform_packages=None,
                 transform_environment=None,
                 blocks_per_read=1,
                 report_progress=2,
                 verbose=0,
                 compute_context=None,
                 xdf_compression_level=1,
                 **kwargs):
    """
    .. rxtitle::

        Parallel External Memory Algorithm for Stochastic Gradient Boosted Decision Trees

    .. rxdescription::

        Fit stochastic gradient boosted decision trees on an ‘.xdf’ file or data frame for
        small or large data using parallel external memory algorithm.

    :param formula: Statistical model using symbolic formulas.
    :param data: Either a data source object, a character string specifying a
        ‘.xdf’ file, or a data frame object.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param output_file: Either an RxXdfData data source object or a character
        string specifying the ‘.xdf’ file for storing the resulting node indices.
        If None, then no node indices are stored to disk. If the input data is a
        data frame, the node indices are returned automatically.
    :param write_model_vars: Bool value. If True, and the output file is
        different from the input file, variables in the model will be written to
        the output file in addition to the node numbers. If variables from the
        input data set are transformed in the model, the transformed variables will
        also be written out.
    :param overwrite: Bool value. If True, an existing output_file with an
        existing column named outColName will be overwritten.
    :param pweights: Character string specifying the variable of numeric values
        to use as probability weights for the observations.
    :param fweights: Character string specifying the variable of integer values
        to use as frequency weights for the observations.
    :param method: Character string specifying the splitting method. Currently,
        only "class" or "anova" are supported. The default is "class" if the
        response is a factor, otherwise "anova".
    :param parms: Optional list with components specifying additional
        parameters for the "class" splitting method, as follows:

        prior: A vector of prior probabilities. The priors must be positive and
            sum to 1. The default priors are proportional to the data counts.
        loss: A loss matrix, which must have zeros on the diagonal and positive
            off-diagonal elements. By default, the off-diagonal elements are set to 1.
        split: The splitting index, either gini (the default) or information.

        If parms is specified, any of the components can be specified or
            omitted. The defaults will be used for missing components.
    :param cost: A vector of non-negative costs, containing one element for
        each variable in the model. Defaults to one for all variables. When
        deciding which split: to choose, the improvement on splitting on a variable
        is divided by its cost.
    :param min_split: The minimum number of observations that must exist in a
        node before a split is attempted. By default, this is sqrt(num of obs). For
        non-XDF data sources, as (num of obs) is unknown in advance, it is wisest
        to specify this argument directly.
    :param min_bucket: The minimum number of observations in a terminal node
        (or leaf). By default, this is min_split/3.
    :param cp: Numeric scalar specifying the complexity parameter. Any split
        that does not decrease overall lack-of-fit by at least cp is not attempted.
    :param max_compete: The maximum number of competitor splits retained in the
        output. These are useful model diagnostics, as they allow you to compare
        splits in the output with the alternatives.
    :param max_surrogate: The maximum number of surrogate splits retained in
        the output. See the Details for a description of how surrogate splits are
        used in the model fitting. Setting this to 0 can greatly improve the
        performance of the algorithm; in some cases almost half the computation
        time is spent in computing surrogate splits.
    :param use_surrogate: An integer specifying how surrogates are to be used
        in the splitting process:
        0: Display-only; observations with a missing value for the primary
            split variable are not sent further down the tree.
        1: Use surrogates,in order, to split observations missing the primary
            split variable. If all surrogates are missing, the observation is not
            split.
        2: Use surrogates, in order, to split observations missing the primary
            split variable. If all surrogates are missing or max_surrogate=0, send
            the observation in the majority direction.
        The 0 value corresponds to the behavior of the tree function, and 2
            (the default) corresponds to the recommendations of Breiman et al.
    :param surrogate_style: An integer controlling selection of a best
        surrogate. The default, 0, instructs the program to use the total number of
        correct classifications for a potential surrogate, while 1 instructs the
        program to use the percentage of correct classification over the
        non-missing values of the surrogate. Thus, 0 penalizes potential surrogates
        with a large number of missing values.
    :param n_tree: A positive integer specifying the number of trees to grow.
    :param m_try: A positive integer specifying the number of variables to
        sample as split candidates at each tree node. The default values is
        sqrt(num of vars) for classification and (num of vars)/3 for regression.
    :param replace: A bool value specifying if the sampling of observations
        should be done with or without replacement.
    :param cutoff: (Classification only) A vector of length equal to the number
        of classes specifying the dividing factors for the class votes. The default
        is 1/(num of classes).
    :param strata: A character string specifying the (factor) variable to use
        for stratified sampling.
    :param sample_rate: A scalar or a vector of positive values specifying the
        percentage(s) of observations to sample for each tree:
        for unstratified sampling: A scalar of positive value specifying the
            percentage of observations to sample for each tree. The default is 1.0
            for sampling with replacement (i.e., replace=True) and 0.632 for
            sampling without replacement (i.e., replace=False).
        for stratified sampling: A vector of positive values of length equal to
            the number of strata specifying the percentages of observations to
            sample from the strata for each tree.
    :param importance: A bool value specifying if the importance of
        predictors should be assessed.
    :param seed: An integer that will be used to initialize the random number
        generator. The default is random. For reproducibility, you can specify the
        random seed either using set.seed or by setting this seed argument as part
        of your call.
    :param compute_oob_error: An integer specifying whether and how to compute
        the prediction error for out-of-bag samples:
        <0: never. This option may reduce the computation time.
        =0: only once for the entire forest.
        >0: once for each addition of a tree. This is the default.
    :param loss_function: Character string specifying the name of the loss function
        to use. The following options are currently supported:
        "gaussian": Regression: for numeric responses.
        "bernoulli": Regression: for 0-1 responses.
        "multinomial": Classification: for categorical responses with two or more levels.
    :param learning_rate: Numeric scalar specifying the learning rate of the boosting procedure.
    :param max_num_bins: The maximum number of bins to use to cut numeric data.
        The default is min(1001, max(101, sqrt(num of obs))). For non-XDF data
        sources, as (num of obs) is unknown in advance, it is wisest to specify
        this argument directly. If set to 0, unit binning will be used instead of
        cutting. See the ‘Details’ section for more information.
    :param max_unordered_levels: The maximum number of levels allowed for an
        unordered factor predictor for multiclass (>2) classification.
    :param remove_missings: Bool value. If True, rows with missing values
        are removed and will not be included in the output data.
    :param use_sparse_cube: Bool value. If True, sparse cube is used.
    :param find_splits_in_parallel: bool value. If True, optimal splits for
        each node are determined using parallelization methods; this will typically
        speed up computation as the number of nodes on the same level is increased.
    :param row_selection: None. Not currently supported, reserved for future use.
    :param transforms: None. Not currently supported, reserved for future use.
    :param transform_objects: None. Not currently supported, reserved for future use.
    :param transform_function: Variable transformation function. The variables used
        in the transformation function must be specified in transform_variables if they
        are not variables used in the model.
    :param transform_variables: List of strings of input data set variables needed
        for the transformation function.
    :param transform_packages: None. Not currently supported, reserved for future use.
    :param transform_environment: None. Not currently supported, reserved for future use.
    :param blocks_per_read: Number of blocks to read for each chunk of data read from
        the data source.
    :param report_progress: integer value with options:
        0: No progress is reported.
        1: The number of processed rows is printed and updated.
        2: Rows processed and timings are reported.
        3: Rows processed and all timings are reported.
    :param verbose: Integer value. If 0, no additional output is printed. If 1,
        additional summary information is printed.
    :param compute_context: A RxComputeContext object for prediction.
    :param kwargs: Additional parameters

    :return: A RxDForestResults object of dtree model.

    .. rxkeywords:: learner, tree

    .. rxseealso::

        :py:func:`rx_predict <revoscalepy.rx_predict>`,
        :py:func:`rx_predict_rx_dforest <revoscalepy.rx_predict_rx_dforest>`.

    .. rxexample::
        import os
        from revoscalepy import rx_btrees, rx_import, RxOptions, RxXdfData
        sample_data_path = RxOptions.get_option("sampleDataDir")
        ds = RxXdfData(os.path.join(sample_data_path, "kyphosis.xdf"))
        kyphosis = rx_import(input_data = ds)

        # classification
        form = "Kyphosis ~ Age + Number + Start"
        seed = 0
        n_trees = 50
        interaction_depth = 4
        n_min_obs_in_node = 1
        shrinkage = 0.1
        bag_fraction = 0.5
        distribution = "bernoulli"

        btree = rx_btrees(form, kyphosis, loss_function=distribution, n_tree=n_trees, learning_rate=shrinkage,
                        sample_rate=bag_fraction, max_depth=interaction_depth, min_bucket=n_min_obs_in_node, seed=seed,
                        replace=False, max_num_bins=200)

        # regression
        ds = RxXdfData(os.path.join(sample_data_path, "airquality.xdf"))
        df = rx_import(input_data = ds)
        df = df[df['Ozone'] != -2147483648]

        form = "Ozone ~ Solar.R + Wind + Temp + Month + Day"
        ntrees = 50
        interaction_depth = 3
        min_obs_in_node = 1
        shrinkage = 0.1
        bag_fraction = 0.5
        distribution = "gaussian"

        btree = rx_btrees(form, df, loss_function=distribution, n_tree=ntrees, learning_rate=shrinkage,
                          sample_rate=bag_fraction, max_depth=interaction_depth, min_bucket=min_obs_in_node, seed=0,
                          replace=False, max_num_bins=200, verbose=2)
    """
    telemetry_capture_call(compute_context)

    input_seed = RANDOM.randint(1, MAX_INT) if seed is None else seed
    # NOTE: compute_oob_error should always be 1 for boosting because we need
    # all predictions to compute the working response for the next iteration.
    if "compute_oob_error" in kwargs:
        if kwargs['compute_oob_error'] <= 0:
            warnings.warn("The 'compute_oob_error' argument is reset to 1 for rx_btrees.")
    # rx_btrees compute_oob_error should always be 1
    kwargs['compute_oob_error'] = 1
    response = rx_dtree(formula=formula, data=data, output_file=output_file,write_model_vars=write_model_vars,
                           overwrite=overwrite, pweights=pweights, fweights=fweights, cost=cost, min_split=min_split,
                           min_bucket=min_bucket, max_depth=max_depth, cp=cp, max_compete=max_compete,
                           max_surrogate=max_surrogate, use_surrogate=use_surrogate, surrogate_style=surrogate_style,
                           x_val=1, n_tree=n_tree, m_try=m_try, replace=replace, strata = strata, sample_rate=sample_rate,
                           importance=importance, seed=input_seed, loss_function=loss_function,
                           learning_rate=learning_rate, max_num_bins=max_num_bins,
                           max_unordered_levels=max_unordered_levels, remove_missings=remove_missings,
                           compute_obs_node_id=compute_obs_node_id, use_sparse_cube=use_sparse_cube,
                           find_splits_in_parallel=find_splits_in_parallel, row_selection=row_selection,
                           transforms=transforms, transform_objects=transform_objects, transform_function = transform_function,
                           transform_variables = transform_variables, transform_packages = transform_packages,
                           transform_environment = transform_environment, blocks_per_read=blocks_per_read,
                           report_progress=report_progress, verbose=verbose, compute_context=compute_context,
                           xdf_compression_level=xdf_compression_level,  compute_cp_table=False,
                           model_type='rxBTrees', **kwargs)

    return response

@on_exit_handler
def rx_dtree(formula,
                data,
                output_file=None,
                output_column_name=".rxNode",
                write_model_vars=False,
                extra_vars_to_write=None,
                overwrite=False,
                pweights=None,
                fweights=None,
                method=None,
                parms=None,
                cost=None,
                min_split=None,
                min_bucket=None,
                max_depth=10,
                cp=0,
                max_compete=0,
                max_surrogate=0,
                use_surrogate=2,
                surrogate_style=0,
                x_val=2,
                max_num_bins=None,
                max_unordered_levels=32,
                remove_missings=False,
                compute_obs_node_id=None,
                use_sparse_cube=False,
                find_splits_in_parallel=True,
                prune_cp=0,
                row_selection=None,
                transforms=None,
                transform_objects=None,
                transform_function=None,
                transform_variables=None,
                transform_packages=None,
                transform_environment=None,
                blocks_per_read=1,
                report_progress=2,
                verbose=0,
                compute_context=None,
                xdf_compression_level=1,
                **kwargs):
    """
    .. rxtitle::

        Parallel External Memory Algorithm for Classification and Regression Trees

    .. rxdescription::

        Fit classification and regression trees on an ‘.xdf’ file or data frame for
        small or large data using parallel external memory algorithm.

    :param formula: Statistical model using symbolic formulas.
    :param data: Either a data source object, a character string specifying a
        ‘.xdf’ file, or a data frame object.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param output_file: Either an RxXdfData data source object or a character
        string specifying the ‘.xdf’ file for storing the resulting node indices.
        If None, then no node indices are stored to disk. If the input data is a
        data frame, the node indices are returned automatically.
    :param output_column_name: Character string to be used as a column name for
        the resulting node indices if output_file is not None. Note that make.names is
        used on outColName to ensure that the column name is valid. If the output_file
        is an RxOdbcData source, dots are first converted to underscores. Thus, the
        default outColName becomes "X_rxNode".
    :param write_model_vars: Bool value. If True, and the output file is
        different from the input file, variables in the model will be written to
        the output file in addition to the node numbers. If variables from the
        input data set are transformed in the model, the transformed variables will
        also be written out.
    :param extra_vars_to_write: None or list of strings of additional
        variables names from the input data or transforms to include in the
        output_file. If writeModelVars is True, model variables will be included as well.
    :param overwrite: Bool value. If True, an existing output_file with an
        existing column named outColName will be overwritten.
    :param pweights: Character string specifying the variable of numeric values
        to use as probability weights for the observations.
    :param fweights: Character string specifying the variable of integer values
        to use as frequency weights for the observations.
    :param method: Character string specifying the splitting method. Currently,
        only "class" or "anova" are supported. The default is "class" if the
        response is a factor, otherwise "anova".
    :param parms: Optional list with components specifying additional
        parameters for the "class" splitting method, as follows:

        prior: A vector of prior probabilities. The priors must be positive and
            sum to 1. The default priors are proportional to the data counts.
        loss: A loss matrix, which must have zeros on the diagonal and positive
            off-diagonal elements. By default, the off-diagonal elements are set to 1.
        split: The splitting index, either gini (the default) or information.

        If parms is specified, any of the components can be specified or
            omitted. The defaults will be used for missing components.
    :param cost: A vector of non-negative costs, containing one element for
        each variable in the model. Defaults to one for all variables. When
        deciding which split to choose, the improvement on splitting on a variable
        is divided by its cost.
    :param min_split: The minimum number of observations that must exist in a
        node before a split is attempted. By default, this is sqrt(num of obs). For
        non-XDF data sources, as (num of obs) is unknown in advance, it is wisest
        to specify this argument directly.
    :param min_bucket: The minimum number of observations in a terminal node
        (or leaf). By default, this is min_split/3.
    :param cp: Numeric scalar specifying the complexity parameter. Any split
        that does not decrease overall lack-of-fit by at least cp is not attempted.
    :param max_compete: The maximum number of competitor splits retained in the
        output. These are useful model diagnostics, as they allow you to compare
        splits in the output with the alternatives.
    :param max_surrogate: The maximum number of surrogate splits retained in
        the output. Setting this to 0 can greatly improve the performance of the
        algorithm; in some cases almost half the computation time is spent in
        computing surrogate splits.
    :param use_surrogate: An integer specifying how surrogates are to be used
        in the splitting process:
        0: Display-only; observations with a missing value for the primary
            split variable are not sent further down the tree.
        1: Use surrogates,in order, to split observations missing the primary
            split variable. If all surrogates are missing, the observation is not
            split.
        2: Use surrogates, in order, to split observations missing the primary
            split variable. If all surrogates are missing or max_surrogate=0, send
            the observation in the majority direction.
        The 0 value corresponds to the behavior of the tree function, and 2
            (the default) corresponds to the recommendations of Breiman et al.
    :param x_val: The number of cross-validations to be performed along with
        the model building. Currently, 1:x_val is repeated and used to identify the
        folds. If not zero, the cptable component of the resulting model will
        contain both the mean (xerror) and standard deviation (xstd) of the
        cross-validation errors, which can be used to select the optimal
        cost-complexity pruning of the fitted tree. Set it to zero if external
        cross-validation will be used to evaluate the fitted model because a value
        of k increases the compute time to (k+1)-fold over a value of zero.
    :param surrogate_style: An integer controlling selection of a best
        surrogate. The default, 0, instructs the program to use the total number of
        correct classifications for a potential surrogate, while 1 instructs the
        program to use the percentage of correct classification over the
        non-missing values of the surrogate. Thus, 0 penalizes potential surrogates
        with a large number of missing values.
    :param max_depth: The maximum depth of any tree node. The computations take
        much longer at greater depth, so lowering max_depth can greatly speed up
        computation time.
    :param max_num_bins: The maximum number of bins to use to cut numeric data.
        The default is min(1001, max(101, sqrt(num of obs))). For non-XDF data
        sources, as (num of obs) is unknown in advance, it is wisest to specify
        this argument directly. If set to 0, unit binning will be used instead of
        cutting.
    :param max_unordered_levels: The maximum number of levels allowed for an
        unordered factor predictor for multiclass (>2) classification.
    :param remove_missings: Bool value. If True, rows with missing values
        are removed and will not be included in the output data.
    :param compute_obs_node_id: Bool value or None. If True, the tree node
        IDs for all the observations are computed and returned. If None, the IDs
        are computed for data.frame with less than 1000 observations and are
        returned as the where component in the fitted rxDTree object.
    :param use_sparse_cube: Bool value. If True, sparse cube is used.
    :param find_splits_in_parallel: Bool value. If True, optimal splits for
        each node are determined using parallelization methods; this will typically
        speed up computation as the number of nodes on the same level is increased.
    :param prune_cp: Optional complexity parameter for pruning. If prune_cp >
        0, prune.rxDTree is called on the completed tree with the specified
        prune_cp and the pruned tree is returned. This contrasts with the cp
        parameter that determines which splits are considered in growing the tree.
        The option prune_cp="auto" causes rxDTree to call the function
        rxDTreeBestCp on the completed tree, then use its return value as the cp
        value for prune.rxDTree.
    :param row_selection: None. Not currently supported, reserved for future use.
    :param transforms: None. Not currently supported, reserved for future use.
    :param transform_objects: None. Not currently supported, reserved for future use.
    :param transform_function: Variable transformation function. The variables used
        in the transformation function must be specified in transform_variables if they
        are not variables used in the model.
    :param transform_variables: List of strings of input data set variables needed
        for the transformation function.
    :param transform_packages: None. Not currently supported, reserved for future use.
    :param transform_environment: None. Not currently supported, reserved for future use.
    :param blocks_per_read: number of blocks to read for each chunk of data read from
        the data source.
    :param report_progress: Integer value with options:
        0: No progress is reported.
        1: The number of processed rows is printed and updated.
        2: Rows processed and timings are reported.
        3: Rows processed and all timings are reported.
    :param verbose: Integer value. If 0, no additional output is printed. If 1,
        additional summary information is printed.
    :param compute_context: A RxComputeContext object for prediction.
    :param kwargs: Additional parameters

    :return: A RxDTreeResults object of dtree model.

    .. rxkeywords:: learner, tree

    .. rxseealso::

        :py:func:`rx_predict <revoscalepy.rx_predict>`,
        :py:func:`rx_predict_rx_dtree <revoscalepy.rx_predict_rx_dtree>`.

    .. rxexample::
        import os
        from revoscalepy import rx_dtree, rx_import, RxOptions, RxXdfData
        sample_data_path = RxOptions.get_option("sampleDataDir")
        ds = RxXdfData(os.path.join(sample_data_path, "kyphosis.xdf"))
        kyphosis = rx_import(input_data = ds)

        # classification
        formula = "Kyphosis ~ Number + Start"
        method = "class"
        parms = {'prior': [0.8, 0.2], 'loss': [0, 2, 3, 0], 'split': "gini"}
        cost = [2,3]
        dtree = rx_dtree(formula, data = kyphosis, pweights = "Age", method = method, parms = parms, cost = cost, max_num_bins = 100)

        # regression
        formula = "Age ~ Number + Start"
        method = "anova"
        parms = {'prior': [0.8, 0.2], 'loss': [0, 2, 3, 0], 'split': "gini"}
        cost = [2,3]
        dtree = rx_dtree(formula, data = kyphosis, pweights = "Kyphosis", method = method, parms = parms, cost = cost, max_num_bins = 100)
    """
    telemetry_capture_call(compute_context)

    data = RxComputeContext.current().validate_in_data(data, "data")
    output_file = RxComputeContext.current().validate_out_data(output_file, "output_file")

    if transform_function is not None:
        transform_function = _resolve_transform_function(transform_function=transform_function)

    py_args = locals()
    for key in kwargs:
        py_args[key] = kwargs[key]
    # For spark compute context, always turn off allow_disk_write for speeding up rx_dtree.
    # Not affect rx_dforest and rx_btrees, since compute_oob_error will still trigger disk write.
    py_args['allow_disk_write'] = False
    py_args.pop("kwargs", None)
    py_function_name = "rx_dtree"
    py_module_name = __name__

    if (compute_context is None):
        compute_context = RxComputeContext.current()

    # todo need to refactor the spark workflow with Keith
    if isinstance(compute_context, RxSpark):
        return compute_context._execute(py_module_name, py_function_name, py_args)

    function_name = "DTree"
    call_parameters = {
                        'RemoveMissings': False
                      }
    results = {}

    # TODO Handle Transforms
    # TODO R-code inline for things that need done


    # manage expression-like input arguments
    # rowSelection <- if(missing(subset)) rxSubstitute(rowSelection) else rxSubstitute(subset)
    # rowSelection < - rxSubstitute(rowSelection)
    n_tree = 1 if "n_tree" not in kwargs else kwargs['n_tree']
    if row_selection is not None and isinstance(data, str) and isinstance(output_file, str) and data == output_file:
        raise ValueError("If rowSelection is performed, then the output file cannot be the same as the input file")

    nr = -1

    if isinstance(data, DataFrame):
        nr = len(data.index)
    else:
        try:
            nr = data.number_rows
        except AttributeError:
            print("Number of observations not available for this data source. 'numObs' set to 1e6.\n")
            print("For this data source, explicitly set 'max_num_bins' and 'min_split', otherwise innacurate models can result.")
            print("Please see rx_dtree 'max_num_bins' and 'min_split' for details.")
            nr = 1e+06


    if 'control' not in kwargs or kwargs['control'] is None:
        control = RxDTreeControl(nr,
                                 min_split,
                                 min_bucket,
                                 max_depth,
                                 cp,
                                 max_compete,
                                 max_surrogate,
                                 use_surrogate,
                                 surrogate_style,
                                 x_val)
        call_parameters['control'] = control.serialize_bxl()
    else:
        call_parameters['control'] = kwargs['control'].serialize_bxl()

    call_parameters['Formula'] = _rx_handle_formula(RxFunction.assign_or_throw(formula, False, str), pweights, fweights)
    call_parameters['OutColName'] = RxFunction.assign_or_throw(output_column_name, False, str)
    call_parameters['maxUnorderedLevels'] = RxFunction.assign_or_throw(max_unordered_levels, False, int)
    call_parameters['RemoveMissings'] = RxFunction.assign_or_throw(remove_missings, False, bool)
    call_parameters['useSparseCube'] = RxFunction.assign_or_throw(use_sparse_cube, False, bool)
    call_parameters['findSplitsInParallel'] = RxFunction.assign_or_throw(find_splits_in_parallel, False, bool)
    call_parameters['method'] = RxFunction.assign_or_throw(method, True, str)
    call_parameters['blocksPerRead'] = RxFunction.assign_or_throw(blocks_per_read, False, int)
    call_parameters['ReportProgress'] = RxFunction.assign_or_throw(report_progress, False, int)
    call_parameters['transforms'] = RxFunction.assign_or_throw(transforms, not_implemented=True)
    call_parameters['transformObjects'] = RxFunction.assign_or_throw(transform_objects, True)
    call_parameters['transformFunc'] = RxFunction.assign_or_throw(transform_function, True)
    transform_vars = RxFunction.assign_or_throw(transform_variables, True)
    call_parameters['transformVars'] = transform_vars
    call_parameters['transformEnvir'] = RxFunction.assign_or_throw(transform_environment, True)
    call_parameters['transformPackages'] = RxFunction.assign_or_throw(transform_packages, True)
    call_parameters['nTree'] = n_tree
    call_parameters['xdfCompressionLevel'] = RxFunction.assign_or_throw(xdf_compression_level, False, int)
    call_parameters['mTry'] = 0 if 'm_try' not in kwargs else kwargs['m_try']
    call_parameters['maxNumBins'] = min(1001, max(101, int(ceil(sqrt(nr))))) if max_num_bins is None else max_num_bins
    call_parameters['computeCpTable'] = True if 'compute_cp_table' not in kwargs else kwargs['compute_cp_table']
    call_parameters['replace'] = True if 'replace' not in kwargs else kwargs['replace']
    call_parameters['importance'] = True if 'importance' not in kwargs else kwargs['importance']
    call_parameters['batchLength'] = -1
    # internal hidden argument
    # compute_oob_error < 0: never; 0: only once for the entire forest; >0: once for each tree.
    # set to -1 (which can save time)
    # when a user wants to do error estimation using cross-validation instead of oob prediction
    if "compute_oob_error" in kwargs:
        call_parameters['computeOobError'] = kwargs['compute_oob_error']
    else:
        call_parameters['computeOobError'] = 1
    # internal hidden argument
    # allow_disk_write set to TRUE to write sNodeIdVarName and sRowSelectionVarName to disk,
    # which can speed up building deep trees
    if "allow_disk_write" in kwargs:
        call_parameters['allowDiskWrite'] = kwargs['allow_disk_write']
    else:
        call_parameters['allowDiskWrite'] = True
    call_parameters['lossFunction'] = None if 'loss_function' not in kwargs else kwargs['loss_function']
    call_parameters['learningRate'] = 0.1 if 'learning_rate' not in kwargs else kwargs['learning_rate']
    call_parameters['quantileScale'] = 10
    call_parameters['strata'] = None if 'strata' not in kwargs else kwargs['strata']
    call_parameters['gatherValidObsOnce'] = True
    call_parameters['numThreads'] = RxOptions.get_option("NumCoresToUse")
    call_parameters['seed'] = RANDOM.randint(1, MAX_INT) if 'seed' not in kwargs else kwargs['seed']
    call_parameters['sampRate'] = 0.0 if 'sample_rate' not in kwargs else kwargs['sample_rate']
    call_parameters['parms'] = parms
    call_parameters['maxRowsInMemory'] = -1
    call_parameters['cost'] = cost
    call_parameters['computeObsNodeId'] = True
    call_parameters['ReportTiming'] = 1
    call_parameters['extraVarsToWrite'] = extra_vars_to_write
    call_parameters['WriteModelVars'] = write_model_vars
    call_parameters['overwrite'] = overwrite
    call_parameters['model_type'] = 'rxDTree' if 'model_type' not in kwargs else kwargs['model_type']
    if output_file is not None:
        output_file = rx_get_out_data_source(output_file)
        call_parameters['OutDataSource'] = RxFunction.resolve_datasource(output_file)

    if isinstance(data, DataFrame):
        call_parameters['..rxDataFrame'] = resolve_dataframe(data)

        faddcp_param = OrderedDict()
        faddcp_param['bComputeLowHigh'] = True
        call_parameters['FaddCP'] = faddcp_param

        call_parameters['readVars'] = data.columns.tolist()
    else:
        call_parameters['DataSource'] = RxFunction.resolve_datasource(data)
        call_parameters['readVars'] = _rx_construct_read_vars(_rx_get_formula_components(formula), pweights, fweights, transform_vars)

    #
    # capture sample data if its a remote compute context call
    #
    #print("Using compute context {0}".format(type(compute_context).__name__))

    if isinstance(compute_context, RxRemoteComputeContext):
        job = compute_context.create_job()
        _rx_create_sample_data_file(
            input_data=data,
            sample_data_file=job.sample_data_file
            )
        if isinstance(data, RxSqlServerData):
            job.set_input_data_query(data.input_data_query)
            job.set_use_sql_satellite(data.use_fast_read)
            data.set_in_sql_server(True)

    #
    # make the C++ call
    #
    if isinstance(compute_context, RxRemoteComputeContext):
        def process_results(ret):
            ret = ret if len(ret) == 0 else ret[0]

            return RxDForestResults.create_results(formula, data, ret, call_parameters)

        return rx_exec(function=rx_native_call, args={'functionname': function_name, 'params': call_parameters},
                       compute_context=compute_context, callback=process_results)
    else:
        results = rx_native_call(functionname=function_name, params=call_parameters)

    return RxDForestResults.create_results(formula, data, results, call_parameters)
