#
# Copyright (c) Microsoft Corporation.  All rights reserved.
#
from typing import Callable, Union, List

from .. RxSerializable import _resolve_transform_function
from .. computecontext.RxComputeContext import RxComputeContext
from .. RxFunction import validate_argument

def _get_local_state(local_state_arg):

    _localstate = {}
    _localstate.update(locals())
    _localstate.update(local_state_arg)
    del _localstate['local_state_arg']
    del _localstate['_localstate']

    return _localstate


def _get_num_arg_groups(args):
    list_lengths = []
    contains_list = False

    for value in args.values():
        if isinstance(value, list):
            length = len(value)
            if(length == 0):
                raise ValueError("args cannot contain empty lists as values")
            list_lengths.append(len(value))

    if(len(list_lengths) > 0):
        contains_list = True
        if len(set(list_lengths)) > 1:
            raise ValueError("args has mismatch in lengths of arguments!")

    if contains_list:
        num_arg_groups = list_lengths[0]
    else:
        num_arg_groups = 1 if args != {} else 0

    return num_arg_groups    


def _kwargs_to_list_of_dicts(inlinekwargs):

    num_arg_groups = _get_num_arg_groups(inlinekwargs)

    for key, value in inlinekwargs.items():
        if not isinstance(value, list):
            inlinekwargs[key] = [value for x in range(num_arg_groups)]

    # Prepared args will be a list of dict object containing parameters for each execution
    return [dict(zip(inlinekwargs,t)) for t in zip(*inlinekwargs.values())]

def _validate_and_prepare_args(function, args, times_to_run, inlinekwargs, task_chunk_size):
    '''
     The args variable can take different forms, so this is a (messy) function that accounts 
     for all of them. The function returns prepared_args and prepared_times. prepared_args
     will be a list of dict objects. Each dict object is the parameters for a function
     execution, and len(prepared_args) always equals prepared_times. Ensuring this equality 
     here makes the code much cleaner down the line.

     This function also checks to make sure parameters are coherent. Lists in the args dict must have the
     same length, and if times_to_run is specified it must be equal to that length.
    '''

    if task_chunk_size is not None and task_chunk_size <= 0:
        raise ValueError("Task chunk size must be greater than zero!")

    if len(inlinekwargs) > 0 and args is not None:
        raise ValueError("Cannot specify both args and kwargs parameter!")

    using_kwargs = args is None and inlinekwargs != {}
    if using_kwargs: args = _kwargs_to_list_of_dicts(inlinekwargs)

    validate_argument(function, 'function', Callable)
    validate_argument(args, 'args', (dict, list), default = {})

    if args is None: args = {}
    if isinstance(args, dict): args = [args]
    num_arg_groups = len(args)

    if num_arg_groups > 1 and times_to_run != -1 and times_to_run != num_arg_groups:
        raise ValueError("Length of args and times_to_run are not the same!")

    # Update unspecified times_to_run to equal the number of argument groups
    if times_to_run == -1:
        times_to_run = 1 if num_arg_groups == 0 else num_arg_groups

    # A single argument group or no arguments and specified times_to_run
    if times_to_run != -1 and len(args) == 1:
        args = [args[0] for i in range(times_to_run)]

    return times_to_run, args


#################################################################################
#
# executes user defined function locally or remotely
#
def rx_exec(function: Callable, args: Union[dict, List[dict]] = None, times_to_run: int = -1,
            task_chunk_size: int = None, compute_context=None, local_state: dict = {}, callback: Callable = None,
            continue_on_failure: bool = True, **kwargs):

    """
    .. rxtitle::

        Run a Function With Different Sets of Arguments

    .. rxdescription::

        Allows distributed execution of a function in parallel across nodes 
        (computers) or cores of a "compute context" such as a cluster.

    :param function: The function to be executed; the nodes or cores on which it 
        is run are determined by the currently-active compute context and by the 
        other arguments of rx_exec.
    :param args: Arguments passed to the function 'function' each time it is executed. 
        for multiple executions with different parameters, args should be a list where each element is a dict.
        each dict element contains the set of named parameters to pass the function on each execution.
    :param times_to_run: Specifies the number of times 'function' should be executed. in
        case of different parameters for each execution, the length of args must equal 
        times_to_run. in case where length of args equals 1 or args is not specified, 
        function will be called times_to_run number of times with the same arguments.
    :param task_chunk_size: Specifies the number of tasks to be executed per compute element. by submitting tasks
        in chunks, you can avoid some of the overhead of starting new python processes over and over. for example,
        if you are running thousands of identical simulations on a cluster, it makes sense to specify the
        task_chunk_size so that each worker can do its allotment of tasks in a single python process.
    :param compute_context: A RxComputeContext object.
    :param local_state: Dictionary with variables to be passed to a RxInSqlServer compute context
        not supported in RxSpark compute context.
    :param callback: An optional function to be called with the results from each execution
        of 'function'
    :param continue_on_failure: None or logical value. If True, the default, then if an
        individual instance of a job fails due to a hardware or network failure, an 
        attempt will be made to rerun that job. (Python errors, however, will cause 
        immediate failure as usual.) Furthermore, should a process instance of a job 
        fail due to a user code failure, the rest of the processes will continue, and 
        the failed process will produce a warning when the output is collected. 
        Additionally, the position in the returned list where the failure occured will 
        contain the error as opposed to a result.
    :param **kwargs: Contains named arguments to be passed to function. Alternative to using args param.
        For a function foo that takes two arguments named 'x' and 'y', kwargs can be used like:
        rx_exec(function=foo, x=5, y=4) (for a single execution with x=5, y=4)
        rx_exec(function=foo, x=[1,2,3,4,5], y=6) (for multiple executions with different values of x and y=6)
        rx_exec(function=foo, x=[1,2,3], y=[4,5,6]) (for multiple executions with different values of x and y)
        Note: if you want to pass in a list as an argument, you must wrap it in another list. For a function foobar
        that takes an argument named 'the_list', kwargs can be used like:
        rx_exec(function=foobar, the_list=[ [0,1,2,3] ]) (single execution with the_list = [0,1,2,3])
        rx_exec(function=foobar, the_list=[ [0,1,2],[3,4,5],[6,7,8] ]) (multiple executions)


    :return: If local compute context or a waiting compute context is active, a list containing the return
        value of each execution of 'function' with specified parameters. If a non-waiting compute context is active, a jobInfo
        object, or a list of a jobInfo objects for multiple executions.

    .. rxkeywords:: parallel execution, multiple executions, remote execution, sql

    .. rxseealso::

        :py:func:`RxComputeContext <revoscalepy.RxComputeContext>`,
        :py:func:`RxLocalSeq <revoscalepy.RxLocalSeq>`,
        :py:func:`RxLocalParallel <revoscalepy.RxLocalParallel>`,
        :py:func:`RxInSqlServer <revoscalepy.RxInSqlServer>`,
        :py:func:`rx_get_compute_context <revoscalepy.rx_get_compute_context>`,
        :py:func:`rx_set_compute_context <revoscalepy.rx_set_compute_context>`.

    .. rxexample:: 

        ###
        # Local parallel rx_exec with different parameters for each execution
        ###
        import os
        from revoscalepy import RxLocalParallel, rx_set_compute_context, rx_exec, rx_btrees, RxOptions, RxXdfData

        sample_data_path = RxOptions.get_option("sampleDataDir")
        in_mort_ds = RxXdfData(os.path.join(sample_data_path, "mortDefaultSmall.xdf"))
        rx_set_compute_context(RxLocalParallel())
        formula = "creditScore ~ yearsEmploy"
        args = [
            {'data' : in_mort_ds, 'formula' : formula, 'n_tree': 3},
            {'data' : in_mort_ds, 'formula' : formula, 'n_tree': 5},
            {'data' : in_mort_ds, 'formula' : formula, 'n_tree': 7}
        ]

            models = rx_exec(rx_btrees, args = args)
            # Alternatively
            models = rx_exec(rx_btrees, data=in_mort_ds, formula=formula, n_tree=[3,5,7])

        ###
        ## SQL rx_exec
        ###

        from revoscalepy import RxSqlServerData, RxInSqlServer, rx_exec
        formula = "ArrDelay ~ CRSDepTime + DayOfWeek"
        connection_string="Driver=SQL Server;Server=.;Database=RevoTestDB;Trusted_Connection=TRUE"
        query="select top 100 [ArrDelay],[CRSDepTime],[DayOfWeek] FROM airlinedemosmall"

        ds = RxSqlServerData(sql_query = query, connection_string = connection_string)
        cc = RxInSqlServer(
            connection_string = connection_string,
            num_tasks = 1,
            auto_cleanup = False,
            console_output = True,
            execution_timeout_seconds = 0,
            wait = True
            )

        def remote_call(dataset):
            from revoscalepy import rx_data_step
            df = rx_data_step(dataset)
            return len(df)

        results = rx_exec(function = remote_call, args={'dataset': ds}, compute_context = cc)
        print(results)

        ###
        ## Run rx_exec in RxSpark compute context
        ###
        from revoscalepy import *

        # start Spark app
        spark_cc = rx_spark_connect()

        # define function to conditional check
        def my_check_fun(id_param):
            if id_param == 13:
                raise Exception("Ensure to fail")
            return id_param

        # prepare args for each run with total 20 runs
        elem_arg = []
        for i in range(0,20):
            elem_arg.append({"id_param": i})

        # get result and print
        results = rx_exec(my_check_fun, elem_arg, continue_on_failure=True)
        print(results)

        # stop Spark app
        rx_spark_disconnect(spark_cc)

        ## End(Not run)
    """

    _localstate = _get_local_state(local_state)

    prepared_times, prepared_args = _validate_and_prepare_args(function, args, times_to_run, kwargs, task_chunk_size)

    if compute_context is None:
        compute_context = RxComputeContext.current()

    # Resolve user given function to __main__ module if needed for remote
    prepared_function = _resolve_transform_function(function)

    return compute_context.execute_tasks(prepared_function, prepared_args, prepared_times, task_chunk_size, callback,
                                         local_state=_localstate, continue_on_failure=continue_on_failure)
