import shutil
import tempfile
import warnings

from pandas import DataFrame
from typing import Union, Callable, List
from os import path

from revoscalepy import RxComputeContext, RxRemoteComputeContext, RxLocalSeq, RxLocalParallel, RxInSqlServer
from revoscalepy import RxDataSource, RxTextData, RxXdfData, RxOdbcData, RxSqlServerData
from revoscalepy import rx_set_compute_context, rx_exec, rx_import, rx_data_step
from revoscalepy.RxFunction import RxFunction, validate_argument
from revoscalepy.RxSerializable import rx_native_call, _resolve_transform_function
from revoscalepy.computecontext.RxSpark import RxSpark
from revoscalepy.datasource.RxFileData import RxFileData
from revoscalepy.datasource.RxSparkData import RxSparkData
from revoscalepy.functions.RxPartition import rx_get_partitions, rx_partition, _rx_partition_odbc_data_source
from revoscalepy.utils.RxOptions import RxOptions
from revoscalepy.utils.RxUtils import rx_get_in_data_source, _rx_create_temp_table_hive_data_with_guid_name, _rx_is_master


class RxExecByResults:
    def __init__(self, dataframe: DataFrame):
        column_names = dataframe.columns.tolist()
        if len(column_names) < 3:
            raise ValueError("result should contain key columns, 'result' column and 'status' column")

        self._dataframe = dataframe

        # use set_index with inplace=True to avoid creating a new object
        self._dataframe.set_index(column_names[:-2], inplace=True)

    def __repr__(self):
        return self._dataframe.__repr__()

    def __str__(self):
        return self._dataframe.__str__()

    def to_dataframe(self) -> DataFrame:
        return self._dataframe

    def _to_list(self) -> List:
        result = []
        key_names = list(self._dataframe.index.names)

        for index, row in self._dataframe.iterrows():
            data = {'keys': dict(zip(key_names, index))}
            data.update(row.to_dict())
            result.append(data)

        return result


def rx_exec_by(input_data: Union[RxDataSource, DataFrame, str],
               keys: List[str] = None,
               function: Callable = None,
               function_parameters: dict = None,
               filter_function: Callable = None,
               compute_context: RxComputeContext = None,
               **kwargs) -> RxExecByResults:
    """
    .. rxtitle::

        Partition Data by Key Values and Execute User Function on Each Partition

    .. rxdescription::

        Partition input data source by keys and apply user defined function on 
        individual partitions. If input data source is already partitioned, apply 
        user defined function on partitions directly. Currently supported in local,
        localpar, RxInSqlServer and RxSpark compute contexts.

    :param input_data: A data source object supported in currently active compute context,
        e.g. 'RxSqlServerData' for 'RxInSqlServer'. In 'RxLocalSeq'
        and 'RxLocalParallel', a character string specifying a '.xdf' file, or a data frame object
        can be also used.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param keys: List of strings of variable names to be used for partitioning 
        the input data set.
    :param function: The user function to be executed. The user function takes 'keys' 
        and 'data' as two required input arguments where 'keys' determines the partitioning 
        values and 'data' is a data source object of the corresponding partition. 
        'data' can be a RxXdfData object or a RxODBCData object, which can be transformed 
        to a pandas data frame by using rx_data_step. 'keys' is a dict where keys of the dict
        are variable names and values are variable values of the partition.
        The nodes or cores on which it is running are determined by the currently active compute context.
    :param function_parameters: A dict which defines a list of additional arguments for 
        the user function func.
    :param filter_function: An user function that takes a Panda data frame of keys/values as 
        an input argument, applies filter to the keys/values and returns a data frame 
        containing rows whose keys/values satisfy the filter conditions. The input data frame
        has similar format to the results returned by rx_partition which comprises of partitioning
        variables and an additional variable of partition data source. This filter_function 
        allows user to control what data partitions to be applied by the user function 'function'.
        'filter_function' currently is not supported in RxHadoopMR and RxSpark compute contexts.
    :param compute_context: A RxComputeContext object.
    :param kwargs: Additional arguments.

    :return: A RxExecByResults object inherited from dataframe with each row to be the result of each partition.
        The indexes of dataframe are keys, columns are 'result' and 'status'.
        'result': the object returned from the user function from each partition.
        'status': 'OK', if the user function on each partition runs success, otherwise, the exception object.

    .. rxkeywords:: execby, groupby 

    .. rxseealso::

        :py:func:`rx_partition <revoscalepy.rx_partition>`.
        :py:func:`RxXdfData <revoscalepy.RxXdfData>`.

    .. rxexample:: 
        ###
        # Run rx_exec_by in local compute context
        ###
        import os
        from revoscalepy import RxLocalParallel, rx_set_compute_context, rx_exec_by, RxOptions, RxXdfData
        data_path = RxOptions.get_option("sampleDataDir")

        input_file = os.path.join(data_path, "claims.xdf")
        input_ds = RxXdfData(input_file)

        # count number of rows
        def count(data, keys):
            from revoscalepy import rx_data_step
            df = rx_data_step(data)
            return len(df)

        rx_set_compute_context(RxLocalParallel())
        local_cc_results = rx_exec_by(input_data = input_ds, keys = ["car.age", "type"], function = count)
        print(local_cc_results)

        ###
        # Run rx_exec_by in SQL Server compute context with data table specified
        ###
        from revoscalepy import RxInSqlServer, rx_set_compute_context, rx_exec_by, RxOptions, RxSqlServerData
        connection_string = 'Driver=SQL Server;Server=.;Database=RevoTestDb;Trusted_Connection=True;'
        ds = RxSqlServerData(table = "AirlineDemoSmall", connection_string=connection_string)

        def count(keys, data):
            from revoscalepy import rx_import
            df = rx_import(data)
            return len(df)

        # filter function
        def filter_weekend(partitioned_df):
            return (partitioned_df.loc[(partitioned_df.DayOfWeek == "Saturday") | (partitioned_df.DayOfWeek == "Sunday")])

        sql_cc = RxInSqlServer(connection_string = connection_string, num_tasks = 4)
        rx_set_compute_context(sql_cc)
        sql_cc_results = rx_exec_by(ds, keys = ["DayOfWeek"], function = count, filter_function = filter_weekend)
        print(sql_cc_results)

        ###
        # Run rx_exec_by in RxSpark compute context
        ###
        from revoscalepy import *

        # start Spark app
        spark_cc = rx_spark_connect()

        # define function to compute average delay
        def average_delay(keys, data):
            df = rx_data_step(data)
            return df['ArrDelay'].mean(skipna=True)

        # define colInfo
        col_info = {
            'ArrDelay': {'type': "numeric"},
            'CRSDepTime': {'type': "numeric"},
            'DayOfWeek': {'type': "string"}
        }

        # create text data source with airline data
        text_data = RxTextData(
            file = "/share/sample_data/AirlineDemoSmall.csv",
            first_row_is_column_names = True,
            column_info = col_info,
            file_system = RxHdfsFileSystem())

        # group text_data by day of week and get average delay on each day
        result = rx_exec_by(text_data, keys = ["DayOfWeek"], function = average_delay)

        # get result in pandas dataframe format and sort multiindex
        result_dataframe = result.to_dataframe()
        result_dataframe.sort_index()
        print(result_dataframe)

        # stop Spark app
        rx_spark_disconnect(spark_cc)
    """

    # validate input arguments
    input_data = RxComputeContext.current().validate_in_data(input_data, "input_data") 
    input_data = rx_get_in_data_source(input_data)

    num_tasks = RxOptions.get_option('numTasks')

    is_partitioned_xdf = isinstance(input_data, RxXdfData) and input_data.create_partition_set
    keys = validate_argument(keys, 'keys', List, "" if is_partitioned_xdf else None)

    if function is None or not callable(function):
        raise ValueError("'function' must be a callable function.")

    function_parameters = validate_argument(function_parameters, 'function_parameters', dict, {})

    if filter_function is not None and not callable(filter_function):
        raise ValueError("'filter_function' must be a callable function.")

    compute_context = validate_argument(compute_context, 'compute_context', RxComputeContext, RxComputeContext.current())

    if isinstance(compute_context, RxRemoteComputeContext):
        # for remote compute context
        if isinstance(compute_context, RxSpark):
            if not (isinstance(input_data, RxFileData) or isinstance(input_data, RxSparkData)):
                raise ValueError("'input_data' must be RxXdfData, RxTextData, RxParquetData, RxOrcData, " +
                                 "RxHiveData or RxSparkDataFrame data source when running in RxSpark compute context.")

            if filter_function is not None:
                warnings.warn("'filter_function' currently is not supported in RxSpark compute context " +
                              "and is ignored for computation in RxSpark.")

            return _rx_exec_by_spark(
                input_data=input_data,
                keys=keys,
                function=function,
                function_parameters=function_parameters,
                compute_context=compute_context)
        elif isinstance(compute_context, RxInSqlServer):
            if isinstance(input_data, RxOdbcData) and input_data.table is not None:
                curr_num_tasks = compute_context.num_tasks
                compute_context.num_tasks = 1

                # call rx_exec to run in SQL Server side:
                args = {'input_data': input_data,
                        'keys': keys,
                        'function': _resolve_transform_function(function),
                        'function_parameters': function_parameters,
                        'filter_function': _resolve_transform_function(filter_function),
                        'num_tasks': curr_num_tasks}

                # use query rewrite approach
                results = rx_exec(function = _localpar_execution, args = args, compute_context = compute_context)
                results = results[0]
            else:
                # use general approach
                job = compute_context.create_job()
                rx_import(
                    input_data = input_data,
                    output_file = RxXdfData(file=job.sample_data_file),
                    number_rows = 100,
                    verbose = 0,
                    report_progress = 0
                    )

                if isinstance(input_data, RxSqlServerData):
                    job.set_input_data_query(input_data.input_data_query)
                    job.set_use_sql_satellite(input_data.use_fast_read)
                    input_data.set_in_sql_server(True)
                else:
                    raise ValueError("input_data must be an RxSqlServerData data source for this compute context.")

                call_parameters = {}
                call_parameters['DataSource'] = RxFunction.resolve_datasource(input_data)
                call_parameters['ReportProgress'] = False
                call_parameters['Print'] = False
                call_parameters['ReportTiming'] = 0
                call_parameters['blocksPerRead'] = -1

                def process_results(ret):
                    result_file = ret if len(ret) == 0 else ret[0]

                    # Convert the temporary compositie XDF file to a partitioned XDF file.
                    partitions_df = _partition_composite_xdf(keys, result_file, filter_function)
                    partitions_list = _prepare_partitions_list(partitions_df, filter_function)

                    # Setup and call rxExec with the user function to spawn a new sub cluster.
                    return _execute_user_function(partitions_list['keys'], partitions_list['data'], function, function_parameters, num_tasks)

                results = rx_exec(function=rx_native_call, 
                                  args={'functionname': "ExecBy", 'params': call_parameters},
                                  compute_context=compute_context, 
                                  callback=process_results)
        else:
            raise ValueError("The compute context {} is currently not supported.".format(compute_context.__str__()))
    else:
        # for local
        if isinstance(input_data, RxSqlServerData) and _rx_is_master():
            call_parameters = {}
            call_parameters['DataSource'] = RxFunction.resolve_datasource(input_data)
            call_parameters['ReportProgress'] = False
            call_parameters['Print'] = False
            call_parameters['ReportTiming'] = 0
            call_parameters['blocksPerRead'] = -1

            # Call into BXL server to receive the data.
            result_file = rx_native_call(functionname="ExecBy", params = call_parameters)
            if result_file is None:
                raise ValueError("Failed to read the input data source.")

            # Convert the temporary compositie XDF file to a partitioned XDF file.
            partitions_df = _partition_composite_xdf(keys, result_file, filter_function)
            partitions_list = _prepare_partitions_list(partitions_df, filter_function)

            # Setup and call rxExec with the user function to spawn a new sub cluster.
            results = _execute_user_function(partitions_list['keys'], partitions_list['data'], function, function_parameters, num_tasks)
        else:
            results = _localpar_execution(input_data, keys, function, function_parameters, filter_function, num_tasks)

    return results


def _partition_composite_xdf(keys: list = None, input_xdf_file = None, filter_function:Callable = None):
    """
        Convert the input XDF file to a partitioned XDF file partitioning on keys.
    """

    original_context = rx_set_compute_context(RxLocalSeq())

    # TEMPORARY - This is inefficient and will be replaced by a parallel conversion.
    # Convert the composite XDF file to an partitioned composite XDF file.
    source_ds = RxXdfData(file = input_xdf_file['OutputXDFName'])
    dest_ds = RxXdfData(file = "{}/PartitionedXdf".format(input_xdf_file['OutputXDFDir']), create_partition_set = True)
    partitions_df = rx_partition(input_data = source_ds, output_data = dest_ds, vars_to_partition = keys, append = "none", overwrite = True)

    rx_set_compute_context(original_context)

    return partitions_df


def _execute_user_function(keys: list = None, partitions_list: list = None, 
                           function: Callable = None,
                           function_parameters: dict = None,
                           num_tasks: int = None):
    """
        Call the user defined function on all of the partitions.
    """

    # @TODO: set numCoresToUse and numTasks

    # construct args to pass to rx_exec
    args = list()
    for k, d in zip(keys, partitions_list):
        function_parameters['keys'] = k
        function_parameters['data'] = d
        args.append(function_parameters.copy())

    #print("call rx_exec with args = {}".format(args))

    # call rx_exec
    results = rx_exec(function = function, 
                      args = args,
                      compute_context = RxLocalParallel())

    # format the results as a dataframe
    results_list = []
    status_list = []
    for res in results:
        results_list.append(res if not isinstance(res, Exception) else None)
        status_list.append(res if isinstance(res, Exception) else "OK")

    final_results_df = DataFrame(keys)
    final_results_df['result'] = results_list
    final_results_df['status'] = status_list

    return RxExecByResults(final_results_df)


def _localpar_execution(input_data: Union[RxDataSource, DataFrame, str] = None, 
                        keys: Union[list, str] = None, 
                        function: Callable = None,
                        function_parameters: dict = None,
                        filter_function: Callable = None,
                        num_tasks: int = None):
    num_cores_in_options = RxOptions.get_option('numCoresToUse')

    temp_partitioned_xdf_file = None
    if isinstance(input_data, RxOdbcData) and input_data.table is not None:
        connection_string = RxOdbcData.fix_connection_string(input_data.connection_string)
        if connection_string is None:
            raise ValueError("Connection string is missing. Please specify it either in SQL Server Data Source object or in SQL Server Compute Context.")

        partitions_df = _rx_partition_odbc_data_source(input_data, vars_to_partition = keys, connection_string = connection_string)
    else:
        if isinstance(input_data, RxXdfData) and input_data.create_partition_set is True:
            # if input_data is a partitioned Xdf, don't need to repartition, just call rx_get_partitions.
            partitions_df = rx_get_partitions(input_data)

            if keys is not "":
                # keys are specified, verify that the specified keys match with the keys in the existing partitioned Xdf
                column_names = partitions_df.columns.tolist()
                del column_names[-1]
                if keys != column_names:
                    raise ValueError("The specified 'keys' doesn't match with those in the input_data. Set 'keys = None' or use matching keys.")
        else:
            # create temporary partitioned xdf
            temp_partitioned_xdf_file = path.join(tempfile._get_default_tempdir(), "tmpPartitions")
            temp_partitioned_xdf = RxXdfData(temp_partitioned_xdf_file, create_partition_set = True)
            partitions_df = rx_partition(input_data = input_data, output_data = temp_partitioned_xdf, vars_to_partition = keys, append = "none", overwrite = True)

    partitions_list = _prepare_partitions_list(partitions_df, filter_function)
    
    results = _execute_user_function(partitions_list['keys'], partitions_list['data'], function, function_parameters, num_tasks)

    # clean up
    if temp_partitioned_xdf_file is not None:
        shutil.rmtree(temp_partitioned_xdf_file)

    return results


def _prepare_partitions_list(partitions_df: DataFrame = None, filter_function: Callable = None):
    """
        form the list of arguments to feed
    """
    partitions_df = filter_function(partitions_df) if filter_function is not None else partitions_df

    number_partitions = len(partitions_df)
    data_column_index = len(partitions_df.columns) - 1
    number_part_columns = data_column_index
    if number_partitions == 0:
        raise ValueError("There is no partitions returned. Please check the input arguments 'input_data', 'keys' and 'filter_function'.")

    keys_list = list()
    data_list = list()
    for i in range(number_partitions):
        # each keys to be passed to the user function is a dict
        keys_i_dict = {partitions_df.columns[j]: partitions_df.iloc[i][j] for j in range(number_part_columns)}
        keys_list.append(keys_i_dict)

        data_source = partitions_df.iloc[i][data_column_index]
        data_source = RxXdfData(data_source) if isinstance(data_source, str) else data_source
        data_list.append(data_source)

    results_list = {'keys': keys_list, 'data': data_list}

    return results_list


def _rx_exec_by_spark(input_data: RxDataSource,
                      keys: List[str] = None,
                      function: Callable = None,
                      function_parameters: dict = None,
                      sample_replace=None,
                      sample_fraction=None,
                      sample_seed=None,
                      sample_partitions=None,
                      compute_context: RxComputeContext = None) -> RxExecByResults:
    if not isinstance(compute_context, RxSpark):
        raise ValueError("'compute_context' must be RxSpark.")

    if isinstance(input_data, RxTextData):
        tmp_data = _rx_create_temp_table_hive_data_with_guid_name()
        rx_data_step(input_data, tmp_data)
        input_data = tmp_data
        input_data.cache = False
    
    py_function_name = "execby_v1"
    py_args = {
        'inData': input_data.extract_info(),
        'keys': keys,
        'sampleEnable': keys is None,
        'sampleReplace': sample_replace,
        'sampleFraction': sample_fraction,
        'sampleSeed': sample_seed,
        'samplePartitions': sample_partitions
    }
    py_data = {
        'func': _resolve_transform_function(function),
        'funcParams': function_parameters
    }

    return RxExecByResults(compute_context._execute_spark_hpc(py_function_name, py_args, py_data))
