# -*- coding: utf-8 -*-
import datetime
import re
import warnings
import os
import numpy as np
from collections import UserDict

from pandas import DataFrame
from collections import OrderedDict
from revoscalepy import rx_data_step
from revoscalepy.functions.RxReadXdf import rx_read_xdf
from revoscalepy.utils.RxFileUtils import file_exists
from revoscalepy.RxFunction import RxFunction
from revoscalepy.utils.RxUtils import _rx_set_verbose_flags
from revoscalepy.datasource import RxNativeFileSystem
from revoscalepy.datasource import RxDataSource
from revoscalepy.datasource import RxFileData
from revoscalepy.datasource import RxOdbcData
from revoscalepy.datasource import RxXdfData
from revoscalepy.datasource import RxSparkData
from revoscalepy.RxSerializable import rx_native_call
from revoscalepy.utils.RxUtils import _rx_resolve_var_list
from revoscalepy import RxComputeContext
from revoscalepy import RxLocalSeq
from revoscalepy import RxRemoteComputeContext
from .. computecontext.RxSpark import RxSpark
from revoscalepy.utils import on_exit_handler
from revoscalepy.utils import telemetry_capture_call

# Converts python types to bxl storage types
def _get_storage_type(dtype):
    storage_map = {
        'float64':'float64',
        'str':'string',
        'category':'uint32',
        'int32':'int32'
    }

    if dtype.name in storage_map:
        return storage_map[dtype.name]
    return dtype.name

# Converts python types to bxl var types
def _get_var_type(dtype):
    var_map = {
        'float64':'Float',
        'str':'String',
        'category':'UInt',
        'int32':'Int',
        'bool':'Logical'
    }

    if dtype.name in var_map:
        return var_map[dtype.name]
    return dtype.name

# Converts the bxl var type to the var info type (for printing and for using as the column_info)
# This mapped var type is used later by BxlServer for reading
def _get_var_type_from_attribute(type):
    var_type_map = {
        "Int":"integer",
        "UInt":"factor",
        "UInt32":"factor",
        "factor":"factor",
        "Float":"numeric",
        "Double":"numeric",
        "String":"character",
        "Logical":"logical",
        "Short":"integer",
        "UShort":"integer",
        "LongLong":"integer",
        "ULongLong":"integer",
        "UChar":"raw",
        "Date":"Date",
        "POSIXct": "POSIXct"
    }

    if type in var_type_map:
        return var_type_map[type]
    return type

class GetVarInfoResults(OrderedDict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __output(self):
        idx = 1
        output_str = str("")
        for key in self.keys():
            column = self[key]
            output_str += "Var {0}: {1}".format(idx, key)
            if 'description' in column:
                output_str += ", {0}".format(column['description'])

            var_type = ""
            if 'varType' in column:
               var_type = _get_var_type_from_attribute(column['varType'])

            if len(var_type) > 0:
                if var_type != 'factor':
                    output_str += ", Type: {0}".format(var_type)
                    if 'storage' in column:
                        output_str += ", Storage: {0}".format(column['storage'])
                    if ('low' in column and 'high' in column):
                        if var_type != 'Date' and not np.isnan(column['low']) and not np.isnan(column['high']) :
                            output_str += ", Low/High: ({0:.4f}, {1:.4f})".format(column['low'], column['high'])
                        elif var_type == 'Date' and isinstance(column['low'], datetime.date) and isinstance(column['high'], datetime.date):
                            output_str += ", Low/High: ({0}, {1})".format(column['low'], column['high'])
                else:
                    if 'levels' in column:
                        levels = column['levels']
                        output_str += "\n\t{0} factor levels: {1}".format(len(levels), str(levels))
                    else:
                        output_str += ", Type: {0}, no factor levels available".format(var_type)

            output_str += "\n"
            idx+=1
        return output_str

    def __str__(self):
        return self.__output()

    def __repr__(self):
        return self.__output()


class GetVarNamesResults(list):
    def __output(self):
        output_str = "\n"
        output_str += " ".join(self)
        return output_str

    def __str__(self):
        return self.__output()

    def __repr__(self):
        return self.__output()


class GetInfoResults(OrderedDict):
    def __output(self):
        output_str = str("")

        if 'fileName' in self:
            output_str += "File name:{0}\n".format(self['fileName'])
        if 'connectionString' in self:
            output_str += "Connection string:{0}\n".format(self['connectionString'])
        if 'description' in self:
            output_str += "File description:{0}\n".format(self['description'])
        if 'dataSource' in self:
            output_str += "Data Source:{0}\n".format(self['dataSource'])

        if 'length' in self:
            output_str += "Length:{0}\n".format(self['length'])

        if 'numRows' in self:
            output_str += "Number of observations:{0}\n".format(self['numRows'])
        if 'numVars' in self:
            output_str += "Number of variables:{0}\n".format(self['numVars'])
        if 'numBlocks' in self:
            output_str += "Number of blocks:{0}\n".format(self['numBlocks'])

        # TODO: handle displaying the rows.

        if 'compressionType' in self:
            output_str += "Compression type:{0}\n".format(self['compressionType'])
        if 'varInfo' in self:
            output_str += "Variable information: \n"
            output_str += str(self['varInfo'])
        if 'data' in self:
            df = self['data']
            output_str += ('Data ({0} rows starting with row {1}):\n'.format(len(df), self['startRow']))
            output_str += str(df)
            output_str += "\n"

        return output_str


    def __str__(self):
        return self.__output()

    def __repr__(self):
        return self.__output()

def _handle_float__properties(info):
    # This ensures that these properties are always represented as a float
    if 'numVars' in info:
        info['numVars'] = float(info['numVars'])
    if 'numRows' in info:
        info['numRows'] = float(info['numRows'])

def _handle_date_var(info) :
    # Info is a dict of information about vars of Date Type
    info['VarType'][0] = 'Date'
    info['VarType'][1] = 'Date'
    if '.rxLowHigh' in info and not np.isnan(info[".rxLowHigh"][0]) and not np.isnan(info[".rxLowHigh"][1]):
        low = info['.rxLowHigh'][0]
        high = info['.rxLowHigh'][1]
        rx_low_high = []
        convert_to_date = lambda val : datetime.date.fromordinal(datetime.date(1970,1,1).toordinal() + int(val))
        rx_low_high.append(convert_to_date(low))
        rx_low_high.append(convert_to_date(high))
        info['.rxLowHigh'] = np.asarray(rx_low_high, dtype = datetime.date)

def _rx_get_info_distributed(data,
                             get_var_info,
                             get_block_sizes,
                             get_value_labels,
                             vars_to_keep,
                             vars_to_drop,
                             start_row,
                             num_rows,
                             compute_info,
                             all_nodes,
                             verbose):

    # Save the current compute context
    compute_context = RxComputeContext.current()

    try:
        if not all_nodes:
 
            py_args = locals()
            py_function_name = "rx_get_info"
            py_module_name = __name__ 

            if isinstance(compute_context, RxSpark):
                if not (isinstance(data, RxFileData) or isinstance(data, RxSparkData)):
                    raise ValueError("Parameter 'data' for rx_get_info must be " +
                                     "RxXdfData, RxTextData, RxParquetData, " +
                                     "RxOrcData, RxHiveData or RxSparkDataFrame data source")
                return compute_context._execute(py_module_name, py_function_name, py_args)
             
            # Change the compute context to local
            RxComputeContext.set_current(RxLocalSeq())
            info = rx_get_info(data, get_var_info, get_block_sizes, get_value_labels, vars_to_keep, vars_to_drop,
                               start_row, num_rows, compute_info=False, verbose=verbose)

            var_info = None
            if 'varInfo' in info:
                var_info = info['varInfo']
            if compute_info == True:
                var_info = _rx_get_var_info_compute_in_database(var_info, data, get_value_labels, compute_context)

            if var_info is not None:
                info['numVars'] = len(var_info.keys())
                if hasattr(var_info,'numRows'):
                    info['numRows'] = getattr(var_info,'numRows')

            return info
        else:
            raise ValueError("Parameters 'all_nodes' is currently not supported.")

    finally:
        RxComputeContext.set_current(compute_context)


def _rx_get_info_local(
        info_data,
        get_var_info: bool,
        get_block_sizes: bool,
        get_value_labels: bool,
        vars_to_keep: list,
        vars_to_drop: list,
        start_row: int,
        num_rows: int,
        compute_info: bool,
        all_nodes:bool,
        verbose: int):

    file_system = RxNativeFileSystem()
    if isinstance(info_data, RxFileData):
        file_system = info_data.file_system

    # Check if this is a file path
    if isinstance(info_data, (str, RxXdfData)):
        call_parameters = dict()
        if isinstance(info_data, str):
            call_parameters['File'] = info_data
        else:
            call_parameters['File'] = info_data.file
        call_parameters['DataSource'] = None
        call_parameters['ColNames'] = None
        call_parameters['ShowValueLabels'] = False
        call_parameters['ShowBlockSizes'] = get_block_sizes
        call_parameters['ShowDetailedBlockInfo'] = False
        call_parameters['ShowFirstNumberOfRows'] = 0
        call_parameters['StartingFromRowNumber'] = 0
        _rx_set_verbose_flags(call_parameters, verbose = verbose, report_progress = 0)
        call_parameters['ReturnObject'] = True
        call_parameters['fileSystem'] = file_system.to_serialize
        results = rx_native_call(
            functionname='GetInfo', params=call_parameters)
        info_list = GetInfoResults(results)
    elif isinstance(info_data, DataFrame):
        info_list = GetInfoResults()
        info_list['class'] = "DataFrame"
        info_list['numRows'] = len(info_data)
        info_list['numVars'] = len(info_data.columns)
    else:
        # Handle data sources
        data_source_type = re.sub("Rx|Data", '', info_data.class_id)
        info_list = GetInfoResults()
        # If the data source contains a file
        if isinstance(info_data, RxFileData):
            info_list['fileName'] = info_data.file
        elif isinstance(info_data, RxOdbcData) and info_data.connection_string is not None:
            info_list['connectionString'] = info_data.connection_string
            info_list['dataSource'] = data_source_type
        elif isinstance(info_data, RxOdbcData) and info_data.dfname is not None:
            info_list['dfName'] = info_data.df_name

        info_list['dataSource'] = data_source_type

    if get_var_info is True:
        if get_value_labels is None:
            get_value_labels = True
        else:
            get_value_labels = False

        var_info = rx_get_var_info(info_data, get_value_labels=get_value_labels, vars_to_keep=vars_to_keep,
                                   vars_to_drop=vars_to_drop, compute_info=compute_info)
        info_list['varInfo'] = var_info

        # If the number of rows was calculated, it will come back as an attribute on var_info.
        # If present, use this value
        calc_num_rows = getattr(var_info, 'numRows', None)
        if calc_num_rows is not None:
            info_list['numRows'] = calc_num_rows

        if 'numVars' not in info_list and (vars_to_keep is None and vars_to_drop is None):
            info_list['numVars'] = len(var_info.keys())

    if num_rows > 0:
        info_df = None
 
        if isinstance(info_data, str):
            info_df = rx_data_step(info_data, number_rows_read=num_rows, start_row=start_row, vars_to_keep=vars_to_keep,
                                   vars_to_drop=vars_to_drop, report_progress=False)
        else:
            if start_row > 1:
                warnings.warn("'start_row' greater than 0 not supported for non-xdf file sources")
                start_row = 1
            info_df = rx_data_step(info_data, number_rows_read=num_rows, start_row=1, vars_to_keep=vars_to_keep,
                                   vars_to_drop=vars_to_drop, report_progress=0)

        info_list['data'] = info_df
        info_list['startRow'] = start_row

    return info_list

@on_exit_handler
def rx_get_info(
        data,
        get_var_info: bool=False,
        get_block_sizes: bool=False,
        get_value_labels: bool=None,
        vars_to_keep: list=None,
        vars_to_drop: list=None,
        start_row: int = 1,
        num_rows: int = 0,
        compute_info: bool=False,
        all_nodes:bool=False,
        verbose: int=0):
    """
    .. rxtitle::

        Get Data Source Information

    .. rxdescription::

        Get basic information about an revoscalepy data source or data frame

    :param data: A data frame, a character string specifying an ".xdf", or an
        RxDataSource object. If a local compute context is being used, this
        argument may also be a list of data sources, in which case the output will
        be returned in a named list. See the details section for more information.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param get_var_info: Bool value. If True, variable information is
        returned.
    :param get_block_sizes: Bool value. If True, block sizes are returned in
        the output for an ".xdf" file, and when printed the first 10 block sizes
        are shown.
    :param get_value_labels: Bool value. If True and get_var_info is True or
        None, factor value labels are included in the output.
    :param vars_to_keep: List of strings of variable names for which
        information is returned. If None or get_var_info is False, argument is
        ignored. Cannot be used with vars_to_drop.
    :param vars_to_drop: List of strings of variable names for which
        information is not returned. If None or get_var_info is False, argument is
        ignored. Cannot be used with vars_to_keep.
    :param start_row: Starting row for retrieval of data if a data frame or
        ".xdf" file.
    :param num_rows: Number of rows of data to retrieve if a data frame or
        ".xdf" file.
    :param compute_info: Bool value. If True, and get_var_info is True,
        variable information (e.g., high/low values) for non-xdf data sources will
        be computed by reading through the data set. If True, and get_var_info is
        False, the number of variables will be gotten from from non-xdf data
        sources (but not the number of rows).
    :param all_nodes: Bool value. Ignored if the active RxComputeContext
        compute context is local or RxForeachDoPar. Otherwise, if True, a list
        containing the information for the data set on each node in the active
        compute context will be returned. If False, only information on the data
        set on the master node will be returned. Note that the determination of the
        master node is not controlled by the end user.
    :param verbose: Integer value. If 0, no additional output is printed. If 1,
        additional summary information is printed for an ".xdf" file.

    :return: List containing the following possible elements:

        fileName: Character string containing the file name and path (if an
            ".xdf" file).
        objName: Character string containing object name (if not an ".xdf" file).
        class: Class of the object.
        length: Length of the object if it is not an ".xdf" file or data frame.
        numCompositeFiles: Number of composite data files(if a composite ".xdf"
            file).
        numRows: Number of rows in the data set.
        numVars: Number of variables in the data set.
        numBlocks: Number of blocks in the data set.
        varInfo: List of variable information where each element is a list
            describing a variable.
        rowsPerBlock: Integer vector containing number of rows in each block
            (if get_block_sizes is set to True). Set to None if data is a data frame.
        data: Data frame containing the data (if num_rows > 0)

    .. rxseealso::

        :py:func:`rx_data_step <revoscalepy.rx_data_step>`,
        :py:func:`rx_get_var_info <revoscalepy.rx_get_var_info>`.

    .. rxkeywords:: xdf

    .. rxexample::
        :execute:
        :print_output:

        import os
        from revoscalepy import RxOptions, RxXdfData, rx_get_info

        sample_data_path = RxOptions.get_option("sampleDataDir")
        claims_ds = RxXdfData(os.path.join(sample_data_path, "claims.xdf"))
        info = rx_get_info(claims_ds, get_var_info=True)
        print(info)
    """
    telemetry_capture_call()
    if isinstance(data, list) and len(data) > 0:
        warnings.warn("List of sources are not supported in python, the first item will be used")
        data = data[0]
 
    data = RxComputeContext.current().validate_in_data(data, "data") 
    
    if isinstance(data, str) and not data.endswith(".xdf") and not os.path.isdir(data):
        raise ValueError("Parameter 'data' must be a path to an xdf file, data frame or an RxDataSource")

    return_info = None
    if isinstance(RxComputeContext.current(), RxRemoteComputeContext):
        return_info = _rx_get_info_distributed(data, get_var_info, get_block_sizes, get_value_labels, vars_to_keep,
                                            vars_to_drop, start_row, num_rows, compute_info, all_nodes, verbose)
    else:
        return_info = _rx_get_info_local(data, get_var_info, get_block_sizes, get_value_labels, vars_to_keep, vars_to_drop,
                                    start_row, num_rows, compute_info, all_nodes, verbose)


    _handle_float__properties(return_info)
    return return_info


def _get_var_info_data_frame(data,
                             vars_to_keep:list=None,
                             vars_to_drop:list=None,
                             compute_info:bool=False):
    if not isinstance(data, DataFrame):
        raise ValueError("Parameter 'data' must be a dataframe")

    rxVarInfo = getattr(data, "..rxVarInfo", None)

    columns = data.columns.tolist()
    columns_to_keep = _rx_resolve_var_list(data, vars_to_keep, vars_to_drop)
    if columns_to_keep is not None and len(columns_to_keep) > 0:
        columns = columns_to_keep

    var_info = OrderedDict()
    for column in columns:
        var_info[column] = {}
        column_data = dict() if rxVarInfo is None or column not in rxVarInfo else rxVarInfo[column]

        # Assign var info to what is in column_data. This will then capture other tags like
        # value labels, description, etc -- things that are not naturally on the dataframe
        var_info[column] = column_data

        if 'VarType' not in column_data:
            storage_type = _get_storage_type(data[column].dtype)
            var_type = _get_var_type(data[column].dtype)
            var_info[column]['VarType'] = [storage_type, var_type]

        # even if compute_info is false, low, high keys present in var_info but as nans
        if compute_info:
            if '.rxLowHigh' not in column_data and \
               (var_type == 'Float' or var_type == 'Int'):
                   var_info[column]['.rxLowHigh'] = [data[column].min(), data[column].max()]
        else:
            var_info[column]['.rxLowHigh'] = [np.nan, np.nan]

        if str(data[column].dtype) == "category" and 'levels' not in column_data:
            # TO DO Remove the list cast and keep numpy.array once bug#83644 handling numpy array is fixed
            levels = list(data[column].cat.categories.values)
            var_info[column]['levels'] = levels
            var_info[column]['.rxLowHigh'] = [0,len(levels)-1]

    return var_info


def _rx_get_var_info_base(data,
                         vars_to_keep: list=None,
                         vars_to_drop: list=None,
                         compute_info: bool=False):

    call_parameters = dict()
    call_parameters['DataSource'] = RxFunction.resolve_datasource(data)
    call_parameters['VarsToKeep'] = vars_to_keep
    call_parameters['VarsToDrop'] = vars_to_drop
    call_parameters['computeInfo'] = compute_info

    results = rx_native_call("DataSourceGetMetaInfo", call_parameters)

    # DataSourceGetMetaInfo will add a numRows column -- move this to  and attribute on var_info
    num_rows = results.pop('numRows', None)
    if hasattr(results, 'var_info') and num_rows is not None:
        var_info = getattr(results, 'var_info')
        var_info = UserDict(var_info)
        setattr(var_info, 'numRows', num_rows)
        setattr(results, 'var_info', var_info)

    # Order matters and the var_info is not ordered, therefore, order the results
    # based on the columns in the dataframe before continuing
    info = getattr(results, "var_info", None)
    if info is None:
        return results

    results_info = OrderedDict()

    # results_info will only contain the columns that we are displaying
    # if there is additional data that came through with the var_info, then
    # that data will be added as an attribute to the results_info structure.
    for idx in results.keys():
        if idx in info:
            results_info[idx] = info[idx]
            # info from BxlServer for Date type comes as a numeric
            # representing number of days since January 1, 1970.
            if 'bIsDateVar' in info[idx] and info[idx]['bIsDateVar'] and 'VarType' in info[idx]:
                _handle_date_var(results_info[idx])

    extra_items = [x for x in info if x not in results_info]
    for idx in extra_items:
        setattr(results_info, str(idx), idx)

    # Copy the num_rows attribute
    if hasattr(info, 'numRows'):
        setattr(results_info, 'numRows', getattr(info, 'numRows'))

    setattr(results,'var_info', results_info)
    return results


# computes the var info within the database
def _rx_get_var_info_compute_in_database(var_info,
                                         data,
                                         get_value_labels,
                                         compute_context):
    from revoscalepy import rx_summary

    # Set the compute context - this should be InSqlServer
    if compute_context is not None:
        RxComputeContext.set_current(compute_context)

    var_names = var_info.keys()

    formula = "~" + "+".join(var_names)
    summary_stat = rx_summary(formula, data, summary_stats=["Min","Max"])

    results = summary_stat._results
    min_max_dataframe = summary_stat.summary_data_frame
    num_objects = 0
    if summary_stat.valid_observations is not None:
        num_objects = summary_stat.valid_observations
    if summary_stat.missing_observations is not None:
        num_objects = num_objects + summary_stat.missing_observations

    idx = 0
    for item in var_info.keys():
        # Copy the values of the min/max from the dataframe
        if min_max_dataframe is not None:
            min_max_row = min_max_dataframe.iloc[idx,]
            if not np.isnan(min_max_row['Min']) and \
               not np.isnan(min_max_row['Max']):
                var_info[item]['low'] = min_max_row['Min']
                var_info[item]['high'] = min_max_row['Max']

        # Copy the factor levels if they are present
        if (get_value_labels and 'categorical'  in results
                # Even when there are no categories, 'categorical' may be present
                # in results as an empty dictionary, hence check if it is an instance of list
                and isinstance(results['categorical'], list)
                and len(results['categorical']) > 0
                and len(results['categorical'][0]) > 0
                and item in results['categorical'][0][0]):
            result_dict = results['categorical'][0][0]
            result_var_info = getattr(result_dict, 'var_info', None)
            if result_var_info is not None and item in result_var_info and 'levels' in result_var_info[item]:
                # TO DO Remove the list cast and keep numpy.array once bug#83644 handling numpy array is fixed
                levels = list(result_var_info[item]['levels'])
                var_info[item]['levels'] = levels
                var_info[item]['low'] = 0
                var_info[item]['high'] = len(levels) - 1
        idx+=1

    setattr(var_info, 'numRows', num_objects)
    return var_info


def _rx_get_var_info_distributed(data,
                                 get_value_labels,
                                 vars_to_keep,
                                 vars_to_drop,
                                 compute_info,
                                 all_nodes):

    # Save the current compute context
    compute_context = RxComputeContext.current()

    try:
        if not all_nodes:
            # Change the compute context to local
            RxComputeContext.set_current(RxLocalSeq())
            var_info = rx_get_var_info(data, vars_to_keep = vars_to_keep, vars_to_drop = vars_to_drop, compute_info=False)

            if compute_info and var_info is not None:
                var_info = _rx_get_var_info_compute_in_database(var_info, data, get_value_labels, compute_context)

            return var_info
        else:
            raise ValueError("Parameters 'all_nodes' is currently not supported.")

    finally:
        RxComputeContext.set_current(compute_context)

@on_exit_handler
def rx_get_var_info(data,
                    get_value_labels: bool=True,
                    vars_to_keep: list=None,
                    vars_to_drop: list=None,
                    compute_info: bool=False,
                    all_nodes: bool=False):
    """
    .. rxtitle::

        Get Variable Information for a Data Source

    .. rxdescription::

        Get variable information for a revoscalepy data source or data frame,
        including variable names, descriptions, and value labels

    :param data: a data frame, a character string specifying an ".xdf", or an
        RxDataSource object. If a local compute context is being used, this
        argument may also be a list of data sources, in which case the output will
        be returned in a named list. See the details section for more information.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param get_value_labels: bool value. If True and get_var_info is True or
        None, factor value labels are included in the output.
    :param vars_to_keep: list of strings of variable names for which
        information is returned. If None or get_var_info is False, argument is
        ignored. Cannot be used with vars_to_drop.
    :param vars_to_drop: list of strings of variable names for which
        information is not returned. If None or get_var_info is False, argument is
        ignored. Cannot be used with vars_to_keep.
    :param compute_info: bool value. If True, variable information (e.g., high/low values) 
        for non-xdf data sources will be computed by reading through the data set.
    :param all_nodes: bool value. Ignored if the active RxComputeContext
        compute context is local or RxForeachDoPar. Otherwise, if True, a list
        containing the information for the data set on each node in the active
        compute context will be returned. If False, only information on the data
        set on the master node will be returned. Note that the determination of the
        master node is not controlled by the end user.

    :return: list with named elements corresponding to the variables in the data set.
    Each list element is also a list with with following possible elements:

        description: character string specifying the variable description
        varType: character string specifying the variable type
        storage: character string specifying the storage type
        low: numeric giving the low values, possibly generated through a
            temporary factor transformation F()
        high: numeric giving the high values, possibly generated through a
            temporary factor transformation F()
        levels: (factor only) a list of strings containing the factor levels
        valueInfoCodes: list of strings of value codes, for informational
            purposes only
        valueInfoLabels: list of strings of value labels that is the same
            length as valueInfoCodes, used for informational purposes only

    .. rxseealso::

        :py:func:`rx_data_step <revoscalepy.rx_data_step>`,
        :py:func:`rx_get_info <revoscalepy.rx_get_info>`.

    .. rxkeywords:: info

    .. rxexample::
        :execute:
        :print_output:

        import os
        from revoscalepy import RxOptions, RxXdfData, rx_get_var_info

        sample_data_path = RxOptions.get_option("sampleDataDir")
        claims_ds = RxXdfData(os.path.join(sample_data_path, "claims.xdf"))
        info = rx_get_var_info(claims_ds)
        print(info)
    """
    telemetry_capture_call()
    if isinstance(data, list) and len(data) > 0:
        warnings.warn("List of sources are not supported in python, the first item will be used")
        data = data[0]

    data = RxComputeContext.current().validate_in_data(data, "data")

    if isinstance(data, str) and not data.endswith(".xdf") and not os.path.isdir(data):
        raise ValueError("Parameter 'data' must be a path to an xdf file, data frame or an RxDataSource")
 
    py_args = locals()
    py_function_name = "rx_get_var_info"
    py_module_name = __name__

    compute_context = RxComputeContext.current()

    # todo need to refactor the spark workflow with Keith
    if isinstance(compute_context, RxSpark):
        return compute_context._execute(py_module_name, py_function_name, py_args)

    if isinstance(RxComputeContext.current(), RxRemoteComputeContext):
        return _rx_get_var_info_distributed(data, get_value_labels, vars_to_keep, vars_to_drop,compute_info, all_nodes)
    else:
        results_info = None
        if isinstance(data, DataFrame):
            results_info = _get_var_info_data_frame(data, vars_to_keep, vars_to_drop, compute_info)
        elif isinstance(data, RxDataSource) or isinstance(data,str):
            results = _rx_get_var_info_base(data, vars_to_keep, vars_to_drop, compute_info)
            results_info = getattr(results, 'var_info', None)
            if results_info is None:
                raise ValueError("The given parameters yield no var information")

        # Create a dictionary of values converting the data specified in results_info
        # to what will be returned in value_info
        var_info = GetVarInfoResults()
        for k in results_info.keys():
            info_dict = dict()
            info = results_info[k]
            # Process description
            if ".rxDescription" in info:
                info_dict["description"] = info[".rxDescription"]
            # Process variable type
            if "VarType" in info:
                info_dict["storage"] = info["VarType"][0]
                # For factors, BxlServer will return a UInt, therefore we convert this to a type factor
                # Need to get the mapped var_type from attribute because BxlServer accepts the mapped 
                # var_types when they are sent as col_info; this is consistent with what R does
                info_dict["varType"] = _get_var_type_from_attribute(info["VarType"][1])
            if ".rxLowHigh" in info:
                info_dict["low"] = info[".rxLowHigh"][0]
                info_dict["high"] = info[".rxLowHigh"][1]
            if "levels" in info and get_value_labels:
                info_dict["levels"] = info["levels"]
            if ".rxValueInfoCodes" in info and get_value_labels:
                info_dict["valueInfoCodes"] = info[".rxValueInfoCodes"]
            if ".rxValueInfoLabels" in info and get_value_labels:
                info_dict["valueInfoLabels"] = info[".rxValueInfoLabels"]

            var_info[k] = info_dict

        # Copy the num_rows attribute if present
        if hasattr(results_info, 'numRows'):
            setattr(var_info, 'numRows', getattr(results_info, 'numRows'))

    _handle_float__properties(var_info)
    return var_info

@on_exit_handler
def rx_get_var_names(data):
    """
    .. rxtitle::

        Variable names for a data source or data frame

    .. rxdescription::

        Read the variable names for data source or data frame

    :param data: an RxDataSource object, a character string specifying the ".xdf" file, or a data frame.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.

    :return: list of strings containing the names of the variables in the data source or data frame.

    .. rxseealso::

        :py:func:`rx_data_step <revoscalepy.rx_data_step>`,
        :py:func:`rx_get_info <revoscalepy.rx_get_info>`,
        :py:func:`rx_get_var_info <revoscalepy.rx_get_var_info>`.

    .. rxkeywords:: variables

    .. rxexample::
        :execute:
        :print_output:

        import os
        from revoscalepy import RxOptions, RxXdfData, rx_get_var_names

        sample_data_path = RxOptions.get_option("sampleDataDir")
        claims_ds = RxXdfData(os.path.join(sample_data_path, "claims.xdf"))
        info = rx_get_var_names(claims_ds)
        print(info)
    """
    telemetry_capture_call()
    var_info = rx_get_var_info(data)
    return GetVarNamesResults(var_info.keys())

@on_exit_handler
def rx_set_var_info(var_info, data):
    """
    .. rxtitle::

        Set Variable Information for .xdf File or Data Frame

    .. rxdescription::

        Set the variable information for an ".xdf" file, including variable
        names, descriptions, and value labels, or set attributes for variables in
        a data frame

    :param var_info: list containing lists of variable information for variables
        in the XDF data source or data frame.
    :param data: an RxDataSource object, a character string specifying the ".xdf"
        file, or a data frame.

    :return: If the input data is a data frame, a data frame is returned
        containing variables with the new attributes. If the input data represents an
        ".xdf" file or composite file, an RxXdfData object representing the modified
        file is returned.

    .. rxseealso::

        :py:func:`rx_get_info <revoscalepy.functions.RxGetInfoXdf.rx_get_info>`,
        :py:func:`rx_get_var_info <revoscalepy.functions.RxGetInfoXdf.rx_get_var_info>`.

    .. rxkeywords:: variables

    .. rxexample::
        :execute:
        :print_output:

        import os
        import tempfile
        from shutil import copy2
        from revoscalepy import RxOptions, RxXdfData, rx_import, rx_get_var_info, rx_set_var_info

        sample_data_path = RxOptions.get_option("sampleDataDir")
        src_claims_file = os.path.join(sample_data_path, "claims.xdf")
        dest_claims_file = os.path.join(tempfile.gettempdir(), "claims_new.xdf")
        copy2(src_claims_file, dest_claims_file)
        var_info = rx_get_var_info(src_claims_file)
        print(var_info)

        claims_labels = ['Type1', 'Type2', 'Type3', 'Type4']
        var_info['type']['levels'] = claims_labels
        var_info['cost']['low'] = 10
        var_info['cost']['high'] = 1000
        rx_set_var_info(var_info, dest_claims_file)

        new_var_info = rx_get_var_info(dest_claims_file)
        print(new_var_info)
    """
    telemetry_capture_call()
    from revoscalepy.RxSerializable import resolve_dataframe

    data = RxFunction.assign_or_throw(data, False, (DataFrame, str, RxXdfData))
    var_info = RxFunction.assign_or_throw(var_info, False, dict)

    if isinstance(data, DataFrame):
        original_info = rx_get_var_info(data)
    elif isinstance(data, (str, RxXdfData)):
        results = _rx_get_var_info_base(data)
        original_info = getattr(results, "var_info", None)

    # Replace in the original info
    for k, v in var_info.items():
        if k not in original_info.keys():
            raise ValueError("Parameter 'var_info' contains a column that doesn't exist in data. Column {:s}".format(k))
        new_entry = v
        orig_entry = original_info[k]
        low_high_processed = False

        for new_key, new_value in new_entry.items():
            if new_key == "description":
                orig_entry[".rxDescription"] = new_value
            elif new_key == "low" or new_key == "high":
                if not low_high_processed and not isinstance(data, DataFrame):
                    lowhigh = np.float64([np.finfo(np.float64).min, np.finfo(np.float64).max])
                    if ".rxLowHigh" in orig_entry:
                        lowhigh = orig_entry['.rxLowHigh']
                    if "low" in new_entry:
                        lowhigh[0] = new_entry['low']
                    if "high" in new_entry:
                        lowhigh[1] = new_entry['high']

                    low_high_processed = True
            elif new_key == "levels":
                if "levels" in orig_entry and len(new_entry["levels"]) != len(orig_entry["levels"]):
                    raise ValueError("Parameter 'var_info' contains a mismatch of levels for column {:s}".format(k))
                orig_entry["levels"] = new_entry["levels"]
                orig_entry["bIsFactor"] = True
                orig_entry["bSubtractOne"] = True
            else:
                orig_entry[new_key] = new_value

        if "list" not in orig_entry:
            if "VarType" in orig_entry and ("Date" in orig_entry["VarType"] or "PosixCt" in orig_entry["VarType"]):
                orig_entry["bIsDateVar"] = True
            else:
                orig_entry["bIsDateVar"] = False

    if isinstance(data, DataFrame):
        setattr(data, "..rxVarInfo", original_info)
        return data

    in_data = data
    if isinstance(in_data, str):
        in_data = RxXdfData(data)

    new_data_frame = DataFrame.from_dict(results)
    # Call set info, set data to be the results dataframe with the new var info attached as an attribute
    setattr(new_data_frame,"var_info", original_info)

    call_parameters = {}
    call_parameters["data"] = resolve_dataframe(new_data_frame)
    call_parameters["inSource"] = in_data.to_serialize
    call_parameters["colMapping"] = None

    results = rx_native_call(
         functionname="SetVarInfoXdf", params=call_parameters)

    return results
