import subprocess
import sys
from revoscalepy.utils.RxOptions import RxOptions
from subprocess import CompletedProcess
from typing import Union


class RxHadoopCommandResults:
    """
    .. rxtitle::

        Class RxHadoopCommandResults

    .. rxdescription::

        Returned result from rx hadoop commands

    .. rxkeywords:: Hadoop Command

    """
    def __init__(self, results: CompletedProcess):
        self._results = results

    @property
    def return_code(self):
        return self._results.returncode

    @property
    def stdout(self) -> list:
        return self._results.stdout.decode('utf-8').splitlines()

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return_code = self._results.returncode is 0
        return_code_str = "\n"+return_code.__str__()
        if self._results.stdout is None or len(self._results.stdout) is 0:
            return return_code_str

        return str.format("\n".join(self._results.stdout.decode('utf-8').splitlines())) + return_code_str


def rx_hadoop_command(cmd: str) -> RxHadoopCommandResults:
    """
    .. rxtitle::

        Execute Hadoop Commands

    .. rxdescription::

        Execute arbitrary Hadoop commands and perform standard file operations in Hadoop.

    :param cmd: A character string containing a valid Hadoop command, that
        is, the cmd portion of Hadoop Command. Embedded quotes are not permitted.

    :return: Class RxHadoopCommandResults

    .. rxkeywords:: Hadoop Command

    .. rxseealso::

        :py:func:`rx_hadoop_make_dir <revoscalepy.rx_hadoop_make_dir>`
        :py:func:`rx_hadoop_copy_from_local <revoscalepy.rx_hadoop_copy_from_local>`
        :py:func:`rx_hadoop_remove_dir <revoscalepy.rx_hadoop_remove_dir>`
        :py:func:`rx_hadoop_file_exists <revoscalepy.rx_hadoop_file_exists>`
        :py:func:`rx_hadoop_list_files <revoscalepy.rx_hadoop_list_files>`

    .. rxexample::
       from revoscalepy import rx_hadoop_command
       rx_hadoop_command("version")
    """
    # todo: after implement client server mode, suport parameters:
    # sshUsername=NULL, sshHostname=NULL, sshSwitches=NULL,sshProfileScript=NULL
    hadoop_cmd = " ".join([RxOptions.get_option("mrsHadoopPath"), cmd])

    if sys.platform == "win32" or sys.platform == "win64":
        raise TypeError('rxHadoopCommand not supported in non-Hadoop compute contexts on Windows')
    else:
        result = subprocess.run(hadoop_cmd.split(), stdout=subprocess.PIPE)
        return RxHadoopCommandResults(result)
    # todo: after implement client server mode, we need to implement with rx_remote_command and rx_local_command


def rx_hadoop_copy_from_local(source: str, dest: str):
    """
    .. rxtitle::

        Execute Hadoop Copy From Local Commands

    .. rxdescription::

        rx_hadoop_copy_from_local wraps the Hadoop fs -copyFromLocal command.

    :param source: A character string specifying file(s) in Local File System

    :param dest: A character string specifying the destination of a copy in HDFS
        If `source` includes more than one file, `dest` must be a directory.

    .. rxkeywords:: Hadoop Command

    .. rxexample::
       from revoscalepy import rx_hadoop_copy_from_local
       rx_hadoop_copy_from_local("/tmp/foo.txt", "/user/RevoShare/newUser")

    """
    if type(source) is list:
        source = " ".join(source)
    cmd = " ".join(["fs -copyFromLocal", source, dest])
    return rx_hadoop_command(cmd)


def rx_hadoop_copy_to_local(source: str, dest: str):
    """
    .. rxtitle::

        Execute Hadoop Copy To Local Commands

    .. rxdescription::

        rx_hadoop_copy_to_local wraps the Hadoop fs -copyToLocal command.

    :param source: A character string specifying file(s) to be copied in HDFS

    :param dest: A character string specifying the destination of a copy in Local File System
        If `source` includes more than one file, `dest` must be a directory.

    .. rxkeywords:: Hadoop Command

    .. rxexample::
       from revoscalepy import rx_hadoop_copy_to_local
       rx_hadoop_copy_to_local("/user/RevoShare/foo.txt", "/tmp/foo.txt")

    """
    if type(source) is list:
        source = " ".join(source)
    cmd = " ".join(["fs -copyToLocal", source, dest])
    return rx_hadoop_command(cmd)


def rx_hadoop_file_exists(path: Union[list, str]):
    """
    .. rxtitle::

        Execute Hadoop Files Exists Commands

    .. rxdescription::

        rx_hadoop_file_exists wraps the Hadoop fs -test -e command.

    :param path: Character string or list. A list of paths or A character string specifying location of one or more files
        or directories.
    :return: A bool value specifying whether file exists.

    .. rxkeywords:: Hadoop Command

    .. rxexample::
       from revoscalepy import rx_hadoop_file_exists
       rx_hadoop_file_exists("/user/RevoShare/newUser/foo.txt")

    """
    if type(path) is list:
        if len(path) != 1:
            raise ValueError("can handle only one path at a time")
        path = " ".join(path)
    cmd = " ".join(["fs -test -e", path])
    return rx_hadoop_command(cmd)


def rx_hadoop_list_files(path: Union[list, str], recursive=False):
    """
    .. rxtitle::

        Execute Hadoop List Files Commands

    .. rxdescription::

        rx_hadoop_list_files wraps the Hadoop fs -ls or -lsr command.

    :param path: Character string or list. A list of paths or A character string specifying location of one or more files
        or directories.

    :param recursive: Bool value. If True, directory listings are recursive.

    .. rxkeywords:: Hadoop Command

    .. rxexample::
       from revoscalepy import rx_hadoop_list_files
       rx_hadoop_list_files("/user/RevoShare/newUser")

    """
    if type(path) is list:
        path = " ".join(path)

    if recursive:
        cmd = " ".join(["fs -ls -R", path])
    else:
        cmd = " ".join(["fs -ls", path])
    return rx_hadoop_command(cmd)


def rx_hadoop_make_dir(path: Union[list, str]):
    """
    .. rxtitle::

        Execute Hadoop Make Directory Commands

    .. rxdescription::

        rx_hadoop_make_dir wraps the Hadoop fs -mkdir -p command.

    :param path: Character string or list. A list of paths or A character string specifying location of one or more
        directories.

    .. rxkeywords:: Hadoop Command

    .. rxexample::
       from revoscalepy import rx_hadoop_make_dir
       rx_hadoop_make_dir("/user/RevoShare/newUser")

    """
    if type(path) is list:
        path = " ".join(path)
    cmd = " ".join(["fs -mkdir -p", path])
    return rx_hadoop_command(cmd)


def rx_hadoop_copy(source: str, dest: str):
    """
    .. rxtitle::

        Execute Hadoop Copy Commands

    .. rxdescription::

        rx_hadoop_copy wraps the Hadoop fs -cp command.

    :param source: A character string specifying file(s) to be copied in HDFS

    :param dest: A character string specifying the destination of a copy in HDFS
        If `source` includes more than one file, `dest` must be a directory.

    .. rxkeywords:: Hadoop Command

    .. rxexample::
       from revoscalepy import rx_hadoop_copy
       rx_hadoop_copy("/user/RevoShare/newUser/foo.txt","/user/RevoShare/newUser/bar.txt")

    """
    if type(source) is list:
        source = " ".join(source)
    cmd = " ".join(["fs -cp", source, dest])
    return rx_hadoop_command(cmd)


def rx_hadoop_move(source: str, dest: str):
    """
    .. rxtitle::

        Execute Hadoop Move Commands

    .. rxdescription::

        rx_hadoop_move wraps the Hadoop fs -mv command.

    :param source: A character string specifying file(s) to be moved in HDFS

    :param dest: A character string specifying the destination of move in HDFS
        If `source` includes more than one file, `dest` must be a directory.

    .. rxkeywords:: Hadoop Command

    .. rxexample::
       from revoscalepy import rx_hadoop_move
       rx_hadoop_move("/user/RevoShare/newUser/foo.txt","/user/RevoShare/newUser/bar.txt")

    """
    if type(source) is list:
        source = " ".join(source)
    cmd = " ".join(["fs -mv", source, dest])
    return rx_hadoop_command(cmd)


def rx_hadoop_remove(path: Union[list, str], skip_trash=False):
    """
    .. rxtitle::

        Execute Hadoop Remove Commands

    .. rxdescription::

        rx_hadoop_remove wraps the Hadoop fs -rm or fs -rm -skipTrash command.

    :param path: Character string or list. A list of paths or A character string specifying location of one or more
        files.
    :param skip_trash. If True, removal bypasses the trash folder, if one has been set up.

    .. rxkeywords:: Hadoop Command

    .. rxexample::
       from revoscalepy import rx_hadoop_remove
       rx_hadoop_remove("/user/RevoShare/newUser/foo.txt", skip_trash = True)

    """
    if type(path) is list:
        path = " ".join(path)

    if skip_trash:
        cmd = " ".join(["fs -rm", "-skipTrash", path])
    else:
        cmd = " ".join(["fs -rm", path])
    return rx_hadoop_command(cmd)


def rx_hadoop_remove_dir(path: Union[list, str], skip_trash=False):
    """
    .. rxtitle::

        Execute Hadoop Remove Directory Commands

    .. rxdescription::

        rx_hadoop_remove_dir wraps the Hadoop fs -rm -r  or fs -rm -r -skipTrash command.

    :param path: Character string or list. A list of paths or A character string specifying location of one or more
        directories.
    :param skip_trash. If True, removal bypasses the trash folder, if one has been set up.

    .. rxkeywords:: Hadoop Command

    .. rxexample::
       from revoscalepy import rx_hadoop_remove_dir
       rx_hadoop_remove_dir("/user/RevoShare/newUser", skip_trash = True)

    """
    if type(path) is list:
        path = " ".join(path)

    if skip_trash:
        cmd = " ".join(["fs -rm -r", "-skipTrash", path])
    else:
        cmd = " ".join(["fs -rm -r", path])
    return rx_hadoop_command(cmd)
