from typing import Union
from collections import OrderedDict

from pandas import DataFrame
from numpy import ndarray

from revoscalepy.datasource import RxXdfData
from revoscalepy.datasource import RxSqlServerData
from revoscalepy.utils.RxUtils import _rx_set_verbose_flags

from ..RxFunction import RxFunction
from .. RxFunction import RxModelResults
from ..RxSerializable import RxParameter
from ..RxSerializable import rx_native_call
from ..RxSerializable import resolve_dataframe, _resolve_transform_function
from ..computecontext.RxInSqlServer import RxRemoteComputeContext
from .. computecontext.RxSpark import RxSpark
from ..functions.RxExec import rx_exec
from ..computecontext.RxComputeContext import RxComputeContext
from ..etl.RxImport import rx_import
from ..utils.RxUtils import rx_print_header, _rx_construct_read_vars, _rx_get_formula_components, _rx_handle_formula
from ..utils.RxOptions import RxOptions
from ..helpers.Match import get_call_as_string
from revoscalepy.utils import on_exit_handler, telemetry_capture_call
from revoscalepy.etl.RxImport import _rx_create_sample_data_file

@on_exit_handler
def rx_lin_mod(formula,
               data,
               pweights=None,
               fweights=None,
               cube: bool = False,
               cube_predictions: bool = False,
               row_selection: str = None,
               transforms: dict = None,
               transform_objects: dict = None,
               transform_function: Union[str, callable] = None,
               transform_variables: list = None,
               transform_packages: list = None,
               transform_environment=None,
               drop_first: bool = False,
               drop_main: bool = True,
               cov_coef: bool = False,
               cov_data: bool = False,
               blocks_per_read: int = 1,
               report_progress: int = None,
               verbose: int = 0,
               compute_context=None,
               **kwargs):
    """
    .. rxtitle::

        Linear Models

    .. rxdescription::

        Fit linear models on small or large data.

    :param formula: Statistical model using symbolic formulas.
    :param data: Either a data source object, a character string specifying a
        .xdf file, or a data frame object.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param pweights: Character string specifying the variable to use as probability
        weights for the observations.
    :param fweights: Character string specifying the variable to use as frequency
        weights for the observations.
    :param cube: Bool flag. If True and the first term of the predictor variables
        is categorical (a factor or an interaction of factors), the regression is
        performed by applying the Frisch-Waugh-Lovell Theorem, which uses a partitioned
        inverse to save on computation time and memory.
    :param cube_predictions: Bool flag. If True and cube is True the predicted
        values are computed and included in the countDF component of the returned
        value. This may be memory intensive.
    :param row_selection: None. Not currently supported, reserved for future use.
    :param transforms: None. Not currently supported, reserved for future use.
    :param transform_objects: None. Not currently supported, reserved for future use.
    :param transform_function: Variable transformation function. The variables used
        in the transformation function must be specified in transformVars if they
        are not variables used in the model.
    :param transform_variables: List of strings of input data set variables needed
        for the transformation function.
    :param transform_packages: None. Not currently supported, reserved for future use.
    :param transform_environment: None. Not currently supported, reserved for future use.
    :param drop_first: Bool flag. If False, the last level is dropped in all sets
        of factor levels in a model. If that level has no observations (in any of the
        sets), or if the model as formed is otherwise determined to be singular, then
        an attempt is made to estimate the model by dropping the first level in all sets
        of factor levels. If True, the starting position is to drop the first level. Note
        that for cube regressions, the first set of factors is excluded from these rules
        and the intercept is dropped.
    :param drop_main: Bool value. If True, main-effect terms are dropped before their
        interactions.
    :param cov_coef: Bool flag. If True and if cube is False, the variance-covariance
        matrix of the regression coefficients is returned.
    :param cov_data: Bool flag. If True and if cube is False and if constant term is
        included in the formula, then the variance-covariance matrix of the data is
        returned.
    :param blocks_per_read: Number of blocks to read for each chunk of data read from
        the data source.
    :param report_progress: Integer value with options:
        0: No progress is reported.
        1: The number of processed rows is printed and updated.
        2: Rows processed and timings are reported.
        3: Rows processed and all timings are reported.
    :param verbose: Integer value. If 0, no additional output is printed. If 1,
        additional summary information is printed.
    :param compute_context: A RxComputeContext object for prediction.
    :param kwargs: Additional parameters

    :return: A RxLinModResults object of linear model.

    .. rxkeywords:: linear

    .. rxseealso::

        :py:func:`rx_logit <revoscalepy.rx_logit>`.

    .. rxexample::
        import os
        import tempfile
        from revoscalepy import RxOptions, RxXdfData, rx_lin_mod

        sample_data_path = RxOptions.get_option("sampleDataDir")
        in_mort_ds = RxXdfData(os.path.join(sample_data_path, "mortDefaultSmall.xdf"))

        lin_mod = rx_lin_mod("creditScore ~ yearsEmploy", in_mort_ds)
    """
    telemetry_capture_call(compute_context)

    data = RxComputeContext.current().validate_in_data(data, "data")

    if transform_function is not None:
        transform_function = _resolve_transform_function(transform_function=transform_function)

    py_args = locals()
    for key in kwargs:
        py_args[key] = kwargs[key]
    py_args.pop("kwargs", None)
    py_function_name = "rx_lin_mod"
    py_module_name = __name__

    if compute_context is None:
        compute_context = RxComputeContext.current()

    # todo need to refactor the spark workflow with Keith
    if isinstance(compute_context, RxSpark):
        return compute_context._execute(py_module_name, py_function_name, py_args)

    function_name = "LinMod"
    call_parameters = {
        'RemoveMissings': True,
        'varTol': 1e-12,
        'rCondTol': 1e-14
    }
    results = {}

    # PARAMS
    call_parameters['Formula'] = _rx_handle_formula(RxFunction.assign_or_throw(formula, False, str), pweights, fweights)
    call_parameters['Cube'] = RxFunction.assign_or_throw(cube, False, bool)
    call_parameters['CubePredictions'] = RxFunction.assign_or_throw(cube_predictions, False, bool)
    call_parameters['rowSelection'] = RxFunction.assign_or_throw(row_selection, True, str)
    call_parameters['transforms'] = RxFunction.assign_or_throw(transforms, not_implemented=True)
    call_parameters['transformObjects'] = RxFunction.assign_or_throw(transform_objects, True)
    call_parameters['transformFunc'] = RxFunction.assign_or_throw(transform_function, True)
    transform_vars = RxFunction.assign_or_throw(transform_variables, True)
    call_parameters['transformVars'] = transform_vars
    call_parameters['transformEnvir'] = RxFunction.assign_or_throw(transform_environment, True)
    call_parameters['transformPackages'] = RxFunction.assign_or_throw(transform_packages, True)
    call_parameters['dropFirst'] = RxFunction.assign_or_throw(drop_first, False, bool)
    call_parameters['covCoef'] = RxFunction.assign_or_throw(cov_coef, False, bool)
    call_parameters['covData'] = RxFunction.assign_or_throw(cov_data, False, bool)
    call_parameters['dropMain'] = RxFunction.assign_or_throw(drop_main, False, bool)
    call_parameters['blocksPerRead'] = RxFunction.assign_or_throw(blocks_per_read, False, int)

    _rx_set_verbose_flags(call_parameters, verbose = verbose, report_progress = report_progress)

    call_parameters['dropPolicy'] = "Competitive"

    if isinstance(data, DataFrame):
        call_parameters['..rxDataFrame'] = resolve_dataframe(data)

        faddcp_param = OrderedDict()
        faddcp_param['bComputeLowHigh'] = True
        call_parameters['FaddCP'] = faddcp_param

        call_parameters['readVars'] = data.columns.tolist()
    else:
        call_parameters['DataSource'] = RxFunction.resolve_datasource(data)
        call_parameters['readVars'] = _rx_construct_read_vars(_rx_get_formula_components(formula), pweights, fweights, transform_vars)

    #
    # capture sample data if its a remote compute context call
    #
    # print("Using compute context {0}".format(type(compute_context).__name__))

    if isinstance(compute_context, RxRemoteComputeContext):
        job = compute_context.create_job()
        _rx_create_sample_data_file(
            input_data=data,
            sample_data_file=job.sample_data_file
            )
        if isinstance(data, RxSqlServerData):
            job.set_input_data_query(data.input_data_query)
            job.set_use_sql_satellite(data.use_fast_read)
            data.set_in_sql_server(True)

    # print("Calling function {0}".format(function_name))
    # print("Serializing parameters: ")

    #
    # make the C++ call
    #
    # print(compute_context)

    call = get_call_as_string()

    if isinstance(compute_context, RxRemoteComputeContext):
        def process_results(ret):
            if len(ret) > 0:
                ret = ret[0]

            return RxLinModResults(formula, data, ret, call_parameters, compute_context, call)

        return rx_exec(function=rx_native_call, args={'functionname': function_name, 'params': call_parameters},
                       compute_context=compute_context, callback=process_results)
    else:
        results = rx_native_call(functionname=function_name, params=call_parameters)

        return RxLinModResults(formula, data, results, call_parameters, compute_context, call)


class RxLinModResults(RxModelResults):
    def __init__(self, formula, data, results, params, compute_context, call):
        super().__init__(model_type = "rxLinMod", formula = formula, params = params, data = data, compute_context = compute_context, call = call)
        self._results = results
        self.__cached_countDF = None

    @property
    def y_names(self) -> list:
        if 'y.names' in self._results:
            return [self._results['y.names']] if isinstance(self._results['y.names'], str) else self._results['y.names']

    @property
    def coef_names(self) -> list:
        if 'coef.names' in self._results:
            return [self._results['coef.names']] if isinstance(self._results['coef.names'], str) else self._results['coef.names']

    @property
    def valid_observations(self) -> float:
        if 'valid.obs' in self._results:
            return int(self._results['valid.obs'])

    @property
    def missing_observations(self) -> float:
        if 'missing.obs' in self._results:
            return int(self._results['missing.obs'])

    @property
    def coefficients(self) -> Union[DataFrame, None]:
        if 'coefficients' in self._results:
            coefs = self._results['coefficients'] if isinstance(self._results['coefficients'],ndarray) else [self._results['coefficients']]
            ret = DataFrame(coefs)
            ret.index = self.coef_names
            return ret
        return None

    @property
    def covariance_coefficients(self) -> Union[DataFrame, None]:
        if 'covCoef' in self._results:
            return DataFrame(self._results['covCoef'])
        return None

    @property
    def covariance_data(self) -> Union[DataFrame, None]:
        if 'covData' in self._results:
            return DataFrame(self._results['covData'])
        return None

    @property
    def aliased(self) -> Union[DataFrame, None]:
        if 'aliased' not in self._results:
            return None

        alias_list = self._results['aliased'] if hasattr(self._results['aliased'], '__getitem__') else [
            self._results['aliased']]
        ret_list = [0] * len(self.coef_names)

        for i in alias_list:
            ret_list[i] = 1

        ret = DataFrame(ret_list)
        ret.index = self.coef_names
        return ret

    @property
    def cube(self) -> bool:
        return self._params['Cube']

    @property
    def rank(self) -> int:
        return self._results['rank']

    @property
    def residual_squares(self) -> float:
        return self._results['residual.squares']

    @property
    def condition_number(self) -> float:
        return self._results['condition.number']

    @property
    def coefficient_standard_error(self) -> Union[ndarray, None]:
        if 'coef.std.error' not in self._results:
            return None
        return self._results['coef.std.error']

    @property
    def coefficient_t_value(self) -> Union[ndarray, None]:
        if 'coef.t.value' not in self._results:
            return None
        return self._results['coef.t.value']

    @property
    def coefficient_p_value(self) -> Union[ndarray, None]:
        if 'coef.p.value' not in self._results:
            return None
        return self._results['coef.p.value']

    @property
    def total_squares(self) -> Union[ndarray, float]:
        return self._results['total.squares']

    @property
    def y_var(self) -> Union[ndarray, float]:
        return self._results['y.var']

    @property
    def sigma(self) -> Union[ndarray, float]:
        return self._results['sigma']

    @property
    def residual_variance(self) -> Union[float]:
        return self._results['residual.variance']

    @property
    def r_squared(self) -> DataFrame:
        return DataFrame(self._results['r.squared'])

    @property
    def f_pvalue(self) -> float:
        return self._results['f.pvalue']

    @property
    def partitions(self):
        if 'partitions' not in self._results:
            return None
        return self._results['partitions']

    @property
    def df(self):
        return self._results['df']

    @property
    def f_statistics(self):
        if 'fstatistics' not in self._results:
            return None
        return self._results['fstatistics']

    @property
    def adjusted_r_squared(self):
        if 'adj.r.squared' not in self._results:
            return None
        return self._results['adj.r.squared']

    @property
    def number_valid_observations(self):
        return self._results['valid.obs']

    @property
    def number_missing_observations(self):
        return self._results['missing.obs']

    @property
    def deviance(self) -> float:
        return self._results['deviance']

    @property
    def anova(self):
        if 'anova' not in self._results:
            return None
        return self._results['anova']

    @property
    def formula_base(self):
        if 'formulaBase' not in self._results:
            return None
        return self._results['formulaBase']

    @property
    def countDF(self) -> Union[DataFrame, None]:
        if self.__cached_countDF is not None:
            return self.__cached_countDF

        if 'countDF' not in self._results or 'countDF' not in self._results['countDF']:
            return None

        ret = DataFrame(self._results['countDF']['countDF'])

        for x in ret:
            if x in self._results['xlevels']:
                ret[x] = [self._results['xlevels'][x][y-1] for y in ret[x]]
                ret[x] = ret[x].astype('category', categories=self._results['xlevels'][x])

        self.__cached_countDF = ret
        return ret

    def print(self):
        if not len(self._results) == 0:
            print(self._results)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        # TODO print DataSet information
        header = rx_print_header(self.cube,
                                 self.formula,
                                 self.valid_observations,
                                 self.missing_observations,
                                 "Linear regression",
                                 number_of_independent_variables=len(self.coef_names),
                                 number_dropped=0,
                                 data_source=self._data,
                                 call=self._call,
                                 dependent_vars=self.y_names)

        df = DataFrame(self.coefficients) if self.coefficients is not None else DataFrame()

        if self.y_names is not None and any(self.y_names):
            df.columns = self.y_names

        if self.coef_names is not None and any(self.coef_names):
            df.insert(0, 'coefficients', self.coef_names)

        return header + df.__str__()

    def summary(self, header=True, footer=True, digits=3) -> str:
        header_str = ''
        footer_str = ''
        if RxOptions.get_option("NumDigits") is not None:
            digits = max(digits, RxOptions.get_option("NumDigits") - 3)
        if header:
            header_str = rx_print_header(self.cube,
                                         self.formula,
                                         self.valid_observations,
                                         self.missing_observations,
                                         "Linear regression",
                                         number_of_independent_variables=len(self.coef_names),
                                         number_dropped=0,
                                         data_source=self._data,
                                         call=self._call,
                                         dependent_vars=self.y_names)

        df = DataFrame(self.coefficients)
        df.columns = self.y_names
        df.insert(0, 'coefficients', self.coef_names)
        if footer:
            sigma = self._results['sigma']
            degrees_freedom = self._results['df']
            if sigma is not None:
                footer_str += "\nResidual standard error: {0} on {1} degrees of freedom".format(
                    round(sigma, digits), degrees_freedom)
            # TODO: Handle na.action
            if 'r.squared' in self._results:
                footer_str += "\nMultiple R-squared:{0}".format(round(self._results['r.squared'][0][0], digits))
            if 'adj.r.squared' in self._results:
                footer_str += "\nAdjusted R-squared:{0}".format(round(self._results['adj.r.squared'], digits))
            # TODO: Handle p-value
            footer_str += "\nF-statistic:{0} on {1} and {2} DF".format(
                round(self._results['f.value'], digits),
                round(self._results['f.numdf'], 0),
                round(self._results['df'], 0)
            )
            footer_str += "\nCondition number:{0}".format(round(1 / self._results['condition.number'], digits))
            # TODO: Handle correlation
        return header_str + df.__str__() + footer_str
