from typing import Union
from collections import OrderedDict

from pandas import DataFrame
from numpy import ndarray
from numpy import array

from revoscalepy.datasource import RxDataSource
from revoscalepy.datasource import RxXdfData
from revoscalepy.datasource import RxSqlServerData
from revoscalepy.utils import RxUtils
from revoscalepy.utils.RxUtils import _rx_set_verbose_flags

from ..RxFunction import RxFunction
from ..RxFunction import RxModelResults
from ..RxSerializable import RxParameter
from ..computecontext.RxComputeContext import RxComputeContext
from .. computecontext.RxSpark import RxSpark
from ..computecontext.RxInSqlServer import RxRemoteComputeContext
from ..utils.RxOptions import RxOptions
from ..etl.RxImport import rx_import
from ..RxSerializable import rx_native_call
from .. RxSerializable import resolve_dataframe, _resolve_transform_function
from ..functions.RxExec import rx_exec

from ..utils.RxUtils import rx_print_header, _rx_construct_read_vars, _rx_get_formula_components, _rx_handle_formula
from ..helpers.Match import get_call_as_string

from revoscalepy.utils import on_exit_handler
from revoscalepy.utils import telemetry_capture_call
from revoscalepy.etl.RxImport import _rx_create_sample_data_file

# TODO: Let's pull transformation input arguments into an RxTransform class

@on_exit_handler
def rx_logit(formula: str,
                data: Union[RxDataSource, DataFrame],
                pweights: str = None,
                fweights: str = None,
                cube: bool = False,
                cube_predictions: bool = False,
                variable_selection: list = None,
                row_selection: str = None,
                transforms: dict = None,
                transform_objects: dict = None,
                transform_function: Union[str, callable] = None,
                transform_variables: list = None,
                transform_packages: list = None,
                transform_environment=None,
                drop_first: bool = False,
                drop_main: bool = None,
                cov_coef: bool = False,
                cov_data: bool = False,
                initial_values: Union[None, list, int] = None,
                coef_label_style: str = None,
                blocks_per_read: int = None,
                max_iterations: int = None,
                coefficent_tolerance: float = 1e-06,
                gradient_tolerance: float = 1e-06,
                objective_function_tolerance: float = 1e-08,
                report_progress: int = None,
                verbose: int = 0,
                compute_context: RxComputeContext = None,
                **kwargs) -> 'RxLogitResults':
    """
    .. rxtitle::

        Logistic Regression

    .. rxdescription::

        Use rx_logit to fit logistic regression models for small or large data.

    :param formula: Statistical model using symbolic formulas. Dependent variable must 
        be binary. It can be a bool variable, a factor with only two categories,
        or a numeric variable with values in the range (0,1). In the latter case 
        it will be converted to a bool.
    :param data: Either a data source object, a character string specifying a 
        ‘.xdf’ file, or a data frame object.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param pweights: Character string specifying the variable to use as probability 
        weights for the observations.
    :param fweights: Character string specifying the variable to use as frequency 
        weights for the observations.
    :param cube: Bool flag. If True and the first term of the predictor variables 
        is categorical (a factor or an interaction of factors), the regression is 
        performed by applying the Frisch-Waugh-Lovell Theorem, which uses a partitioned 
        inverse to save on computation time and memory.
    :param cube_predictions: Bool flag. If True and cube is True the estimated 
        model is evaluated (predicted) for each cell in the cube, fixing the non-cube 
        variables in the model at their mean values, and these predictions are included 
        in the countDF component of the returned value. This may be time and memory 
        intensive for large models.
    :param variable_selection: A list specifying various parameters that control 
        aspects of stepwise regression. If it is an empty list (default), no stepwise 
        model selection will be performed. If not, stepwise regression will be 
        performed and cube must be False.
    :param row_selection: None. Not currently supported, reserved for future use.
    :param transforms: None. Not currently supported, reserved for future use. 
    :param transform_objects: None. Not currently supported, reserved for future use. 
    :param transform_function: Variable transformation function.
    :param transform_variables: List of strings of input data set variables needed 
        for the transformation function.
    :param transform_packages: None. Not currently supported, reserved for future use. 
    :param transform_environment: None. Not currently supported, reserved for future use. 
    :param drop_first: Bool flag. If False, the last level is dropped in all sets 
        of factor levels in a model. If that level has no observations (in any of the 
        sets), or if the model as formed is otherwise determined to be singular, then 
        an attempt is made to estimate the model by dropping the first level in all 
        sets of factor levels. If True, the starting position is to drop the first 
        level. Note that for cube regressions, the first set of factors is excluded 
        from these rules and the intercept is dropped.
    :param drop_main: Bool value. If True, main-effect terms are dropped before 
        their interactions.
    :param cov_coef: Bool flag. If True and if cube is False, the variance-covariance 
        matrix of the regression coefficients is returned.
    :param cov_data: Bool flag. If True and if cube is False and if constant term 
        is included in the formula, then the variance-covariance matrix of the data 
        is returned.
    :param initial_values: Starting values for the Iteratively Reweighted Least 
        Squares algorithm used to estimate the model coefficients. 
    :param coef_label_style: Character string specifying the coefficient label style. 
        The default is "Revo".
    :param blocks_per_read: Number of blocks to read for each chunk of data read 
        from the data source.
    :param max_iterations: Maximum number of iterations.
    :param coefficent_tolerance: Convergence tolerance for coefficients. If the 
        maximum absolute change in the coefficients (step), divided by the maximum 
        absolute coefficient value, is less than or equal to this tolerance at the 
        end of an iteration, the estimation is considered to have converged. To 
        disable this test, set this value to 0.
    :param gradient_tolerance: This argument is deprecated. 
    :param objective_function_tolerance: Convergence tolerance for the objective function. 
    :param report_progress: Integer value with options:
        0: No progress is reported.
        1: The number of processed rows is printed and updated.
        2: Rows processed and timings are reported.
        3: Rows processed and all timings are reported.
    :param verbose: Integer value.
    :param compute_context: A RxComputeContext object for prediction.
    :param kwargs: Additional parameters

    :return: A RxLogitResults object of linear model.

    .. rxkeywords:: logic

    .. rxseealso::

        :py:func:`rx_lin_mod <revoscalepy.rx_lin_mod>`.

    .. rxexample:: 
        import os
        from revoscalepy import rx_logit, RxOptions, RxXdfData
        sample_data_path = RxOptions.get_option("sampleDataDir")
        kyphosis = RxXdfData(os.path.join(sample_data_path, "kyphosis.xdf"))

        logit = rx_logit('Kyphosis ~ Age + Number + Start', data = kyphosis)
    """
    telemetry_capture_call(compute_context)

    data = RxComputeContext.current().validate_in_data(data, "data")

    if blocks_per_read is None:
        blocks_per_read = RxOptions.get_option("BlocksPerRead")

    if drop_main is None:
        drop_main = RxOptions.get_option("DropMain")

    if coef_label_style is None:
        coef_label_style = RxOptions.get_option("CoefLabelStyle")

    if transform_function is not None:
        transform_function = _resolve_transform_function(transform_function=transform_function)

    py_args = locals()
    for key in kwargs:
        py_args[key] = kwargs[key]
    py_args.pop("kwargs", None)
    py_function_name = "rx_logit"
    py_module_name = __name__

    if compute_context is None:
        compute_context = RxComputeContext.current()

    # todo need to refactor the spark workflow with Keith
    if isinstance(compute_context, RxSpark):
        return compute_context._execute(py_module_name, py_function_name, py_args)

    function_name = "Logit"
    call_parameters = {
        'removeMissings': True,
        'varTol': 1e-12,
        'rCondTol': 1e-14
    }
    results = {}

    # PARAMS
    call_parameters['Formula'] = _rx_handle_formula(RxFunction.assign_or_throw(formula, False, str), pweights, fweights)
    call_parameters['Cube'] = RxFunction.assign_or_throw(cube, False, bool)
    call_parameters['CubePredictions'] = RxFunction.assign_or_throw(cube_predictions, False, bool)
    call_parameters['rowSelection'] = RxFunction.assign_or_throw(row_selection, True, str)
    call_parameters['transforms'] = RxFunction.assign_or_throw(transforms, not_implemented=True)
    call_parameters['transformObjects'] = RxFunction.assign_or_throw(transform_objects, True)
    call_parameters['transformFunc'] = RxFunction.assign_or_throw(transform_function, True)
    transform_vars = RxFunction.assign_or_throw(transform_variables, True)
    call_parameters['transformVars'] = transform_vars
    call_parameters['transformEnvir'] = RxFunction.assign_or_throw(transform_environment, True)
    call_parameters['transformPackages'] = RxFunction.assign_or_throw(transform_packages, True)
    call_parameters['dropFirst'] = RxFunction.assign_or_throw(drop_first, False, bool)
    call_parameters['covCoef'] = RxFunction.assign_or_throw(cov_coef, False, bool)
    call_parameters['covData'] = RxFunction.assign_or_throw(cov_data, False, bool)
    call_parameters['initial_values'] = RxFunction.assign_or_throw(initial_values, True)
    call_parameters['dropMain'] = RxFunction.assign_or_throw(drop_main, False, bool)
    call_parameters['blocksPerRead'] = RxFunction.assign_or_throw(blocks_per_read, False, int)
    call_parameters['MaxIterations'] = 25 if max_iterations is None else RxFunction.assign_or_throw(
        max_iterations, False, int)
    call_parameters['bUserSetMaxIterations'] = max_iterations is not None
    call_parameters['CoeffTolerance'] = RxFunction.assign_or_throw(coefficent_tolerance, False, float)
    call_parameters['ObjectiveFunctionTolerance'] = RxFunction.assign_or_throw(objective_function_tolerance, False, float)
    call_parameters['GradientTolerance'] = RxFunction.assign_or_throw(gradient_tolerance, False, float)

    if isinstance(data, DataFrame):
        call_parameters['..rxDataFrame'] = resolve_dataframe(data)

        faddcp_param = OrderedDict()
        faddcp_param['bComputeLowHigh'] = True
        call_parameters['FaddCP'] = faddcp_param

        call_parameters['readVars'] = data.columns.tolist()
    else:
        call_parameters['DataSource'] = RxFunction.resolve_datasource(data)
        call_parameters['readVars'] = _rx_construct_read_vars(_rx_get_formula_components(formula), pweights, fweights, transform_vars)

    #
    # capture sample data if its a remote compute context call
    #

    if isinstance(compute_context, RxRemoteComputeContext):
        job = compute_context.create_job()
        _rx_create_sample_data_file(
            input_data=data,
            sample_data_file=job.sample_data_file
            )
        if isinstance(data, RxSqlServerData):
            job.set_input_data_query(data.input_data_query)
            job.set_use_sql_satellite(data.use_fast_read)
            data.set_in_sql_server(True)

    _rx_set_verbose_flags(call_parameters, verbose = verbose, report_progress = report_progress)

    #
    # make the C++ call
    #
    call = get_call_as_string()
    if isinstance(compute_context, RxRemoteComputeContext):
        def process_results(ret):
            ret = ret if len(ret) == 0 else ret[0]

            return RxLogitResults(formula, data, ret, call_parameters, compute_context, call)

        return rx_exec(function=rx_native_call, args={'functionname': function_name, 'params': call_parameters},
                       compute_context=compute_context, callback=process_results)
    else:
        results = rx_native_call(
            functionname=function_name, params=call_parameters)

    return RxLogitResults(formula, data, results, call_parameters, compute_context, call)


class RxLogitResults(RxModelResults):
    def __init__(self, formula, data, results, params, compute_context, call):
        super().__init__(model_type = "rxLogit", formula = formula, params = params, data = data, compute_context = compute_context, call = call)
        self._results = results

    @property
    def y_names(self) -> list:
        if 'y.names' in self._results:
            return [self._results['y.names']] if isinstance(self._results['y.names'], str) else self._results['y.names']

    @property
    def coef_names(self) -> list:
        if 'coef.names' in self._results:
            return [self._results['coef.names']] if isinstance(self._results['coef.names'], str) else self._results['coef.names']

    @property
    def valid_observations(self) -> float:
        if 'valid.obs' in self._results:
            return float(self._results['valid.obs'])

    @property
    def missing_observations(self) -> float:
        if 'missing.obs' in self._results:
            return float(self._results['missing.obs'])

    @property
    def coefficients(self) -> Union[DataFrame, None]:
        if 'coefficients' in self._results:
            coefs = self._results['coefficients'] if isinstance(self._results['coefficients'],ndarray) else [self._results['coefficients']]
            ret = DataFrame(coefs)
            ret.index = self.coef_names
            return ret
        return None
        
    @property
    def coefficient_standard_error(self) -> Union[ndarray, None]:
        if 'coef.std.error' not in self._results:
            return None
        return self._results['coef.std.error']

    @property
    def coefficient_t_value(self) -> Union[ndarray, None]:
        if 'coef.t.value' not in self._results:
            return None
        return self._results['coef.t.value']

    @property
    def coefficient_p_value(self) -> Union[ndarray, None]:
        if 'coef.p.value' not in self._results:
            return None
        return self._results['coef.p.value']

    @property
    def f_pvalue(self) -> float:
        return self._results['f.pvalue']

    @property
    def df(self):
        return self._results['df']

    @property
    def countDF(self) -> Union[DataFrame, None]:
        if 'countDF' not in self._results:
            return None
        return DataFrame(self._results['countDF'])

    @property
    def deviance(self) -> float:
        return self._results['deviance']

    @property
    def dispersion(self) -> float:
        return self._results['dispersion']

    @property
    def partitions(self):
        if 'partitions' not in self._results:
            return None
        return self._results['partitions']

    @property
    def rank(self):
        if 'rank' not in self._results:
            return None
        return self._results['rank']

    def aic(self) -> Union[float, None]:
        if 'aic' not in self._results:
            return None
        return self._results['aic']

    @property
    def total_squares(self) -> Union[ndarray, float]:
        return self._results['total.squares']

    @property
    def covariance_coefficients(self) -> Union[DataFrame, None]:
        if 'covCoef' in self._results:
            return DataFrame(self._results['covCoef'])
        return None

    @property
    def condition_number(self) -> float:
        if 'condition.number' in self._results:
            return self._results['condition.number']

    @property
    def covariance_data(self) -> Union[DataFrame, None]:
        if 'covCoef' in self._results:
            return DataFrame(self._results['covData'])
        return None

    @property
    def fstatistics(self) -> list:
        return [self.df]

    @property
    def aliased(self) -> Union[DataFrame, None]:
        if 'aliased' not in self._results:
            return None

        alias_list = self._results['aliased'] if hasattr(self._results['aliased'], '__getitem__') else [
            self._results['aliased']]
        ret_list = [0] * len(self.coef_names)

        for i in alias_list:
            ret_list[i] = 1

        ret = DataFrame(ret_list)
        ret.index = self.coef_names
        return ret

    @property
    def cube(self) -> bool:
        return self._params['Cube']

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        header = rx_print_header(self.cube, self.formula, [self.valid_observations], [self.missing_observations],
                                 "Logistic Regression", len(self.coef_names), 0,
                                 data_source=self._data,
                                 call=self._call,
                                 dependent_vars=self.y_names)
        return header + "Coefficients:\n{0}".format(self.coefficients)

    def summary(self, digits=3, header=True) -> str:
        if RxOptions.get_option("NumDigits") is not None:
            digits = max(digits, RxOptions.get_option("NumDigits")-3)
        header_str = ''
        if header:
            header_str = rx_print_header(self.cube,
                                         self.formula,
                                         [self.valid_observations],
                                         [self.missing_observations],
                                         "Logistic Regression",
                                         len(self.coef_names),
                                         0,
                                         data_source=self._data,
                                         call=self._call,
                                         dependent_vars=self.y_names)
            header_str += "-2*LogLikelihood: {0} (Residual deviance on {1} degrees of freedom)".format(
                round(self._results['deviance'], digits),
                round(self._results['df'], digits)
            )

        coef_df = DataFrame(OrderedDict([
                    ("Estimate", self.coefficients.ix[:, 0]),
                    ("Std. Error", self._results['coef.std.error']),
                    ("z value", self._results['coef.t.value']),
                    ("Pr(>|z|)", self._results['coef.p.value'])
                ]))
        footer_str = "\n\nCoefficients:\n{0}\n".format(coef_df)
        footer_str += "\n\nCondition number of final variance-covariance matrix: {0}".format(
            round(1/self._results['condition.number'], digits))
        if 'iterations' in self._results:
            footer_str += "\nNumber of iterations: {0}".format(self._results['iterations'])

        return header_str + footer_str
