import os
import shutil
from pandas import DataFrame, Series
from typing import Union, List

from revoscalepy import rx_get_compute_context, rx_set_compute_context, RxLocalSeq
from revoscalepy.utils.RxUtils import rx_get_in_data_source, rx_get_out_data_source
from revoscalepy import RxDataSource, RxXdfData, RxOdbcData
from revoscalepy import rx_import
from revoscalepy.RxFunction import RxFunction, validate_argument
from revoscalepy.RxSerializable import rx_native_call, resolve_dataframe
from revoscalepy.functions.RxGetInfoXdf import rx_get_info

def rx_partition(input_data: Union[RxDataSource, DataFrame, str] = None,
                 output_data: RxXdfData = None,
                 vars_to_partition: Union[List[str], str] = None,
                 append: str = "rows",
                 overwrite: bool = False,
                 **kwargs) -> DataFrame:
    """
    .. rxtitle::

        Partition Data by Key Values and Save the results to a Partitioned .Xdf

    .. rxdescription::

        Partition input data sources by key values and save the results to a partitioned Xdf on disk.

    :param input_data: Either a data source object, a character string specifying 
        a ‘.xdf’ file, or a Pandas data frame object.
    :param output_data: A partitioned data source object created by RxXdfData 
        with createPartitionSet = True.
    :param vars_to_partition: List of strings of variable names to be used for 
        partitioning the input data set.
    :param append: Either "none" to create a new file or "rows" to append rows to 
        an existing file. If output_data exists and append is "none", the overwrite 
        argument must be set to True.
    :param overwrite: Bool value. If True, an existing output_data will be overwritten. 
        overwrite is ignored if appending rows.
    :param kwargs: Additional arguments.

    :return: A Pandas data frame of partitioning values and data sources, each row in the 
        Pandas data frame represents one partition and the data source in the last variable 
        holds the data of a specific partition.

    .. rxkeywords:: partition 

    .. rxseealso::

        :py:func:`rx_exec_by <revoscalepy.rx_exec_by>`.
        :py:func:`RxXdfData <revoscalepy.RxXdfData>`.

    .. rxexample:: 
        import os, tempfile
        from revoscalepy import RxOptions, RxXdfData, rx_partition
        data_path = RxOptions.get_option("sampleDataDir")
        
        # input xdf data source
        xdf_file = os.path.join(data_path, "claims.xdf")
        xdf_ds = RxXdfData(xdf_file)

        # create a partitioned xdf data source object
        out_xdf_file = os.path.join(tempfile._get_default_tempdir(), "outPartitions")
        out_xdf = RxXdfData(out_xdf_file, create_partition_set = True)

        # do partitioning for input data set
        partitions = rx_partition(input_data = xdf_ds, output_data = out_xdf, vars_to_partition = ["car.age","type"])
        print(partitions)
    """

    # force to run in local compute context
    current_cc = rx_get_compute_context()

    try:
        local_cc = RxLocalSeq();
        if current_cc is not local_cc:
            rx_set_compute_context(local_cc)

        # validate input arguments
        validate_argument(input_data, 'input_data', (str, DataFrame, RxDataSource), None)
        input_data = rx_get_in_data_source(input_data)

        output_data = validate_argument(output_data, 'output_data', RxXdfData, None)
        if not output_data.create_partition_set:
            raise ValueError("'output_data' is an invalid Partition RxXdfData object. Please make sure the object was created using RxXdfData with create_partition_set = True'")

        vars_to_partition = validate_argument(vars_to_partition, 'vars_to_partition', (str, list), None)
        append = validate_argument(append, 'append', (str, list), "rows")
        append_list = ["none", "rows"]
        if append not in append_list:
            raise ValueError("'append' must be either 'rows' or 'none'.")
        if 'rows' in append:
            append = True
        else:
            append = False

        overwrite = validate_argument(overwrite, 'overwrite', bool, False)

        if not append and overwrite:
            file_path, extension = os.path.splitext(output_data.file)
            if os.path.isdir(file_path):
                shutil.rmtree(file_path)

        # setup parameters and make the C++ call
        call_parameters = {}
        call_parameters['inSource'] = None
        call_parameters['..rxDataFrame'] = None
        if isinstance(input_data, RxDataSource):
            call_parameters['inSource'] = RxFunction.resolve_datasource(input_data)
        elif isinstance(input_data, DataFrame):
            call_parameters['inSource'] = {'..rxDataFrame': resolve_dataframe(input_data)}

        call_parameters['outSource'] = RxFunction.resolve_datasource(output_data)
        call_parameters['varsToPartition'] = vars_to_partition
        call_parameters['append'] = append
        call_parameters['overwrite'] = overwrite

        result = rx_native_call(functionname="Partition", params = call_parameters)

    finally:
        # reset compute context
        rx_set_compute_context(current_cc)

    return result


def _gen_key_value_str(key_name, value):
    import math
    if value is None or (isinstance(value, float) and math.isnan(value)):
        ret_str = "{} IS NULL".format(key_name)
    elif isinstance(value, str):
        ret_str = "{} = '{}'".format(key_name, value)
    else:
        ret_str = "{} = {}".format(key_name, value)

    return ret_str


def _rx_partition_odbc_data_source(input_data: RxOdbcData,
                                   vars_to_partition: Union[list, str],
                                   connection_string: str):
    """
    Read ODBC/SQL Server Data Source and partition it.
    """
    # validate input arguments
    validate_argument(input_data, 'input_data', RxOdbcData, None)
    vars_to_partition = validate_argument(vars_to_partition, 'vars_to_partition', (str, list), None)
    vars_to_partition = [vars_to_partition] if isinstance(vars_to_partition, str) else vars_to_partition
    connection_string = validate_argument(connection_string, 'connection_string', str, input_data.connection_string)

    table = input_data.table
    if table is None or table is "":
        raise ValueError("Table name must be specified in the Data Source.")

    # Construct a query that returns all distinct values of the given vars_to_partition
    partition_columns = ",".join(vars_to_partition)
    partition_query = "SELECT DISTINCT {} FROM {}".format(partition_columns, table)

    # Create ODBC data source object and load it into a data frame
    partition_odbc_ds = RxOdbcData(sql_query = partition_query, connection_string = connection_string)

    partition_df = rx_import(partition_odbc_ds)
    number_of_partitions = len(partition_df)
    number_of_columns = len(vars_to_partition)

    # for each row in partition_df, construct a query
    ds_list = Series()
    for i in range(number_of_partitions):
        # construct WHERE clause
        key_name = vars_to_partition[0]
        keys_str = _gen_key_value_str(key_name, partition_df.iloc[i][key_name])
        if number_of_columns > 1:
            for j in range(1,number_of_columns):
                key_name = vars_to_partition[j]
                keys_str = keys_str + " AND {}".format(_gen_key_value_str(key_name, partition_df.iloc[i][key_name]))

        # construct the query
        partition_i_query = "SELECT * FROM {} WHERE {}".format(table, keys_str)

        # create ODBC data source
        partition_i_odbc_ds = RxOdbcData(sql_query = partition_i_query, connection_string = connection_string)

        # append to the list
        ds_list.set_value(i, partition_i_odbc_ds)

    partition_df['DataSource'] = ds_list
    return partition_df


def rx_get_partitions(input_data: RxXdfData = None, **kwargs):
    """
    .. rxtitle::

        Get Partitions of a Partitioned Xdf Data Source

    .. rxdescription::

        Get partitions enumeration of a partitioned Xdf data source.

    :param input_data: An existing partitioned data source object which was 
        created by RxXdfData with create_partition_set = True and 
        constructed by rx_partition.

    :return: A Pandas data frame with (n+1) columns, the first n columns are 
        partitioning columns specified by vars_to_partition in rx_partition 
        and the (n+1)th column is a data source column where each row contains 
        an Xdf data source object of one partition.

    .. rxkeywords:: partition, execby,

    .. rxseealso::

        :py:func:`rx_exec_by <revoscalepy.rx_exec_by>`.
        :py:func:`RxXdfData <revoscalepy.RxXdfData>`.
        :py:func:`rx_partition <revoscalepy.rx_partition>`.

    .. rxexample:: 
        import os, tempfile
        from revoscalepy import RxOptions, RxXdfData, rx_partition, rx_get_partitions
        data_path = RxOptions.get_option("sampleDataDir")

        # input xdf data source
        xdf_file = os.path.join(data_path, "claims.xdf")
        xdf_ds = RxXdfData(xdf_file)

        # create a partitioned xdf data source object
        out_xdf_file = os.path.join(tempfile._get_default_tempdir(), "outPartitions")
        out_xdf = RxXdfData(out_xdf_file, create_partition_set = True)

        # do partitioning for input data set
        partitions = rx_partition(input_data = xdf_ds, output_data = out_xdf, vars_to_partition = ["car.age","type"], append = "none", overwrite = True)

        # use rx_get_partitions to load an existing partitioned xdf
        out_xdf_1 = RxXdfData(out_xdf_file)
        partitions_1 = rx_get_partitions(out_xdf_1)
        print(partitions_1)
    """
    # force to run in local compute context
    current_cc = rx_get_compute_context()
    local_cc = RxLocalSeq();
    if current_cc is not local_cc:
        rx_set_compute_context(local_cc)

    # validate input arguments
    input_data = validate_argument(input_data, 'input_data', (RxXdfData, str), None)
    error_message = "'input_data' must be a partitioned RxXdfData object or a non-empty character string of a partitioned xdf file."
    if isinstance(input_data, str):
        file_path, extension = os.path.splitext(input_data)
        if extension is not None and extension is not "" and extension is not "xdf":
            raise ValueError(error_message)

        input_data = RxXdfData(file_path, create_partition_set = True)

    if input_data.create_partition_set is False:
        raise ValueError(error_message)

    # setup parameters and make the C++ call
    call_parameters = {}
    call_parameters['inSource'] = RxFunction.resolve_datasource(input_data)
    result = rx_native_call(functionname="GetPartitions", params = call_parameters)

    # verify each partition
    number_of_partitions = len(result)
    for i in range(number_of_partitions):
        partition_ds = RxXdfData(file = result.iloc[i][len(result.columns)-1])
        rx_get_info(data = partition_ds, num_rows = 1)

    # reset compute context
    rx_set_compute_context(current_cc)

    return result
