﻿from typing import Union
from revoscalepy.computecontext.RxComputeContext import RxComputeContext
from revoscalepy.datasource.RxDataSource import RxDataSource
from revoscalepy.functions.RxDTree import RxDTreeResults
from revoscalepy.functions.RxPredictBase import RxPredictDefault, RxPredictDTree, RxPredictDForest
from revoscalepy.utils.RxOptions import RxOptions

from revoscalepy.utils import on_exit_handler
from revoscalepy.utils import telemetry_capture_call


###############################################################################
# default rx_predict
###############################################################################
@on_exit_handler
def rx_predict_default(model_object = None,
                  data: RxDataSource = None,
                  output_data: Union[RxDataSource, str] = None,
                  compute_standard_errors: bool = False,
                  interval: Union[list, str] = "none",
                  confidence_level: float = 0.95,
                  compute_residuals: bool = False,
                  type: Union[list, str] = None,
                  write_model_vars: bool = False,
                  extra_vars_to_write: Union[list, str] = None,
                  remove_missings: bool = False,
                  append: Union[list, str] = None,
                  overwrite: bool = False,
                  check_factor_levels: bool = True,
                  predict_var_names: Union[list, str] = None,
                  residual_var_names: Union[list, str] = None,
                  interval_var_names: Union[list, str] = None,
                  std_errors_var_names: Union[list, str] = None,
                  blocks_per_read: int = 0,
                  report_progress: int = None,
                  verbose: int = 0,
                  xdf_compression_level: int = 0,
                  compute_context: RxComputeContext = None,
                  **kwargs):
    """
    .. rxtitle::

        Predicted Values and Residuals for rx_lin_mod and rx_logit

    .. rxdescription::

        Compute predicted values and residuals using rx_lin_mod and
        rx_logit objects.

    :param model_object: Object returned from a call to rx_lin_mod and rx_logit.
        Objects with multiple dependent variables are not supported.
    :param data: a data frame or an RxXdfData data source object to be used for predictions.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param output_data: A character string specifying the output ‘.xdf’ file, a
        RxXdfData object, RxTextData object, a RxOdbcData data source, or a
        RxSqlServerData data source to store predictions.
    :param compute_standard_errors: Bool value. If True, the standard errors
        for each dependent variable are calculated.
    :param interval: Character string defining the type of interval calculation
        to perform. Supported values are "none", "confidence", and "prediction".
    :param confidence_level: Numeric value representing the confidence level on
        the interval [0, 1].
    :param compute_residuals: bool Balue. If True, residuals are computed.
    :param type: The type of prediction desired. Supported choices are: "response"
        and "link". If type = "response", the predictions are on the scale of the
        response variable. For instance, for the binomial model, the predictions
        are in the range (0,1). If type = "link", the predictions are on the scale
        of the linear predictors. Thus for the binomial model, the predictions are
        of log-odds.
    :param write_model_vars: Bool value. If True, and the output data set is
        different from the input data set, variables in the model will be written
        to the output data set in addition to the predictions (and residuals,
        standard errors, and confidence bounds, if requested). If variables from
        the input data set are transformed in the model, the transformed variables
        will also be included.
    :param extra_vars_to_write: None or list of strings of additional variables
        names from the input data or transforms to include in the output_data. If
        write_model_vars is True, model variables will be included as well.
    :param remove_missings: Bool value. If True, rows with missing values are removed.
    :param append: either "none" to create a new files or "rows" to append rows
        to an existing file. If output_data exists and append is "none", the overwrite
        argument must be set to True. Ignored for data frames.
    :param overwrite: Bool value. If True, an existing output_data will be overwritten.
        overwrite is ignored if appending rows. Ignored for data frames.
    :param check_factor_levels: Bool value.
    :param predict_var_names: List of strings specifying name(s) to give to the prediction results.
    :param residual_var_names: List of strings specifying name(s) to give to the residual results.
    :param interval_var_names: None or a list of strings defining low and high
        confidence interval variable names, respectively. If None, the strings
        "_Lower" and "_Upper" are appended to the dependent variable names to
        form the confidence interval variable names.
    :param std_errors_var_names: None or a list of strings defining variable
        names corresponding to the standard errors, if calculated. If None, the
        string "_StdErr" is appended to the dependent variable names to form the
        standard errors variable names.
    :param blocks_per_read: Number of blocks to read for each chunk of data read
        from the data source. If the data and output_data are the same file,
        blocks_per_read must be 1.
    :param report_progress: Integer value with options:
        0: No progress is reported.
        1: The number of processed rows is printed and updated.
        2: Rows processed and timings are reported.
        3: Rows processed and all timings are reported.
    :param verbose: Integer value. If 0, no additional output is printed. If 1,
        additional summary information is printed.
    :param xdf_compression_level: Integer in the range of -1 to 9 indicating the
        compression level for the output data if written to an .xdf file.
    :param compute_context: A RxComputeContext object for prediction.
    :param kwargs: Additional parameters

    :return: A data frame or a data source object of prediction results.

    .. rxkeywords:: predict

    .. rxseealso::

        :py:func:`rx_predict <revoscalepy.rx_predict>`.

    .. rxexample::
        import os
        from revoscalepy import RxOptions, RxXdfData, rx_lin_mod, rx_predict_default, rx_data_step

        sample_data_path = RxOptions.get_option("sampleDataDir")
        mort_ds = RxXdfData(os.path.join(sample_data_path, "mortDefaultSmall.xdf"))
        mort_df = rx_data_step(mort_ds)

        lin_mod = rx_lin_mod("creditScore ~ yearsEmploy", mort_df)
        pred = rx_predict_default(lin_mod, data = mort_df, compute_residuals = True, write_model_vars = True)
        pred.head()
    """
    telemetry_capture_call(compute_context)
    
    data = RxComputeContext.current().validate_in_data(data, "data")
    output_data = RxComputeContext.current().validate_out_data(output_data, "output_data")

    py_args = locals()
    for key in kwargs:
        py_args[key] = kwargs[key]

    model_object_parameters = model_object._params
    if "transformFunc" in model_object_parameters:
        py_args['transform_function'] = model_object_parameters['transformFunc']

    py_args.pop("kwargs", None)
    kwargs['py_args'] = py_args
    kwargs['py_function_name'] = "rx_predict_default"
    kwargs['py_module_name'] = __name__

    rx_predict = RxPredictDefault(model_object, data, output_data, compute_standard_errors,
                                  interval, confidence_level, compute_residuals, type,
                                  write_model_vars, extra_vars_to_write, remove_missings,
                                  append, overwrite, check_factor_levels, predict_var_names,
                                  residual_var_names, interval_var_names, std_errors_var_names,
                                  blocks_per_read, report_progress, verbose,
                                  xdf_compression_level, compute_context, **kwargs)

    return rx_predict.get_results()


###############################################################################
# rx_predict_rx_dtree
###############################################################################
@on_exit_handler
def rx_predict_rx_dtree(model_object = None,
                  data: RxDataSource = None,
                  output_data: Union[RxDataSource, str] = None,
                  predict_var_names: list = None,
                  write_model_vars: bool = False,
                  extra_vars_to_write: list = None,
                  append: Union[list, str] = "none",
                  overwrite: bool = False,
                  type: Union[list, str] = None,
                  remove_missings: bool = False,
                  compute_residuals: bool = False,
                  residual_type: Union[list, str] = "usual",
                  residual_var_names: list = None,
                  blocks_per_read: int = None,
                  report_progress: int = None,
                  verbose: int = 0,
                  xdf_compression_level: int = None,
                  compute_context = None,
                  **kwargs):
    """
    .. rxtitle::

        Prediction for Large Data Classification and Regression Trees

    .. rxdescription::

        Calculate predicted or fitted values for a data set from an rx_dtree object.

    :param model_object: Object returned from a call to rx_dtree.
    :param data: A data frame or an RxXdfData data source object to be used for predictions.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param output_data: An RxXdfData data source object or existing data frame
        to store predictions.
    :param predict_var_names: List of strings specifying name(s) to give to the prediction results
    :param write_model_vars: Bool value. If True, and the output data set is
        different from the input data set, variables in the model will be written
        to the output data set in addition to the predictions (and residuals,
        standard errors, and confidence bounds, if requested). If variables from
        the input data set are transformed in the model, the transformed variables
        will also be included.
    :param extra_vars_to_write: None or list of strings of additional variables
        names from the input data or transforms to include in the output_data. If
        write_model_vars is True, model variables will be included as well.
    :param append: Either "none" to create a new files or "rows" to append rows
        to an existing file. If output_data exists and append is "none", the overwrite
        argument must be set to True. Ignored for data frames.
    :param overwrite: Bool value. If True, an existing output_data will be overwritten.
        overwrite is ignored if appending rows. Ignored for data frames.
    :param type: the type of prediction desired. Supported choices are: "vector",
        "prob", "class", and "matrix".
    :param remove_missings: Bool value. If True, rows with missing values are removed.
    :param compute_residuals: Bool value. If True, residuals are computed.
    :param residual_type: Indicates the type of residual desired.
    :param residual_var_names: List of strings specifying name(s) to give to the residual results.
    :param blocks_per_read: Number of blocks to read for each chunk of data read
        from the data source. If the data and output_data are the same file,
        blocks_per_read must be 1.
    :param report_progress: Integer value with options:
        0: No progress is reported.
        1: The number of processed rows is printed and updated.
        2: Rows processed and timings are reported.
        3: Rows processed and all timings are reported.
    :param verbose: Integer value. If 0, no additional output is printed. If 1,
        additional summary information is printed.
    :param xdf_compression_level: Integer in the range of -1 to 9 indicating the
        compression level for the output data if written to an .xdf file.
    :param compute_context: A RxComputeContext object for prediction.
    :param kwargs: Additional parameters

    :return: A data frame or a data source object of prediction results.

    .. rxkeywords:: predict

    .. rxseealso::

        :py:func:`rx_predict <revoscalepy.rx_predict>`.

    .. rxexample::
        import os
        from revoscalepy import rx_dtree, rx_predict_rx_dtree, rx_import, RxOptions, RxXdfData
        sample_data_path = RxOptions.get_option("sampleDataDir")
        ds = RxXdfData(os.path.join(sample_data_path, "kyphosis.xdf"))
        kyphosis = rx_import(input_data = ds)

        # classification
        formula = "Kyphosis ~ Number + Start"
        method = "class"
        parms = {'prior': [0.8, 0.2], 'loss': [0, 2, 3, 0], 'split': "gini"}
        cost = [2,3]
        dtree = rx_dtree(formula, data = kyphosis, pweights = "Age", method = method, parms = parms, cost = cost, max_num_bins = 100)
        rx_pred = rx_predict_rx_dtree(dtree, data = kyphosis)
        rx_pred.head()

        # regression
        formula = "Age ~ Number + Start"
        method = "anova"
        parms = {'prior': [0.8, 0.2], 'loss': [0, 2, 3, 0], 'split': "gini"}
        cost = [2,3]
        dtree = rx_dtree(formula, data = kyphosis, pweights = "Kyphosis", method = method, parms = parms, cost = cost, max_num_bins = 100)
        rx_pred = rx_predict_rx_dtree(dtree, data = kyphosis)
        rx_pred.head()
    """
    telemetry_capture_call(compute_context)

    data = RxComputeContext.current().validate_in_data(data, "data")
    output_data = RxComputeContext.current().validate_out_data(output_data, "output_data")

    if type is None:
        type = ["vector", "prob", "class", "matrix"]

    if blocks_per_read is None:
        blocks_per_read = RxOptions.get_option("BlocksPerRead")

    if xdf_compression_level is None:
        xdf_compression_level = RxOptions.get_option("XdfCompressionLevel")

    py_args = locals()
    for key in kwargs:
        py_args[key] = kwargs[key]

    model_object_parameters = model_object._params 
    if "transformFunc" in model_object_parameters:
        py_args['transform_function'] = model_object_parameters['transformFunc']

    py_args.pop("kwargs", None)
    kwargs['py_args'] = py_args
    kwargs['py_function_name'] = "rx_predict_rx_dtree"
    kwargs['py_module_name'] = __name__

    if compute_context is None:
        compute_context = RxComputeContext.current()

    rx_predict = RxPredictDTree(model_object, data, output_data, predict_var_names,
                                write_model_vars, extra_vars_to_write, append,
                                overwrite, type, remove_missings,
                                compute_residuals, residual_type, residual_var_names,
                                blocks_per_read, report_progress, verbose,
                                xdf_compression_level, compute_context, **kwargs)

    return rx_predict.get_results()


###############################################################################
# rx_predict_rx_dforest
###############################################################################
@on_exit_handler
def rx_predict_rx_dforest(model_object = None,
                  data: RxDataSource = None,
                  output_data: Union[RxDataSource, str] = None,
                  predict_var_names: list = None,
                  write_model_vars: bool = False,
                  extra_vars_to_write: list = None,
                  append: Union[list, str] = "none",
                  overwrite: bool = False,
                  type: Union[list, str] = None,
                  cutoff: list = None,
                  remove_missings: bool = False,
                  compute_residuals: bool = False,
                  residual_type: Union[list, str] = "usual",
                  residual_var_names: list = None,
                  blocks_per_read: int = None,
                  report_progress: int = None,
                  verbose: int = 0,
                  xdf_compression_level: int = None,
                  compute_context = None,
                  **kwargs):
    """
    .. rxtitle::

        Prediction for Large Data Classification and Regression Forests

    .. rxdescription::

        Calculate predicted or fitted values for a data set from an rx_dforest or rx_btrees object.

    :param model_object: object returned from a call to rx_dtree.
    :param data: a data frame or an RxXdfData data source object to be used for predictions.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param output_data: an RxXdfData data source object or existing data frame
        to store predictions.
    :param predict_var_names: list of strings specifying name(s) to give to the prediction results
    :param write_model_vars: bool value. If True, and the output data set is
        different from the input data set, variables in the model will be written
        to the output data set in addition to the predictions (and residuals,
        standard errors, and confidence bounds, if requested). If variables from
        the input data set are transformed in the model, the transformed variables
        will also be included.
    :param extra_vars_to_write: None or list of strings of additional variables
        names from the input data or transforms to include in the output_data. If
        write_model_vars is True, model variables will be included as well.
    :param append: either "none" to create a new files or "rows" to append rows
        to an existing file. If output_data exists and append is "none", the overwrite
        argument must be set to True. Ignored for data frames.
    :param overwrite: bool value. If True, an existing output_data will be overwritten.
        overwrite is ignored if appending rows. Ignored for data frames.
    :param type: the type of prediction desired. Supported choices are: "response",
        "prob", and "vote".
    :param cutoff: (Classification only) a vector of length equal to the number of
        classes specifying the dividing factors for the class votes. The default is
        the one used when the decision forest is built.
    :param remove_missings: bool value. If True, rows with missing values are removed.
    :param compute_residuals: bool value. If True, residuals are computed.
    :param residual_type: Indicates the type of residual desired.
    :param residual_var_names: list of strings specifying name(s) to give to the residual results.
    :param blocks_per_read: number of blocks to read for each chunk of data read
        from the data source. If the data and output_data are the same file,
        blocks_per_read must be 1.
    :param report_progress: integer value with options:
        0: no progress is reported.
        1: the number of processed rows is printed and updated.
        2: rows processed and timings are reported.
        3: rows processed and all timings are reported.
    :param verbose: integer value. If 0, no additional output is printed. If 1,
        additional summary information is printed.
    :param xdf_compression_level: integer in the range of -1 to 9 indicating the
        compression level for the output data if written to an .xdf file.
    :param compute_context: a RxComputeContext object for prediction.
    :param kwargs: additional parameters

    :return: a data frame or a data source object of prediction results.

    .. rxkeywords:: predict

    .. rxseealso::

        :py:func:`rx_predict <revoscalepy.rx_predict>`.

    .. rxexample::
        import os
        from revoscalepy import rx_dforest, rx_predict_rx_dforest, rx_import, RxOptions, RxXdfData
        sample_data_path = RxOptions.get_option("sampleDataDir")
        ds = RxXdfData(os.path.join(sample_data_path, "kyphosis.xdf"))
        kyphosis = rx_import(input_data = ds)

        # classification
        formula = "Kyphosis ~ Number + Start"
        method = "class"
        parms = {'prior': [0.8, 0.2], 'loss': [0, 2, 3, 0], 'split': "gini"}

        dforest = rx_dforest(formula, data = kyphosis, pweights = "Age", method = method,
                             parms = parms, cost = [2, 3], max_num_bins = 100, importance = True,
                             n_tree = 3, m_try = 2, sample_rate = 0.5,
                             replace = False, seed = 0, compute_oob_error = 1)
        rx_pred = rx_predict_rx_dforest(dforest, data = kyphosis)
        rx_pred.head()

        # regression
        formula = "Age ~ Number + Start"
        method = "anova"
        parms = {'prior': [0.8, 0.2], 'loss': [0, 2, 3, 0], 'split': "gini"}

        dforest = rx_dforest(formula, data = kyphosis, pweights = "Kyphosis", method = method,
                             parms = parms, cost = [2, 3], max_num_bins = 100, importance = True,
                             n_tree = 3, m_try = 2, sample_rate = 0.5,
                             replace = False, seed = 0, compute_oob_error = 1)
        rx_pred = rx_predict_rx_dforest(dforest, data = kyphosis)
        rx_pred.head()
    """
    telemetry_capture_call(compute_context)

    data = RxComputeContext.current().validate_in_data(data, "data")
    output_data = RxComputeContext.current().validate_out_data(output_data, "output_data")

    if type is None:
        type = ["response", "prob", "vote"]

    if blocks_per_read is None:
        blocks_per_read = RxOptions.get_option("BlocksPerRead")

    if xdf_compression_level is None:
        xdf_compression_level = RxOptions.get_option("XdfCompressionLevel")

    py_args = locals()
    for key in kwargs:
        py_args[key] = kwargs[key]

    model_object_parameters = model_object._params 
    if "transformFunc" in model_object_parameters:
        py_args['transform_function'] = model_object_parameters['transformFunc']

    py_args.pop("kwargs", None)
    kwargs['py_args'] = py_args
    kwargs['py_function_name'] = "rx_predict_rx_dforest"
    kwargs['py_module_name'] = __name__

    if compute_context is None:
        compute_context = RxComputeContext.current()

    rx_predict = RxPredictDForest(model_object, data, output_data, predict_var_names,
                                  write_model_vars, extra_vars_to_write, append,
                                  overwrite, type, cutoff, remove_missings,
                                  compute_residuals, residual_type, residual_var_names,
                                  blocks_per_read, report_progress, verbose,
                                  xdf_compression_level, compute_context, **kwargs)
    return rx_predict.get_results()


###############################################################################
# rx_predict
###############################################################################
@on_exit_handler
def rx_predict(model_object = None, data = None, output_data = None, **kwargs):
    """
    .. rxtitle::

        Predicted Values and Residuals for rx_lin_mod, rx_logit, rx_dtree,
        rx_dforest and rx_btrees

    .. rxdescription::

        Generic function to compute predicted values and residuals using
        rx_lin_mod, rx_logit, rx_dtree, rx_dforest and rx_btrees objects.

        For specific functions, see:
            rx_predict_default for rx_lin_mod and rx_logit objects.
            rx_predict_rx_dtree for rx_dtree object.
            rx_predict_rx_dforest for rx_dforest and rx_btrees objects.

    :param model_object: object returned from a call to rx_lin_mod, rx_logit, rx_dtree,
        rx_dforest and rx_btrees. Objects with multiple dependent variables are not
        supported.
    :param data: a data frame or an RxXdfData data source object to be used for predictions.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param output_data: an RxXdfData data source object or existing data frame to store predictions.
    :param kwargs: additional parameters

    :return: a data frame or a data source object of prediction results.

    .. rxseealso::

        :py:func:`rx_predict_default <revoscalepy.rx_predict_default>`,
        :py:func:`rx_predict_rx_dtree <revoscalepy.rx_predict_rx_dtree>`,
        :py:func:`rx_predict_rx_dforest <revoscalepy.rx_predict_rx_dforest>`.

    .. rxkeywords:: predict

    .. rxexample::
        :execute:
        :print_output:

        import os
        from revoscalepy import RxOptions, RxXdfData, rx_lin_mod, rx_predict, rx_data_step

        sample_data_path = RxOptions.get_option("sampleDataDir")
        mort_ds = RxXdfData(os.path.join(sample_data_path, "mortDefaultSmall.xdf"))
        mort_df = rx_data_step(mort_ds)

        lin_mod = rx_lin_mod("creditScore ~ yearsEmploy", mort_df)
        pred = rx_predict(lin_mod, data = mort_df)
        print(pred.head())
    """
    telemetry_capture_call()
    if model_object is None:
        raise ValueError("'model_object' argument is required")

    params = {
        'predict_var_names': None,
        'write_model_vars': False,
        'extra_vars_to_write': None,
        'append': ["none"],
        'overwrite': False,
        'remove_missings': False,
        'compute_residuals': False,
        'residual_var_names': None,
        'blocks_per_read': RxOptions.get_option("BlocksPerRead"),
        'report_progress': RxOptions.get_option("ReportProgress"),
        'verbose': 0,
        'xdf_compression_level': RxOptions.get_option("XdfCompressionLevel"),
        'compute_context': RxComputeContext.current()
        }

    other_params = {}

    from revoscalepy.functions.RxLinMod import RxLinModResults
    from revoscalepy.functions.RxLogit import RxLogitResults
    from revoscalepy.functions.RxDTree import RxDTreeResults
    from revoscalepy.functions.RxDTree import RxDForestResults

    if type(model_object) == RxLinModResults or \
       type(model_object) == RxLogitResults:
        extra_params = {
            'compute_standard_errors': False,
            'interval': "none",
            'confidence_level': 0.95,
            'type': None,
            'check_factor_levels': True,
            'interval_var_names': None,
            'std_errors_var_names': None,
        }

        params.update(extra_params)
        for key, value in kwargs.items():
            if key in params:
                params[key] = value
            else:
                other_params[key] = value

        return rx_predict_default(model_object, data, output_data,
                                 params['compute_standard_errors'], params['interval'],
                                 params['confidence_level'], params['compute_residuals'],
                                 params['type'], params['write_model_vars'],
                                 params['extra_vars_to_write'], params['remove_missings'],
                                 params['append'], params['overwrite'], params['check_factor_levels'],
                                 params['predict_var_names'], params['residual_var_names'],
                                 params['interval_var_names'], params['std_errors_var_names'],
                                 params['blocks_per_read'], params['report_progress'], params['verbose'],
                                 params['xdf_compression_level'], params['compute_context'], **other_params)
    elif type(model_object) == RxDTreeResults:
        # dtree
        extra_params = {
            'type': ["vector", "prob", "class", "matrix"],
            'residual_type': "usual",
        }

        params.update(extra_params)
        for key, value in kwargs.items():
            if key in params:
                params[key] = value
            else:
                other_params[key] = value

        return rx_predict_rx_dtree(model_object, data, output_data,
                                   params['predict_var_names'], params['write_model_vars'],
                                   params['extra_vars_to_write'], params['append'],
                                   params['overwrite'], params['type'], params['remove_missings'],
                                   params['compute_residuals'], params['residual_type'],
                                   params['residual_var_names'], params['blocks_per_read'],
                                   params['report_progress'], params['verbose'],
                                   params['xdf_compression_level'], params['compute_context'],
                                   **other_params)
    elif type(model_object) == RxDForestResults:
        #dforest
        extra_params = {
            'type': ["response", "prob", "vote"],
            'cutoff': None,
            'residual_type': ["usual"],
        }

        params.update(extra_params)
        for key, value in kwargs.items():
            if key in params:
                params[key] = value
            else:
                other_params[key] = value

        return rx_predict_rx_dforest(model_object, data, output_data, params['predict_var_names'],
                                     params['write_model_vars'], params['extra_vars_to_write'], params['append'],
                                     params['overwrite'], params['type'], params['cutoff'], params['remove_missings'],
                                     params['compute_residuals'], params['residual_type'], params['residual_var_names'],
                                     params['blocks_per_read'], params['report_progress'], params['verbose'],
                                     params['xdf_compression_level'], params['compute_context'], **other_params)
    else:
        raise ValueError("'model_object' is not valid")

