import tempfile
import numpy
from numpy import cumsum
from os import path
from pandas import DataFrame
from pandas import Series
from patsy import ModelDesc
from typing import Union

from revoscalepy.computecontext.RxComputeContext import RxComputeContext
from revoscalepy.computecontext.RxInSqlServer import RxRemoteComputeContext
from revoscalepy.computecontext.RxSpark import RxSpark
from revoscalepy.computecontext.RxLocalSeq import RxLocalSeq
from revoscalepy.datasource.RxDataSource import RxDataSource
from revoscalepy.datasource.RxSqlData import RxSqlServerData
from revoscalepy.datasource.RxXdfData import RxXdfData
from revoscalepy.etl.RxImport import rx_import
from revoscalepy.functions.RxExec import rx_exec
from revoscalepy.functions.RxDTree import RxDTreeResults
from revoscalepy.utils.RxUtils import rx_get_in_data_source, rx_get_out_data_source
from revoscalepy.utils.RxUtils import _rx_set_verbose_flags
from revoscalepy.utils.RxUtils import _rx_get_dependent_vars, _rx_get_input_vars
from revoscalepy.utils.RxOptions import RxOptions

from revoscalepy.RxFunction import RxFunction
from revoscalepy.RxSerializable import rx_native_call, resolve_dataframe
from revoscalepy.RxSerializable import parse_result_dataframe


###############################################################################
# helper functions
###############################################################################
def compute_prediction_names(model_object,
                             predict_variable_names=None,
                             residual_variable_names=None,
                             compute_residuals=False):
    ret = []

    if predict_variable_names is None:
        if len(model_object.y_names) > 1:
            raise ValueError("Predictions with more than one dependent variable are not supported.")
        ret.append(model_object.y_names[0] + "_Pred")

    if residual_variable_names is None and compute_residuals:
        if len(model_object.y_names) > 1:
            raise ValueError("Predictions with more than one dependent variable are not supported.")
        ret.append(model_object.y_names[0] + "_Resid")

    return ret


def check_list_or_str(var_name, var, value_list: list = None, error_str: str = ""):
    """
        check the given var:
            1: var is a type of list or str
            2: var must be one of the values in value_list
        otherwise, raise error with the given error_str
    """
    if error_str == "":
        error_str = "'{0}' must be a character string or a list of character strings.".format(var_name)

    if not (var is None or isinstance(var, str) or isinstance(var, list)):
        raise ValueError(error_str)

    if var is not None and value_list is not None:
        if isinstance(var, list):
            if len(var) > 1:
                raise ValueError(error_str)
            var = var[0]

        if var not in value_list:
            raise ValueError(error_str)


###############################################################################
# rx_predict class
###############################################################################
class RxPredict:
    """
        Base class for rxPredict function
    """
    def __init__(self, model_object = None,
                 data: RxDataSource = None,
                 output_data: Union[RxDataSource, str] = None,
                 compute_residuals: bool = False,
                 predict_var_names: Union[list, str] = None,
                 residual_var_names: Union[list, str] = None,
                 write_model_vars: bool = False,
                 extra_vars_to_write: Union[list, str] = None,
                 type: Union[list, str] = None,
                 remove_missings: bool = False,
                 append: Union[list, str] = None,
                 overwrite: bool = False,
                 report_progress: int = None,
                 verbose: int = 0,
                 xdf_compression_level: int = 0,
                 compute_context: RxComputeContext = None,
                 **kwargs):

        if 'py_args' in kwargs:
            self._py_args = kwargs['py_args']
        if 'py_function_name' in kwargs:
            self._py_function_name = kwargs['py_function_name']
        if 'py_module_name' in kwargs:
            self._py_module_name = kwargs['py_module_name']

        super().__init__()

        # model_object
        if model_object is None:
            raise ValueError("'model_object' argument is required")
        self._model_object = model_object

        # data and output_data
        if data is None:
            raise ValueError("'data' argument is required")
        data = rx_get_in_data_source(data)
        output_data = rx_get_out_data_source(output_data)

        self._data = data
        self._output_data = output_data

        self._compute_residuals = RxFunction.assign_or_throw(compute_residuals, False, bool)

        # extra_vars_to_write, predict_var_names, residual_var_names
        var_names = {}
        var_names['extra_vars_to_write'] = extra_vars_to_write
        var_names['predict_var_names'] = predict_var_names
        var_names['residual_var_names'] = residual_var_names
        for key, value in var_names.items():
            check_list_or_str(key, value, None, "")

        if predict_var_names is not None and residual_var_names is not None \
           and isinstance(predict_var_names, list) and isinstance(residual_var_names, list) \
           and len(predict_var_names) != len(residual_var_names):
            raise ValueError("Length of 'predict_var_names' must equal that of 'residual_var_names'")

        self._predict_var_names = predict_var_names
        self._residual_var_names = residual_var_names
        self._extra_vars_to_write = extra_vars_to_write
        # TODO: validate extra vars in dependent vars and data columns

        self._write_model_vars = write_model_vars
        self._type = type[0] if isinstance(type, list) else type
        self._remove_missings = remove_missings
        
        # append_eval
        append_list = ["none", "rows", "columns", "cols"]
        check_list_or_str("append", append, append_list, 
                          "'append' must be a character string or a list of character strings of 'none', 'rows', or 'cols'.")
        self._append_eval = ["none", "rows"] if append is None else append
        self._overwrite = overwrite

        self._report_params = _rx_set_verbose_flags({}, verbose = verbose, report_progress = report_progress)

        self._xdf_compression_level = xdf_compression_level
        self._compute_context = RxComputeContext.current() if compute_context is None else compute_context

    @property
    def function_name(self):
        raise NotImplementedError

    @property
    def call_parameters(self):
        raise NotImplementedError

    @property
    def results(self):
        raise NotImplementedError

    @property
    def model_object(self):
        return self._model_object

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        """ setter for _data """
        self._data = value

    @property
    def output_data(self):
        return self._output_data

    @output_data.setter
    def output_data(self, value):
        """ setter for _output_data """
        self._output_data = value

    @property
    def compute_residuals(self):
        return self._compute_residuals

    @property
    def predict_var_names(self):
        return self._predict_var_names

    @property
    def residual_var_names(self):
        return self._residual_var_names

    @residual_var_names.setter
    def residual_var_names(self, value):
        self._residual_var_names = value

    @property
    def extra_vars_to_write(self):
        return self._extra_vars_to_write

    @property
    def write_model_vars(self):
        return self._write_model_vars

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def remove_missings(self):
        return self._remove_missings

    @property
    def append_eval(self):
        return self._append_eval

    @property
    def overwrite(self):
        return self._overwrite

    @property
    def report_params(self):
        return self._report_params

    @property
    def xdf_compression_level(self):
        return self._xdf_compression_level

    @property
    def compute_context(self):
        return self._compute_context

    def get_results(self):
        """
            format the results to return
        """
        if self.results is None:
            # input data is a DataSource and output_data is None
            # then, return the modified input DataSource
            if isinstance(self.data, RxDataSource) and self.output_data is None:
                return (self.data)
            elif isinstance(self.output_data, RxDataSource):
                return (self.output_data)

        # Single Scalar values must be in a list
        if isinstance(self.results, dict):
            for x in self.results:
                if not hasattr(self.results[x], '__iter__'):
                    self.results[x] = [self.results[x]]

        # output data is DataFrame
        # then, output is a dataframe combining output data and predict results.
        if isinstance(self.output_data, DataFrame):
            self.output_data = self.output_data.join(DataFrame(self.results))
            self.results = self.output_data
            return (self.results)

        if isinstance(self.results, dict):
            return (DataFrame(self.results))

        return self.results 


###############################################################################
# rx_predict subclass for default method for LinMod, Logit and Glm
###############################################################################
class RxPredictDefault(RxPredict):
    """
        Subclass for rxPredict default function
    """
    function_name = "Predict"
    call_parameters = {}
    results = {}

    def __init__(self,
                 model_object = None,
                 data: RxDataSource = None,
                 output_data: Union[RxDataSource, str] = None,
                 compute_standard_errors: bool = False,
                 interval: Union[list, str] = None,
                 confidence_level: float = 0.95,
                 compute_residuals: bool = False,
                 type: Union[list, str] = None,
                 write_model_vars: bool = False,
                 extra_vars_to_write: Union[list, str] = None,
                 remove_missings: bool = False,
                 append: Union[list, str] = None,
                 overwrite: bool = False,
                 check_factor_levels: bool = True,
                 predict_var_names: Union[list, str] = None,
                 residual_var_names: Union[list, str] = None,
                 interval_var_names: Union[list, str] = None,
                 std_errors_var_names: Union[list, str] = None,
                 blocks_per_read: int = 0,
                 report_progress: int = None,
                 verbose: int = 0,
                 xdf_compression_level: int = 0,
                 compute_context: RxComputeContext = None,
                 **kwargs):

        super().__init__(model_object, data, output_data,
                         compute_residuals, predict_var_names,
                         residual_var_names, write_model_vars,
                         extra_vars_to_write, type, remove_missings,
                         append, overwrite, report_progress, verbose,
                         xdf_compression_level, compute_context, **kwargs)

        if model_object.coefficients is None or len(model_object.coefficients) == 0:
            raise ValueError("The model object does not contain any estimated coefficients")

        # compute_standard_errors or se.fit
        compute_standard_errors = RxFunction.assign_or_throw(compute_standard_errors, False, bool)
        if compute_standard_errors is not None and 'se.fit' in kwargs:
            raise ValueError("Must specify only one of 'compute_standard_errors' or 'se.fit', not both.")

        # interval
        interval_list = ["none", "confidence", "prediction"]
        interval = RxFunction.assign_or_throw(interval, False, str)
        if not (interval in interval_list):
            interval = None

        # throw a warning if the user has specified to not compute standard errors but has implied
        # the return of confidenceintervals
        if interval in ["prediction", "confidence"] and compute_standard_errors == False:
            print("Warning: Confidence/tolerance intervals require 'compute_standard_errors = True'.")

        # confidence_level
        if confidence_level is not None and 'level' in kwargs:
            raise ValueError("Must specify only one of 'confidence_level' or 'level', not both.")

        if 'level' in kwargs:
            confidence_level = kwargs['level']

        confidence_level = RxFunction.assign_or_throw(confidence_level, False, float)
        if confidence_level < 0.0 or confidence_level > 1.0:
            raise ValueError("'confidence_level' must be between 0.0 and 1.0")

        # interval_var_names, std_errors_var_names
        var_names = {}
        var_names['interval_var_names'] = interval_var_names
        var_names['std_errors_var_names'] = std_errors_var_names
        for key, value in var_names.items():
            check_list_or_str(key, value, None, "")

        if interval_var_names is not None and isinstance(interval_var_names, list):
            interval_var_names_len = len(interval_var_names)
            if interval_var_names_len < 1 or interval_var_names_len > 2:
                raise ValueError("Length of 'interval_var_names' must be either 1 or 2.'")

        if len(self.model_object.coefficients) == 0:
            raise ValueError("The model object does not contain any estimated coefficients")

        # TODO: checkFactorLevel and coefLabelStyle

        self._compute_standard_errors = compute_standard_errors
        self._interval = interval
        self._confidence_level = confidence_level
        self._interval_var_names = interval_var_names
        self._check_factor_levels = check_factor_levels
        self._std_errors_var_names = std_errors_var_names
        self._blocks_per_read = blocks_per_read
        self._other_arguments = kwargs

        # dependent_var_names
        self._dependent_var_names = self.model_object.y_names
        if len(self._dependent_var_names) > 1:
            raise ValueError("Predictions for multiple dependent variables",
                self._dependent_var_names, "are currently not supported")

        # type
        type_list = ["response", "link"]
        type_error_msg = "'type' must be a character string of either 'response' or 'link'."
        check_list_or_str("type", self.type, type_list, type_error_msg)


    def __prepare_parameters(self):
        # TODO: fill out information for F() that is not there
        # modelObject$params$Formula <- addLowHighToF( modelObject )

        slim_data = None
        if isinstance(self.data, DataFrame):
            # Only send the needed variables to C++
            # Note that 'data' may not have the dependent variable or transformed variables
            base_model_vars = self.model_object.coef_names + self.model_object.y_names

            if self.extra_vars_to_write is not None:
                base_model_vars = base_model_vars + self.extra_vars_to_write

            for column in self.data.columns:
                if not (column in base_model_vars):
                    if slim_data is None:
                        slim_data = self.data.copy(deep = True)
                    slim_data.drop(column, axis = 1, inplace = True)

            #TODO: RE-enable after making work with factor levels
            #if slim_data is not None:
                #self.data = slim_data
        else:
            if not isinstance(self.data, RxXdfData) and self.output_data is None:
                raise ValueError("'output_data' must be specified if the input data is not a data frame or .xdf file")

        # @TODO: rxNetFeedForward??

        # std_errors_var_names
        if self._std_errors_var_names is None and self._compute_standard_errors:
            self._std_errors_var_names = []
            for var_name in self._dependent_var_names:
                self._std_errors_var_names.append(var_name + "_StdErr")

        # residual_var_names
        if self.residual_var_names is None and self.compute_residuals:
            self.residual_var_names = []
            for var_name in self._dependent_var_names:
                self.residual_var_names.append(var_name + "_Resid")

        # interval_var_names
        if self._interval in ["confidence", "prediction"]:
            low_high_str = ["_Lower", "_Upper"]
            if self._interval_var_names is None:
                self._interval_var_names = [self._dependent_var_names[0] + low_high_str[0], self._dependent_var_names[0] + low_high_str[1]]
            elif isinstance(self._interval_var_names, str):
                self._interval_var_names = [self._interval_var_names + low_high_str[0], self._interval_var_names + low_high_str[1]]
            elif isinstance(self._interval_var_names, list) and len(self._interval_var_names) == 1:
                self._interval_var_names = [self._interval_var_names[0] + low_high_str[0], self._interval_var_names[0] + low_high_str[1]]
        else:
            self._interval_var_names = None

    def __assign_call_parameters(self):
        """
            put things together into call_parameters
        """
        call_parameters = self.call_parameters
        call_parameters['PredNames'] = compute_prediction_names(self.model_object,
                                                                self.predict_var_names,
                                                                self.residual_var_names,
                                                                self.compute_residuals)

        call_parameters['DepVarNames'] = self._dependent_var_names[0]
        call_parameters['ResidVarNames'] = self.residual_var_names
        call_parameters['ComputeResiduals'] = self.compute_residuals
        call_parameters['WriteModelVars'] = RxFunction.assign_or_throw(self.write_model_vars, False, bool)
        call_parameters['extraVarsToWrite'] = self.extra_vars_to_write
        call_parameters['RemoveMissings'] = RxFunction.assign_or_throw(self.remove_missings, False, bool)
        call_parameters['blocksPerRead'] = RxFunction.assign_or_throw(self._blocks_per_read, False, int)
        call_parameters['checkFactorLevels'] = RxFunction.assign_or_throw(self._check_factor_levels, False, bool)
        call_parameters['appendRows'] = self.append_eval is not None and "rows" in self.append_eval
        call_parameters['appendCols'] = self.append_eval is not None and "cols" in self.append_eval
        call_parameters['overwrite'] = RxFunction.assign_or_throw(self.overwrite, False, bool)
        call_parameters['Betas'] = self.model_object.coefficients.iloc[:,0].tolist()
        call_parameters['InterceptIncluded'] = self.model_object.coefficients.iloc[0].name == "(Intercept)"
    #    call_parameters['Aliased'] = model_object.aliased
        call_parameters['CoefNames'] = self.model_object.coefficients.index.tolist()
        call_parameters['Cube'] = self.model_object.cube
        call_parameters['IntervalVarNames'] = RxFunction.assign_or_throw(self._interval_var_names, True, list)
        call_parameters['StdErrorsVarNames'] = RxFunction.assign_or_throw(self._std_errors_var_names, True, list)
        call_parameters['ComputeStdErrors'] = RxFunction.assign_or_throw(self._compute_standard_errors, False, bool)
        call_parameters['Interval'] = RxFunction.assign_or_throw(self._interval, False, str)
        call_parameters['ConfLevel'] = RxFunction.assign_or_throw(self._confidence_level, False, float)
        call_parameters['Type'] = self.type

        # model parameters
        call_parameters['ModelType'] = self.model_object.model_type
        call_parameters['Formula'] = self.model_object.formula
        call_parameters['Binary'] = {'':[self.model_object.binary]} if hasattr(self.model_object, 'binary') else None
        
        model_object_parameters = self.model_object._params
        model_parameters_list = ["env", "NewVars", "RowSelection", "transformFunc", "transformVars", "Transforms", "readVars"]
        for param in model_parameters_list:
            call_parameters[param] = model_object_parameters[param] if param in model_object_parameters else None

        call_parameters['OutDataSource'] = self.output_data
        if isinstance(self.output_data, RxDataSource):
            call_parameters['OutDataSource'] = RxFunction.resolve_datasource(self.output_data)
        elif isinstance(self.output_data, DataFrame):
            call_parameters['OutDataSource'] = None

        call_parameters['DataSource'] = None
        call_parameters['..rxDataFrame'] = None
        if isinstance(self.data, RxDataSource):
            call_parameters['DataSource'] = RxFunction.resolve_datasource(self.data)
        elif isinstance(self.data, DataFrame):
            call_parameters['..rxDataFrame'] = resolve_dataframe(self.data)

        call_parameters['CovCoef'] = None
        if self.model_object.covariance_coefficients is not None \
           and isinstance(self.model_object.covariance_coefficients, DataFrame):
            call_parameters['CovCoef'] = self.model_object.covariance_coefficients.as_matrix()

        # report parameters
        call_parameters.update(self.report_params)
        call_parameters['xdfCompressionLevel'] = RxFunction.assign_or_throw(self.xdf_compression_level, False, int)


    def __invoke_native_call(self):
        """
            invoke native call
        """
        #print("Using compute context {0}".format(type(compute_context).__name__))
        if isinstance(self.compute_context, RxRemoteComputeContext):
            job = self.compute_context.create_job()
            rx_import(
                input_data = self.data,
                output_file = RxXdfData(file=job.sample_data_file),
                number_rows = 100,
                verbose = 0,
                report_progress = 0
                )

            if isinstance(self.data, RxSqlServerData):
                job.set_input_data_query(self.data.input_data_query)
                job.set_use_sql_satellite(self.data.use_fast_read)
                self.data.set_in_sql_server(True)

        #print("Calling function {0}".format(function_name))
        #print("Serializing parameters: ")
        #print(self.call_parameters)

        #
        # make the C++ call
        #
        #print(compute_context)

        if isinstance(self.compute_context, RxRemoteComputeContext):
            def process_results(ret):
                self.results = ret if len(ret) == 0 else ret[0]

                return self.results

            self.results = rx_exec(function=rx_native_call, args={'functionname': self.function_name, 'params': self.call_parameters},
                                   compute_context=self.compute_context, callback=process_results)
        else:
            self.results = rx_native_call(functionname = self.function_name, params = self.call_parameters)

    def get_results(self):
        if isinstance(self._compute_context, RxSpark):
            return self._compute_context._execute(self._py_module_name, self._py_function_name, self._py_args)
        self.__prepare_parameters()
        self.__assign_call_parameters()
        self.__invoke_native_call()

        return (super().get_results())


###############################################################################
# rx_predict subclass for DTree and BTrees
###############################################################################
class RxPredictDTree(RxPredict):
    """
        Subclass for rxPredict.rxDTree/BTrees function
    """
    function_name = "DTreePredict"
    call_parameters = {}
    results = {}

    def __init__(self,
                 model_object=None,
                 data: RxDataSource=None,
                 output_data: Union[RxDataSource, str]=None,
                 predict_var_names: Union[list, str]=None,
                 write_model_vars: bool=False,
                 extra_vars_to_write: Union[list, str]=None,
                 append: Union[list, str]=None,
                 overwrite: bool=False,
                 type: Union[list, str]=None,
                 remove_missings: bool=False,
                 compute_residuals: bool=False,
                 residual_type: Union[list, str]=None,
                 residual_var_names: Union[list, str]=None,
                 blocks_per_read: int=0,
                 report_progress: int=None,
                 verbose: int=0,
                 xdf_compression_level: int=0,
                 compute_context: RxComputeContext=None,
                 **kwargs):

        self._mtype = True if isinstance(type, list) and len(type) > 1 else False # missing_type = True if type is a list.

        super().__init__(model_object, data, output_data,
                         compute_residuals, predict_var_names,
                         residual_var_names, write_model_vars,
                         extra_vars_to_write, type, remove_missings,
                         append, overwrite, report_progress, verbose,
                         xdf_compression_level, compute_context, **kwargs)

        # model_object
        if not isinstance(model_object, RxDTreeResults):
            # TODO: check on DForest isinstance(model_object, RxDForestResults):
            raise ValueError("Not legitimate tree or forest")

        # prepareModel:
        #   0 - regular rxPredict call
        #   1 - prepare the model for real-time scoring, return it in
        #   binary format
        #   2 - return the model parameters after they are linearized
        self._prepare_model = 0
        if 'prepare_model' in kwargs \
            and kwargs['prepare_model'] is not None:
            self._prepare_model = kwargs['prepare_model']

        if self._prepare_model > 0 and not isinstance(self.compute_context, RxLocalSeq):
            raise ValueError("rxSerializeModel only supported in sequential local compute context (RxLocalSeq).")

        # type
        type_list = ["vector", "prob", "class", "matrix"]
        type_error_msg = "'type' must be a character string of 'vector', 'prob', 'class', or 'matrix'."
        check_list_or_str("type", self.type, type_list, type_error_msg)

        # residual_type
        resid_type_list = ["usual", "pearson", "deviance"]
        resid_type_error_msg = "'residual_type' must be a character string of 'usual', 'pearson', or 'deviance'."
        check_list_or_str("residual_type", residual_type, resid_type_list, resid_type_error_msg)
        self._residual_type = residual_type

        self._blocks_per_read = blocks_per_read
        self._other_arguments = kwargs

        self._method = None
        self._loss = kwargs['loss_function'] if 'loss_function' in kwargs else None
        self._use_surrogate = None
        self._model_vars = None
        self._ylevels = None
        self._dependent_var_names = None
        self._target_vars = None
        self._input_vars = None
        self._read_vars = None

        # for forest prediction
        self._itree = kwargs['itree'] if 'itree' in kwargs else 1 # the index of the current tree
        self._ntree = kwargs['ntree'] if 'ntree' in kwargs else 1 # the total number of trees in the forest
        self._nforest = kwargs['nforest'] if 'nforest' in kwargs else 1 # the total number of partial forests in the forest
        if 'cutoff' in kwargs:
            self._cutoff = kwargs['cutoff']
        else:
            self._cutoff = self.model_object._results['cutoff'] if 'cutoff' in self.model_object._results else None

        # for boosting
        self._loss_function = kwargs['loss_function'] if 'loss_function' in kwargs else None
        self._learning_rate = kwargs['learning_rate'] if 'learning_rate' in kwargs else 0.1
        self._init_pred = kwargs['init_pred'] if 'init_pred' in kwargs else 0.0
        self._cur_yvar_level = kwargs['cur_yvar_level'] if 'cur_yvar_level' in kwargs else 1

    def __prepare_parameters(self):
        """
            prepare parameters for rxPredict rxDTree/BTrees call
        """
        is_tree = isinstance(self.model_object, RxDTreeResults)
        is_forest = 'forest' in self.model_object._results

        # nclass and ylevels
        nclass = 0
        if is_forest and 'ylevels' in self.model_object._results['forest'][0] \
           and self.model_object._results['forest'][0]['ylevels'] is not None:
            self._ylevels = self.model_object._results['forest'][0]['ylevels']
            nclass = len(self._ylevels)
        elif 'ylevels' in self.model_object._results:
            self._ylevels = self.model_object._results['ylevels']
            nclass = len(self._ylevels)

        if self._mtype and nclass > 0:
            self.type = "prob"

        # fill out information for F() that is not there
        # TODO: modelObject$params$Formula <- addLowHighToF( modelObject )

        # remove rowSelection transformation information from model
        # NOTE: removeRowSelection must run after rxExec, otherwise
        # it is done it twice, once locally and once on remote node.
        # TODO: modelObject <- removeRowSelection(modelObject)

        # data
        #if isinstance(self.data, DataFrame):
        #    # Only send the needed variables to C++
        #    # Note that 'data' may not have the dependent variable or transformed variables
        #    base_model_vars = self.model_object._params['readVars']

        #    if self.extra_vars_to_write is not None:
        #        base_model_vars = base_model_vars + self.extra_vars_to_write

        #    slim_data = None
        #    for column in self.data.columns:
        #        if not (column in base_model_vars):
        #            if slim_data is None:
        #                slim_data = self.data.copy(deep = True)
        #            slim_data.drop(column, axis = 1, inplace = True)

        #    #TODO: RE-enable after resolving factor level issues
        #    #if slim_data is not None:
        #        #self.data = slim_data

        # output_data
        if self._prepare_model > 0:
            self.output_data = None

        # common parameters
        model_object_parameters = self.model_object._params
        formula = self.model_object._formula
        # TODO: temporarily use dependent_var_names and xlevels for TargetVars and InputVars
        #depVarNames <- modelObjectParams[["formulaVars"]][["mapped"]][["depVars"]]
        #TargetVars <- modelObjectParams[["formulaVars"]][["original"]][["depVars"]]
        #InputVars <- modelObjectParams[["formulaVars"]][["original"]][["indVars"]] 
        #ReadVars <- modelObjectParams[["readVars"]]
        self._dependent_var_names = _rx_get_dependent_vars(formula)
        self._input_vars = _rx_get_input_vars(formula)
        self._target_vars = self._dependent_var_names

        if 'readVars' in self.model_object._params \
            and self.model_object._params['readVars'] is not None:
            self._read_vars = self.model_object._params['readVars']

        # residual_var_names
        if self.residual_var_names is None and self.compute_residuals:
            self.residual_var_names = []
            for var_name in self._dependent_var_names:
                self.residual_var_names.append(var_name + "_Resid")

        is_boosting = self._loss_function is not None

        model_object1 = self.model_object
        if is_forest:
            forest = self.model_object._results['forest']
            tree = forest[0]
            model_object1 = RxDTreeResults(self.model_object._formula,
                                             self.model_object._data,
                                             tree,
                                             self.model_object._params)
        self._method = model_object1._results['method']
        self._loss = None
        if self._method == 'class' and model_object1.parms is not None \
           and 'loss' in model_object1.parms:
            self._loss = model_object1.parms['loss']

        self._use_surrogate = model_object1._results['controls']['usesurrogate']

        if is_tree and not is_forest:
            ntree1 = 1
            ntree2 = 1
            nylevel = 1
        else:
            ## ntree2 differs from ntree1 only in case of multinomial boosting.
            ntree1 = self._ntree
            ntree2 = model_object_parameters['nTree']
            is_multi_boosting = self._loss_function == "multinomial" and is_boosting
            nylevel = nclass if is_multi_boosting else 1

        # map model-specific variables
        model_vars = {
            'FrameNumNode': None,
            'FrameNodeInfo': None,
            'FrameNodeId': None,
            'FrameYVal': None,
            'FrameYVal2': None,
            'Splits': None,
            'SplitsDim': None,
            'SplitsVar': None,
            'CSplits': None,
            'CSplitsDim': None}
        self._model_vars = model_vars
        for key in model_vars.keys():
            model_vars[key] = {}
            for i in range(ntree2):
                tree_str = "tree{0}".format(i+1)
                model_vars[key][tree_str] = None

        for itree1 in range(int(ntree1)):
            for curr_yval_level1 in range(nylevel):
                # itree2 differs from itree1 only in case of multinomial boosting.
                # note that itree1 and curr_yval_level1 starting from 0.

                #itree1 = 0
                #nylevel = 1
                #curr_yval_level1 = 0
                #my_model_object = dtree
                #from pandas import Series
                #from numpy import cumsum

                itree2 = itree1 * nylevel + curr_yval_level1
                itree2_str = "tree{0}".format(itree2+1)

                my_model_object = self.model_object._results if (is_tree and not is_forest) else self.model_object._results['forest'][itree2]

                # frame
                frame = my_model_object['frame']
                nc = frame[list(['ncompete', 'nsurrogate'])]
                s = Series(1)
                s = s.append(cumsum(DataFrame((frame['var'] != "<leaf>"))['var'] + nc['ncompete'] + nc['nsurrogate']) + 1)
                s = s.drop(s.index[len(frame)])
                s.index = frame.index
                for i in frame.index:
                    if frame['var'][i] == "<leaf>":
                        s[i] = 0

                frame['index'] = s

                model_vars['FrameNumNode'][itree2_str] = [frame.shape[0]]
                model_vars['FrameNodeInfo'][itree2_str] = [int(i) for i in ((frame[["n", "ncompete", "nsurrogate", "index"]].unstack()).tolist())]
                model_vars['FrameNodeId'][itree2_str] = frame.index.tolist()
                model_vars['FrameYVal'][itree2_str] = frame[["yval"]].unstack().tolist()
                model_vars['FrameYVal2'][itree2_str] = frame.filter(regex =("yval2.*")).unstack().tolist() \
                    if frame.filter(regex =("yval2.*")) is not None and len(frame.filter(regex =("yval2.*")).columns) > 0 else numpy.empty([0,0], dtype=int)

                # splits
                splits = my_model_object['splits']
                model_vars['Splits'][itree2_str] = splits.unstack().tolist() if len(splits.unstack()) > 0 else numpy.empty([0,0], dtype=int)
                model_vars['SplitsDim'][itree2_str] = [splits.shape[0],splits.shape[1]]
                model_vars['SplitsVar'][itree2_str] = [str(s) for s in splits.index.tolist()] if len(splits.index) > 0 else [""]

                # csplit
                csplit = my_model_object['csplit']
                # For default value empty list, need to use numpy for creating empty list.
                # Cannot simply use [] or [0].
                # Otherwise it will break pyLink logic or spark .input object deserialization on worker.
                model_vars['CSplits'][itree2_str] = [int(i) for i in (csplit - 2).unstack().tolist()] if isinstance(csplit, DataFrame) else numpy.empty([0,0], dtype=int)
                model_vars['CSplitsDim'][itree2_str] = [csplit.shape[0],csplit.shape[1]] if isinstance(csplit, DataFrame) else [0, 0]

    def __assign_call_parameters(self):
        """
            put things together into call_parameters
        """
        call_parameters = self.call_parameters

        #call_parameters['modelObject'] = self.model_object
        model_object_parameters = self.model_object._params
        model_parameters_list = ["env", "NewVars", "RowSelection", "transformFunc", "transformVars", "Transforms", "readVars"]
        for param in model_parameters_list:
            call_parameters[param] = model_object_parameters[param] if param in model_object_parameters else None

        call_parameters['OutDataSource'] = self.output_data
        if isinstance(self.output_data, RxDataSource):
            call_parameters['OutDataSource'] = RxFunction.resolve_datasource(self.output_data)
        elif isinstance(self.output_data, DataFrame):
            call_parameters['OutDataSource'] = None

        call_parameters['DataSource'] = None
        call_parameters['..rxDataFrame'] = None
        if isinstance(self.data, RxDataSource):
            call_parameters['DataSource'] = RxFunction.resolve_datasource(self.data)
        elif isinstance(self.data, DataFrame):
            call_parameters['..rxDataFrame'] = resolve_dataframe(self.data)

        call_parameters['DepVarNames'] = self._dependent_var_names[0]
        call_parameters['PredVarNames'] = self.predict_var_names
        call_parameters['ResidType'] = self._residual_type
        call_parameters['ResidVarNames'] = self.residual_var_names
        call_parameters['ComputeResiduals'] = RxFunction.assign_or_throw(self.compute_residuals, False, bool)
        call_parameters['WriteModelVars'] = RxFunction.assign_or_throw(self.write_model_vars, False, bool)
        call_parameters['extraVarsToWrite'] = self.extra_vars_to_write
        call_parameters['RemoveMissings'] = RxFunction.assign_or_throw(self.remove_missings, False, bool)
        call_parameters['Type'] = self.type
        call_parameters['missingData'] = True if self.data is None else False
        call_parameters['blocksPerRead'] = RxFunction.assign_or_throw(self._blocks_per_read, False, int)

        call_parameters['appendRows'] = self.append_eval is not None and "rows" in self.append_eval
        call_parameters['appendCols'] = self.append_eval is not None and "cols" in self.append_eval
        call_parameters['overwrite'] = RxFunction.assign_or_throw(self.overwrite, False, bool)

        call_parameters['Method'] = self._method
        call_parameters['Loss'] = self._loss
        call_parameters['UseSurrogate'] = self._use_surrogate

        # FrameNodeInfo, FrameNumNode, FrameNodeId, FrameYVal, FrameYVal2
        # Splits, SplitsDim, SplitsVar
        # CSplits, CSplitsDim
        for key, value in self._model_vars.items():
            if value is not None:
                call_parameters[key] = value

        call_parameters['TargetVarLevels'] = self._ylevels
        call_parameters['TargetVars'] = self._target_vars
        call_parameters['InputVars'] = self._input_vars
        call_parameters['ReadVars'] = self._read_vars

        call_parameters['itree'] = self._itree
        call_parameters['ntree'] = self._ntree
        call_parameters['nforest'] = self._nforest
        call_parameters['cutoff'] = self._cutoff

        # for boosting prediction
        call_parameters['lossFunction'] = self._loss_function
        call_parameters['learningRate'] = self._learning_rate
        call_parameters['initPred'] = self._init_pred
        call_parameters['curYVarLevel'] = self._cur_yvar_level

        call_parameters['numThreads'] = -1
        call_parameters['fakeThreading'] = False

        # useWriteDataSource
        # if > 0 (TRUE), WriteDataSource will be used for saving intermediate results; if < 0 (FALSE), several data sources will be used; 
        # if ==0, its value will be decided in the C++ code: TRUE for hadoop/teradata, FALSE otherwise.
        call_parameters['useWriteDataSource'] = 0

        # doAutoCleanup
        # delete temporary files at the end? set to false for debugging.
        # NOTE: autoCleanup would be intercepted by the autoCleanup argument of rxExec for distributed computing.
        call_parameters['doAutoCleanup'] = True
        call_parameters['tempFileName'] = path.join(tempfile._get_default_tempdir(), ".rxPredict.rxDTree.xdf")

        # Binary
        # TODO: check self.model_object._results['binary']
        call_parameters['Binary'] = {'': [self.model_object.binary]} if hasattr(self.model_object, 'binary') else None

        call_parameters.update(self.report_params)
        call_parameters['xdfCompressionLevel'] = RxFunction.assign_or_throw(self.xdf_compression_level, False, int)


    def __invoke_native_call(self):
        """
            invoke native call
        """
        if isinstance(self.compute_context, RxRemoteComputeContext):
            job = self.compute_context.create_job()
            rx_import(input_data = self.data,
                                 output_file = RxXdfData(file=job.sample_data_file),
                                 number_rows = 100, verbose = 0, report_progress = False)

            if isinstance(self.data, RxSqlServerData):
                job.set_input_data_query(self.data.input_data_query)
                job.set_use_sql_satellite(self.data.use_fast_read)
                self.data.set_in_sql_server(True)

        # print parameters to be sent to Bxl
        #for key, value in self.call_parameters.items():
        #    print(key)
        #    print(value)

        # make the C++ call
        if isinstance(self.compute_context, RxRemoteComputeContext):
            def process_results(ret):
                self.results = ret if len(ret) == 0 else ret[0]
                return self.results

            self.results = rx_exec(function=rx_native_call,
                                   args={'functionname': self.function_name, 'params': self.call_parameters},
                                   compute_context=self.compute_context, callback=process_results)
        else:
            if self._prepare_model > 0:
                # return self.call_parameters
                self.results = self.call_parameters
            else:
                self.results = rx_native_call(functionname = self.function_name, params = self.call_parameters)
    
    def get_results(self):
        if isinstance(self._compute_context, RxSpark):
            return self._compute_context._execute(self._py_module_name, self._py_function_name, self._py_args)
        self.__prepare_parameters()
        self.__assign_call_parameters()
        self.__invoke_native_call()

        if self.results is not None and isinstance(self.results, list) \
           and len(self.results) == 1:
            self.results = parse_result_dataframe(self.results[0])
        
        if (isinstance(self.results, DataFrame) and self.results.empty) \
           or (isinstance(self.results, dict) and len(self.results.keys()) == 0):
            self.results = None

        if self._prepare_model > 0:
            return self.results

        return (super().get_results())




###############################################################################
# rx_predict subclass for DTree and BTrees
###############################################################################
class RxPredictDForest(RxPredict):
    """
        Subclass for rxPredict.rxDForest function
    """
    function_name = "DTreePredict"
    call_parameters = {}
    results = {}

    def __init__(self,
                 model_object=None,
                 data: RxDataSource=None,
                 output_data: Union[RxDataSource, str]=None,
                 predict_var_names: Union[list, str]=None,
                 write_model_vars: bool=False,
                 extra_vars_to_write: Union[list, str]=None,
                 append: Union[list, str]=None,
                 overwrite: bool=False,
                 type: Union[list, str]=None,
                 cutoff: Union[list, str]=None,
                 remove_missings: bool=False,
                 compute_residuals: bool=False,
                 residual_type: Union[list, str]=None,
                 residual_var_names: Union[list, str]=None,
                 blocks_per_read: int=0,
                 report_progress: int=None,
                 verbose: int=0,
                 xdf_compression_level: int=0,
                 compute_context: RxComputeContext=None,
                 **kwargs):

        super().__init__(model_object, data, output_data,
                         compute_residuals, predict_var_names,
                         residual_var_names, write_model_vars,
                         extra_vars_to_write, type, remove_missings,
                         append, overwrite, report_progress, verbose,
                         xdf_compression_level, compute_context, **kwargs)

        # model_object
        if not isinstance(model_object, RxDTreeResults):
            # TODO: check on DForest isinstance(model_object, RxDForestResults):
            raise ValueError("Not legitimate tree or forest")

        # prepareModel:
        #   0 - regular rxPredict call
        #   1 - prepare the model for real-time scoring, return it in
        #   binary format
        #   2 - return the model parameters after they are linearized
        self._prepare_model = 0
        if 'prepare_model' in kwargs \
            and kwargs['prepare_model'] is not None:
            self._prepare_model = kwargs['prepare_model']

        if self._prepare_model > 0 and not isinstance(self.compute_context, RxLocalSeq):
            raise ValueError("rxSerializeModel only supported in sequential local compute context (RxLocalSeq).")

        # type (prediction type)
        type_list = ["response", "prob", "vote", "freq", "class", "link"]
        type_error_msg = "'type' must be a character string of 'response', 'prob', 'vote', 'freq', 'class', or 'link'."
        check_list_or_str("type", self.type, type_list, type_error_msg)

        # tree type
        self._tree_type = self.model_object._results['type']

        # residual_type
        resid_type_list = ["usual", "prob", "class"]
        resid_type_error_msg = "'residual_type' must be a character string of 'usual', 'prob', or 'class'."
        check_list_or_str("residual_type", residual_type, resid_type_list, resid_type_error_msg)
        self._residual_type = residual_type

        # compute_residuals
        if self._compute_residuals and self._residual_type == "usual" \
           and self.type != "prob" and self._tree_type == "class":
            raise ValueError("residual_type '{0}' is valid only for type 'prob'".format(self._residual_type))

        # cutoff
        self._cutoff = cutoff
        check_list_or_str("cutoff", cutoff, value_list = None, error_str = "")
        # TODO-1: 'cutoff'?
        if self._cutoff is None:
            self._cutoff = self.model_object._results['cutoff'] if 'cutoff' in self.model_object._results else None
        else:
            if len(self._cutoff != nclass) or sum(self._cutoff) > 1 \
               or any(self._cutoff) <= 0:
                raise ValueError("Incorrect 'cutoff' specified.")

        self._append = append
        self._blocks_per_read = blocks_per_read
        self._other_arguments = kwargs

        self._ylevels = None
        self._loss_function = None
        self._learning_rate = None

        self._method = None
        
        self._use_surrogate = None
        self._model_vars = None
        
        self._dependent_var_names = None
        self._target_vars = None
        self._input_vars = None
        self._read_vars = None
        self._itree = None
        self._ntree = None
        self._nforest = None
        self._init_pred = None
        self._verbose = verbose

        if report_progress is None:
            report_progress = RxOptions.get_option("ReportProgress")

        self._report_progress = report_progress

    def __prepare_parameters(self):
        """
            prepare parameters for rxPredict DForest call
        """
        is_forest = 'forest' in self.model_object._results and len(self.model_object._results['forest']) > 0

        # nclass and ylevels
        nclass = 0
        if is_forest and 'ylevels' in self.model_object._results['forest'][0] \
           and self.model_object._results['forest'][0]['ylevels'] is not None:
            self._ylevels = self.model_object._results['forest'][0]['ylevels']
            nclass = len(self._ylevels)

        # dependent_var_names
        formula = self.model_object._formula
        # TODO: temporarily use dependent_var_names from lhs of formula
        #depVarNames <- modelObjectParams[["formulaVars"]][["mapped"]][["depVars"]]
        self._dependent_var_names = _rx_get_dependent_vars(formula)
        self._input_vars = _rx_get_input_vars(formula)

        #-------------
        # for boosting
        #-------------
        self._loss_function = None
        if self.model_object._params is not None \
           and 'lossFunction' in self.model_object._params:
            self._loss_function = self.model_object._params['lossFunction']

        # TODO-1: verify if DForest model has 'learning_rate' attr.
        self._learning_rate = None
        if self.model_object._params is not None \
           and 'learningRate' in self.model_object._params:
            self._learning_rate = self.model_object._params['learningRate']

        is_boosting = self._loss_function is not None
        self._init_pred = self.model_object._results['init.pred'] if 'init.pred' in self.model_object._results else None

        nylevel = 1
        if is_boosting and self._loss_function == "multinomial":
            nylevel = len(self._ylevels)

        #-------------------------------------------------------
        # convert forest prediction type to tree prediction type
        #-------------------------------------------------------
        if self._tree_type == "class":
            forest_tree_type_mapping = {
                'response': 'vector', #vector
                'prob': 'prob',       #probabilities
                'vote': 'class',      #votes
                'freq': 'matrix'     #raw counts of obs, experimental
                }
            if self._type in forest_tree_type_mapping.keys():
                self._type = forest_tree_type_mapping[self._type]
            else:
                self._type = "vector"
        elif self._tree_type == "anova":
            if is_boosting:
                forest_tree_type_mapping = {
                    'response': 'prob', #the predictions are on the scale of the response variable
                    'prob': 'prob',     #probabilities
                    'link': 'matrix'   #the predictions are on the scale of the working response
                    }
                if self._type in forest_tree_type_mapping.keys():
                    self._type = forest_tree_type_mapping[self._type]
                else:
                    self._type = 'prob'
            else:
                if self._type != "response":
                    print("Warning: The only valid type for regression forest is 'response'.")
                self._type = "vector"
        else:
            raise ValueError("Wrong forest prediction type: {0}".format(self._type))

        # ntree2 equals ntree * nylevel for multinomial boosting,
        # 	which is the only case where ntree2 and ntree differ.
        model_object_parameters = self.model_object.parms
        ntree2 = self.model_object._results['ntree']
        self._ntree = ntree2 / nylevel
        forest = self.model_object._results['forest']
        self._nforest = 1
        if 'nforest' in self.model_object._results and self.model_object._results['nforest'] is not None:
            self._nforest = self.model_object._results['nforest']

        # schedule_once argument has been defuncted in R
        schedule_once = 2

    def get_results(self):
        if isinstance(self._compute_context, RxSpark):
            return self._compute_context._execute(self._py_module_name, self._py_function_name, self._py_args)
        self.__prepare_parameters()

        rxpredict = RxPredictDTree(self.model_object,
                                   self.data,
                                   self.output_data,
                                   self.predict_var_names,
                                   self.write_model_vars,
                                   self.extra_vars_to_write,
                                   self._append,
                                   self.overwrite,
                                   self.type,
                                   self.remove_missings,
                                   self.compute_residuals,
                                   self._residual_type,
                                   self.residual_var_names,
                                   self._blocks_per_read,
                                   self._report_progress,
                                   self._verbose,
                                   self.xdf_compression_level,
                                   self.compute_context,
                                   ntree=self._ntree,
                                   nforest=self._nforest,
                                   cutoff=self._cutoff,
                                   loss_function=self._loss_function,
                                   learning_rate=self._learning_rate,
                                   init_pred=self._init_pred,
                                   prepare_model=self._prepare_model)
        return rxpredict.get_results()


