# -*- coding: utf-8 -*-
import numpy as np

from revoscalepy.RxSerializable import rx_native_call
from revoscalepy.utils.RxUtils import _rx_set_verbose_flags
from revoscalepy.datasource import RxNativeFileSystem
from revoscalepy.datasource import RxXdfData


def rx_read_xdf(file: str,
                vars_to_keep: list = None,
                vars_to_drop: list = None,
                row_var_name: str = None,
                start_row: int = 1,
                num_rows: int = None,
                return_data_frame: bool = True,
                strings_as_factors: bool = False,
                max_rows_by_columns: int = None,
                report_progress: int = None,
                read_by_block:bool = False,
                cpp_interp: list = None):
    """
    .. rxtitle::

        Read .xdf File

    .. rxdescription::

        Read data from an ".xdf" file into a data frame.

    :param file: Either an RxXdfData object or a character string specifying 
        the ".xdf" file.
    :param vars_to_keep: List of strings of variable names to include when 
        reading from the input data file. If None, argument is ignored. Cannot be 
        used with vars_to_drop.
    :param vars_to_drop: List of strings of variable names to exclude when 
        reading from the input data file. If None, argument is ignored. Cannot be 
        used with vars_to_keep.
    :param row_var_name: Optional character string specifying the variable in 
        the data file to use as row names for the output data frame.
    :param start_row: Starting row for retrieval.
    :param num_rows: Number of rows of data to retrieve. If -1, all are read.
    :param return_data_frame: Bool indicating whether or not to create a 
        data frame. If False, a list is returned.
    :param strings_as_factors: Bool indicating whether or not to convert 
        strings into factors.
    :param max_rows_by_columns: The maximum size of a data frame that will be 
        read in, measured by the number of rows times the number of columns. If the 
        numer of rows times the number of columns being extracted from the ".xdf" 
        file exceeds this, a warning will be reported and a smaller number of rows 
        will be read in than requested. If max_rows_by_columns is set to be too large, 
        you may experience problems from loading a huge data frame into memory. To 
        extract a subset of rows and/or columns from an ".xdf" file, use rx_data_step.
    :param report_progress: Integer value with options:
        0: No progress is reported.
        1: The number of processed rows is printed and updated.
        2: Rows processed and timings are reported.
        3: Rows processed and all timings are reported.
    :param read_by_block: Read data by blocks. This argument is deprecated.
    :param cpp_interp: List of information sent to C++ interpreter.

    :return: a data frame.

    .. rxseealso::
    
        :py:func:`rx_import <revoscalepy.rx_import>`, 
        :py:func:`rx_data_step <revoscalepy.rx_data_step>`.

    .. rxkeywords:: xdf

    .. rxexample:: 
        :execute:
        :print_output:

        import os
        from revoscalepy import RxOptions, rx_read_xdf

        sample_data_path = RxOptions.get_option("sampleDataDir")
        mort_file = os.path.join(sample_data_path, "mortDefaultSmall.xdf")
        mort_df = rx_read_xdf(mort_file, num_rows = 10)
        print(mort_df)
    """
    if vars_to_keep is not None and vars_to_drop is not None:
        raise ValueError("Parameter 'vars_to_keep' and 'vars_to_drop' both parameters can not be valid.")

    file_system = RxNativeFileSystem()
    xdf_file = file
    if isinstance(xdf_file, RxXdfData):
        xdf_file = file.file

    from revoscalepy.utils.RxUtils import _rx_resolve_var_list
    columns = _rx_resolve_var_list(xdf_file, vars_to_keep, vars_to_drop)

    keep_vars_string = None if columns is None else ','.join(columns)
    call_parameters = dict()
    call_parameters["File"] = xdf_file
    call_parameters["ColNames"] = keep_vars_string
    call_parameters["GetRowNames"] = None
    call_parameters["StartRow"] = start_row
    call_parameters["NumRows"] = num_rows
    call_parameters["RowSelection"] = None  # requires rxCheckCppInterp
    call_parameters["NewVars"] = None  # requires rxCheckCppInterp
    call_parameters["Transforms"] = None  # requires rxCheckCppInterp
    call_parameters["ReturnDataFrame"] = return_data_frame
    call_parameters["ReadByBlock"] = read_by_block
    _rx_set_verbose_flags(call_parameters, verbose = 0, report_progress = report_progress)
    call_parameters["fileSystem"] = file_system.to_serialize

    results = rx_native_call(functionname="ReadXdf", params=call_parameters)

    return results
