from pandas import DataFrame
import revoscalepy
from revoscalepy import rx_predict
from ..RxSerializable import rx_native_call
from revoscalepy.utils import on_exit_handler
from revoscalepy.utils import telemetry_capture_call

@on_exit_handler
def rx_serialize_model(model,
                       realtime_scoring_only:bool) -> bytes:
    """
    .. rxtitle::

        Serialize the python model.

    .. rxdescription::

        Serialize the given python model.

    :param model: A python model supported for real-time scoring. Supported models
        are "rx_logit", "rx_lin_mod", "rx_dtree", "rx_btrees", "rx_dforest" and MicrosoftML models.

    :param realtime_scoring_only: Boolean flag indicating if model serialization is for real-time only. Currently, the
        serialization is supported only for real-time scoring purposes.

    :return: Bytes of the serialized model.

    .. rxkeywords:: serialization

    .. rxexample::
        import os
        from revoscalepy import RxOptions, RxXdfData, rx_serialize_model, rx_lin_mod
        sample_data_path = RxOptions.get_option("sampleDataDir")
        ds = RxXdfData(os.path.join(sample_data_path, "AirlineDemoSmall.xdf"))
        linmod = rx_lin_mod("ArrDelay~DayOfWeek", ds)
        s_linmod = rx_serialize_model(linmod, realtime_scoring_only = True)
    """
    telemetry_capture_call()

    if model is None:
        raise ValueError("model cannot be none.")
    if not realtime_scoring_only:
        raise ValueError("rx_serialize_model only supports serialization for real-time scoring. Set ‘realtime_scoring_only’ to True.")


    # is model is a one of the regressionModels
    if (isinstance(model, revoscalepy.functions.RxLogit.RxLogitResults)
        or isinstance(model, revoscalepy.functions.RxLinMod.RxLinModResults)):

        param_dict = ({'coefNames': model._results['coef.names'],
                       'interceptIncluded': (len(model._results['coef.names']) > 0
                                             and model._results['coef.names'][0] == '(Intercept)'),
                       'aliased': model.aliased[0].values.tolist(),
                       'betas': model.coefficients[0].apply(float).tolist(),
                       'modelType': model.model_type,
                       'factorLevels': dict(model._results['xlevels']),
                       'residualDF': model.df,
                       'formula': model.formula.replace(' ', ''),
                       'runtimeId': '0e472e9c-2635-4afc-8a44-74cfd49695c6',
                       'rmodel': None,
                       'metadata': None,
                       'depVarNames': model._results['y.names']})
    # else if model is rxDTree, rxDForest or rxBTrees
    elif (isinstance(model, revoscalepy.functions.RxDTree.RxDTreeResults)
          or isinstance(model, revoscalepy.functions.RxDTree.RxDForestResults)):
        param_dict = rx_predict(model, data=DataFrame(), prepare_model=1)
        param_dict['runtimeId'] = '0e472e9c-2635-4afc-8a44-74cfd49695c6'
        param_dict['modelType'] = model.model_type
        is_forest = isinstance(model, revoscalepy.functions.RxDTree.RxDForestResults)
        # if model is a rxDForest or rxBTrees
        if is_forest:
            param_dict['factorLevels'] = dict(model.forest[0].x_levels)
        # rxDTress
        else:
            param_dict['factorLevels'] = dict(model.x_levels)
    # else if model is an MML model
    elif model.__module__.startswith('microsoftml'):
        bin_model = model.model_['binaryModel'].tobytes()
        param_dict = ({'runtimeId': '6581DC2C-FECD-4144-B1E7-327AF22CFABF',
                       'modelType': 'raw',
                       'rawModel': {'': [bin_model]}})
    else:
        raise ValueError("Model type not supported by rx_serialize_model.")

    function_name = "SerializeModel"
    s_model = rx_native_call(functionname=function_name, params=param_dict)
    return s_model['blob'][0] if s_model is not None and 'blob' in s_model else None
