from ..RxFunction import RxFunction
from ..RxSerializable import RxParameter
from ..utils.RxOptions import RxOptions
from typing import Union
from collections import OrderedDict
import math
from pandas import DataFrame, Index, Panel
import numpy as np

from revoscalepy.datasource import RxXdfData
from revoscalepy.datasource import RxSqlServerData

from ..RxFunction import RxFunction
from .. RxFunction import RxModelResults
from ..RxSerializable import rx_native_call
from .. RxSerializable import resolve_dataframe, _resolve_transform_function
from ..computecontext.RxInSqlServer import RxRemoteComputeContext
from .. computecontext.RxSpark import RxSpark
from ..functions.RxExec import rx_exec
from ..computecontext.RxComputeContext import RxComputeContext
from ..etl.RxImport import rx_import
from ..utils.RxUtils import rx_print_header, _rx_handle_formula
from revoscalepy.utils.RxUtils import _rx_set_verbose_flags
from ..helpers.Match import get_call_as_string

from revoscalepy.utils import on_exit_handler
from revoscalepy.utils import telemetry_capture_call
from revoscalepy.etl.RxImport import _rx_create_sample_data_file

"""
Make sure for formula, you are using the "C" type, with no tilde before the dependent variables.
ie "X+Y" instead of ~X+Y.
"""
@on_exit_handler
def rx_summary(
        formula: str,  # Formula
        data,  # Input data
        by_group_out_file=None,
        summary_stats: list = None,
        by_term: bool = True,
        pweights=None,
        fweights=None,
        row_selection: str = None,  # Row selection
        transforms=None,
        transform_objects=None,
        transform_function=None,
        transform_variables=None,
        transform_packages=None,
        transform_environment=None,
        overwrite: bool = False,
        use_sparse_cube: bool = None,
        remove_zero_counts: bool = None,
        blocks_per_read: int = None,
        rows_per_block: int = 100000,
        report_progress: int = None,
        verbose: int = 0,
        compute_context=None,
        **kwargs):
    """
    .. rxtitle::

        Object Summaries

    .. rxdescription::

        Produce univariate summaries of objects in revoscalepy.

    :param formula: Statistical model using symbolic formulas. The formula typically
        does not contain a response variable, i.e. it should be of the form ~ terms.
    :param data: either a data source object, a character string specifying a
        ‘.xdf’ file, or a data frame object to summarize.
        If a Spark compute context is being used, this argument may also be an RxHiveData,
        RxOrcData, RxParquetData or RxSparkDataFrame object or a Spark data frame object from pyspark.sql.DataFrame.
    :param by_group_out_file: None, a character string or vector of character
        strings specifying .xdf file names(s), or an RxXdfData object or list of
        RxXdfData objects. If not None, and the formula includes computations by
        factor, the by-group summary results will be written out to one or more
        ‘.xdf’ files. If more than one .xdf file is created and a single character
        string is specified, an integer will be appended to the base by_group_out_file
        name for additional file names. The resulting RxXdfData objects will be
        listed in the categorical component of the output object.
    :param summary_stats: A list of strings containing one or more of the
        following values: "Mean", "StdDev", "Min", "Max", "ValidObs", "MissingObs",
        "Sum".
    :param by_term: bool variable. If True, missings will be removed by term
        (by variable or by interaction expression) before computing summary
        statistics. If False, observations with missings in any term will be
        removed before computations.
    :param pweights: Character string specifying the variable to use as
        probability weights for the observations.
    :param fweights: Character string specifying the variable to use as
        frequency weights for the observations.
    :param row_selection: None. Not currently supported, reserved for future use.
    :param transforms: None. Not currently supported, reserved for future use.
    :param transform_objects: None. Not currently supported, reserved for
        future use.
    :param transform_function: Variable transformation function.
    :param transform_variables: List of strings of input data set variables
        needed for the transformation function.
    :param transform_packages: None. Not currently supported, reserved for
        future use.
    :param transform_environment: None. Not currently supported, reserved for
        future use.
    :param overwrite: Bool value. If True, an existing byGroupOutFile will
        be overwritten. overwrite is ignored byGroupOutFile is None.
    :param use_sparse_cube: Bool value. If True, sparse cube is used.
    :param remove_zero_counts: Bool flag. If True, rows with no observations
        will be removed from the output for counts of categorical data. By default,
        it has the same value as useSparseCube. For large summary computation, this
        should be set to True, otherwise the Python interpreter may run out of
        memory even if the internal C++ computation succeeds.
    :param blocks_per_read: Number of blocks to read for each chunk of data
        read from the data source.
    :param rows_per_block: Maximum number of rows to write to each block in the
        by_group_out_file (if it is not None).
    :param report_progress: Integer value with options:
        0: No progress is reported.
        1: The number of processed rows is printed and updated.
        2: Rows processed and timings are reported.
        3: Rows processed and all timings are reported.
    :param verbose: Integer value. If 0, no additional output is printed. If 1,
        additional summary information is printed.
    :param compute_context: A valid RxComputeContext object.
    :param kwargs: Additional arguments to be passed directly to the Revolution
        Compute Engine.

    :return: An RxSummary object containing the following elements:
        nobs.valid: Number of valid observations.
        nobs.missing: Number of missing observations.
        sDataFrame: Data frame containing summaries for continuous variables.
        categorical: List of summaries for categorical variables.
        categorical.type: Types of categorical summaries: can be "counts", or "cube" (for crosstab counts) or "none" (if there is no categorical summaries).
        formula: Formula used to obtain the summary.

    .. rxkeywords:: summary

    .. rxexample::
        import os
        from revoscalepy import rx_summary, RxOptions, RxXdfData
        sample_data_path = RxOptions.get_option("sampleDataDir")
        ds = RxXdfData(os.path.join(sample_data_path, "AirlineDemoSmall.xdf"))
        summary = rx_summary("ArrDelay+DayOfWeek", ds)
        print(summary)
    """
    telemetry_capture_call(compute_context)

    data = RxComputeContext.current().validate_in_data(data, "data")

    if use_sparse_cube is None:
        use_sparse_cube = RxOptions.get_option("UseSparseCube")

    if remove_zero_counts is None:
        remove_zero_counts = use_sparse_cube

    if blocks_per_read is None:
        blocks_per_read = RxOptions.get_option("BlocksPerRead")

    if summary_stats is None:
        summary_stats = ["Mean", "StdDev", "Min", "Max", "ValidObs", "MissingObs"]

    if blocks_per_read is None:
        blocks_per_read = -1

    if transform_function is not None:
        transform_function = _resolve_transform_function(transform_function=transform_function)

    py_args = locals()
    for key in kwargs:
        py_args[key] = kwargs[key]
    py_args.pop("kwargs", None)
    py_function_name = "rx_summary"
    py_module_name = __name__

    if compute_context is None:
        compute_context = RxComputeContext.current()

    # todo need to refactor the spark workflow with Keith
    if isinstance(compute_context, RxSpark):
        return compute_context._execute(py_module_name, py_function_name, py_args)

    function_name = "Summary"
    call_parameters = {"Cube": False}
    results = {}

    call_parameters['Formula'] = _rx_handle_formula(RxFunction.assign_or_throw(formula, False, str), pweights, fweights)
    call_parameters['byTerm'] = RxFunction.assign_or_throw(by_term, False, bool)
    call_parameters['useSparseCube'] = RxFunction.assign_or_throw(use_sparse_cube, False, bool)
    call_parameters['removeZeroCounts'] = RxFunction.assign_or_throw(remove_zero_counts, False, bool)
    call_parameters['rowSelection'] = RxFunction.assign_or_throw(row_selection, True, str)
    call_parameters['transforms'] = RxFunction.assign_or_throw(transforms, True)
    call_parameters['transformObjects'] = RxFunction.assign_or_throw(transform_objects, True)
    call_parameters['transformFunc'] = RxFunction.assign_or_throw(transform_function, True)

    call_parameters['transformVars'] = RxFunction.assign_or_throw(transform_variables, True)
    call_parameters['transformEnvir'] = RxFunction.assign_or_throw(transform_environment, True)
    call_parameters['transformPackages'] = RxFunction.assign_or_throw(transform_packages, True)
    call_parameters['blocksPerRead'] = RxFunction.assign_or_throw(blocks_per_read, False, int)
    call_parameters['overwrite'] = RxFunction.assign_or_throw(overwrite, False, bool)
    call_parameters['rowsPerBlock'] = RxFunction.assign_or_throw(rows_per_block, False, int)
    call_parameters['byGroupOutFile'] = RxFunction.assign_or_throw(by_group_out_file, True)
    call_parameters['summaryStats'] = RxFunction.assign_or_throw(summary_stats, False, list )

    # Get all verbose parameters
    _rx_set_verbose_flags(call_parameters, verbose = verbose, report_progress = report_progress)

    if isinstance(data, DataFrame):
        call_parameters['..rxDataFrame'] = resolve_dataframe(data)

        faddcp_param = OrderedDict()
        faddcp_param['bComputeLowHigh'] = True
        call_parameters['FaddCP'] = faddcp_param

        call_parameters['readVars'] = data.columns.tolist()
    else:
        call_parameters['DataSource'] = RxFunction.resolve_datasource(data)

    #
    # capture sample data if its a remote compute context call
    #
    #print("Using compute context {0}".format(type(compute_context).__name__))

    #
    # TODO: This should be handled by formula parser
    #
    fixed_up_formula = call_parameters['Formula'].strip().lstrip('~')

    if fixed_up_formula == "":
        fixed_up_formula = "."

    if ".rxALL" not in fixed_up_formula:
        fixed_up_formula = fixed_up_formula.replace(".", ".rxALL")

    call_parameters['Formula'] = fixed_up_formula

    if isinstance(compute_context, RxRemoteComputeContext):
        job = compute_context.create_job()
        _rx_create_sample_data_file(
            input_data=data,
            sample_data_file=job.sample_data_file
            )
        if isinstance(data, RxSqlServerData):
            job.set_input_data_query(data.input_data_query)
            job.set_use_sql_satellite(data.use_fast_read)
            data.set_in_sql_server(True)

    #print("Calling function {0}".format(function_name))
    #print("Serializing parameters: ")

    #
    # make the C++ call
    #
    #print(compute_context)

    call = get_call_as_string()

    if isinstance(compute_context, RxRemoteComputeContext):
        def process_results(ret):
            if len(ret) > 0:
                ret = ret[0]

            return RxSummaryResults(formula, data, ret, call_parameters, compute_context, call)

        return rx_exec(function=rx_native_call, args={'functionname': function_name, 'params': call_parameters},
                          compute_context=compute_context, callback=process_results)
    else:
        results = rx_native_call(functionname=function_name, params=call_parameters)

    return RxSummaryResults(formula, data, results, call_parameters, compute_context, call)




class RxSummaryResults(RxModelResults):
    def __init__(self, formula, data, results, params, compute_context, call):
        super().__init__(model_type = 'rxSummary', formula = formula, params = params, data = data, compute_context = compute_context, call = call)
        self._results = results

        if isinstance(data, DataFrame):
            # drop all rows and just retain column info which
            # is needed in make_index() for categorical cols
            self._data_column_info = data.drop(data.index) 

        self._sDataFrame = None
        self._categorical = None

    @property
    def y_names(self) -> list:
        if 'y.names' in self._results:
            return [self._results['y.names']] if isinstance(self._results['y.names'], str) else self._results['y.names']



    @property
    def categorical(self) -> list:
        ##Apparently this property can represent two different things depending on formula.
        if self._categorical is not None:
            return self._categorical
        elif 'categorical' in self._results:
            categorical_list = self._results['categorical']
            if 'categorical.name' in self._results:
                name_list = self._results['categorical.name']
                if len(name_list) == len(categorical_list) and len(name_list) > 0 and len(name_list[0]) is not 0:
                    self._categorical = [self.make_categorical_name(x,y) for (x,y) in zip(categorical_list,name_list)]
                    return self._categorical
            self._categorical = [self.categorical_counts(x) for x in categorical_list]
            return self._categorical

    def categorical_counts(self, x):
        categorical = x[0]
        counts = categorical.pop('Counts') if 'Counts' in categorical else 0
        if len(categorical.keys()) is 1:
            return self.make_one_dimensional(categorical, counts)
        else:
            return self.make_higher_dimensional(categorical, counts)

    def make_categorical_name(self, x, y):
        if len(y) is 0:
            return self.categorical_counts(x)
        return DataFrame(x[0], index=y)

    def make_one_dimensional(self, categorical, counts):
        (values,name) = self.make_index(name = list(categorical.keys())[0],values = list(categorical.values())[0] )
        index = Index(values, name=name)
        return DataFrame({"Counts": counts}, index=index)

    def make_index(self, name, values):
        try:
            #Special case to include cat name for dataframes
            #For other data sources, we need to fix the bug in BXLServer
            cat_list = self._data_column_info[name].cat.categories
            values = [cat_list[x-1] for x in values]
        finally:
            return (values, name)

    @property
    def valid_observations(self) -> float:
        if 'nobs.valid' in self._results:
            _valid_observations = self._results['nobs.valid']
            return float(_valid_observations) if not np.isnan(_valid_observations) else None

    @property
    def missing_observations(self) -> float:
        if 'nobs.missing' in self._results:
            result = self._results['nobs.missing']
            return result if not math.isnan(result) else None


    @property
    def summary_data_frame(self) -> DataFrame:
        if self._sDataFrame is not None:
            return self._sDataFrame
        elif 'sDataFrame' in self._results:
            dictionary = self._results['sDataFrame']['sDataFrame'] #It is in this format because it's a named list in R.
            names = dictionary['Name']
            self._sDataFrame = DataFrame(dictionary) if isinstance(names, list) else DataFrame([dictionary])
            return self._sDataFrame

    @property
    def cube(self) -> bool:
        return self._params['Cube']

    def __str__(self):

        header = rx_print_header(self.cube,
                                 self.formula,
                                 self.valid_observations,
                                 self.missing_observations,
                                 "Summary Statistics",
                                 data_source=self._data,
                                 call=self._call
                                 )

        return header +"\n"+ self.__remove_categorical().__str__() + "\n" + self.__categorical_string()

    def __categorical_string(self):
        return "\n".join([self.__format_one_categorical(categorical) for categorical in self.categorical])


    def __format_one_categorical(self, categorical : DataFrame) -> str:
        if 'Counts' in categorical.columns:
            str = "\nCategory Counts for {0}".format(categorical.index.name)
            str += "\nNumber of categories: {0}".format(len(categorical))
            try:
                str += "\nNumber of valid observations: {0}".format(self.summary_data_frame["ValidObs"][categorical.index.name])
            except KeyError:
                pass
            try:
                str += "\nNumber of missing observations: {0}".format(
                    self.summary_data_frame["MissingObs"][categorical.index.name])
            except KeyError:
                pass
            return str + "\n\n" + categorical.__str__()
        else:
            return "\nStatistics by category ({0} categories):\n\n".format(len(categorical)) + categorical.__str__()


    def __remove_categorical(self) -> dict:
        df = self.summary_data_frame
        if 'Mean' in df.columns:
            return df[~np.isnan(df['Mean'])]
        elif len(df.columns) > 0:
            return df[~np.isnan(df[df.columns[1]])] # columns[0] is 'Name' column, so start from [1]
        return df

    def make_higher_dimensional(self, categorical, counts):
        index = [self.make_index(name = i,values= list(set(categorical[i]))) for i in
                 sorted(categorical.keys())]
        index = [Index(data,name = name) for (data,name) in index]
        dim = len(index)
        if dim is 2:
            positions = [categorical[i] for i in sorted(categorical.keys())]
            a = np.empty((len(index[0]),len(index[1])))
            for (x,y,val) in zip(positions[0],positions[1],counts):
                a[x-1][y-1] = val
            return DataFrame(data=a, index=index[0],columns=index[1])
        if isinstance(categorical, OrderedDict):
            return DataFrame(categorical)
        if dim is 3:
            raise NotImplementedError("Do not support more than 1 interaction")
            #return {"Counts":Panel(data=np.reshape(counts,(len(index[0]),len(index[1]),len(index[2]))),items=index[0],
            #            major_axis=index[1],minor_axis=index[2])} ##Not Tested
        if dim > 3:
            raise NotImplementedError("Do not support more than 3 dimensions of counts")
