import json
import os
import platform
import sys

from typing import Union
import uuid

from .. import RxSerializable
from ..utils.RxOptions import RxOptions
from ..computecontext.RxComputeContext import RxComputeContext

# Module globals
__telemetry_state = None


def get_opt_in_file_path():
    from ..utils.RxFileUtils import get_exec_dir
    from ..utils.RxFileUtils import get_r_install_dir
    install_dir = get_r_install_dir()
    return install_dir if install_dir is not None else get_exec_dir()


class CallInfo:
    def __init__(self, operation_name, compute_context):
        self._operation_name = operation_name

        self._compute_context = "Unknown"
        if isinstance(compute_context, str):
            self._compute_context = compute_context
        elif isinstance(compute_context, RxComputeContext):
            self._compute_context = compute_context.description
        elif compute_context is None and RxComputeContext.current() is None:
            self._compute_context = RxComputeContext.current().description

    @property
    def operation_name(self):
        return self._operation_name

    @property
    def compute_context(self):
        return self._compute_context


class TelemetryState:
    def __init__(self):
        self._opt_in = False
        self._pause = False
        self._call_info = None
        self._telemetry_id = None
        self._telemetry_id_file = None

        self._load_opt_in()
        if self._opt_in:
            self._load_id()

    def _load_opt_in(self):
        # Determines the opt-in state at start-up
        self._opt_in = False
        optin_file = os.path.join(get_opt_in_file_path(), '.optIn')
        if os.path.exists(optin_file):
            self._opt_in = True

        if 'MS_PY_TELEMETRY' in os.environ and os.environ['MS_PY_TELEMETRY'] == 1:
            print("Detected MS_PY_TELEMETRY environment variable!")
            self._opt_in = True

    def _create_id(self, id_file: str) -> Union[str, None]:
        __telemetry_version = "2.0"
        telemetry_id = self._generate_id()

        to_serialize = {'version': __telemetry_version, 'id': telemetry_id}

        # noinspection PyBroadException
        try:
            with open(id_file, 'w') as out:
                json.dump(to_serialize, out)
                return telemetry_id
        except:
            return None

    @staticmethod
    def _generate_id():
        return str(uuid.uuid4()).replace('-', '').upper()

    def _load_id(self):
        from ..utils.RxFileUtils import get_local_config_path
        if get_local_config_path() is None:
            return None

        telemetry_id_file = os.path.join(get_local_config_path(), "ml-msid.json")
        telemetry_id = None

        if not os.path.exists(telemetry_id_file):
            # File not created, create the telemetry id
            telemetry_id = self._create_id(telemetry_id_file)
        else:
            try:
                # File does exist, attempt to load the id from the file
                with open(telemetry_id_file) as idfile:
                    file_contents = json.load(idfile)

                if file_contents is not None and 'id' in file_contents:
                    self._telemetry_id_file = telemetry_id_file
                    telemetry_id = file_contents['id']
                else:
                    telemetry_id = self._generate_id()
            except:
                # json file failed to load, therefore generate an id to use
                telemetry_id = self._generate_id()

        self._telemetry_id_file = telemetry_id_file
        self._telemetry_id = telemetry_id

    def pause(self):
        self._pause = True

    def unpause(self):
        self._pause = False

    def clear_call_info(self):
        self._call_info = None

    @property
    def telemetry_id(self):
        return self._telemetry_id

    @property
    def telemetry_id_file(self):
        return self._telemetry_id_file

    @property
    def active(self):
        return self._opt_in and not self._pause

    @property
    def opt_in(self):
        return self._opt_in

    @property
    def call_info(self):
        return self._call_info

    @call_info.setter
    def call_info(self, value: CallInfo):
        if self._call_info is None:
            self._call_info = value

def __get_hadoop_info():
    if os.path.exists("/usr/hdp"):
        return "Hadoop;Hortonworks"
    if os.path.exists("/opt/mapr"):
        return "Hadoop;Mapr"
    if os.path.exists("/opt/cloudera"):
        return "Hadoop;Cloudera"
    return ""

def __telemetry_version_event():
    from ..utils.RxFileUtils import get_python_version

    global __telemetry_state
    if not __telemetry_state or not __telemetry_state.active:
        return

    params = {
               "runtime": "python",
               "runtimeVersion": get_python_version(),
               "revoVersion": "9.2.1.0",
               "installedPackages": list(sys.modules.keys()),
               "osVersion": platform.platform(),
               "clusterInfo": __get_hadoop_info(),
               "sessionInternal": False,
               "telemetryId": __telemetry_state.telemetry_id
             }

    RxSerializable.rx_native_call("TelemetryVersionEvent", params)


def __reset_telemetry():
    global __telemetry_state
    __telemetry_state = TelemetryState()


def __update_opt_in_file_state(enable: bool):
    opt_in_file = os.path.join(get_opt_in_file_path(), ".optIn")
    file_exists = os.path.exists(opt_in_file)

    if enable and file_exists:
        return
    if not enable and not file_exists:
        return

    if enable:
        open(opt_in_file, 'a').close()
    else:
        os.remove(opt_in_file)


def __telemetry_handle_call_exit():
    global __telemetry_state

    # Do not log the telemetry call on the master node
    # Therefore check if the nodeType is set to master in the RxOptions
    nodeType = RxOptions.get_option("nodeType")
    if nodeType is "master":
        __telemetry_state.clear_call_info()
        return

    if not __telemetry_state.active:
        # clear the call info
        __telemetry_state.clear_call_info()
        return

    __telemetry_call_event(__telemetry_state.call_info)

    # clear the call info
    __telemetry_state.clear_call_info()


def __telemetry_call_event(call_info:CallInfo):
    if call_info is None:
        return

    params = {
        "telemetryCallInfo": {
            "operationName":  call_info.operation_name,
            "computeContext": call_info.compute_context
        }
    }

    RxSerializable.rx_native_call("TelemetryCallEvent", params)

# Captures a function call information for telemetry collection
def telemetry_capture_call(compute_context=None, name=None):
    import inspect
    from revoscalepy.utils.RxFunctionUtils import ExitHandler
    from revoscalepy.utils.RxFunctionUtils import __on_exit_handler

    global __telemetry_state

    if __telemetry_state.call_info is None:
        frame = inspect.currentframe().f_back
        operation_name = name
        if operation_name is None:
            # If the name is not set, then walk the stack to retrieve it
            frame_info = inspect.getframeinfo(frame)

            operation_name = frame_info.function
        __telemetry_state.call_info = CallInfo(operation_name, compute_context)

        # Go back one more frame to get the exit handler
        __on_exit_handler(ExitHandler(__telemetry_handle_call_exit), frame.f_back)
        return True
    return False


def telemetry_pause():
    __telemetry_state.pause()


def telemetry_unpause():
    __telemetry_state.unpause()



def rx_privacy_control(enable: bool = None):
    """
    .. rxtitle::

        Changes the opt in state for anonymous usage collection

    .. rxdescription::

        rx_privacy_control(opt_in, bool=None)

    :param opt_in: A bool value that specifies to opt in True or opt out False with anonymous data usage collection.
        If not specified, the value is returned.

    .. rxkeywords:: privacy, telemetry

    .. rxexample::
        import revoscalepy
        revoscalepy.rx_privacy_control(False)
    """
    if enable is None:
        return __telemetry_state.opt_in

    __update_opt_in_file_state(enable)
    __reset_telemetry()

    global __telemetry_state

    if __telemetry_state.opt_in:
        print("You've opted to permit Microsoft to gather anonymous user data.")
    else:
        print("You've opted ouf of anonymous data collection. No user data will be collected.")
    return __telemetry_state.opt_in


def __init():
    global __telemetry_state
    __telemetry_state = TelemetryState()
    __telemetry_version_event()
