import textwrap
import re
import uuid

from typing import List, Union
from os import path
from pandas import DataFrame
from patsy import ModelDesc, PatsyError, Term

from revoscalepy.datasource.RxFileData import RxFileData
from revoscalepy import RxNativeFileSystem
from revoscalepy import RxDataSource, RxXdfData, RxTextData, RxHiveData

from revoscalepy.utils.RxOptions import RxOptions
from revoscalepy.datasource.RxSparkData import RxSparkData
from revoscalepy.datasource.constants import PropertyConstants as pc


CONSOLE_WIDTH = 120


def rx_formula_string_deblank(x: str):
    # Strips white space from beginning and end,
    # removes all returns, and replaces multiple
    # spaces in the body with a single space, e.g.,
    # > rxFormulaStringDeblank(" my string  \n \t \r  has spaces removed from \r \nthe ends    ")
    # [1] "my string has spaces removed from the ends"
    return ' '.join(x.split())


def data_name_string(dataName, dataSource, fileName, textWrapper):
    isNonXdfDataSource = False
    dataNameStr = ""
    if dataSource is not None:
        try:
            fileName = dataSource.file
        except AttributeError:
            pass
        if dataName is not None and not (dataName == ""):
            dataNameStr += "Data: {0} ({1} Data Source)\n".format(dataName, dataSource.class_name())
            dataName = None
            isNonXdfDataSource = True
    if dataName is not None:
        dataNameStr += "Data: {0}\n".format(dataName)
    if fileName is not None:
        file_name_str = textWrapper.fill("File name: {0}".format(fileName))
        if isNonXdfDataSource:
            dataNameStr += file_name_str
        else:
            dataNameStr = file_name_str
    return dataNameStr


def analysis_string(cube, formula, textWrapper, type):
    analysis_str = None
    if formula is not None and type is not None:
        analysis_str = "Cube " if cube else ""
        analysis_str += type
        analysis_str += " Results for: "
        analysis_str += formula
        analysis_str = textWrapper.fill(analysis_str)
    return analysis_str


def weight_string(formulaString):
    strDict = {"p": "Probability", "f": "Frequency"}
    pat = re.compile("(p|f)weight=(\s*)")
    match = pat.search(formulaString)
    if (match):
        return "{0} weights: {1}".format(strDict[match.group(1)], match.group(2))
    return None


def independent_variable_string(number_dropped, number_of_independent_variables):
    ind_var_str = None
    if number_of_independent_variables is not None and number_dropped is not None:
        ind_var_str = "Total independent variables: {0}".format(number_of_independent_variables)
        if number_dropped > 0:
            ind_var_str += " (Including number dropped: {0})".format(number_dropped)
    return ind_var_str


def make_list(x):
    return [x]


def rx_print_header(cube: bool = False, formula: str = None, n_valid_obs: Union[List[Union[int,float]], int, float] = None,
                    n_missing_obs:  Union[List[Union[int,float]], int, float] = None,
                    rx_operation_type: str = None, number_of_independent_variables: int = None,
                    number_dropped: int = None, data_name: str = None, file_name: str = None,
                    data_source: RxFileData = None,
                    formula_string: str = None, call: str = None, dependent_vars=None) -> str:
    textWrapper = textwrap.TextWrapper(CONSOLE_WIDTH, subsequent_indent=4 * ' ', break_long_words=False)
    callStr = "Call:\n{0}\n".format(textWrapper.fill(call)) if call is not None else None
    analysisStr = analysis_string(cube, formula, textWrapper, rx_operation_type)
    dataNameStr = data_name_string(data_name, data_source, file_name, textWrapper)
    weightsStr = weight_string(formula_string) if formula_string is not None else None
    depvarStr = "Dependent variable(s): {0}".format(dependent_vars) if dependent_vars is not None and len(dependent_vars) != 0 else None

    indVarStr = independent_variable_string(number_dropped, number_of_independent_variables)

    n_valid_obs = make_list(n_valid_obs) if n_valid_obs is not None and not hasattr(n_valid_obs, "__iter__") else n_valid_obs
    n_missing_obs = make_list(n_missing_obs) if n_missing_obs is not None and not hasattr(n_missing_obs, "__iter__") else n_missing_obs

    validobsStr = "Number of valid observations: {0}".format(
        ", ".join([str(x) for x in n_valid_obs])) if n_valid_obs is not None else None
    missingobsStr = "Number of missing observations: {0}".format(
        ", ".join([str(x) for x in n_missing_obs])) if n_missing_obs is not None else None

    return "\n".join(filter(None.__ne__, [callStr, analysisStr, dataNameStr,
                                          weightsStr, depvarStr, indVarStr, validobsStr, missingobsStr])) + "\n"


def rx_get_in_data_source(data, file_system = None):
    in_source = data
    if (data is None or isinstance(data, DataFrame)):
        return in_source

    # if data is a character string, it must be a file path
    if (isinstance(data, str)):
        if (len(data) <= 0):
            return None
        else:
            file_system = RxNativeFileSystem()
            filename, file_extension = path.splitext(data)

            # if there is no file_extension, assume it's a .xdf
            if (file_extension is None or len(file_extension) <= 0):
                file_extension = '.xdf'

            if (file_extension == '.xdf'):
                in_source = RxXdfData(data, file_system = file_system)
            elif (file_extension == '.csv' or file_extension == '.txt'):
                in_source = RxTextData(data)
            elif (file_extension == '.sav'):
                raise ValueError('.sav files have not been supported.')
                #TODO: in_source = RxSpssData(data)
            elif (file_extension in ['.sas7bdat', '.sd7']):
                raise ValueError('.sas7bdat or .sd7 files have not been supported.')
                #TODO: in_source = RxSasData(data)
            elif (file_extension in ['.sqlite', '.sql', '.dbf']):
                raise ValueError('.sqlite, .sql or .dbf files have not been supported.')
                #TODO: in_source = RxOdbcData(data, useFastRead = True)
            else:
                raise ValueError('Unrecognized file extension. Please specify an input data source.')

    # in_source now must be one of the supported data source objects
    if not isinstance(in_source, RxDataSource):
        raise ValueError('Unrecognized data source. Please specify a valid input data source.')

    return (in_source)


def rx_get_out_data_source(data):
    if data is None or isinstance(data, DataFrame):
        return data

    if not isinstance(data, (str, RxDataSource)):
        raise ValueError('Unrecognized data source. Please specify a valid output data source.')

    # if data is a character string, it must be a file path
    if isinstance(data, str) and len(data) <= 0:
        return None

    if isinstance(data, RxTextData):
        if data.to_serialize[pc.delimiter] is None:
            data.to_serialize[pc.delimiter] = ","

        if data.to_serialize[pc.use_fast_read] is None:
            data.to_serialize[pc.use_fast_read] = True
        elif data.to_serialize[pc.use_fast_read] is False:
            raise ValueError("'use_fast_read' must be TRUE in order to write to a text data source.")

        if data.to_serialize[pc.is_fixed_format] is None:
            data.to_serialize[pc.is_fixed_format] = False
        elif data.to_serialize[pc.is_fixed_format] is True:
            raise ValueError("Cannot write to a fixed-format text data source.")

        return data

    out_source = data
    if isinstance(data, str):
        file_system = RxOptions.get_option("FileSystem")

        _, ext = path.splitext(data.lower())
        if ext in (".xdf", ""):
            out_source = RxXdfData(data, file_system=file_system)
        elif ext in (".txt", ".csv"):
            out_source = RxTextData(file=data, use_fast_read=True, is_fixed_format=False,
                                    delimiter=",", file_system=file_system)
        else:
            raise ValueError('Unrecognized file extension for output data source.')

    return out_source


def _rx_set_verbose_flags(call_params : dict, verbose : int, report_progress : int) -> dict:
    print = False
    show_iterations = False
    show_detailed_iterations = False
    show_progress = False
    report_timing = 0

    verbose = 0 if verbose is None else verbose
    report_progress = RxOptions.get_option("ReportProgress") if report_progress is None else report_progress

    if not (isinstance(report_progress, int)):
        raise ValueError("'report_progress' must be an integer value between 0 and 3.")

    if (report_progress == 0):
        show_progress = False
    elif (report_progress == 1):
        show_progress = True
    elif (report_progress == 2):
        show_progress = True
        report_timing = 1
    elif (report_progress == 3):
        show_progress = True
        report_timing = 2
    elif (report_progress == 4):
        # For debugging purposes - not documented
        show_progress = True
        report_timing = 2
        verbose = True
    else:
        raise ValueError("'report_progress' must be an integer value between 0 and 3.")

    # verbose
    if (verbose is not None):
        if verbose >= 1:
            print = True
            show_iterations = True
            if verbose >= 2:
                show_detailed_iterations = True

    call_params['Verbose'] = verbose;
    call_params['Print'] = print
    call_params['ReportProgress'] = show_progress
    call_params['ReportTiming'] = report_timing
    call_params['ShowIterations'] = show_iterations
    call_params['ShowDetailedIterations'] = show_detailed_iterations

    return call_params


def _rx_resolve_var_list(data, vars_to_keep:list=None, vars_to_drop:list=None):
    import revoscalepy.RxFunction as arg
    arg.validate_argument(data, 'data', (str, DataFrame, RxXdfData, RxTextData, RxSparkData), False)
    if vars_to_keep is None and vars_to_drop is None:
        return list()
    if vars_to_keep is not None and vars_to_drop is not None:
        raise ValueError("Parameter 'vars_to_keep' and 'vars_to_drop' both parameters cannot be valid")
    if isinstance(data, str) and not data.endswith('.xdf'):
        raise ValueError("Parameter 'data' must reference an xdf file")

    from revoscalepy.functions.RxGetInfoXdf import rx_get_var_names

    #Get the columns
    columns = None
    if isinstance(data, DataFrame):
        columns = data.columns.values.tolist()
    elif isinstance(data, (RxXdfData, RxTextData, RxSparkData)):
        columns = rx_get_var_names(data)
    elif isinstance(data, str):
        columns = rx_get_var_names(data)
    else:
        import warnings
        warnings.warn("'vars_to_keep' and 'vars_to_drop' are not supported")
        return list()

    if vars_to_drop is not None:
      for x in vars_to_drop:
        if x in columns:
            columns.remove(x)
    else:
        items_to_remove = list()
        for x in columns:
            if x not in vars_to_keep:
                items_to_remove.append(x)

        for item in items_to_remove:
            columns.remove(item)

    if columns is not None:
        return columns

    return list()


def _rx_handle_formula(formula: str = None, pweights: str = None, fweights: str = None) -> str:
    # TODO: This logic should be extracted to formula parser when we have it
    # TODO: Also, may need the logic similar to rxLinkFormulaComponents() in rxLink.R
    #       to pre-process the formula.
    if formula is None or formula == "":
        return formula

    if '*' in formula:
        from patsy import ModelDesc
        desc = ModelDesc.from_formula(formula)
        formula = desc.describe()

    if isinstance(pweights, str):
        formula += (",pweight=" + pweights)

    if isinstance(fweights, str):
        formula += (",fweight=" + fweights)

    return formula

def extract_name(iterm : Term = None) -> str :
    return iterm.name() if isinstance(iterm,Term) else ''

def _rx_get_dependent_vars(formula:str = None) -> list :
    #TODO This function depends on patsy's capabilities and is not as rich as R's rxLinkFormulaComponents
    dependent_vars = []
    try:
        formula_desc = ModelDesc.from_formula(formula)
        lhs_termlist = formula_desc.lhs_termlist
        length = len(lhs_termlist)
        dependent_vars = list(map(extract_name, lhs_termlist[0:length]))

    except PatsyError as formula_exception:
        print("Exception occurred while parsing formula : {0}".format(str(formula_exception)))

    return dependent_vars

def _rx_get_input_vars(formula: str = None) -> list :
    #TODO This function depends on patsy's capabilities and is not as rich as R's rxLinkFormulaComponents
    input_vars = []
    try:
        formula_desc = ModelDesc.from_formula(formula)
        rhs_termlist = formula_desc.rhs_termlist
        length = len(rhs_termlist)
        input_vars = list(map(extract_name, rhs_termlist[1:length]))

    except PatsyError as formula_exception:
        print("Exception occurred while parsing formula : {0}".format(str(formula_exception)))

    return input_vars


def _rx_get_formula_components(formula: str = None) -> list :
    formula_components = []

    dependent_vars = _rx_get_dependent_vars(formula)
    if dependent_vars :
        formula_components.extend(dependent_vars)

    input_vars = _rx_get_input_vars(formula)
    if input_vars :
        formula_components.extend(input_vars)

    return formula_components

def _rx_construct_read_vars(formula_components: list = None, pweights: str = None, fweights: str = None, transform_variables: list = None) -> list :
    read_vars = set()
    if formula_components is not None:
        read_vars = read_vars.union(set(formula_components))

    if isinstance(pweights, str) :
        read_vars.add(pweights)

    if isinstance(fweights, str):
        read_vars.add(fweights)

    if isinstance(transform_variables, list) :
        read_vars = read_vars.union(set(transform_variables))

    return list(read_vars)


def _rx_create_temp_table_hive_data_with_guid_name():
    # Spark 1.6 has bug on dropping temp table, if table name starts with {number}E{number} pattern.
    # Add prefix letter "T" to workaround this bug.
    tmp_table_name = "T" + uuid.uuid1().hex
    return RxHiveData(table=tmp_table_name, save_as_temp_table=True)


def _rx_is_master():
    nodeType = RxOptions.get_option("nodeType")
    return (nodeType is "master")


def _rx_is_worker():
    nodeType = RxOptions.get_option("nodeType")
    return (nodeType is "worker")
